/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define LOG 	1
#define SAVE	3


#import "AlertPanel.h"
#import "NXImage.h"
#import "NXTextField.h"
#import "cstrings/cstrings.h"


static AlertPanel* alert_panel = nil;
static NSImage*    red_icon    = nil;
static NSImage*    yellow_icon = nil;


@implementation AlertPanel

-(void)awakeFromNib 
{
	XLogIn("AlertPanel:awakeFromNib");
	alert_panel = self;
	[self setFloatingPanel:yes];
	[self setWorksWhenModal:yes];
	[self setBecomesKeyOnlyIfNeeded:yes];
}

- (IBAction)doButton1:(id)sender
{
	XLogIn("Button 1 clicked");
	[self orderOut:nil];
}

- (IBAction)doButton2:(id)sender
{
	XLogIn("Button 2 clicked");
	[self orderOut:nil];
}

- (IBAction)doButton3:(id)sender
{
	XLogIn("Button 3 clicked");
	[self orderOut:nil];
}


@end


static void load_icons()
{
	if( !red_icon ) 
	{
		yellow_icon = [alert_panel->icon image];	// = [[NXImage alloc] initWithContentsOfResourceFile:"attention.gif"];
		red_icon    = [[NXImage alloc] initWithContentsOfResourceFile:"stop.gif"];
	}
}


void ShowAlert( cstr msg, ... )
{
	XLogIn("AlertPanel:ShowAlert(...)");

// we may be called on an arbitrary thread
// => assert an autorelease pool exists
// TODO: auf einem anderen Thread knallt es aber wg. thread != mainThread ...
// deswegen ist das Textfeld ein NSTextField, kein NXTextField,
// das zeigt (vielleicht) nicht richtig an, aber es knallt nicht...
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init]; 
	OMEM(pool);

		va_list va;
		va_start(va,msg);
		str s = VaUsing(msg,va);
		va_end(va);

		LogLine(s);

		if( alert_panel )	// we may be called before AlertPanel:awakeFromNib!
		{
			load_icons();
		
			[alert_panel->message setStringValue:[NSString stringWithUTF8String:s]];
			[alert_panel->icon    setImage:red_icon];
			[alert_panel->button1 setStringValue:@"OK"];
			[alert_panel->button2 setHidden:yes];
			[alert_panel->button3 setHidden:yes];
			[alert_panel center];
			[alert_panel orderFront:nil];
		}
		
	[pool release];
}


void ShowWarning( cstr msg, ... )
{
	XLogIn("AlertPanel:ShowWarning(...)");
	XXXASSERT( alert_panel );
	
// we may be called on an arbitrary thread
// => assert an autorelease pool exists
// TODO: auf einem anderen Thread knallt es aber wg. thread != mainThread ...
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init]; 
	OMEM(pool);

		va_list va;
		va_start(va,msg);
		str s = VaUsing(msg,va);
		va_end(va);

		LogLine(s);

		if( alert_panel )	// we may be called before AlertPanel:awakeFromNib!
		{
			load_icons();
			
			[alert_panel->message setStringValue:[NSString stringWithUTF8String:s]];
			[alert_panel->icon    setImage:yellow_icon];
			[alert_panel->button1 setStringValue:@"OK"];
			[alert_panel->button2 setHidden:yes];
			[alert_panel->button3 setHidden:yes];
			[alert_panel center];
			[alert_panel orderFront:nil];
		}
		
	[pool release];
}








