/*	Copyright  (c)	Günter Woigk 2005 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#define	SAFE	3
#define	LOG		1

#import "kio/kio.h"
#import "CheckUpdateController.h"
#import "MenuController.h"
#import "NXTextField.h"
#import "Preferences.h"
#import "utilities.h"
#import "cstrings/cstrings.h"
#import "unix/file_utilities.h"




@implementation CheckUpdateController

- (id)init
{
	XLogIn("CheckUpdateController:init");			// note: wird nach App.-Start ausgeführt!

	self = [super init];	
	if(self)
	{
		urlState		= urlClosed;
		urlRequest		= NULL;
		urlConnection	= NULL;
		urlData			= NULL;
		fileNameUpdate	= NULL;
		fileNamePreview	= NULL;
		baseUrl			= NULL;
		saveFilePath	= NULL;
	}
	return self;
}


- (void)awakeFromNib
{
	XLogIn("CheckUpdateController:awakeFromNib");

	[progress_window setDelegate:self];
	checkPreview = [preferences getBoolForKey:"check_update_includes_preview_version"];
	[check_preview_switch setState:checkPreview];
	baseUrl = [[NSURL alloc] initWithString:@"http://k1.dyndns.org/Develop/projects/zxsp-osx/distributions/"];

	bool f = [preferences getBoolForKey:"startup_check_update"];
	[automatic_update_switch setState:f];
	[automatic_update_group setEnabled:f];
	int d = [preferences getIntForKey:"startup_check_update_intervall"];
	[automatic_update_group selectCellWithTag:d];

	if (!f) { XLogLine("automatic update disabled"); return; }
	int n = [preferences getIntForKey:"startup_check_update_next_date"];
	int today = (int)([NSDate timeIntervalSinceReferenceDate] / (24*60*60));
	if(n>today) { XLogLine("next automatic update in %i day%s",n-today,n-today==1?"":"s"); return; }

// automatic check update
	[self automaticCheckUpdate];

// set next date for automatic check update:
	[preferences setInt:today+d forKey:"startup_check_update_next_date"];
}



/*	Parse directory listing to get latest release and pre-release version 
*/
-(void)extractLatestVersions
{
	XLogIn("CheckUpdateController:extractLatestVersions");
	
	if( !urlData ) { XXLogLine("--> no urlData"); return; }
	delete[] fileNameUpdate;  fileNameUpdate = NULL;	
	delete[] fileNamePreview; fileNamePreview = NULL;	

	NSString* s = [[[NSString alloc] initWithBytes:[urlData bytes] length:[urlData length] encoding:NSUTF8StringEncoding] autorelease];
	NSArray*  a = [s componentsSeparatedByString:@"\n"];
	
	for( uint i=0; i<[a count]-1; i++ )
	{
		NSString* fn  = [a objectAtIndex:i]; 
		bool      pre = [fn rangeOfString:@"pre"].location != NSNotFound;		// flag: pre-version
		cstr      s   = NewCopy([fn UTF8String]); 
		if(pre) fileNamePreview = s; else fileNameUpdate = s;
	}
	
	XLog("Response: %s", [s UTF8String]);
	if( XLOG && fileNamePreview ) Log("    new preview version: %s\n",fileNamePreview);
	if( XLOG && fileNameUpdate )  Log("    new regular version: %s\n",fileNameUpdate);
}



-(void)setState:(UrlState)newState 
{
	XLogIn("CheckUpdateController:setState");
	
	if( newState==urlClosed ) 
	{
xx:		if(urlConnection)	[urlConnection cancel];		
		if(urlConnection)	[urlConnection release];	urlConnection = NULL; 
		if(urlRequest)		[urlRequest release];		urlRequest = NULL; 
		if(urlData)			[urlData release];			urlData = NULL; 
		delete[] fileNameUpdate;  fileNameUpdate  = NULL; 
		delete[] fileNamePreview; fileNamePreview = NULL; 

		[check_update_menu_item	setEnabled:1];
		urlState = urlClosed;
		return; 
	}

	[check_update_menu_item	setEnabled:0];
	cstr newMessage = NULL;

	if( newState==urlGotList )
	{
		if(urlState==urlGetList||urlState==urlSilentGetList) [self extractLatestVersions]; 
	
		if( checkPreview && fileNamePreview )	newMessage = CatStr("New preview: ",BasenameFromPath(fileNamePreview));
		else if( fileNameUpdate )				newMessage = CatStr("New version: ",BasenameFromPath(fileNameUpdate));
		else if( urlState==urlSilentGetList )	{ XLogLine("no update available"); goto xx; }
		else									newState   = urlNoUpdate;
	}
	
	if( newState==urlGotFile )
	{
		bool okflag = [ urlData writeToFile:saveFilePath atomically:no ]; 
		if( !okflag ) { newMessage = "Error: writing to file failed."; newState = urlError; goto x; }		
	}

	if( newState==urlGetList||newState==urlSilentGetList||newState==urlGetFile )
	{
		NSURL* url = baseUrl;			
		if( newState==urlGetFile )		// urlGetFile
		{
			bool pre = checkPreview && fileNamePreview;
			NSString* fn  = ToNSString( pre?fileNamePreview:fileNameUpdate );
			NSString* ufn = [fn stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding];
			url = [NSURL URLWithString:ufn relativeToURL:baseUrl];

			NSSavePanel* dialog = [NSSavePanel savePanel];
			[dialog setRequiredFileType:NULL/*@"zip"*/];
			int result = [dialog runModalForDirectory:NULL file:fn];		
			if( result!=NSFileHandlingPanelOKButton ) return;		// cancel

 			if(saveFilePath) [saveFilePath release];
 			saveFilePath = [[dialog filename] retain];
 		}
 		else							// urlGetList || urlSilentGetList
		{								// append query string => get text list, not html dir
			cstr query = CatStr("?checkupdate=", appl_version_str, appl_pre_version_str);
			url = [NSURL URLWithString:ToNSString(query) relativeToURL:baseUrl];	
		}

		[close_button		setTitle:@"Cancel"];
		[progress_bar		setIndeterminate:1];
		[progress_bar		startAnimation:self];
		newMessage = "Connecting to http://k1.dyndns.org";		

		if( urlRequest ) [urlRequest release];
		urlRequest =  [[NSURLRequest alloc] initWithURL:url cachePolicy:NSURLRequestReloadIgnoringCacheData timeoutInterval:60.0 ];
		if( !urlRequest ) { newState = urlError; newMessage = "Error: creating urlRequest failed!"; goto x; }		

		if( urlConnection ) [urlConnection release];
		urlConnection = [[NSURLConnection alloc] initWithRequest:urlRequest delegate:self];
		if( !urlConnection ) { newState = urlError; newMessage = "Error: creating urlConnection failed!"; goto x; }
	}
	else
	{
x:		[close_button		setTitle:@"Close"];
		[progress_bar		setIndeterminate:0];
		[progress_bar		setDoubleValue:newState==urlGotFile?[progress_bar doubleValue]:0];
	}

	if(!newMessage)	newMessage	= newState==urlNoUpdate ? "There is no new version available." 
								: newState==urlGotFile  ? "Download finished successfully." 
								: "Connection did fail with error";
	
	[message_view			setStrValue:newMessage];
	[download_button		setEnabled:newState==urlGotList];	
	if(newState!=urlSilentGetList) [progress_window	orderFront:self];
	
	urlState = newState;	
}



/* event from 'check update' in 'help' menu: 
*/
- (IBAction)doOpen:(id)sender		
{
	XLogIn("CheckUpdateController:open");			
	[self setState:urlGetList]; 
}

/* event from the 'download' button: 
*/
- (IBAction)doDownload:(id)sender	
{
	XLogIn("CheckUpdateController:doDownload");	
	[self setState:urlGetFile]; 
}

/* event from 'close' button:
*/
-(IBAction)doClose:(id)sender		
{
	XLogIn("CheckUpdateController:close");		
	[progress_window close]; 
}

/* notification for window closing:
*/
- (void)windowWillClose:(NSNotification*)dummy	
{ 
	XLogIn("CheckUpdateController:windowWillClose"); 
	[self setState:urlClosed]; 
}


/* event from the "include 'pre' versions" checkbox: 
*/
- (IBAction)setCheckPreview:(id)sender
{
	XLogIn("CheckUpdateController:setCheckPreview");

	checkPreview = [sender state];
	[preferences setBool:checkPreview forKey:"check_update_includes_preview_version"];
	
	if( urlState==urlGotList || urlState==urlNoUpdate || urlState==urlGotFile ) [self setState:urlGotList];
}









// Delegate Methods for NSURLConnection:

//- (void)connection:(NSURLConnection *)connection didCancelAuthenticationChallenge:(NSURLAuthenticationChallenge *)challenge;
//- (void)connection:(NSURLConnection *)connection didReceiveAuthenticationChallenge:(NSURLAuthenticationChallenge *)challenge;

-(NSURLRequest*)connection:(NSURLConnection*)connection willSendRequest:(NSURLRequest*)request redirectResponse:(NSURLResponse*)response { return request; }
-(NSCachedURLResponse*)connection:(NSURLConnection*)connection willCacheResponse:(NSCachedURLResponse*)response { return NULL; /* don't cache */ }


-(void)connection:(NSURLConnection*)connection didReceiveResponse:(NSURLResponse*)response
{
	XLogIn("CheckUpdateController:connection:didReceiveResponse");

	if( urlData ) { [urlData release]; urlData = NULL; }
	
	if( urlState==urlGetList||urlState==urlSilentGetList) 
		 [message_view setStrValue:"Retrieving directory listing..."];	
	else [message_view setStrValue:"Retrieving update..."];	

	if( [response expectedContentLength]>0 ) 
	{ 
		[progress_bar setMinValue:0]; 
		[progress_bar setMaxValue:[response expectedContentLength]]; 
		[progress_bar setDoubleValue:0]; 
		[progress_bar setIndeterminate:0]; 
	}
}

-(void)connection:(NSURLConnection*)connection didReceiveData:(NSData*)newData
{
	XLogIn("CheckUpdateController:connection:didReceiveData");

	[progress_bar incrementBy:[newData length]]; 	
	if( urlData )	[urlData appendData:newData];
	else			urlData = [[NSMutableData alloc] initWithData:newData];
}


/*	either connection:didFailWithError: or connectionDidFinishLoading: is sent as last delegate callback
	unless 'cancel' was sent to the NSURLConnection 
*/
-(void)connection:(NSURLConnection *)connection didFailWithError:(NSError*)error
{
	XLogIn("CheckUpdateController:connection:didFailWithError");

	[self setState:urlError];	
}

-(void)connectionDidFinishLoading:(NSURLConnection*)connection
{
	XLogIn("CheckUpdateController:connectionDidFinishLoading");

	[self setState:urlState==urlGetList||urlState==urlSilentGetList ? urlGotList : urlGotFile];
}



- (IBAction)toggleAutomaticUpdate:(id)sender
{
	XLogIn("CheckUpdateController:toggleAutomaticUpdate");
	
	bool f = [sender state];
	[automatic_update_group setEnabled:f];
	[preferences setBool:f forKey:"startup_check_update"];
	if(f)
	{
		int d = [preferences getIntForKey:"startup_check_update_intervall"];
		d += (int)([NSDate timeIntervalSinceReferenceDate] / (24.0*60*60));			
		[preferences setInt:d forKey:"startup_check_update_next_date"];
	}
}


- (IBAction)setAutomaticUpdateIntervall:(id)sender
{
	XLogIn("CheckUpdateController:setAutomaticUpdateIntervall");
	id cell = [sender selectedCell];
	int n = [cell tag];
	[preferences setInt:n forKey:"startup_check_update_intervall"];
	n += (int)([NSDate timeIntervalSinceReferenceDate] / (24*60*60));
	[preferences setInt:n forKey:"startup_check_update_next_date"];
}


-(void)automaticCheckUpdate
{
	XLogIn("CheckUpdateController:automaticCheckUpdate");
	[self setState:urlSilentGetList];
}




@end












