/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#import "kio/kio.h"
#import "DspTime.h"


typedef float	Sample;			typedef Sample const		cSample;
typedef class	StereoSample;	typedef StereoSample const	cStereoSample;

class StereoSample
{
	Sample	l,r;

public:
				StereoSample( )						throw()							{}
				StereoSample( Sample m )			throw()		: l(m), r(m)		{}
				StereoSample( Sample l, Sample r )	throw()		: l(l), r(r)		{}
				StereoSample( cStereoSample& q )	throw()		: l(q.l), r(q.r)	{}
		
StereoSample&	operator=	( Sample q )			throw()		{ l=r=q;			return *this; }
StereoSample&	operator=	( cStereoSample& q )	throw()		{ l=q.l;  r=q.r;	return *this; }
StereoSample&	operator+=	( Sample q )			throw()		{ l+=q;	  r+=q;		return *this; }
StereoSample&	operator+=	( cStereoSample& q )	throw()		{ l+=q.l; r+=q.r;	return *this; }
StereoSample&	operator-=	( Sample q )			throw()		{ l-=q;	  r-=q;		return *this; }
StereoSample&	operator-=	( cStereoSample& q )	throw()		{ l-=q.l; r-=q.r;	return *this; }
StereoSample&	operator*=	( float f )				throw()		{ l*=f;	  r*=f;		return *this; }
StereoSample	operator*	( float f )				throw()		{ return StereoSample(l*f,r*f); }
bool			operator>=	( Sample q )	const	throw()		{ return Mono()>=q; }
bool			operator<=	( Sample q )	const	throw()		{ return Mono()<=q; }
bool			operator<	( Sample q )	const	throw()		{ return Mono()<q;  }

Sample&			operator[]	(int i)					throw(index_error) { XXXINDEX(i,2); return (&l)[i]; }
Sample&			Left		()						throw()		{ return l; }
Sample&			Right		()						throw()		{ return r; }
cSample&		Left		()				const	throw()		{ return l; }
cSample&		Right		()				const	throw()		{ return r; }
Sample			Mono		()				const	throw()		{ return (l+r)/2.0; }
};



namespace DSP
{
extern	StereoSample	audio_out_buffer[];
extern	Sample			audio_in_buffer[];

extern	bool	audio_input_device_present;
extern	bool	audio_input_device_enabled;
extern	bool	audio_output_device_enabled;
extern	Sample	audio_output_volume;

extern	cstr	audio_in_filename;
extern	cstr	audio_out_filename;

extern	void	StartCoreAudio				( );

inline	bool	AudioInputDevicePresent		( )			{ return audio_input_device_present; }
inline	bool	AudioInputDeviceEnabled		( )			{ return audio_input_device_enabled; }
extern	void	EnableAudioInputDevice		( bool f );

inline	bool	AudioOutputDevicePresent	( )			{ return yes; }
inline	bool	AudioOutputDeviceEnabled	( )			{ return audio_output_device_enabled; }
extern	void	EnableAudioOutputDevice		( bool f );

extern	void	SetOutputVolume				( Sample volume );
inline	Sample	GetOutputVolume				( )			{ return audio_output_volume; }

extern	void	SetPlaythrough				( int mode );	
		enum 
		{		playthrough_minus10dB, 
				playthrough_minus30dB, 
			 	playthrough_off 
		};

extern	void	OutputSamples				( Sample l, Sample r, Time start, Time end );

extern	void	StopAudioInputFromFile();
extern	void	StartAudioInputFromFile();
extern	void	StopAudioOutputToFile();
extern	void	StartAudioOutputToFile();


// ---- Utilities ----

// apply high pass filter
// returns new center sample
extern	Sample	HighPassFilter			( Sample* z, Sample const* q, int count, Sample center, float damping );
};




















