/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
	
	
	NOTE:
	
				wird z.Zt. wg. 10.5 mit -Wno-error übersetzt!		ERSTMAL NICHT mEHR..
	
	
*/

#define	LOG		1
#define	SAFE	3

#import <CoreServices/CoreServices.h>
#import <CoreAudio/CoreAudio.h>
#import <AudioToolbox/ExtendedAudioFile.h>



#import "unix/pthreads.h"
#import "unix/file_utilities.h"
#import "unix/tempmem.h"
#import "cstrings/cstrings.h"

#import "globals.h"
#import "Dsp.h"
#import "Machine.h"
#import "TapeRecorder.h"
#import "Preferences.h"
#import "AlertPanel.h"


Time			system_time			= 0.0;
Frequency		samples_per_second	= 44100;



namespace DSP
{

 
// ---- core audio interrupt ----

static PLock	audio_callback_lock;		// 2006-11-16 kio:	needed if in and out devices are different

bool			audio_input_device_present	= no;	// preset to no if error
bool			audio_input_device_enabled	= no;	// preset to no if error
bool			audio_output_device_enabled = yes;	// must be present
Sample			audio_output_volume			= 0.3;
Sample			play_through_damping		= 0.0;	// no playthrough during startup: no 'klick' if audio-in disabled in prefs
StereoSample	audio_out_center			= 0.0;	// for audio out High-pass filter
Sample			audio_in_center				= 0.0;	// for audio in High-pass filter

StereoSample	audio_out_buffer [DSP_SAMPLES_PER_BUFFER+DSP_SAMPLES_STITCHING]; // Sample dummy1_l[DSP_SAMPLES_GUARDZONE];
Sample	 		audio_in_buffer	 [DSP_SAMPLES_PER_BUFFER+DSP_SAMPLES_STITCHING]; // Sample dummy2[DSP_SAMPLES_GUARDZONE];


cstr			audio_in_filename			= NULL;
cstr			audio_out_filename			= NULL;
PLock			audio_input_from_file_lock;
PLock			audio_output_to_file_lock;
static ExtAudioFileRef audio_in_file		= NULL; 
static ExtAudioFileRef audio_out_file		= NULL; 
static AudioBufferList audio_in_data		= { 1/*numBuffers*/, 
												{ 2/*numChannels*/, DSP_SAMPLES_PER_BUFFER*2*sizeof(Sample)/*byteSize*/, NULL/*data*/ } };


static PLock			device_list_lock;
//static int				device_list_seed;				// incremented when device list changed
//static int				num_devices = 0;
static AudioDeviceID	//device_list[16],				// available devices. updated by AHPropertyListenerProc()
						//std_input_device_id=0,			// standard audio devices 
						//std_output_device_id=0,
						input_device_id=0,				// actually used audio devices 
						output_device_id=0;


#if 0
static void GetAudioDevices() throw()
{
/*	in:  device_list_lock should be locked.
	out: get list of all devices
		 get standard input and output device
		 validate current input and output device
		 increment device_list_seed
	
*/
	OSStatus status = noErr;
	UInt32 	 size;
	int		 count,i;
    
	device_list_seed++;

// get number of available devices:
	status = AudioHardwareGetPropertyInfo( kAudioHardwarePropertyDevices, &size, NULL );	
	if(status) { size=0; Log( "GetAudioDevices:GetCount:Error: $%8X\n", (int)status ); }

	count = size / sizeof(AudioDeviceID);
	if( count < 1 ) { count=0; Log( "GetAudioDevices: No audio devices found!\n" ); }
    else if( count > NELEM(device_list) ) 
	{ count=NELEM(device_list); size=sizeof(device_list); Log( "GetAudioDevices: Too many devices. List truncated.\n" ); }
	
// get device list:
	status = AudioHardwareGetProperty( kAudioHardwarePropertyDevices, &size, (void*)device_list );
	if(status) { count=0; Log( "GetAudioDevices:GetList:Error: $%8X\n", (int)status ); }
	num_devices = count;

// get standard output device:
	size = sizeof(std_output_device_id);
	status = AudioHardwareGetProperty( kAudioHardwarePropertyDefaultOutputDevice, &size, &std_output_device_id );
	if(status) { std_output_device_id=kAudioDeviceUnknown; Log( "GetAudioDevices:GetStdOut:Error: $%8X\n", (int)status ); }
	if( std_output_device_id == kAudioDeviceUnknown ) Log( "GetAudioDevices: standard output device = Unknown.\n" ); 

// validate current output device:
	if( output_device_id==kAudioDeviceUnknown ) output_device_id=std_output_device_id;
	for( i=0; i<count; i++ ) { if(output_device_id==device_list[i]) break; } 
	if(i==count) { output_device_id=std_output_device_id; Log( "GetAudioDevices: Current output device not in list.\n" );  }

// get standard input device:
	size = sizeof(std_input_device_id);
	status = AudioHardwareGetProperty( kAudioHardwarePropertyDefaultInputDevice, &size, &std_input_device_id );
	if(status) { std_input_device_id=kAudioDeviceUnknown; Log( "GetAudioDevices:GetStdIn:Error: $%8X\n", (int)status ); }
	if( std_input_device_id == kAudioDeviceUnknown ) Log( "GetAudioDevices: standard input device = Unknown.\n" ); 

// validate current input device:
	if( input_device_id==kAudioDeviceUnknown ) input_device_id=std_input_device_id;
	for( i=0; i<count; i++ ) { if(input_device_id==device_list[i]) break; } 
	if(i==count) { input_device_id=std_input_device_id; Log( "GetAudioDevices: Current input device not in list.\n" );  }
}


static void UpdateDeviceList()
{
	TODO();
	PLocker z(device_list_lock);
	device_list_seed++;
}


static OSStatus AHPropertyListenerProc( AudioHardwarePropertyID n, void* in_client_data )
{
    switch(n)
	{
//	case kAudioHardwarePropertyDefaultInputDevice:
//		fprintf(stderr, "AHPropertyListenerProc: default input device changed\n"); break;
			
//	case kAudioHardwarePropertyDefaultOutputDevice:
//		fprintf(stderr, "AHPropertyListenerProc: default output device changed\n"); break;
			
//	case kAudioHardwarePropertyDefaultSystemOutputDevice:
//		fprintf(stderr, "AHPropertyListenerProc: default system output device changed\n"); break;

	case kAudioHardwarePropertyDevices:
		UpdateDeviceList();
		//[menu_controller performSelectorOnMainThread:@selector(updateAudioDeviceList) withObject:nil waitUntilDone:NO];
		break;
			
	default:
		LogLine("AHPropertyListenerProc: unknown message"); break;
    }
	return noErr;
}
#endif



inline Sample HighPassFilter ( Sample* p, int count, Sample center, float damping )
{	
	for( Sample* e = p+count; p < e; p++ )
	{
		center += (*p-=center)*damping;
	}
	return center;
}


inline StereoSample HighPassFilter ( StereoSample* p, int count, StereoSample center, float damping )
{	
	for( StereoSample* e = p+count; p < e; p++ )
	{
		center += (*p-=center)*damping;
	}
	return center;
}


static void stop_audio_output_to_file()		// assumes locked lock!
{
	delete[] audio_out_filename; audio_out_filename = NULL;
	if( audio_out_file ) { /* OSStatus err = */ ExtAudioFileDispose(audio_out_file); }
	audio_out_file = NULL;
}


static void stop_audio_input_from_file()	// assumes locked lock!
{
	delete[] audio_in_filename; audio_in_filename = NULL;
	if( audio_in_file ) { /* OSStatus err = */ ExtAudioFileDispose(audio_in_file); }
	audio_in_file = NULL;
}



/*	abhängig von der Hardware und dem gewählten Default-Input- und -Output-Device
	sind drei Betriebsmodi möglich:
		
	combined io:	input und output gehen über das selbe Device.
					Es gibt _einen_ Callback für input und output.
	
	output only:	es gibt kein Audio-Input-Device. 
					Der Inputbuffer muss im Audio-Output-Callback gelöscht werden.
					
	split io:		Audio-Input und -Output laufen über unterschiedliche Devices.
					Es gibt für Input und Output jeweils einen eigenen Callback.
					*Falls* sich diese Interrupts nicht gegenseitig unterbrechen können,
						genügt es, die Input-Aktionen im Input-Callback auszuführen und
						die Output-Aktionen im Output-Callback, gerade so, als habe man hier selbst
						zuvor die Input-Aktionen durchgeführt.
					*Falls* sie sich gegenseitig unterbrechen können, müssen die Input-Daten
						im Input-Callback in einen Puffer zwischengespeichert werden.
					*Falls* Input- und Output-Datenrate unterschiedlich sind, muss der Input resampelt werden.
					*Falls* Input- und Output-Puffer unterschiedlich groß sind, muss Input zwischengespeichert werden.

	Wenn ein Input-Device vorhanden ist, kann dieses durch mutwilliges Umschalten auf 'output_only' deaktiviert werden.
*/

 

/*	Der In- und Outpuffer haben 4 Samples extra am Ende.

Output:
	Die Maschinen können nicht auf Sample-Grenze gestoppt werden, müssen aber den Outpuffer komplett füllen.
	Also werden sie immer etwas über den Endzeitpunkt hinauslaufen und mindestens noch ins nächste Sample schreiben.
	4 Samples @ 44.100 Hz ergeben einen Überlaufpuffer von 90.7 µs, bei einem CPU-Takt von 3.5 MHz sind das bspw. 317.5 T-Zyklen
	oder für einen PSG-Takt von 1,75 MHz sind das 159 T-Zyklen, nach internem Vorteiler 1/16 noch 9.9 Takte.
	Die Überlauf-Samples werden beim nächsten Audio-Interrupt an den Start des Puffers kopiert, so dass die Maschinen
	ihren Sound jetzt nahtlos anschließen können. Dazu müssen sich die Maschinen natürlich immer den Zeitpunkt exakt merken,
	bis zu dem sie bereits gelaufen sind.

Input:	
	Für den Audioin-Puffer gilt ähnliches. Nur da die Maschinen ggf. etwas über das Pufferende hinausschießen, werden sie
	aus dem Überlaufpuffer lesen, dieser muss also gültige Daten enthalten. Deshalb wird der Audioin-Puffer vom Interrupt 
	ab Offset +DSP_SAMPLES_STITCHING bis zum Ende der Überlaufsamples bei Pufferende+DSP_SAMPLES_STITCHING gelesen. 
	Vorher werden die alten Samples aus dem Überlaufbereich an den Audioin-Puffer-Anfang kopiert. Diese Samples konnten 
	also entweder beim Letzten Lauf aus dem Überlaufbereich oder bei diesem Interrupt aus dem Pufferanfang gelesen werden, 
	jenachdem wie weit die Mashcine beim letzten Interrupt über das Pufferende hinaus schoss.
	
Latenz:
	Liest eine Maschine den Audio-Input und kopiert ihn sofort danach wieder nach Output, ergibt sich folgende Latenz:
	2 Buffer für das Betriebssystem 
		Das BS hat 2 Puffer für Audio-In und zwei für Audio-Out. Jeweils ein Puffer wird gerade von der Audio-HW beackert,
		und jeweils ein Puffer ist der, den wir per Audio-Interrupt zum Befüllen/Auslesen übergeben bekommen.
		Das 1. Sample im Audio-In-Puffer is "1 Puffer alt". Nach Kopieren in den Audio-Out-Puffer ist das 1. Sample im 
		Output-Puffer also "1 Puffer alt".
		Der nächste Audio-Interrupt wird kommen, wenn die Puffer an der Audio-HW wieder voll/leer sind, also nach "einem Puffer". 
		In diesem Moment wird "unser" Puffer an die Audio-HW übergeben, und das Sample wird also mit 2 Pufferlängen Verspätung
		wieder abgespielt.
	Hinzu kommen jetzt 1 Buffer + 4 Bytes Überlaufbereich für unsere Audio-Implementierung.		((DSP_SAMPLES_STITCHING=4))
	Grande Totale 3 Puffer + 4 Samples = 3x256+4 = 772 Samples = 17,5 ms bzw. ~57 Hz 

TODO: Man könnte eine Option vorsehen, den Input (gedämpft) direkt zurückzukopieren um die EAR-MIC-Verkoppelung zu simulieren.
	Optional kleinere Puffer verwenden.
*/

/* ---- core audio irpt handler ----
*/
static OSStatus audioDeviceIOProc
(
	AudioDeviceID			inDevice,
	const AudioTimeStamp*	inNow,
	const AudioBufferList*	inInputData,
	const AudioTimeStamp*	inInputTime,
	AudioBufferList*		outOutputData,
	const AudioTimeStamp*	inOutputTime,
	void* 					inClientData 
)
{
/*	1.	line in       -->  audio_in_buffer
	2.	tape recorder -->  audio_in_buffer
	3.	                   audio_in_buffer ---- -10dB --> audio_out_buffer
	4.                     audio_in_buffer -> Machines -> audio_out_buffer
	5.                                                  audio_out_buffer  -->  tape recorder
	6.                                                  audio_out_buffer  -->  line out
*/	

	TempMemPool tmp;							// pool must be explicitely created for external visitor thread!
	PLocker lock(&audio_callback_lock);			// 2006-11-16 kio
	uint buffers, channels;

// === Handle Audio Input ============

//	if no audio-in data is supplied:
//	=>	this is an audio-out-only callback:
//		either there is no audio-in device present or 
//		we get separate callbacks for audio-in and -out in systems with different audio-in and -out device IDs
//
//		if no audio-in device present:
//		=>	we'll never be called for audio-in:
//			=>	handle audio-in here: clear input sample buffer & merge tape recorder signal.
//
//		if audio-in device is present:
//		=>	we'll get a separate callback for audio-in: 
//			=>	skip audio-in handling!
//
//	if audio-in data is supplied:
//	=>	this is an audio-in-only callback in systems with different audio-in and -out device IDs or
//		a combined audio-in and -out callback, if the device ID are the same.
//		still, audio-in may be disabled by the user.
//
//		if audio-in is enabled:
//		=>	handle audio-in: copy audio-in data to input sample buffer & merge tape recorder signal.
//
//		if audio-in is disabled:
//		=>	handle audio-in: clear input sample buffer & merge tape recorder signal.
//

	if( (inInputData && inInputData->mNumberBuffers != 0) || !audio_input_device_present )	
	{

	// shift input buffer stitching samples:

		for( int i=0; i<DSP_SAMPLES_STITCHING; i++ ) 
		{
			audio_in_buffer [i] = audio_in_buffer [DSP_SAMPLES_PER_BUFFER+i]; 
		}

	// if "read audio from file" is selected, 
	// then read data from file and point 'inInputData' to the read data:

		if( audio_in_filename!=NULL )		
		{
			if( audio_input_from_file_lock.TryLock() )
			{			
				if( !audio_in_file )			
				{ stop_audio_input_from_file(); XLogLine("DSP: no audio_in_file"); }
				else
				{
					UInt32 num_frames = DSP_SAMPLES_PER_BUFFER;
					OSStatus err = ExtAudioFileRead( audio_in_file, &num_frames, &audio_in_data );
					if( err!=noErr ) 
					{ stop_audio_input_from_file(); XLogLine("DSP: ExtAudioFileRead error: %s",ErrorStr(err)); }
					else if( num_frames!=DSP_SAMPLES_PER_BUFFER )
					{ stop_audio_input_from_file(); XLogLine("DSP: num_frames!=DSP_SAMPLES_PER_BUFFER"); }	
					else { inInputData = &audio_in_data; }
				}
				audio_input_from_file_lock.Unlock();
			}
		}

	// copy audio-in buffer to audio_in_buffer (stereo -> mono):
	// or clear audio_in_buffer, if audio-in not present or disabled by user:

		Sample* z = audio_in_buffer+DSP_SAMPLES_STITCHING;
		Sample* e = z + DSP_SAMPLES_PER_BUFFER;

		if( audio_in_filename!=NULL || audio_input_device_enabled /* && audioInputDevicePresent (implied) */ )
		{
			XXXASSERT( audio_in_filename!=NULL || audio_input_device_present );

		// new: 2007-07-13

			buffers  = inInputData->mNumberBuffers;
			channels = inInputData->mBuffers[0].mNumberChannels;

			if( XLOG||XXSAFE ) 
			{	static bool f=1; 
				if(f)
				{	f=0;
					XLog( "Dsp:audioDeviceIOProc: %u audio-in buffers.\n",   buffers ); 
					XLog( "Dsp:audioDeviceIOProc: %u channels per buffer.\n", channels );
							
					if(XXSAFE) for( uint i=1; i<buffers; i++ )
					{
						ASSERT( inInputData->mBuffers[i].mNumberChannels == channels );
						ASSERT( inInputData->mBuffers[i].mDataByteSize == (DSP_SAMPLES_PER_BUFFER * channels * sizeof(Sample)) );
						ASSERT( buffers==1 || channels==1 );
					}
				}
			}

			switch( buffers )
			{
			default:	// multiple buffers => assume 1 channel per buffer
				{		//  2 buffers:  stereo?		not yet seen		2007-08-13
						// >2 buffers:  dolby 5.1?	not yet seen		2007-08-13
				
				// ignore other channels:
				//	for( uint i=2; i<inInputData->mNumberBuffers; i++ ) {}
					
				// copy audio-in buffers to audio_in_buffer to  (stereo -> mono):
					Sample const* q0 = (Sample const*)inInputData->mBuffers[0].mData;
					Sample const* q1 = (Sample const*)inInputData->mBuffers[1].mData;
					while(z<e) { *z++ = (*q0++ + *q1++)/2.0; }
				}	break;
				
			case 1:		// 1 buffer => assume interleaved channels
				{		//   1 channel: mono
						//	 2 channels: stereo  (99% of all cases)
						//  >2 channels: dolby5.1?
					
					Sample const* q = (Sample const*)inInputData->mBuffers[0].mData;

					switch( channels )
					{
					case 1:		// mono
						while(z<e) { *z++ = *q++; }		
						break;
						
				//	case 2:		// interleaved stereo ((99% of cases))
				//		while(z<e) { *z++ = (q[0]+q[1])/2.0; q+=2; }
				//		break;

					default:	// e.g. 6 channels: dolby 5.1?  l/r channel seem to be buffer[0/1]
						while(z<e) { *z++ = (q[0]+q[1])/2.0; q+=channels; }
						break;
					}	
					break;
				}			
			}
		}
		else	// audio-in not present or disabled by user:
		{
			while(z<e) { *z++ = 0.0; }		// clear input buffer
		}
		
	// merge sound from virtual tape recorder:
		[tape_recorder readAudioFromTape:audio_in_buffer+DSP_SAMPLES_STITCHING count:DSP_SAMPLES_PER_BUFFER];		

	// High-pass filter: (remove signal bias)
		audio_in_center = HighPassFilter( 
							audio_in_buffer+DSP_SAMPLES_STITCHING, 
							DSP_SAMPLES_PER_BUFFER, audio_in_center, 0.001 /*damping*/ );
	}
	
	else	// audio-in present, enabled, but no audio data supplied:	
			// => this is an audio-out-only callback for systems with different audio-in and -out device IDs
	{
		if(XLOG){ static bool f=1; if(f){f=0;Log("Dsp:audioDeviceIOProc: separate audio-in and -out callbacks.\n");}}
		/* nothing to do */
	}
	

// === Handle Audio Output =======================

// audio output buffer supplied 
// => this is a output-only or combined i/o callback

	if( outOutputData && outOutputData->mNumberBuffers != 0 )
	{

	// shift buffer stitching samples
		for( int i=0; i<DSP_SAMPLES_STITCHING; i++ ) { audio_out_buffer[i] = audio_out_buffer[DSP_SAMPLES_PER_BUFFER+i]; }

	// copy audio_in_buffer to audio_out_buffer
		if( play_through_damping==0.0 )
			for ( uint i=DSP_SAMPLES_STITCHING; i < DSP_SAMPLES_PER_BUFFER+DSP_SAMPLES_STITCHING; i++) 
			{
				audio_out_buffer[i] = 0.0;
			}
		else
			for ( uint i=DSP_SAMPLES_STITCHING; i < DSP_SAMPLES_PER_BUFFER+DSP_SAMPLES_STITCHING; i++ ) 
			{
				audio_out_buffer[i] = audio_in_buffer[i] * play_through_damping;
			}

	// run the machines to read/write sound
		machine_list.RunMachinesForSound();
		
	// High-pass filter: (remove signal bias)
		audio_out_center = HighPassFilter( audio_out_buffer, DSP_SAMPLES_PER_BUFFER, audio_out_center, 0.001 /*damping*/ );
		
	// copy sound to virtual tape recorder:
		[tape_recorder writeAudioToTape:audio_out_buffer count:DSP_SAMPLES_PER_BUFFER];

	// copy audio_out_buffer to audio-out buffer (mono -> stereo):

		buffers  = outOutputData->mNumberBuffers;
		channels = outOutputData->mBuffers[0].mNumberChannels;

		if( XLOG||XXSAFE ) 
		{	static bool f=1; 
			if(f)
			{	f=0;
				XLog( "Dsp:audioDeviceIOProc: %u audio-out buffers.\n",   buffers ); 
				XLog( "Dsp:audioDeviceIOProc: %u channels per buffer.\n", channels );
 						
 				if(XXSAFE) for( uint i=1; i<buffers; i++ )
 				{
 					ASSERT( outOutputData->mBuffers[i].mNumberChannels == channels );
 					ASSERT( outOutputData->mBuffers[i].mDataByteSize == (DSP_SAMPLES_PER_BUFFER * channels * sizeof(Sample)) );
 					ASSERT( buffers==1 || channels==1 );
				}							
			}
		}

	// output to audio-out:

		if( audio_output_device_enabled && audio_output_volume>0.0 )
		{

			StereoSample const* q = audio_out_buffer;
			StereoSample const* e = q + DSP_SAMPLES_PER_BUFFER;
			Sample volume = audio_output_volume;
			
			switch( buffers )
			{
			default:	// multiple buffers => assume 1 channel per buffer
				{		//  2 buffers:  stereo?		not yet seen		2006-10-23
						// >2 buffers:  dolby 5.1?	MOTU 828 kmII		2007-05-18
				
					XXXASSERT(channels==1);
				
				// copy audio_out_buffer to audio-out buffers (mono -> stereo):
					Sample* z0 = (Sample*)outOutputData->mBuffers[0].mData;
					Sample* z1 = (Sample*)outOutputData->mBuffers[1].mData;
					while(q<e) { *z0++ = q->Left()*volume; *z1++ = q->Right()*volume; q++; }

				// clear other channels:
					for( uint i=2; i<buffers; i++ )
					{
						Sample* z = (Sample*)outOutputData->mBuffers[i].mData;
						Sample* e = z + outOutputData->mBuffers[i].mDataByteSize/sizeof(Sample);
						while(z<e) { *z++ = 0.0; }
					}
				}	break;
				
			case 1:		// 1 buffer => assume interleaved channels
				{		//   1 channel: mono
						//	 2 channels: stereo  (99% of all cases)
						//  >2 channels: dolby5.1?
					
					Sample* z = (Sample*)outOutputData->mBuffers[0].mData;

					switch( channels )
					{
					case 1:		// mono
						while(q<e) { *z++ = (q++)->Mono()*volume; }		
						break;
						
					case 2:		// interleaved stereo ((99% of cases))
						while(q<e) { z[0] = q->Left()*volume; z[1] = q->Right()*volume; q++; z+=2; }	
						//memcpy( z, q, DSP_SAMPLES_PER_BUFFER * sizeof(StereoSample) );
						break;

					default:	// e.g. 6 channels: dolby 5.1?  l/r channel seem to be buffer[0/1]
						while(q<e)
						{
							z[0] = q->Left()*volume; z[1] = q->Right()*volume; q++; z+=2;
							for( uint i=2; i<channels; i++ ) { *z++ = 0.0; }
						}
						break;
					}	
					break;
				}			
			}
		}

	// write to file:
	
		if( audio_out_filename!=NULL )
		{
			if( audio_output_to_file_lock.TryLock() )		// should always work
			{
				if( !audio_out_file )
				{ stop_audio_output_to_file(); }
				else
				{			
					UInt32 num_frames = DSP_SAMPLES_PER_BUFFER;
					AudioBufferList* audio_data = outOutputData;
					OSStatus err = ExtAudioFileWriteAsync( audio_out_file, num_frames, audio_data );
					if(err) { stop_audio_output_to_file(); }
				}
				audio_output_to_file_lock.Unlock();
			}
		}
		
		system_time += DSP_SAMPLES_PER_BUFFER / samples_per_second;
	}

    return 0;		// success
}
	


/* ----	start core audio interrupt ----
		abort on any problem with the output device
		disable audio input on any problem with the input device
*/
void StartCoreAudio()
{
	XLogIn("Dsp:StartCoreAudio");
	
    OSStatus status = ok;
	UInt32 bufferByteCount;
	//AudioDeviceID input_device_id, output_device_id;

#if 0
	status = UpdateDeviceList();
	if(status) throw any_error("Dsp:UpdateDeviceList");
	status = AudioHardwareAddPropertyListener(kAudioHardwarePropertyDevices, AHPropertyListenerProc, NULL );
	if(status) throw any_error("AudioHardwareAddPropertyListener");
#endif
	
// block audio interrupt until setup completed:
	PLocker lock(&audio_callback_lock);	

// clear stitching samples: (required?)
	for ( int i=0;i<DSP_SAMPLES_STITCHING; i++ )
	{
		audio_out_buffer[DSP_SAMPLES_PER_BUFFER+i]=0.0;
		audio_in_buffer [DSP_SAMPLES_PER_BUFFER+i]=0.0;
	}


// Setup audio-out:
// abort on any problem with the output device

	try
	{

	// Get output device ID: 
	
		UInt32 propertySize = sizeof(output_device_id);
		status = AudioHardwareGetProperty(kAudioHardwarePropertyDefaultOutputDevice, &propertySize, &output_device_id);
		if(status) throw any_error("GetDefaultOutputDevice");
		if( output_device_id == kAudioDeviceUnknown ) throw any_error("GetDefaultOutputDevice: kAudioDeviceUnknown");

	// Check & print output device status

		AudioStreamBasicDescription	outputStreamBasicDescription;
		propertySize = sizeof(outputStreamBasicDescription);
		status = AudioDeviceGetProperty(output_device_id, 0, NO/*audio-out*/, kAudioDevicePropertyStreamFormat, 
										&propertySize, &outputStreamBasicDescription);
		if(status) throw any_error("GetOutputStreamFormat"); 

		{	XLogIn ( "Dsp: CoreAudio default output device:" );
			XLogLine ( "%.2f mSampleRate",       outputStreamBasicDescription.mSampleRate		);
			XLogLine ( "%4.4s mFormatID",	 (str)&outputStreamBasicDescription.mFormatID		);
			XLogLine ( "%4ld mBytesPerPacket",   outputStreamBasicDescription.mBytesPerPacket	);
			XLogLine ( "%4ld mFramesPerPacket",  outputStreamBasicDescription.mFramesPerPacket	);
			XLogLine ( "%4ld mBytesPerFrame",    outputStreamBasicDescription.mBytesPerFrame	);
			XLogLine ( "%4ld mChannelsPerFrame", outputStreamBasicDescription.mChannelsPerFrame	);
			XLogLine ( "%4ld mBitsPerChannel",   outputStreamBasicDescription.mBitsPerChannel	);
		}
		samples_per_second = (long)outputStreamBasicDescription.mSampleRate;

		UInt32 ofmt = outputStreamBasicDescription.mFormatID;
		UInt32 ocpf = outputStreamBasicDescription.mChannelsPerFrame;
		UInt32 obpf = outputStreamBasicDescription.mBytesPerFrame;
		
		if( ofmt != kAudioFormatLinearPCM ) throw any_error("Default Audio Output Device does not use Linear PCM");
		if( obpf != ocpf * sizeof(Sample) )	throw any_error("Default Audio Output Device has not 4 bytes per sample");

	// Configure output device	

		propertySize = sizeof(bufferByteCount);
		bufferByteCount = DSP_SAMPLES_PER_BUFFER * obpf/*output bytes per frame*/;
		status = AudioDeviceSetProperty(output_device_id, NULL, 0, NO, kAudioDevicePropertyBufferSize, propertySize, &bufferByteCount);
		if(status) throw any_error("SetOutputBufferSize"); 

	// Start audio output interrupt

		status = AudioDeviceAddIOProc(output_device_id, audioDeviceIOProc, NULL);
		if(status) throw any_error("AudioDeviceAddIOProc");

		status = AudioDeviceStart(output_device_id, audioDeviceIOProc);
		if(status) throw any_error("AudioDeviceStart");
	}
	catch(any_error e)
	{
		ShowAlert ( "Audio output setup failed:\n%s.\n"
					"Select another default audio output device in the System Settings and start again.%s",
					status?Using("%s: error %ld",e.Message(),status):e.Message(),
					appl_is_preview?"\nPlease send in the console log: Applications:Utilities:Console.":"" );
		
	//	while( [alert_panel isVisible] ) WaitDelay(1.0);		<-- requires to run the event loop
	//	WaitDelay(5.0);
	//	abort();
		return;
	}


// Setup audio-in:
// disable audio-in on any problem with the input device

	try
	{

	// Get input_device_id  

		UInt32 propertySize = sizeof(input_device_id);
		status = AudioHardwareGetProperty(kAudioHardwarePropertyDefaultInputDevice, &propertySize, &input_device_id);
		if(status) throw any_error("GetDefaultInputDevice");
		if( input_device_id == kAudioDeviceUnknown ) throw any_error("GetDefaultInputDevice: kAudioDeviceUnknown"); 

		XLogLine( "Dsp: Input and output device are %s.", output_device_id==input_device_id?"same":"different" );

	// Check & print the input device status

		AudioStreamBasicDescription	inputStreamBasicDescription;
		propertySize = sizeof(inputStreamBasicDescription);
		status = AudioDeviceGetProperty(input_device_id, 0, YES/*audio-in*/, kAudioDevicePropertyStreamFormat, 
										&propertySize, &inputStreamBasicDescription);
		if(status) throw any_error("GetInputStreamFormat");

		{	XLogIn   ( "Dsp: CoreAudio default input device:" );
			XLogLine ( "%.2f mSampleRate",       inputStreamBasicDescription.mSampleRate		);	// 44100
			XLogLine ( "%4.4s mFormatID",	 (str)&inputStreamBasicDescription.mFormatID		);	// lpcm
			XLogLine ( "%4ld mBytesPerPacket",   inputStreamBasicDescription.mBytesPerPacket	);	// 8
			XLogLine ( "%4ld mFramesPerPacket",  inputStreamBasicDescription.mFramesPerPacket	);	// 1
			XLogLine ( "%4ld mBytesPerFrame",    inputStreamBasicDescription.mBytesPerFrame	);	// 8
			XLogLine ( "%4ld mChannelsPerFrame", inputStreamBasicDescription.mChannelsPerFrame);	// 2
			XLogLine ( "%4ld mBitsPerChannel",   inputStreamBasicDescription.mBitsPerChannel	);	// 32
		}
		UInt32 ifmt = inputStreamBasicDescription.mFormatID;
		UInt32 icpf = inputStreamBasicDescription.mChannelsPerFrame;
		UInt32 ibpf = inputStreamBasicDescription.mBytesPerFrame;
		
		if( ifmt != kAudioFormatLinearPCM)	throw any_error("Default Audio Input Device does not use Linear PCM");
		if( ibpf != icpf*sizeof(Sample) )	throw any_error("Default Audio Input Device has not 4 bytes per sample"); 
		if( samples_per_second != (long)inputStreamBasicDescription.mSampleRate ) 
											throw any_error("Audio-in and -out sampling frequencies differ"); 

	// Configure input device	
		
		if( input_device_id != output_device_id )
		{
			propertySize = sizeof(bufferByteCount);
			bufferByteCount = DSP_SAMPLES_PER_BUFFER * ibpf/*input bytes per frame*/;
			status = AudioDeviceSetProperty(input_device_id, NULL, 0, NO, kAudioDevicePropertyBufferSize, propertySize, &bufferByteCount);
			if(status) throw any_error("SetBufferSize");

		// Start audio input interrupt

			status = AudioDeviceAddIOProc(input_device_id, audioDeviceIOProc, NULL);
			if(status) throw any_error("DeviceAddIOProc");

			status = AudioDeviceStart(input_device_id, audioDeviceIOProc);
			if(status) throw any_error("AudioDeviceStart");
		}

		audio_input_device_present = yes;
		audio_input_device_enabled = [preferences getBoolForKey:"startup_enable_audio_input"];
		SetPlaythrough( [preferences getIntForKey:"runtime_audio_playthrough"] );
	}
	catch( any_error e )
	{
		if( [preferences getBoolForKey:"startup_enable_audio_input"] ) 
		{
			ShowWarning("Audio input setup failed:\n%s.\n"
						"Select another default audio input device in the System Settings "
						"or disable audio input in the preferences.%s",
						status?Using("%s: error %ld",e.Message(),status):e.Message(),
						appl_is_preview?"\nPlease send in the console log: Applications:Utilities:Console.":"" );
		}
		else
		{
			LogLine("Dsp: Audio input setup failed:");
			LogLine( status?"Dsp: %s: error = %ld.":"Dsp: %s.", e.Message(), status );
		}
	}
}


// ------------------------------------------------------------------------

/*	Utility:
	apply high pass filter to sample buffer
	
	center  = running center value of signal
	damping = damping per sample; e.g. 0.001
	
	returns updated center value.
*/
Sample HighPassFilter ( Sample* z, Sample const* q, int count, Sample center, float damping )
{	
	XXLogIn("DSP::HighPassFilter(%.2f)",(float)center);

	for( int n=count; n>0; n-- )
	{
		Sample qv = *q++;
		*z++ = qv - center;
		center += (qv-center)*damping;
	}

	return center;
}



/*	Set audio-in to audio-out playthrough damping:
*/
void SetPlaythrough ( int mode/*as in prefs*/ )
{
	switch(mode)
	{
	case playthrough_minus10dB:		play_through_damping = 1.0/2.0; break;
	case playthrough_minus30dB:		play_through_damping = 1.0/8.0; break;	
	case playthrough_off:			play_through_damping = 0.0;     break;
	default:	IERR();
	}
}


/*	Enable / Disable Audio Input Device
*/
void EnableAudioInputDevice ( bool f )	
{
	if( f && !audio_input_device_present )
	{
		ShowAlert( "No audio input device found." );
	}
	audio_input_device_enabled = f && audio_input_device_present; 
}


/*	Enable / Disable Audio Input Device
*/
void EnableAudioOutputDevice ( bool f )	
{
	audio_output_device_enabled = f;
}


void SetOutputVolume ( Sample volume )
{
	if( volume <= 0.0 )
	{
		audio_output_device_enabled = off; 
	}
	else
	{
		audio_output_volume = Min(volume,(Sample)1.0);
		audio_output_device_enabled = on;
	}
}


/*	store samples in audio_out_buffer: 
*/
void OutputSamples ( Sample sample_l, Sample sample_r, Time aa/*start [seconds]*/, Time ee/*end [seconds]*/ )
{
	aa *= samples_per_second;		// sample-based
	ee *= samples_per_second;		// ""

	long a = long(aa);
	long e = long(ee);

	XXTRAP( a<0 );
	XXTRAP( e<a );	
	XXTRAP( e>=DSP_SAMPLES_PER_BUFFER+DSP_SAMPLES_STITCHING );
	
	StereoSample* pa = audio_out_buffer + a;	
		pa->Left()  -= sample_l * (aa-floor(aa));
		pa->Right() -= sample_r * (aa-floor(aa));

	StereoSample* pe = audio_out_buffer + e;	
		pe->Left()  += sample_l * (ee-floor(ee));
		pe->Right() += sample_r * (ee-floor(ee));

	while( --pe >= pa ) 
	{
		pe->Left() += sample_l;
		pe->Right() += sample_r;
	}
}



/*	stop audio input from file 
	and close file
*/
void StopAudioInputFromFile()
{
	XLogIn("StopAudioInputFromFile()");
	PLocker z(audio_input_from_file_lock);
	stop_audio_input_from_file();
}



/*	stop audio output to file 
	and close file
*/
void StopAudioOutputToFile()
{
	XLogIn("StopAudioOutputToFile()");
	PLocker z(audio_output_to_file_lock);
	stop_audio_output_to_file();
}



/*	open audio input file 
	and start play back from file
	may throw error
*/
void StartAudioInputFromFile()		
{
	if( fullscreen_mode ) { RLOCK; m->SetFullscreen(off); RUNLOCK; }
	
	StopAudioInputFromFile();
	
//	LBLOCK; 

		NSOpenPanel* dialog = [NSOpenPanel openPanel];
		NSArray*	 ftlist = [NSSound soundUnfilteredFileTypes];		// get list of supported file types
		int result = [dialog runModalForTypes:ftlist];					// or: runModalForDirectory:file:types
	
		if( result==NSFileHandlingPanelOKButton )
		{
			PLocker z(audio_input_from_file_lock);
	
			stop_audio_input_from_file();
	
			if( audio_in_data.mBuffers[0].mData==NULL )
			{ audio_in_data.mBuffers[0].mData = new Sample[DSP_SAMPLES_PER_BUFFER*2];	}

			XXXASSERT( audio_in_data.mNumberBuffers==1 );
			XXXASSERT( audio_in_data.mBuffers[0].mNumberChannels==2 );
			audio_in_data.mBuffers[0].mDataByteSize = DSP_SAMPLES_PER_BUFFER*2*sizeof(Sample);	// wird nach 1. File vermüllt!
			//XXXASSERT( audio_in_data.mBuffers[0].mDataByteSize==DSP_SAMPLES_PER_BUFFER*2*sizeof(Sample) );

			OSStatus err;
			FSRef fileRef;
			cstr path = [[dialog filename] fileSystemRepresentation];
			err = FSPathMakeRef( (UInt8*)path, &fileRef, NULL/*isDirectory*/ );
			if(err) throw file_error(err,path);

		// ExtendedAudioFile.h: OSX10.4++
			err = ExtAudioFileOpen( &fileRef, &audio_in_file ); 
			if(err) throw file_error(err,path);
		
		// display info:
			AudioStreamBasicDescription	format;
			UInt32 propertySize = sizeof(format);
			err = ExtAudioFileGetProperty( audio_in_file, kExtAudioFileProperty_FileDataFormat, &propertySize, &format );
			if(err) throw any_error("Get audio file data format");

			{	XLogIn ( "Dsp: Audio file format:" );
				XLogLine ( "%.2f mSampleRate",       format.mSampleRate			);	// 44100
				XLogLine ( "%4.4s mFormatID",	 (str)&format.mFormatID			);	// lpcm
				XLogLine ( "%4ld mBytesPerPacket",   format.mBytesPerPacket		);	// 8
				XLogLine ( "%4ld mFramesPerPacket",  format.mFramesPerPacket	);	// 1
				XLogLine ( "%4ld mBytesPerFrame",    format.mBytesPerFrame		);	// 8
				XLogLine ( "%4ld mChannelsPerFrame", format.mChannelsPerFrame	);	// 2
				XLogLine ( "%4ld mBitsPerChannel",   format.mBitsPerChannel		);	// 32
			}
		// declare audio format as expected by appl.:

			format.mSampleRate			= samples_per_second;
			format.mFormatID			= kAudioFormatLinearPCM;
			format.mFormatFlags			= kAudioFormatFlagsNativeFloatPacked;
			format.mBytesPerPacket		= 2* sizeof(Sample);
			format.mFramesPerPacket		=    1;
			format.mBytesPerFrame		= 2* sizeof(Sample);
			format.mChannelsPerFrame	= 2* 1;
			format.mBitsPerChannel		=    sizeof(Sample)*8;
			format.mReserved			= 0;

			propertySize = sizeof(format);
			err = ExtAudioFileSetProperty( audio_in_file, kExtAudioFileProperty_ClientDataFormat, propertySize, &format );
			if(err) throw any_error(err,CatStr("Set audio file conversion to lpcm: ",ErrorStr(err)));

		// enable file input on audio interrupt:

			audio_in_filename = NewCopy(path);
		}

//	LUNBLOCK; 	
}



/*	open audio output file 
	and start recording to file
*/
void StartAudioOutputToFile()
{
	if( fullscreen_mode ) { RLOCK; m->SetFullscreen(off); RUNLOCK; }
	
//	LBLOCK; 

		StopAudioOutputToFile();

	// Save File requester:

		NSSavePanel* dialog = [NSSavePanel savePanel];
		NSArray*	 ftlist = [NSArray arrayWithObjects:@"aiff",@"aif",@"AIF",@"AIFF",nil];
		[dialog setTitle:@"Save Audio Output to File:"];
		[dialog setAllowedFileTypes:ftlist];
		int result = [dialog runModal];		// or: runModalForDirectory:file
	
		if( result==NSFileHandlingPanelOKButton )
		{
			PLocker z(audio_output_to_file_lock);

			stop_audio_output_to_file();		// security only

		// select format & create file:
			
			OSStatus err;
			FSRef parentDir;
			cstr dirpath = [[dialog directory] fileSystemRepresentation];
			err = FSPathMakeRef( (UInt8*)dirpath, &parentDir, NULL/*out: isDirectory*/ );
			if(err) throw file_error(err,CatStr("FSPathMakeRef for directory path failed: ",ErrorStr(err)));

			cstr path = [[dialog filename] fileSystemRepresentation]; path = FileNameFromPath(path);
			CFStringRef fileName = CFStringCreateWithCString( NULL, path, kCFStringEncodingUTF8 );

			AudioFileTypeID fileType = kAudioFileAIFFType;	// AudioToolbox/AudioFile.h
			/*	kAudioFileAIFFType				= 'AIFF',
				kAudioFileAIFCType				= 'AIFC',
				kAudioFileWAVEType				= 'WAVE',
				kAudioFileSoundDesigner2Type	= 'Sd2f',
				kAudioFileNextType				= 'NeXT',
				kAudioFileMP3Type				= 'MPG3',
				kAudioFileAC3Type				= 'ac-3',
				kAudioFileAAC_ADTSType			= 'adts',
				kAudioFileMPEG4Type             = 'mp4f',
				kAudioFileM4AType               = 'm4af',
				kAudioFileCAFType				= 'caff'	*/
			
			AudioChannelLayout* const channelLayout = NULL;

			AudioStreamBasicDescription format;
			format.mSampleRate			= samples_per_second;
			format.mFormatID			= kAudioFormatLinearPCM;
			format.mFormatFlags			= kAudioFormatFlagIsBigEndian | kAudioFormatFlagIsPacked | kAudioFormatFlagIsSignedInteger;
											// note: kAudioFormatFlagsNativeEndian => Fehler auf Intel.		2008-04-19
			format.mBytesPerPacket		= 2* sizeof(short int);
			format.mFramesPerPacket		=    1;					// uncompressed
			format.mBytesPerFrame		= 2* sizeof(short int);
			format.mChannelsPerFrame	= 2* 1;					// interleaved
			format.mBitsPerChannel		=    sizeof(short int)*8;
			format.mReserved			= 0;
			
			err = ExtAudioFileCreateNew( &parentDir, fileName, fileType, &format, channelLayout, &audio_out_file /*out*/ );
			CFRelease(fileName);
			if(err) {
				audio_out_file=NULL; 
				throw file_error(err,CatStr("ExtAudioFileCreateNew for audio file failed: ",ErrorStr(err))); 
			}

		// declare the data format used to feed ExtAudioFileWrite():

			format.mSampleRate			= samples_per_second;
			format.mFormatID			= kAudioFormatLinearPCM;
			format.mFormatFlags			= kAudioFormatFlagsNativeFloatPacked;
			format.mBytesPerPacket		= 2* sizeof(Sample);
			format.mFramesPerPacket		=    1;
			format.mBytesPerFrame		= 2* sizeof(Sample);
			format.mChannelsPerFrame	= 2* 1;
			format.mBitsPerChannel		=    sizeof(Sample)*8;
			format.mReserved			= 0;

			UInt32 propertySize = sizeof(format);
			err = ExtAudioFileSetProperty( audio_out_file, kExtAudioFileProperty_ClientDataFormat, propertySize, &format );
			if(err) throw any_error(err,CatStr("Set audio file conversion source to lpcm: ",ErrorStr(err)));

		// initialize async buffers:
		
			err = ExtAudioFileWriteAsync( audio_out_file, 0 /*numberFrames*/, NULL /* AudioBufferList*/ );
			if(err) throw file_error(err,CatStr("Starting async write to file failed: ",ErrorStr(err)));

		// enable file output on audio interrupt:

			audio_out_filename = NewCopy(path);
		}
	
//	LUNBLOCK; 	
}


}	// namespace
	







