/*	Copyright  (c)	Günter Woigk 2006 - 2006
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#define	SAFE	3
#define	LOG		1

#import "cstrings/cstrings.h"
#import "KeyboardJoystickPanel.h"
#import "Joystick.h"
#import "KbdPanel.h"
#import "Preferences.h"
#import "NXTextField.h"



// all arrays: FUDLR  ((fire,up,down,left,right))

static unichar	old_keycodes[5];	// old state for 'abort'
static unichar	new_keycodes[5];	// key codes currently in use & displayed in panel

static unichar	old_unicodes[5];	// character codes
static unichar	new_unicodes[5];	


inline void UseKeys ( )
{
	SetKeyboardJoystickKeycodes(new_keycodes);
}


static void RestoreOldKeys ( )
{
	for(uint i=0;i<5;i++) new_keycodes[i] = old_keycodes[i];
	for(uint i=0;i<5;i++) new_unicodes[i] = old_unicodes[i];
	UseKeys();
}


static void AcceptNewKeys ( )
{
	for(uint i=0;i<5;i++) old_keycodes[i] = new_keycodes[i];
	for(uint i=0;i<5;i++) old_unicodes[i] = new_unicodes[i];	
}


static void ReadPreferences()
{
	XLogIn("KeyboardJoystickPanel:ReadPreferences");
	
	NSString* keycodes = [preferences getStringForKey:"startup_keyboard_joystick_keycodes"];
	NSString* unicodes = [preferences getStringForKey:"startup_keyboard_joystick_unicodes"];
	
	if( [keycodes length]!=5 || [unicodes length]!=5 )
	{
		keycodes = @"\x31\x7e\x7d\x7b\x7c";			// basic validation of...
		unicodes = @" UDLR";						// ...broken preferences
	}

	for(uint i=0;i<5;i++) new_keycodes[i] = old_keycodes[i] = [keycodes characterAtIndex:i];
	for(uint i=0;i<5;i++) new_unicodes[i] = old_unicodes[i] = [unicodes characterAtIndex:i];
	UseKeys();
}


static void WritePreferences ( )
{
	XLogIn("KeyboardJoystickPanel:WritePreferences");

	NSString* keycodes = [NSString stringWithCharacters:new_keycodes length:5];
	NSString* unicodes = [NSString stringWithCharacters:new_unicodes length:5];

	[preferences setString:keycodes forKey:"startup_keyboard_joystick_keycodes"];
	[preferences setString:unicodes forKey:"startup_keyboard_joystick_unicodes"];
}


static cstr NameStr ( unichar c )
{
	if(c>' '&&c<0x7f) { return CharStr(c); }
	
	switch(c)
	{
	case 0x0020: return "space";
	case 0x000d: return "return";
	case 0x0009: return "tab";
	case 0xf700: return "up";
	case 0xf701: return "down";
	case 0xf702: return "left";
	case 0xf703: return "right";
	case 0xf72c: return "page up";
	case 0xf729: return "start";
	case 0xf72b: return "end";
	case 0xf728: return "del";
	case 0xf72d: return "page down";
	default: return Using("$%04x",(int)c);
	}
}


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



@implementation KeyboardJoystickPanel

-(id)initWithContentRect:(NSRect)bbox styleMask:(uint)style 
			     backing:(NSBackingStoreType)backing defer:(BOOL)flag
{
	XLogIn("KeyboardJoystickPanel:initWithContentRect");
	
	self = keyboard_joystick_panel = 
	[super initWithContentRect:bbox styleMask:style backing:backing defer:flag];
	
	if(self)
	{
		//edit_view = nil;		<-- macht NXPanel
		currentTextField=0;
	}
	
	return self;
}

- (void)awakeFromNib 
{
	XLogIn("KeyboardJoystickPanel:awakeFromNib"); 
	
	[self setDelegate:self];
	ReadPreferences();
}


-(NXTextField*)keyView:(uint)i
{
	XXXASSERT(i<=5);
	static id keyview[5]  = { fireKeyView, upKeyView, downKeyView, leftKeyView, rightKeyView };
	return keyview[i];
}


-(NXTextField*)charView:(uint)i
{
	XXXASSERT(i<=5);
	static id charview[5] = { fireCharView, upCharView, downCharView, leftCharView, rightCharView };
	return charview[i];
}


-(uint)indexOf:(NXTextField*)f
{
	return [f tag];
}


/*	Update Panel: 
	- call from main event loop
	- for regular panel update
	- called only if panel is visible
	- front machine m is locked for reading (RLOCK/RUNLOCK)
*/
-(void)updatePanel:(Machine*)m
{
	static unichar old[5] = { 0xffff, 0xffff, 0xffff, 0xffff, 0xffff };

	for (uint i=0; i<5; i++)
	{
		if (old[i] != new_keycodes[i])
		{
			old[i] = new_keycodes[i];
			[[self keyView:i]  setStringValue:[NSString stringWithFormat:@"$%2x",int(old[i])]];
			[[self charView:i] setStringValue:[NSString stringWithUTF8String:NameStr(new_unicodes[i])]];
		}
	}
}


- (IBAction)doAbort:(id)sender
{
	RestoreOldKeys();
	if(edit_view) [edit_view setBackgroundColor:[NSColor whiteColor]]; 
	edit_view = fireKeyView; currentTextField=0; 
	[edit_view setBackgroundColor:[NSColor greenColor]];
}

- (IBAction)doFinished:(id)sender
{
	AcceptNewKeys();
	if(edit_view) [edit_view setBackgroundColor:[NSColor whiteColor]]; edit_view=nil;
	if( [prefsCheckbox state]==NSOnState ) WritePreferences();
	[self hidePanel];
}

- (IBAction)togglePrefsCheckbox:(id)sender
{
}


// -(void)keyUp:(NSEvent*)event    { [kbd_panel keyUp:event];    }
// -(void)flagsChanged:(NSEvent*)e { [kbd_panel flagsChanged:e]; }


-(void)keyDown:(NSEvent*)event	
{
	if(!edit_view) { [kbd_panel keyDown:event]; return; }

	NSString* chars = [event charactersIgnoringModifiers];	
	
	int i = currentTextField;
	new_keycodes[i] = [event keyCode];
	new_unicodes[i] = [chars length] ? [chars characterAtIndex:0] : 0x0000;

	[edit_view setBackgroundColor:[NSColor whiteColor]];
	currentTextField = (i+1)%5;
	edit_view = [self keyView:currentTextField];
	[edit_view setBackgroundColor:[NSColor greenColor]];
	UseKeys();
}

- (void)windowDidResignKey:(NSNotification *)aNotification
{
	XLogIn("KeyboardJoystickPanel:windowDidResignKey"); 

	if(edit_view) [edit_view setBackgroundColor:[NSColor whiteColor]];
}

- (void)windowDidBecomeKey:(NSNotification *)aNotification
{
	XLogIn("KeyboardJoystickPanel:windowDidBecomeKey"); 

	if(!edit_view) { edit_view = fireKeyView; currentTextField=0; }
	[edit_view setBackgroundColor:[NSColor greenColor]];
}



@end
