/*	Copyright  (c)	Günter Woigk  2000-2004
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

 
	keyboard and keyboard window handling
	-------------------------------------

	01.Jul.1994	kio	Started work Mac Spectacle version
	01.Jan.1999	kio	Started work on X11 version
	18.mar.1999	kio	started work on QT version
	05.apr.1999	kio seems to work fine
	03.jan.2000	kio	started work on MacOS version (back port)
	09.feb.2000	kio	now seems to work fine ...
	11.feb.2002 kio start of port to cocoa
*/

#import "KbdPanel.h"
#import "MenuController.h"
#import "Machine.h"
#import "globals.h"
#import "cstrings/cstrings.h"




/* -------------------------------------------------------------------
			The Tables
------------------------------------------------------------------- */

// define modifier masks
#define	shiftmask	 NSShiftKeyMask
#define	controlmask	 NSControlKeyMask
#define	altmask		 NSAlternateKeyMask
#define	capslockmask 0						// •••keine modifier mask mehr?•••



// ----	handy names for specci keyboard matrix codes -----------------
//		0x00-70		= key's row
//		0x00-04		= key's column
enum 
{	x  	= 0xFF, 	 		// no mapping
	c  	= 0x80,				// add caps shift key
	s  	= 0x08,				// add symbol shift key

	CSH	= 0x00,Z,X,C,V,		/*	left bottom row	*/	cZ1, cZ2, cZ3,	/*	EDIT, CAPS LOCK, TRUE VIDEO		*/
	A  	= 0x10,S,D,F,G,		/* 	...				*/	cZ4, cZ5, cZ6,	/*	INV.VIDEO, CRSR LEFT, CRSR DOWN	*/
	Q  	= 0x20,W,E,R,T,		/*	...				*/	cZ7, cZ8, cZ9,	/*	CRSR UP, CRSR RIGHT, GRAPHICS	*/
	Z1 	= 0x30,Z2,Z3,Z4,Z5,	/* left top row		*/	cZ0, cSPC,		/*	DELETE, BREAK,					*/
	Z0 	= 0x40,Z9,Z8,Z7,Z6,	/* right top row	*/	sO, sP,			/*	; "								*/
	P  	= 0x50,O,I,U,Y,		/* 	...				*/	sN, sM,			/*	, .								*/
	ENT	= 0x60,L,K,J,H,		/*	...				*/	CSH2, SSH2, EXT,/*	2nd CAPS SHIFT, 2nd SYMBOL SHIFT, EXTENDED MODE */
	SPC	= 0x70,SSH,M,N,B,	/* right bottom row	*/	NOKEY = 0x77
};


/* ----	logical translation -----------------------------

		ascii code + modifier -> specci keyboard matrix code
		for text input, e.g. Basic

		logical mapping should be common to all models
		note: control codes are handled by fallback to physical mapping
*/
static uchar	log2spec[256] = 
{	
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x,
	SPC,  Z1+s, P+s, Z3+s, Z4+s, Z5+s, Z6+s, Z7+s,	// 0x20-0x27: space ! " # $ % & '
	Z8+s, Z9+s, B+s, K+s,  N+s,  J+s,  M+s,  V+s,	// 0x28-0x2F: ()*+,-./
	Z0,	  Z1,	Z2,	 Z3,   Z4,	 Z5,   Z6,	 Z7,	// 0x30-0x37: 01234567
	Z8,	  Z9,	Z+s, O+s,  R+s,  L+s,  T+s,  C+s,	// 0x38-0x3F: 89:;<=>?
	Z2+s, A+c,  B+c, C+c,  D+c,  E+c,  F+c,  G+c,	// 0x40-0x47: @ABCDEFG
	H+c,  I+c,  J+c, K+c,  L+c,  M+c,  N+c,  O+c,	// 0x48-0x4F: HIJKLMNO
	P+c,  Q+c,  R+c, S+c,  T+c,  U+c,  V+c,  W+c,	// 0x50-0x57: PQRSTUVW
	X+c,  Y+c,  Z+c, Y,    D,    U,    H+s,  Z0+s,	// 0x58-0x5F: XYZ[\]^_
	x, 	  A,	B,	 C,	   D,	 E,	   F,	 G,		// 0x60-0x67: `abcdefg
	H,	  I,	J,	 K,	   L,	 M,	   N,	 O,		// 0x68-0x6F: hijklmno
	P,	  Q,	R,	 S,	   T,	 U,	   V,	 W,		// 0x70-0x77: pqrstuvw
	X,	  Y,	Z,	 F+s,  S+s,  G+s,  A+s,  Z0+c,	// 0x78-0x7F: xyz{|}~<del>
	Y+s,  x,    x,   x,    x,    D+s,  U+s,  x,		// 0x80-0x87: Ä....ÖÜ.
	x,    x,    F+s, x,    x,    x,    x,    x,		// 0x88-0x8F: ..ä.....
	x,x,x,x,x,x,x,x, 								// 0x90-0x97:
	x,    x,    S+s, x,    x,    x,    x,    G+s,	// 0x98-0x9F: ..ö....ü
	x,	  x,	x,	 X+s,  x,	 x,	   x,	 x,		// 0xA0-0xA7: sterling
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, 
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, 
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x
};


/* ----	physical translation ------------------------------

		mac key code + modifier -> specci keyboard matrix code
		for games

		because we are only interested in physical key translation for games, 
		emulating the new combined-code keys of ZX Spectrum+ and later is not a real issue.
		what might be left for emulation are 'caps lock' and '.'
		
		keypad keys might be left out because they should be handled fine by 
		fallback to logical translation
*/
static uchar	phys2spec[256] =
{
	A,   S,   D,   F,   H,   G,    Z,   X,		// 0x00-0x07: a   s   d   f   h   g   z   x
	C,   V,   x,   B,   Q,   W,    E,   R,		// 0x08-0x0F: c   v   •   b   q   w   e   r
	Y,   T,   Z1,  Z2,  Z3,  Z4,   Z6,  Z5,		// 0x10-0x17: y   t   1   2   3   4   6   5
	x,   Z9,  Z7,  x,   Z8,  Z0,   x,   O,		// 0x18-0x1F: •   9   7   •   8   0(null)   •   O(oh)
	U,   x,   I,   P,   ENT, L,    J,   x,		// 0x20-0x27: u   •   i   p   enter   l   j   •
	K,   ENT, x,   SSH, x,   N,    M,   SPC,	// 0x28-0x2F: k   ';'   •   ','   •   n   m   '.'
	x,   SPC, CSH, 0+c, x,   SSH+c, x,   x,		// 0x30-0x37: •   space  '<'  delete   •   esc   •   •
	CSH, x,   CSH, SSH, CSH, CSH,  SSH,  x,		// 0x38-0x3F: shift  •  alt  ctrl  shift2  alt2  ctrl2  •
	x,   M+s, x,   B+s, x,   K+s,  x,    x,		// 0x40-0x47: •   [.]   •   [*]   •   [+]   •   •
	x,   x,   x,   V+s, ENT, x,    J+s,  x,		// 0x48-0x4F: •   •   •   [/]   [enter]   •   [-]   •
	x,   L+s, Z0,  Z1,  Z2,  Z3,   Z4,   Z5,	// 0x50-0x57: •   [=]   [0]   [1]   [2]   [3]   [4]   [5]
	Z6,  Z7,  x,   Z8,  Z9,  x,    x,    x,		// 0x58-0x5F: [6]   [7]   •   [8]   [9]   •   •   •
	x,x,x,x,x,x,x,x,							// 0x60-0x67:
	x,x,x,x,x,x,x,x, 							// 0x68-0x6F:
	x,x,x,x,x,x,x,x, 							// 0x70-0x77:
	x,   x,   x,   Z5+c, Z8+c, Z6+c, Z7+c, x,	// 0x78-0x7F: •   •   •   left   right   down   up   •
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, 
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, 
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, 
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x
};


#if 0
static char KeycodeFromScancode ( uchar c )
{	uchar* p = (uchar*)memchr ( phys2spec, c, NELEM(phys2spec) );
	return p ? p-phys2spec : 0xff;
}
static char AsciicodeFromScancode ( uchar c )
{	uchar* p = (uchar*)memchr ( log2spec, c, NELEM(log2spec) );
	return p ? p-log2spec : 0xff;
}
inline uchar ScancodeFromAsciicode ( char c )	{ return log2spec[(uchar)c]; }
inline uchar ScancodeFromKeycode ( char c )		{ return phys2spec[(uchar)c]; }
#endif


/* ----	Ersetze Komposittasten --------------
		sucht im Scancode-Buffer nach Komposittasten-Kombinationen
		und ersetzt diese durch die Komposittasten.
		Aus CAPS SHIFT + SPACE wird so BREAK.
*/
static void ComposeComposedKeys ( int model, uchar bytes[8] )
{
	for(int i=0;i<8;i++) { bytes[i] |= 0xe0; }

	switch(model)
	{
	default:		return; // TODO: other models
	case zx16k:
	case zx48i2:
	case zx48i3:	return;	// no composed keys
	
	case zxplus:
	case zx128:
	case zxplus2:
	case zxplus3:
		#define RES_KEY(SPEC)	  bytes[(SPEC)>>4] |= (1<<((SPEC)&7))
		#define SET_KEY(SPEC)	  bytes[(SPEC)>>4] &=~(1<<((SPEC)&7))
		#define GET_KEY(SPEC)	!(bytes[(SPEC)>>4] &  (1<<((SPEC)&7)))
				
		if( GET_KEY(SSH) )	// SYMBOL SHIFT 
		{
			RES_KEY(SSH);	// Ann.: Komposittaste
			if( GET_KEY(CSH) )	{ RES_KEY(CSH); SET_KEY(EXT);  } else	//	SSH+CSH = EXTENDED MODE
			if( GET_KEY(O) )	{ RES_KEY(O);   SET_KEY(sO);   } else	//	SSH+O	= ;
			if( GET_KEY(P) )	{ RES_KEY(P);   SET_KEY(sP);   } else	//	SSH+P	= "
			if( GET_KEY(N) )	{ RES_KEY(N);   SET_KEY(sN);   } else	//	SSH+N	= ,
			if( GET_KEY(M) )	{ RES_KEY(M);   SET_KEY(sM);   } else	//	SSH+M	= .
								{ SET_KEY(SSH); SET_KEY(SSH2); }			//	doch keine Komposittaste
			return;
		}

		if( GET_KEY(CSH) )	// CAPS SHIFT 
		{
			RES_KEY(CSH);	// Ann.: Komposittaste
			if( GET_KEY(SPC) )		{ RES_KEY(SPC); SET_KEY(cSPC); }		//	CSH+SPACE = BREAK
			else if(bytes[3]!=0xff)
			{
				if( GET_KEY(Z1) )	{ RES_KEY(Z1);  SET_KEY(cZ1);  } else	//	CSH+1	= EDIT
				if( GET_KEY(Z2) )	{ RES_KEY(Z2);  SET_KEY(cZ2);  } else	//	CSH+2	= CAPS LOCK
				if( GET_KEY(Z3) )	{ RES_KEY(Z3);  SET_KEY(cZ3);  } else	//	CSH+3	= TRUE VIDEO
				if( GET_KEY(Z4) )	{ RES_KEY(Z4);  SET_KEY(cZ4);  } else	//	CSH+4	= INV. VIDEO
									{ RES_KEY(Z5);  SET_KEY(cZ5);  }		//	CSH+5	= CRSR LEFT
			}
			else if(bytes[4]!=0xff)
			{
				if( GET_KEY(Z6) )	{ RES_KEY(Z6);  SET_KEY(cZ6);  } else	//	CSH+6	= CRSR DOWN
				if( GET_KEY(Z7) )	{ RES_KEY(Z7);  SET_KEY(cZ7);  } else	//	CSH+7	= CRSR UP
				if( GET_KEY(Z8) )	{ RES_KEY(Z8);  SET_KEY(cZ8);  } else	//	CSH+8	= CRSR RIGHT
				if( GET_KEY(Z9) )	{ RES_KEY(Z9);  SET_KEY(cZ9);  } else	//	CSH+9	= GRAPHICS
									{ RES_KEY(Z0);  SET_KEY(cZ0);  }		//	CSH+0	= DELETE
			}														 
			else					{ SET_KEY(CSH); SET_KEY(CSH2); }		//	doch keine Komposittaste
		}
	}
}
	


/* ---- Berechne Tasten-Boxen für Taste gem. Scancode ----		
		Komposittasten müssen mit ihrem Komposittasten-Pseudoscancode übergeben werden.
		Beachte: Die ENTER-Taste hat manchmal L-Form. Dann wird nur die untere Hälfte zurückgegeben.
		Das Layout von plus/128 und +2/+3 unterscheidet sich nur durch die Breite von EXTENDED MODE und EDIT.
*/
static NSRect KeyRect ( int model, uchar spec )
{
	spec &= 0x77;
	int x,y,w,h;				// result box

	switch(model)
	{
	default:	x=y=w=h=0; goto x;	// TODO: other models

	case zx16k:
	case zx48i2:
	case zx48i3:			
			y = spec>>4;		// 0..7
			x = spec&7;			// 0..4
			if (x>=5) { x=y=w=h=0; goto x; }
			if (y>=4) 
			{	y = 7-y;		// 0..3
				x = 9-x;		// 0..9
			}
			static int x0[]={20,45,33,8};	// x-Wert für 1. Taste einer Reihe (bei CSH: normalisiert auf Std-Tastenbreite)
			x = x0[y]+50*x;
			y = 9+49*y;
			w = 37;
			h = 27;
			if(spec==CSH) { x=8; w=49; }	// CAPS SHIFT ist breiter
			if(spec==SPC) { w=62; }			// ENTER ist breiter
			goto x;
	
	case zxplus:
	case zx128:
	case zxplus2:
	case zxplus3:
		x=0;y=0;w=41;h=41;			// most common values for x, y, and width

	// Komposit- und Sondertasten
		switch(spec)				
		{
		case EXT:/*EXTEND MODE*/	y=80; w=model<zxplus2?61:71; goto x;
		case sO: /* ; */			x=40;				goto x;
		case sP: /* " */            x=80;				goto x;
		case sN: /* , */            x=460;				goto x;
		case sM: /* . */			x=410; y=40;		goto x;
		case cZ1: /*EDIT*/			x=model<zxplus2?60:70; y=80; w=model<zxplus2?51:41;	goto x;
		case cZ2: /*CAPS LOCK*/		x=90; y=40;			goto x;
		case cZ3: /*TRUE VIDEO*/		  y=160;		goto x;
		case cZ4: /*INV VIDEO*/		x=40; y=160;		goto x;
		case cZ5: /*CURSOR LEFT*/	x=120;				goto x;
		case cZ6: /*CURSOR DOWN*/	x=420;				goto x;
		case cZ7: /*CURSOR UP*/		x=380;				goto x;
		case cZ8: /*CURSOR RIGHT*/	x=160;				goto x;
		case cZ9: /*GRAPHICS*/		x=60; y=80;			goto x;
		case cZ0: /*DELETE*/			  y=120; w=61;	goto x;
		case cSPC:/*BREAK*/			x=480;y=160; w=61;	goto x;
		case SSH2:/*2. SSH Taste*/						goto x;
		case CSH2:/*2. CSH Taste*/	      y=40;  w=91;	goto x;
		case CSH:					x=450; y=40; w=91;  goto x;
		case SSH:					x=500;              goto x; 
		case ENT:					x=470; y=80; w=81;  goto x;	// L-förmig!
		case SPC:					x=200;       w=181; goto x;
		}

	// normale Taste
		y = spec>>4;		// 0..7
		x = spec&7;			// 0..4
		if (x>=5) { x=y=w=h=0; goto x; }
		if (y>=4) 
		{	y = 7-y;		// 0..3
			x = 9-x;		// 0..9
		}
		{
			static int x0[]={90,110,100,80};	// x-wert für 1. taste einer reihe (bei CSH-Taste: extrapolierte Position)
			x = x0[y]+40*x; 
			y = 40 + 40*y;
		}
		break;
	}
	
x:	//printf("box($%02x) = (%i,%i),(%i,%i)\n",spec,x,y,w,h);
	NSRect box = { {x,y}, {w,h} };
	return box;
}




@implementation KbdView : NSView

/* ----	replacement function for drawing --------------------
		draw whatever belongs into this box
*/
- (void)drawRect:(NSRect)box
{
	puts("KbdView:drawRect");
	CGContext* context = (CGContext*)[[NSGraphicsContext currentContext] graphicsPort];		// Quartz-2D drawing context

	CGContextSetRGBFillColor (context, 1, 0, 0, 1); 
    CGContextFillRect (context, CGRectMake(0, 0, 200, 100 ));
	
	cstr filename = CatStr(rsrc_path,model_info[model].kbd_filename);
	CFStringRef filepath = CFStringCreateWithCString ( NULL/*alloc*/, filename, kCFStringEncodingUTF8 );
	CFURLRef fileurl = CFURLCreateWithFileSystemPath ( NULL/*alloc*/, filepath, kCFURLPOSIXPathStyle, no/*isDirectory*/ );
	CGDataProviderRef dataprovider = CGDataProviderCreateWithURL ( fileurl );
	CGImageRef image = CGImageCreateWithPNGDataProvider ( dataprovider, NULL/*decode[]*/, no/*shouldInterpolate*/, kCGRenderingIntentDefault );

	CGRect rect = { {0,0}, {541,201} };
	CGContextDrawImage ( context, rect, image );

	CGContextSetGrayFillColor ( context, 0.0/*grey=black*/, 0.5/*alpha*/ );

	for(int i=0;i<8;i++)	// 8 scanbytes
	{
		uchar byte = bytes[i];
		if(byte==0xff) continue;	// no key down
		for( int j=0; j<8; j++ )	// 5 std bits + 3 bits used for composed keys
		{
			if( (byte>>j)&1 ) continue;							// this key is not down
			NSRect rect = KeyRect(model,(i<<4)+j);				// NOTE: NSRect == CGRect
			CGContextFillRect ( context, *(CGRect*)&rect );				
		}
	}
	
	CGImageRelease ( image );
	CGDataProviderRelease ( dataprovider );
	CFRelease(filepath);
	CFRelease(fileurl);
}


/* ---- init ----------------------
*/
-(id)initWithModel:(uint)current_model 
{
	NSRect box = {{0,0},{541,201}};
	self = [super initWithFrame:box];
	if(self)
	{
		model = current_model;
		memset(bytes,0xff,8);	// all up
		mouseKey = NOKEY;
	}
	return self;
}


/* ----	update key matrix state --------------------
*/
-(void)updateBytes:(uchar*)newbytes
{
	uchar zbytes[8];
	memcpy(zbytes,newbytes,8);
	ComposeComposedKeys(model,zbytes);
	
	for (int i=0;i<8;i++)
	{
		uchar c = zbytes[i] ^ bytes[i]; 
		if(c) for(int j=0;j<8;j++)
		{
			if( (c>>j)&1 ) [self setNeedsDisplayInRect:KeyRect(model,(i<<4)+j)];		
		}
	}
	
	memcpy(bytes,zbytes,8);
}

-(uchar)findKeyForPoint:(NSPoint)where
{
	for( uchar key=0; key<0x80; key++ ) 
	{
		NSRect box = KeyRect(model,key);
		if( [self mouse:where inRect:box] ) return key;
	}
	return NOKEY;
}


/* ----	NSResponder: event handling ------------------------------
		Handling Mouse down, Mouse drag and Mouse up
*/
-(BOOL)acceptsFirstMouse:(NSEvent*)event
{
	return yes;
}

-(void)mouseDown:(NSEvent*)event
{
	puts("KbdView:MouseDown:");
	NSPoint where = [event locationInWindow];			// window coordinates
	where = [self convertPoint:where fromView:nil];		// nil == my window
	mouseKey = [self findKeyForPoint:where];
	[kbd_panel setMouseKey:mouseKey];	
}

-(void)mouseDragged:(NSEvent*)event
{
	puts("KbdView:MouseDrag:");
	NSPoint where = [event locationInWindow];			// window coordinates
	where = [self convertPoint:where fromView:nil];		// nil == my window
	if( mouseKey!=NOKEY && [self mouse:where inRect:KeyRect(model,mouseKey)] ) return;	// still over same key

// mouse left previous key
	mouseKey = [self findKeyForPoint:where];
	[kbd_panel setMouseKey:mouseKey];	
}

-(void)mouseUp:(NSEvent*)event
{
	puts("KbdView:MouseUp:");	
	mouseKey=NOKEY;
	[kbd_panel setMouseKey:mouseKey];
}



@end







// =================================================
//
//				IMPLEMENTATION
//
// =================================================

@implementation KbdPanel


- (id)initWithContentRect:(NSRect)contentRect styleMask:(uint)styleMask backing:(NSBackingStoreType)backingType defer:(BOOL)flag
{
	kbd_panel = self = 
	[super initWithContentRect:contentRect styleMask:styleMask backing:backingType defer:flag];
	
	if(self)
	{
	//	kbd_image_view = nil;
	//	kbd_image = nil;
		kbd_view = nil;
		mode = kbdlog;		// logisch ~ ascii code
		stickyshifts = 0;	// use sticky shift keys?
		memset(jkey,0xff,sizeof(jkey));
		mouseKey = NOKEY;
	}
	
	return self;
}

-(void)close
{
	[menu_controller panelClosed:self];
	[super close];
}

/*
- (void)windowWillClose:(NSNotification *)aNotification
{
	[menu_controller toggleKeyboardWindow:self];
}
*/


// ----	initialize ----
//		init the one&only keyboard controller
- (void)awakeFromNib
{
//	SetupTables();
}


// ----	set keyboard for model ----
-(void)updateModel
{
	Machine* machine = specci[0];

	model = [machine getModel];
	TRAP(model==0xffff);
	TRAP([machine getModelInfo].name==nil);
	memset(bytes,0xff,sizeof(bytes));
	caps_cnt = 0;
	symb_cnt = 0;
	capskeysticky = no;
	symbolkeysticky = no;
	capskeydown = no;
	symbolkeydown = no;
	modifiers = 0;

#if 0
	cstr filename = CatStr(rsrc_path,[machine getModelInfo].kbd_filename);
	OMEM( kbd_image = [[NSImage alloc] initByReferencingFile:[NSString stringWithUTF8String:filename]] );
	OMEM( kbd_image_view = [[NSImageView alloc] init] );
	[kbd_image_view setImage:kbd_image];
	[kbd_image_view setEditable:no];
	[kbd_image_view setImageAlignment:NSImageAlignTopLeft];
	[kbd_image_view setImageFrameStyle:NSImageFrameNone];
	[kbd_image_view setImageScaling:NSScaleNone];
#else
	OMEM( kbd_view = [[KbdView alloc] initWithModel:model] );
#endif
	[self setContentView:kbd_view];
	[self setTitle:[NSString stringWithUTF8String:[machine getModelInfo].name]];
}


-(void)setBasicMode
{
	if(mode==kbdlog) return;
	mode = kbdlog;		// logisch ~ ascii code
	[menu_controller setKeyboardBasicMode:self];
}

-(void)setGameMode
{
	if(mode==kbdphys) return;
	mode = kbdphys;		// physical ~ scan code
	[menu_controller setKeyboardGameMode:self];
}


/* --------------------------------------------------------------
	process a keydown or keyup event

	kbdType=kbdPhysical	=> physical mapping is preferred (for games)
	kbdType=kbdLogical	=> logical  mapping is preferred (for basic)

	the table of the preferred method is tried first. if it does not
	contain an entry for the pressed/released key, the other table is tried.
	
	"alt" works as auxiliary "caps shift" keys in both mapping modes.
	they are essential in logical mapping mode to get the "caps shift" + "number key",
	e.g. "caps shift" + "1" to edit a basic line. if you simply press "shift" plus "1"
	this is "!" and this is decoded as "symbol shift" + "1" which results in a "!"
	on the specci. on the other hand, "alt" + "1" works. useful for breaking into 
	running basic programs too, since shift + space results in a simple space too.

	the state of control key is taken from ev->state in both modes, since it does
	not influence the character code returned by XLookupString().

	the state of shift and shift lock are taken from xev->state only in physical 
	mapping mode. they are not decoded separately in logical mode, since the key 
	translation table supplies the corresponding shift and control key state and an 
	additional modifier detected will disturb. e.g. if you press "!" (that is: "shift" + "1") 
	this results in "symbol shift" + "1" which is transposed to "!" by the specci
	keyboard routines just as expected. adding "caps shift" because "shift" is 
	pressed on the real keyboard (just to get the "!") will result in "1" and both
	shift keys pressed to the specci, decoding to nothing (3 keys pressed!)

---------------------------------------------------------------*/

-(void)handleKeycode:(uchar)keycode charCode:(uchar)charcode state:(int)state down:(BOOL)down
{	uchar kspec,lspec;

/*
// test for joystick button
// joystick is only emulated in physical (game) mode
// if a joystick key is detected, ignore anything else
	if (mode==kbdphys) for(int i=0;i<10;i++) 
	{
		if (keycode==jkey[i]) 
		{	
			SendSignal(joy1_up+i,down);
			goto x;
		}
	}
*/

// store modifier state for keyboard mouse down routine
	modifiers = state;
	
// release sticky shift keys
// sticky keys are released always when any key goes up
	if (!down)	// if (up)
	{
		if (capskeysticky)
		{	capskeysticky = false;		// clear stickyness
			caps_cnt -= !capskeydown;	// release if caps key in kbd pict currently not pressed with mouse
		}
		if (symbolkeysticky)
		{	symbolkeysticky = false;
			symb_cnt -= !symbolkeydown;
		}
	}

// get physical and logical mapping
	kspec = phys2spec[keycode];	// physical
	lspec = log2spec[charcode];	// logical
	
// select mapping
	if ((mode!=kbdphys&&lspec!=x)||kspec==x)
	{
		if (lspec==x)
		{
			printf("•unknown key• key=$%02X, ascii=$%02X\n", (int)keycode, (int)charcode );
			lspec = NOKEY;
		}
		kspec = lspec;
	}
	assert((kspec&7)<5||kspec==NOKEY);

// count modifier
// count real keyboard modifier keys and keyboard image modifier keys
	if ((kspec&0x77)==CSH)
	{
		if (down) caps_cnt++;
		else      caps_cnt--;	assert(caps_cnt>=0);
	}
	if ((kspec&0x77)==SSH)
	{
		if (down) symb_cnt++;
		else      symb_cnt--;	assert(symb_cnt>=0);
	}
	
// remove built-in modifiers if the key is going up
	if (!down) kspec &= 0x77;

// apply modifiers to kspec if it has no built-in modifiers 
	if ( !(kspec&0x88) )
	{
		if ( symb_cnt || (down&&(state&   controlmask     )) )	kspec |= s;
		if ( caps_cnt || (down&&(state&(altmask|shiftmask))) )	kspec |= c;
	}
	
	{
		uchar z; int i,m;

	// store symbol shift key state:
		if (kspec&s)	{ z = bytes[7] & ~0x02; /* +symbol */ }
		else			{ z = bytes[7] |  0x02; /* -symbol */ }
		if (z!=bytes[7]) { bytes[7]=z; [specci[0] getKbdData][7] = z; }
		
	// store caps shift key state:
		if (kspec&c)	{ z = bytes[0] & ~0x01; /* +caps */ }
		else			{ z = bytes[0] |  0x01; /* -caps */ }
		if (z!=bytes[0]) { bytes[0]=z; [specci[0] getKbdData][0] = z; }
	
	// store the key itself:
		if((kspec&0x77)!=NOKEY)
		{
			i = (kspec>>4)&7;				// affected io byte
			m = 1<<(kspec&7);				// affected bit mask
			if (down) bytes[i] &= ~m; else bytes[i] |= m;
			assert((kspec&7)<5);
			[specci[0] getKbdData][i] = bytes[i];
		}
	}
	
// update keyboard window
	if( [self isVisible] ) [kbd_view updateBytes:bytes];
}



// ----	replacement handler ----
//		gets all key events to the machine!
//
-(void)keyDown:(NSEvent*)event
{
	NSString* chars = [event characters];
	uchar charcode  = [chars characterAtIndex:0];	// •••unichar•••
	uchar keycode   = [event keyCode];				// •••keyCode ist ushort, nicht uchar!•••
	uint  state     = [event modifierFlags];
	
	[self handleKeycode:keycode charCode:charcode state:state down:yes];
}


// ----	replacement handler ----
//		gets all key events to the machine!
//
-(void)keyUp:(NSEvent*)event
{
	NSString* chars = [event characters];
	uchar charcode  = [chars characterAtIndex:0];	// •••unichar•••
	uchar keycode   = [event keyCode];				// •••keyCode ist ushort, nicht uchar!•••
	uint  state     = [event modifierFlags];
	
	[self handleKeycode:keycode charCode:charcode state:state down:no];
}


// ----	replacement handler ----
//		gets all key events to the machine!
//
-(void)flagsChanged:(NSEvent*)event
{
//	if( mode==kbdlog ) return;		// exaktes Tracken der Shifttasten nur bei phys. Übersetzung sinnvoll ?
	
	uchar charcode  = x;
	uchar keycode	= 0xff;
	uint  state     = [event modifierFlags];
	
	[self handleKeycode:keycode charCode:charcode state:state down:yes/*damit modifier keys ausgewertet werden*/];
}


-(void)setMouseKey:(uchar)theKey
{
	if(theKey==mouseKey) return;
	printf("KbdPanel:setMouseKey:$%02x\n",(int)theKey);

	if(mouseKey!=NOKEY)	// => release
	{
		mouseKey = NOKEY;
	}

	if(theKey!=NOKEY)	// => depress
	{
		mouseKey = theKey;
	}
}





@end













