/*	Copyright  (c)	Günter Woigk  2000-2006
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#import <Cocoa/Cocoa.h>
#import "NXPanel.h"


// preferred keyboard translation:
enum 
{	kbdphys,				// prefer physical translation (from scan code)
	kbdlog					// use logical translation (from character code)
};


// Modifier key masks for realworld keyboard: 
#define	capsmask			NSShiftKeyMask			// realworld shift key
#define	controlmask			NSControlKeyMask		// realworld control key
#define	metamask			NSAlternateKeyMask		// realworld alt/option key: used as an auxilliary caps shift key
#define	capslockmask		NSAlphaShiftKeyMask		// realworld caps lock key

// Modifier key masks for sticky keys on visual keyboard:
#define capsshiftsticky		0x0001					// caps shift key no longer pressed but still considered down
#define capsshift2sticky	0x0002					// for models with 2 caps shift keys
#define	symbolshiftsticky	0x0004					// symbol shift key no longer pressed but still considered down
#define	symbolshift2sticky	0x0008					// for models with 2 caps shift keys


/*	Keyboard view:
	This is used as content view of a KbdPanel
	Displays a picture of the keyboard of the selected model
	Displays hilight blobs over pressed keys as told via updateKeymap
	Handles mouseDown, mouseUp and mouseDragged events:
		determines pressed or released key and reports it to window:setMouseKey
*/
@interface KbdView : NSView
{
	uint		model;
	uchar		keymap[8];				// keyboard map representing pressed (hilighted) keys. combining keys are stored at bits 5..7	
	uchar		mousekey;				// key currently depressed by mouse
}
-(id)initWithModel:(uint)model;
-(void)updateKeymap:(uchar*)bytes;
-(uchar)findKeyForPoint:(NSPoint)point;	
-(uchar*)getKeymap;
-(uchar&)getMousekey;
@end



/*	Keyboard panel:
	This panel shows a picture of the current keyboard.
	It shows the pressed keys and handles mouse clicks on the virtual keys.
	There is only one panel, accessible via global kbd_panel, serving the frontmost machine specci_0.
	This panel handles all key presses, even if not displayed.
	The panel can be set to prefer physical (Games) or logical (Basic) translation of realworld key presses.
	The current keymap is updated to the frontmost machine specci_0.
*/
@interface KbdPanel : NXPanel
{
// settings:
	uint		model;					// emulated specci model
	uint		mode;					// prefer physical or logical key translation ?
	uchar		keymap[8];				// keyboard matrix as seen by ula

// visual keyboard picture:
	KbdView*	kbd_view;

// state:
	int			pressed_keys;
	uchar		os_keycodes[16];		// stored values = ~os_keycode
	uchar		zx_keycodes[16];
	uint		modifiers;				// modifiers down on realworld keyboard
	uchar		stickykeys;				// modifier keys on visual kbd_view sticky down
}

// internal use:
-(void)specciKeyUp:(uchar)speccikey;
-(void)specciKeyDown:(uchar)speccikey;
-(void)updateKeymap;


-(void)allKeysUp;
-(uint)getPressedKeysCount;
-(uchar*)getPressedOSKeycodes;
-(uchar*)getKeymap;
-(uint) getMode;

@end


extern KbdPanel* kbd_panel;


// ==========================================================
// Joystick emulation with keyboard

extern void	 SetKeyboardJoystickKeycodes ( unichar cc[5] );
extern uchar KeyboardJoystickGetState	 ( );






