/*	Copyright  (c)	Günter Woigk 1994 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		1

#import "cstrings/cstrings.h"
#import "unix/file_utilities.h"
#import "KbdPanel.h"
#import "Machine.h"
#import "globals.h"
#import "ZxInfo.h"
#import "Preferences.h"
#import "utilities.h"
#import "Joystick.h"
#import "Mouse.h"


/* ===============================================
		Joystick emulation with Keyboard 
*/

static uchar keyboard_joystick_keycodes[5];		// FUDLR

void SetKeyboardJoystickKeycodes ( unichar cc[5] ) 
{
	for(uint i=0;i<5;i++) keyboard_joystick_keycodes[i] = cc[i]; 
}

/* ---- Get state of joystick buttons ----
		determine joystick buttons state
		for the frontmost machine
*/
uchar KeyboardJoystickGetState ( )
{
	XXLogIn("KeyboardJoystickGetState");

	uint   n    = [kbd_panel getPressedKeysCount];	
	uchar* cc   = [kbd_panel getPressedOSKeycodes];
	uchar  rval = 0;
	uchar* kk   = keyboard_joystick_keycodes;		// FUDLR

	while (n>0)
	{
		uchar c = ~cc[--n];
		XXLog("[$%2x]",int(c));
		if(c==kk[0]) rval |= joystick_button1_mask; else
		if(c==kk[1]) rval |= joystick_paddle_up_mask; else
		if(c==kk[2]) rval |= joystick_paddle_down_mask; else
		if(c==kk[3]) rval |= joystick_paddle_left_mask; else
		if(c==kk[4]) rval |= joystick_paddle_right_mask; 
	}

	return rval;
}







/* -------------------------------------------------------------------
			The Tables
------------------------------------------------------------------- */




// ----	handy names for specci keyboard matrix codes -----------------
//		0x00-70		= key's row
//		0x00-04		= key's column
enum
{	x  	= 0x77,				// no mapping
	c  	= 0x80,				// add caps shift key
	s  	= 0x08,				// add symbol shift key

	CSH	= 0x00,Z,X,C,V,		/*	left bottom row	*/	cZ1, cZ2, cZ3,	/*	EDIT, CAPS LOCK, TRUE VIDEO		*/
	A  	= 0x10,S,D,F,G,		/* 	...				*/	cZ4, cZ5, cZ6,	/*	INV.VIDEO, CRSR LEFT, CRSR DOWN	*/
	Q  	= 0x20,W,E,R,T,		/*	...				*/	cZ7, cZ8, cZ9,	/*	CRSR UP, CRSR RIGHT, GRAPHICS	*/
	Z1 	= 0x30,Z2,Z3,Z4,Z5,	/* left top row		*/	cZ0, cSPC,		/*	DELETE, BREAK,					*/
	Z0 	= 0x40,Z9,Z8,Z7,Z6,	/* right top row	*/	sO, sP,			/*	; "								*/
	P  	= 0x50,O,I,U,Y,		/* 	...				*/	sN, sM,			/*	, .								*/
	ENT	= 0x60,L,K,J,H,		/*	...				*/	CSH2, SSH2, EXT,/*	2nd CAPS SHIFT, 2nd SYMBOL SHIFT, EXTENDED MODE */
	SPC	= 0x70,SSH,M,N,B,	/* right bottom row	*/	NOKEY = 0x77,
	
	DOT = SSH,				// "." on ZX80/ZX81 keyboard
};

uchar const decomposed_zxsp[64] =					/*	decompose composed key:			*/
{	c+x,Z,X,C,V,				c+Z1, c+Z2, c+Z3,	/*	EDIT, CAPS LOCK, TRUE VIDEO		*/
	A  ,S,D,F,G,				c+Z4, c+Z5, c+Z6,	/*	INV.VIDEO, CRSR LEFT, CRSR DOWN	*/
	Q  ,W,E,R,T,				c+Z7, c+Z8, c+Z9,	/*	CRSR UP, CRSR RIGHT, GRAPHICS	*/
	Z1 ,Z2,Z3,Z4,Z5,			c+Z0, c+SPC,x,		/*	DELETE, BREAK,					*/
	Z0 ,Z9,Z8,Z7,Z6,			s+O,  s+P,  x,		/*	; "								*/
	P  ,O,I,U,Y,				s+N,  s+M,  x,		/*	, .								*/
	ENT,L,K,J,H,				c+x,  s+x,  c+s+x,	/*	2nd CAPS SHIFT, 2nd SYMBOL SHIFT, EXTENDED MODE */
	SPC,s+x,M,N,B,				x,	  x,	x		/*  -x- -x- NOKEY					*/
};

#if 0
uchar const composed_with_c_zx128[64] =				/*	composed key with caps shift:	*/
{	CSH,c+Z,c+X,c+C,c+V,		c+Z1, c+Z2, c+Z3,	/*	EDIT, CAPS LOCK, TRUE VIDEO		*/
	c+A,c+S,c+D,c+F,c+G,		c+Z4, c+Z5, c+Z6,	/*	INV.VIDEO, CRSR LEFT, CRSR DOWN	*/
	c+Q,c+W,c+E,c+R,c+T,		c+Z7, c+Z8, c+Z9,	/*	CRSR UP, CRSR RIGHT, GRAPHICS	*/
	cZ1,cZ2,cZ3,cZ4,cZ5,		c+Z0, c+SPC,CSH,	/*	DELETE, BREAK,					*/
	cZ0,cZ9,cZ8,cZ7,cZ6,		c+sO, c+sP, CSH,	/*	; "								*/
	c+P,c+O,c+I,c+U,c+Y,		c+sN, c+sM, CSH,	/*	, .								*/	
	c+ENT,c+L,c+K,c+J,c+H,		CSH2, EXT,  EXT,	/*	2nd CAPS SHIFT, 2nd SYMBOL SHIFT, EXTENDED MODE */	
	cSPC,EXT,c+M,c+N,c+B,		CSH,  CSH,  CSH		/*  -x- -x- NOKEY					*/
};
#endif

#if 0
uchar const composed_with_s_zx128[64] =				/* composed key with symbol shift:	*/
{	EXT,s+Z,s+X,s+C,s+V,		s+cZ1,s+cZ2,s+cZ3,	/*	EDIT, CAPS LOCK, TRUE VIDEO		*/
	s+A,s+S,s+D,s+F,s+G,		s+cZ4,s+cZ5,s+cZ6,	/*	INV.VIDEO, CRSR LEFT, CRSR DOWN	*/
	s+Q,s+W,s+E,s+R,s+T,		s+cZ7,s+cZ8,s+cZ9,	/*	CRSR UP, CRSR RIGHT, GRAPHICS	*/
	s+Z1,s+Z2,s+Z3,s+Z4,s+Z5,	s+cZ0,s+cSPC,SSH,	/*	DELETE, BREAK,					*/
	s+Z0,s+Z9,s+Z8,s+Z7,s+Z6,	sO,   sP,    SSH,	/*	; "								*/
	sP,sO,s+I,s+U,s+Y,			sN,   sM,    SSH,	/*	, .								*/
	s+ENT,s+L,s+K,s+J,s+H,		EXT,  SSH2,  EXT,	/*	2nd CAPS SHIFT, 2nd SYMBOL SHIFT, EXTENDED MODE */
	s+SPC,SSH,sM,sN,s+B,		SSH,  SSH,   SSH	/*  -x- -x- NOKEY					*/
};
#endif


/* ----	logical translation -----------------------------

		ascii code + modifier -> specci keyboard matrix code
		for text input, e.g. Basic

		logical mapping should be common to all models
		note: control codes are handled by fallback to physical mapping
*/
typedef uchar logtab[0x80];
logtab const log2spec_zxsp =						// TODO: ÖÄÜöäü GBP and (C)
{	
//	composed keycodes where possible.
//	also for Jupiter Ace: => []|{}~ are mapped to Symbol Shift + Key
//		on Specci use 'ESC' before these keys to enter extended keyboard mode 
//	there is also a difference in position of '©' (copyright) 
//		but this is no ASCII code and therefore not in this table (as '£' (GBP)).
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x,
	SPC,  Z1+s, sP,  Z3+s, Z4+s, Z5+s, Z6+s, Z7+s,	// 0x20-0x27: space ! " # $ % & '
	Z8+s, Z9+s, B+s, K+s,  sN,   J+s,  sM,   V+s,	// 0x28-0x2F: ()*+,-./
	Z0,	  Z1,	Z2,	 Z3,   Z4,	 Z5,   Z6,	 Z7,	// 0x30-0x37: 01234567
	Z8,	  Z9,	Z+s, sO,   R+s,  L+s,  T+s,  C+s,	// 0x38-0x3F: 89:;<=>?
	Z2+s, A+c,  B+c, C+c,  D+c,  E+c,  F+c,  G+c,	// 0x40-0x47: @ABCDEFG
	H+c,  I+c,  J+c, K+c,  L+c,  M+c,  N+c,  O+c,	// 0x48-0x4F: HIJKLMNO
	P+c,  Q+c,  R+c, S+c,  T+c,  U+c,  V+c,  W+c,	// 0x50-0x57: PQRSTUVW
	X+c,  Y+c,  Z+c, Y+s,  D+s,  U+s,  H+s,  Z0+s,	// 0x58-0x5F: XYZ[\]^_
	x, 	  A,	B,	 C,	   D,	 E,	   F,	 G,		// 0x60-0x67: `abcdefg
	H,	  I,	J,	 K,	   L,	 M,	   N,	 O,		// 0x68-0x6F: hijklmno
	P,	  Q,	R,	 S,	   T,	 U,	   V,	 W,		// 0x70-0x77: pqrstuvw
	X,	  Y,	Z,	 F+s,  S+s,  G+s,  A+s,  cZ0,	// 0x78-0x7F: xyz{|}~<del>
};													// table must be filled to end with 'x'

#if 0
logtab const log2spec_jupiter =
{	
// identical to ZX Spectrum, except: 
//		[]|{} actually accessed with Symbol Shift and 
//		no composed keys:
//	no composed keycodes: => no need for decomposed_jupiter[] table
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x,
	SPC,  Z1+s, P+s, Z3+s, Z4+s, Z5+s, Z6+s, Z7+s,	// 0x20-0x27: space ! " # $ % & '
	Z8+s, Z9+s, B+s, K+s,  N+s,  J+s,  M+s,  V+s,	// 0x28-0x2F: ()*+,-./
	Z0,	  Z1,	Z2,	 Z3,   Z4,	 Z5,   Z6,	 Z7,	// 0x30-0x37: 01234567
	Z8,	  Z9,	Z+s, O+s,  R+s,  L+s,  T+s,  C+s,	// 0x38-0x3F: 89:;<=>?
	Z2+s, A+c,  B+c, C+c,  D+c,  E+c,  F+c,  G+c,	// 0x40-0x47: @ABCDEFG
	H+c,  I+c,  J+c, K+c,  L+c,  M+c,  N+c,  O+c,	// 0x48-0x4F: HIJKLMNO
	P+c,  Q+c,  R+c, S+c,  T+c,  U+c,  V+c,  W+c,	// 0x50-0x57: PQRSTUVW
	X+c,  Y+c,  Z+c, Y+s,  D+s,  U+s,  H+s,  Z0+s,	// 0x58-0x5F: XYZ[\]^_
	x, 	  A,	B,	 C,	   D,	 E,	   F,	 G,		// 0x60-0x67: `abcdefg
	H,	  I,	J,	 K,	   L,	 M,	   N,	 O,		// 0x68-0x6F: hijklmno
	P,	  Q,	R,	 S,	   T,	 U,	   V,	 W,		// 0x70-0x77: pqrstuvw
	X,	  Y,	Z,	 F+s,  S+s,  G+s,  A+s,  Z0+c,	// 0x78-0x7F: xyz{|}~<del>
};													// table must be filled to end with 'x'
#endif

logtab const log2spec_zx80 =
{	
//	no composed keycodes: => no need for decomposed_zx80[] table
//  CSH = SHIFT = c
//	SSH = DOT
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x,
	SPC,  x,    Y+c, x,    c+U,  x,    x,    x,  	// 0x20-0x27: space ! " # $ % & '	
	I+c,  O+c,  P+c, K+c,  DOT+c,J+c,  DOT,  V+c,	// 0x28-0x2F: ()*+,-./
	Z0,	  Z1,	Z2,	 Z3,   Z4,	 Z5,   Z6,	 Z7,	// 0x30-0x37: 01234567
	Z8,	  Z9,	Z+c, X+c,  N+c,  L+c,  M+c,  C+c,	// 0x38-0x3F: 89:;<=>?
	x,    A,    B,   C,    D,    E,    F,    G,	    // 0x40-0x47: @ABCDEFG
	H,    I,    J,   K,    L,    M,    N,    O,  	// 0x48-0x4F: HIJKLMNO
	P,    Q,    R,   S,    T,    U,    V,    W,  	// 0x50-0x57: PQRSTUVW
	X,    Y,    Z,   x,    x,    x,    x,    x,   	// 0x58-0x5F: XYZ[\]^_
	x, 	  A,	B,	 C,	   D,	 E,	   F,	 G,		// 0x60-0x67: `abcdefg
	H,	  I,	J,	 K,	   L,	 M,	   N,	 O,		// 0x68-0x6F: hijklmno
	P,	  Q,	R,	 S,	   T,	 U,	   V,	 W,		// 0x70-0x77: pqrstuvw
	X,	  Y,	Z,	 x,    x,    x,    x,    Z0+c,	// 0x78-0x7F: xyz{|}~<del>
};													// table must be filled to end with 'x'


logtab const log2spec_zx81 =
{	
//	no composed keycodes: => no need for decomposed_zx80[] table
//  CSH = SHIFT = c
//	SSH = DOT
	x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x, x,x,x,x,x,x,x,x,
	SPC,  x,    P+c, x,    c+U,  x,    x,    x,  	// 0x20-0x27: space ! " # $ % & '	
	I+c,  O+c,  B+c, K+c,  DOT+c,J+c,  DOT,  V+c,	// 0x28-0x2F: ()*+,-./
	Z0,	  Z1,	Z2,	 Z3,   Z4,	 Z5,   Z6,	 Z7,	// 0x30-0x37: 01234567
	Z8,	  Z9,	Z+c, X+c,  N+c,  L+c,  M+c,  C+c,	// 0x38-0x3F: 89:;<=>?
	x,    A,    B,   C,    D,    E,    F,    G,	    // 0x40-0x47: @ABCDEFG
	H,    I,    J,   K,    L,    M,    N,    O,  	// 0x48-0x4F: HIJKLMNO
	P,    Q,    R,   S,    T,    U,    V,    W,  	// 0x50-0x57: PQRSTUVW
	X,    Y,    Z,   x,    x,    x,    x,    x,   	// 0x58-0x5F: XYZ[\]^_
	x, 	  A,	B,	 C,	   D,	 E,	   F,	 G,		// 0x60-0x67: `abcdefg
	H,	  I,	J,	 K,	   L,	 M,	   N,	 O,		// 0x68-0x6F: hijklmno
	P,	  Q,	R,	 S,	   T,	 U,	   V,	 W,		// 0x70-0x77: pqrstuvw
	X,	  Y,	Z,	 x,    x,    x,    x,    Z0+c,	// 0x78-0x7F: xyz{|}~<del>
};													// table must be filled to end with 'x'


/* ----	physical translation ------------------------------

		mac key code + modifier -> specci keyboard matrix code
		for games

		because we are only interested in physical key translation for games, 
		emulating the new combined-code keys of ZX Spectrum+ and later is not a real issue.
		what might be left for emulation are 'caps lock' and '.'
		
		keypad keys might be left out because they should be handled fine by 
		fallback to logical translation
*/
typedef uchar phystab[0xc0];
phystab const phys2spec_zxsp =		
{
//	composed keycodes where possible.
//	also used for Jupiter Ace.
	A,   S,   D,   F,   H,   G,    Z,   X,		// 0x00-0x07: a   s   d   f   h   g   z   x
	C,   V,   x,   B,   Q,   W,    E,   R,		// 0x08-0x0F: c   v   •   b   q   w   e   r
	Y,   T,   Z1,  Z2,  Z3,  Z4,   Z6,  Z5,		// 0x10-0x17: y   t   1   2   3   4   6   5
	x,   Z9,  Z7,  x,   Z8,  Z0,   x,   O,		// 0x18-0x1F: •   9   7   •   8   0(null)   •   O(oh)
	U,   x,   I,   P,   ENT, L,    J,   x,		// 0x20-0x27: u   •   i   p   enter   l   j   •
	K,   ENT, x,   SSH, x,   N,    M,   SPC,	// 0x28-0x2F: k   ';'   •   ','   •   n   m   '.'
	x,   SPC, CSH, cZ0, x,   EXT,  x,   x,		// 0x30-0x37: •   space  '<'  delete   •   esc   •   •
	CSH, x,   CSH, SSH, CSH2,CSH2, SSH2,x,		// 0x38-0x3F: shift  •  alt  ctrl  shift2  alt2  ctrl2  •
	x,   sM,  x,   B+s, x,   K+s,  x,   x,		// 0x40-0x47: •   [.]   •   [*]   •   [+]   •   •
	x,   x,   x,   V+s, ENT, x,    J+s, x,		// 0x48-0x4F: •   •   •   [/]   [enter]   •   [-]   •
	x,   L+s, Z0,  Z1,  Z2,  Z3,   Z4,  Z5,		// 0x50-0x57: •   [=]   [0]   [1]   [2]   [3]   [4]   [5]
	Z6,  Z7,  x,   Z8,  Z9,  x,    x,   x,		// 0x58-0x5F: [6]   [7]   •   [8]   [9]   •   •   •
	x,x,x,x,x,x,x,x,							// 0x60-0x67:
	x,x,x,x,x,x,x,x, 							// 0x68-0x6F:
	x,x,x,x,x,x,x,x, 							// 0x70-0x77:
	x,   x,   x,   cZ5, cZ8, cZ6,  cZ7, x,		// 0x78-0x7F: •   •   •   left   right   down   up   •
												// table must be filled to end with 'x'
};

phystab const phys2spec_zx80 =		
{
// no composed keycodes => no need for decomposed_zx80[] table
// used for ZX80 and ZX81:
	A,   S,   D,   F,   H,   G,    Z,   X,		// 0x00-0x07: a   s   d   f   h   g   z   x
	C,   V,   x,   B,   Q,   W,    E,   R,		// 0x08-0x0F: c   v   •   b   q   w   e   r
	Y,   T,   Z1,  Z2,  Z3,  Z4,   Z6,  Z5,		// 0x10-0x17: y   t   1   2   3   4   6   5
	x,   Z9,  Z7,  x,   Z8,  Z0,   x,   O,		// 0x18-0x1F: •   9   7   •   8   0(null)   •   O(oh)
	U,   x,   I,   P,   ENT, L,    J,   x,		// 0x20-0x27: u   •   i   p   enter   l   j   •
	K,   ENT, x,   SSH, x,   N,    M,   SPC,	// 0x28-0x2F: k   ';'   •   ','   •   n   m   '.'
	x,   SPC, CSH, Z0+c,x,   ENT+c,x,   x,		// 0x30-0x37: •   space  '<'  delete   •   esc   •   •
	CSH, x,   CSH, CSH, CSH, CSH,  CSH, x,		// 0x38-0x3F: shift  •  alt  ctrl  shift2  alt2  ctrl2  •
	x,   DOT, x,   B+c, x,   K+c,  x,   x,		// 0x40-0x47: •   [.]   •   [*]   •   [+]   •   •
	x,   x,   x,   V+c, ENT, x,    J+c, x,		// 0x48-0x4F: •   •   •   [/]   [enter]   •   [-]   •
	x,   L+c, Z0,  Z1,  Z2,  Z3,   Z4,  Z5,		// 0x50-0x57: •   [=]   [0]   [1]   [2]   [3]   [4]   [5]
	Z6,  Z7,  x,   Z8,  Z9,  x,    x,   x,		// 0x58-0x5F: [6]   [7]   •   [8]   [9]   •   •   •
	x,x,x,x,x,x,x,x,							// 0x60-0x67:
	x,x,x,x,x,x,x,x, 							// 0x68-0x6F:
	x,x,x,x,x,x,x,x, 							// 0x70-0x77:
	x,   x,   x,   Z5+c,Z8+c,Z6+c, Z7+c, x,		// 0x78-0x7F: •   •   •   left   right   down   up   •
												// table must be filled to end with 'x'
};


/*	these macros work on any  ''uchar[8] keymap;''
*/
#define RES_KEY(KEYMAP,SPEC)	  (KEYMAP[(SPEC)>>4] |= (1<<((SPEC)&7)))
#define SET_KEY(KEYMAP,SPEC)	  (KEYMAP[(SPEC)>>4] &=~(1<<((SPEC)&7)))
#define GET_KEY(KEYMAP,SPEC)	(!(KEYMAP[(SPEC)>>4] &  (1<<((SPEC)&7))))


/* ----	Ersetze Komposittasten --------------
		sucht im Scancode-Buffer nach Komposittasten
		und ersetzt diese durch die normale Tastenkombination,
		wenn es diese Taste auf dem aktuellen Modell nicht gibt.
		Aus BREAK wird so für den ZX Spectrum 48k CAPS SHIFT + SPACE.
*/
static void DecomposeKeys ( int model, uchar keymap[8] )
{
	switch(model)
	{
	case jupiter:			// uses zxsp tables. keyboard virtually identical.
	case pentagon:
	case zx16k:
	case zx48i2:
	case zx48i3:	
		for(int i=0;i<8;i++)				// loop over all bytes in keymap
		{
			uchar byte = keymap[i];					// get byte from keymap
			if((byte>>5)==0x07) continue;			// no composed keys => next
			for(int j=5;j<8;j++)					// loop over all composed keys in byte
			{
				if((byte>>j)&1) continue;					// composed key not set => next
				uchar decomp = decomposed_zxsp[i*8+j];		// decompose composed key
				if(decomp&s) SET_KEY(keymap,SSH);			// if s-flag is set then set SYMBOL SHIFT key
				if(decomp&c) SET_KEY(keymap,CSH);			// if c-flag is set then set CAPS SHIFT key 
				decomp &= 0x77;								// remove c and s flags from key code
				if(decomp!=NOKEY) SET_KEY(keymap,decomp);	// set base key
			}
		}
		((long*)keymap)[0] |= 0xe0e0e0e0;			// reset all composed keys in this byte
		((long*)keymap)[1] |= 0xe0e0e0e0;
		return;
	
	case ts2068:
	case tc2048:
	{	bool csh2 = GET_KEY(keymap,CSH2); RES_KEY(keymap,CSH2);
		bool brk  = GET_KEY(keymap,cSPC); RES_KEY(keymap,cSPC);
		DecomposeKeys(zx48i2,keymap);				// decompose all others, Timex key maps only had above 2 additional keys
		if(csh2) SET_KEY(keymap,CSH2);
		if(brk)  SET_KEY(keymap,cSPC);
		return;
	}

// ZX80 and ZX81 have separate log2spec and phys2spec tabels 
// which have no composed keys:
	case zx80:
	case zx81:
// all currently defined composed keys are those of these models
// so no decomposition can occur:
	case zxplus:
	case zxplus_span:
	case zx128:
	case zx128_span:
	case zxplus2:
	case zxplus2_span:
	case zxplus2_french:
	case zxplus3:
	case zxplus3_span:
	case zxplus2a:
	case zxplus2a_span:		return;							
	}
}
	


/* ---- Berechne Tasten-Boxen für Taste gem. Scancode ----		
		Komposittasten müssen mit ihrem Komposittasten-Pseudoscancode übergeben werden.
		Beachte: Die ENTER-Taste hat manchmal L-Form. Dann wird nur die untere Hälfte zurückgegeben.
		Das Layout von plus/128 und +2/+3 unterscheidet sich nur durch die Breite von EXTENDED MODE und EDIT.
*/
static NSRect KeyRect ( int model, uchar spec )
{
	XXLogIn("KbdPanel:KeyRect(spec=$%2hhX)",spec);

	int x,y,w,h;				// {x,y,w,h} = result box

	spec &= 0x77;				// remove modifier bits  ((should not be set anyway!))
	y = spec>>4;				// 0..7			0..3 = left side, 4..7 = right side
	x = spec&7;					// 0..4			0=outmost key, 4=inmost key

	switch(model)
	{
	case zx80:
	case zx81:
		{
			if (x>=5) goto xx;	// error
			if (y>=4) { y=7-y; x=9-x; }	// right side?  =>  y=0..3  x=5..9  

			int const x0[] = {17+12,17+36,17+24,17};	// left border of leftmost key
		
			w = 40;				// key width
			h = 32;				// key height
			x = x*48 + x0[y];
			y = y*46 + 11;

			if(model==zx81) { h=28; y+=2; }
			break;
		}

	case pentagon:
		{
			if (x>=5) goto xx;				// error
			if (y>=4) { y=7-y; x=9-x; }		// right side?  =>  y=0..3  x=5..9  

			int const x0[]={71,81,71,61};	// x-Wert für 1. Taste einer Reihe

			x = x0[y]+40*x;
			y = 21+40*y;
			w = 41;
			h = 41;
			break;
		}
		
	case zx16k:
	case zx48i2:
	case zx48i3:			
		{
			if (x>=5) goto xx;				// error
			if (y>=4) { y=7-y; x=9-x; }		// right side?  =>  y=0..3  x=5..9  

			int const x0[]={20,45,33,8};	// x-Wert für 1. Taste einer Reihe (bei CSH: normalisiert auf Std-Tastenbreite)

			x = x0[y]+50*x;
			y = 9+49*y;
			w = 37;
			h = 27;

			if(spec==CSH) { x=8; w=49; }	// CAPS SHIFT ist breiter
			if(spec==SPC) { w=62; }			// SPACE ist breiter
			break;
		}
		
	case jupiter:			
		{
			if (x>=5) goto xx;				// error
			if (y>=4) { y=7-y; x=9-x; }		// right side?  =>  y=0..3  x=5..9  

			int const i = 1;						// inset. else the inverting key box has unesthetic corners
			int const x0[]={32+i,54+i,41+i,19+i};	// x-Wert für 1. Taste einer Reihe

			x = x0[y]+48*x;
			y = 15+i+48*y;
			w = 33-2*i;
			h = 30-2*i;

			if(spec==SPC) { w=55-2*i; }			// SPACE ist breiter
			break;
		}

	case ts2068:
	case tc2048:
		{
			h = 39;		// height for all keys
			w = 46;		// default width

		// Komposit- und Sondertasten:
			switch(spec)	
			{
			case CSH:	/* 1. CSH */	x=13;      y=37;   w=58;  goto x;	// 1. CAPS SHIFT Taste: breiter ((die linke))
			case CSH2:	/* 2. CSH */	x=70+45*9; y=37;   w=58;  goto x;	// 2. CAPS SHIFT Taste: zus. Taste ((die rechte))
			case ENT:	/* ENTER  */	x=47+45*9; y=37+39;w=81;  goto x;	// ENTER: breiter
			case SPC:	/* SPACE  */	x=115;     y=0;    w=316; goto x;	// SPACE: unter allen anderen Tasten und richtig breit
			case cSPC:	/* BREAK  */	x=70+45*8; y=37;   w=46;  goto x;	// BREAK: neue Komposittaste
			}
		
		// normale Taste:
			if (x>=5) goto xx;				// error
			if (y>=4) { y=7-y; x=9-x; }		// right side?  =>  y=0..3  x=5..9  
			int const x0[]={25,47,36,13};	// x-wert für 1. taste einer reihe (bei CSH-Taste: extrapolierte Position)
			x = 45*x + x0[y];
			y = 39*y + 37;
			break;
		}
		
	case zxplus:
	case zxplus_span:
	case zx128:
	case zx128_span:
	case zxplus2:
	case zxplus2_span:
	case zxplus2_french:
	case zxplus3:
	case zxplus3_span:
	case zxplus2a:
	case zxplus2a_span:
		{
			w = 41;			// default width
			h = 41;			// height for all keys

		// Komposit- und Sondertasten:
			switch(spec)				
			{
			case EXT:/*EXTEND MODE*/	y=80; w=model<zxplus2?61:71; goto x;
			case sO: /* ; */			x=40;  y=0;			 goto x;
			case sP: /* " */            x=80;  y=0;			 goto x;
			case sN: /* , */            x=460; y=0;			 goto x;
			case sM: /* . */			x=410; y=40;		 goto x;
			case cZ1: /*EDIT*/			x=model<zxplus2?60:70; y=80; w=model<zxplus2?51:41;	goto x;
			case cZ2: /*CAPS LOCK*/		x=90;  y=40;		 goto x;
			case cZ3: /*TRUE VIDEO*/	x=0;   y=160;		 goto x;
			case cZ4: /*INV VIDEO*/		x=40;  y=160;		 goto x;
			case cZ5: /*CURSOR LEFT*/	x=120; y=0;			 goto x;
			case cZ6: /*CURSOR DOWN*/	x=420; y=0;			 goto x;
			case cZ7: /*CURSOR UP*/		x=380; y=0;			 goto x;
			case cZ8: /*CURSOR RIGHT*/	x=160; y=0;			 goto x;
			case cZ9: /*GRAPHICS*/		x=60;  y=120;		 goto x;
			case cZ0: /*DELETE*/		x=0;   y=120; w=61;	 goto x;
			case cSPC:/*BREAK*/			x=480; y=160; w=61;	 goto x;
			case SSH2:/*2. SSH Taste*/	x=0;   y=0;			 goto x;
			case CSH2:/*2. CSH Taste*/	x=0;   y=40;  w=91;	 goto x;
			case CSH:					x=450; y=40;  w=91;  goto x;
			case SSH:					x=500; y=0;			 goto x;
			case ENT:					x=470; y=80;  w=81;  goto x;	// L-förmig!
			case SPC:					x=200; y=0;   w=181; goto x;
			}

		// normale Taste:
			if (x>=5) goto xx;				// error
			if (y>=4) { y=7-y; x=9-x; }		// right side?  =>  y=0..3  x=5..9  
			int const x0[]={90,110,100,80};	// x-wert für 1. taste einer reihe (bei CSH-Taste: extrapolierte Position)
			x = 40*x + x0[y]; 
			y = 40*y + 40;
			break;
		}

	default:				// error exit:
		{
			static bool f=1; if(f) { LogLine("KbdPanel:KeyRect: model %s not yet supported",zx_info[model].name); f=0; }
xx:			x=y=w=h=0;
			break;
		}
	}
	
x:	//printf("box($%02x) = (%i,%i),(%i,%i)\n",spec,x,y,w,h);
	NSRect box = { {x,y}, {w,h} };
	return box;
}



/* ----	add implicitely connected keys ----
		if you press 3 keys where 2 are in the same row and 2 in the same comumn, 
		the resulting 4th key seems to be pressed too. 
*/
static void AddGhostKeys ( uchar keymap[8] )
{
	uchar* e=keymap+7;
	while(e>keymap&&*e==0xff) e--;
	if(e<=keymap) return;					// only keys in 1 row or no keys pressed at all
p:
	for( uchar*a=keymap; a<e; a++ )			// loop over lines 1 to 7
	{
		if(*a==0xff) continue;				// no keys pressed in this row 

		for( uchar*b=e; b>a; b-- )			// compare to other lines
		{
			uchar xo = *a|*b;				// xo = same bits 0 ((keys down)) in both rows
			if( xo==0xff ) continue;		// no 2 keys in same column
			uchar xa = *a&*b;				// xa = all bits 0 ((keys down)) in any of both rows
			if( xa==xo ) continue;			// no 3rd key down
		// ghost key(s) detected:
		// rows a and b are connected via 2 keys on same column wire
		// => keys from a flow to b and vice versa
			*a = *b = xa;					
			goto p;							// ghost key may connect a or b to already tested wires
		}
	}
}




// ######################################################################################################
// ######################################################################################################




#import "MenuController.h"


@implementation KbdView : NSView


/* -------------------------------------------------------------------------------------------------------
	context menu handling:
*/

/* ----	call backs for the menu items:
*/
- (void) setKbdLog:(id)s	{ [menu_controller setKeyboardBasicMode:s]; /* machine->SetKeyboardMode(kbdlog); */ }
- (void) setKbdPhys:(id)s	{ [menu_controller setKeyboardGameMode:s]; /* machine->SetKeyboardMode(kbdphys); */ }

/* ----	UP: create context menu:
*/
- (void) addContextMenu
{
	NSMenu*	theMenu	  = [[[NSMenu alloc] initWithTitle:@"Keyboard mode"] autorelease];
	NSMenuItem* item1 = [[[NSMenuItem alloc] initWithTitle:@"Basic mode" action:@selector(setKbdLog:) keyEquivalent:@"b"] autorelease];
						[item1 setToolTip:@"Use logical keyboard translation: preserve the keys' meaning"];
	NSMenuItem* item2 = [[[NSMenuItem alloc] initWithTitle:@"Game mode" action:@selector(setKbdPhys:) keyEquivalent:@"g"] autorelease];
						[item2 setToolTip:@"Use physical keyboard translation: preserve the keys' location"];

    [theMenu insertItem:item1 atIndex:0];
    [theMenu insertItem:item2 atIndex:1];
    [self setMenu:theMenu];		// set the default menu to be returned by  [NSView menu]
}

/* ----	override for [NSView menuForEvent]
		which is called by the GUI to get the context menu via [NSView menu]:
*/
- (NSMenu*) menuForEvent:(NSEvent*)e
{
	KbdPanel* window = (KbdPanel*)[self window];
	NSMenu* menu = [self menu];		// [[self class] defaultMenu];
	[[menu itemAtIndex:0] setState:[window getMode] == kbdlog];
	[[menu itemAtIndex:1] setState:[window getMode] == kbdphys];
	return menu;
}		




/* ----	replacement function for drawing --------------------
		draw whatever belongs into this box
*/
- (void)drawRect:(NSRect)box
{
	XXLogIn("KbdView:drawRect");
	CGContext* context = (CGContext*)[[NSGraphicsContext currentContext] graphicsPort];		// Quartz-2D drawing context
	
	cstr filename = zx_info[model].kbd_filename;
	ulong ext = Peek4X(LowerStr(ExtensionFromPath(filename)));
	filename = CatStr(rsrc_path,filename);

	CFStringRef filepath = CFStringCreateWithCString ( NULL/*alloc*/, filename, kCFStringEncodingUTF8 );
	CFURLRef fileurl = CFURLCreateWithFileSystemPath ( NULL/*alloc*/, filepath, kCFURLPOSIXPathStyle, no/*isDirectory*/ );
	CGDataProviderRef dataprovider = CGDataProviderCreateWithURL ( fileurl );
	CGImageRef image = 
		ext=='.png' ? CGImageCreateWithPNGDataProvider  ( dataprovider, NULL/*decode[]*/, no/*shouldInterpolate*/, kCGRenderingIntentDefault )
	  :	ext=='.jpg' ? CGImageCreateWithJPEGDataProvider ( dataprovider, NULL/*decode[]*/, no/*shouldInterpolate*/, kCGRenderingIntentDefault )
	  : nil;	// bummer

	CGRect rect = { {0,0}, {541,201} };
	CGContextDrawImage ( context, rect, image );

	CGContextSetGrayFillColor ( context, 0.0/*grey=black*/, 0.5/*alpha*/ );

	for(int i=0;i<8;i++)	// 8 scanbytes
	{
		uchar byte = keymap[i];
		if(byte==0xff) continue;	// no key down
		for( int j=0; j<8; j++ )	// 5 std bits + 3 bits used for composed keys
		{
			if( (byte>>j)&1 ) continue;							// this key is not down
			NSRect rect = KeyRect(model,(i<<4)+j);				// NOTE: NSRect == CGRect
			CGContextFillRect ( context, *(CGRect*)&rect );				
		}
	}
	
	CGImageRelease ( image );
	CGDataProviderRelease ( dataprovider );
	CFRelease(filepath);
	CFRelease(fileurl);
}


/* ---- init ----------------------
*/
-(id)initWithModel:(uint)currentmodel 
{
	XLogIn("KbdView:initWithModel: %s",zx_info[currentmodel].name);

	NSRect box = {{0,0},{541,201}};
	self = [super initWithFrame:box];
	if(self)
	{
		model = currentmodel;
		memset(keymap,0xff,8);	// all up
		mousekey = NOKEY;
		[self addContextMenu];
	}
	return self;
}


/* ----	update visually displayed keys --------------------
*/
-(void)updateKeymap:(uchar*)newkeymap
{
	XXXASSERT( [NSThread currentThread] == main_thread );		// setNeedsDisplayInRect darf nur im Main Thread aufgerufen werden!

	DecomposeKeys(model,newkeymap);

	for (int i=0;i<8;i++) 
	{
		uchar c = newkeymap[i] ^ keymap[i]; 
		if(c) 
		{
			keymap[i] = newkeymap[i];
			for(int j=0;j<8;j++)
			{
				if( (c>>j)&1 ) [self setNeedsDisplayInRect:KeyRect(model,(i<<4)+j)];		
			}
		}
	}
}


/* ----	find key for mouse pointer position ----
*/
-(uchar)findKeyForPoint:(NSPoint)where
{
	for( int i=0;i<8;i++ ) 
	for( int j=0;j<8;j++ )
	{
		uchar key = i*16 + j;
		NSRect box = KeyRect(model,key);
		if( [self mouse:where inRect:box] ) return key;
	}
	XXLogLine("point outside any key");
	return NOKEY;
}


/* ----	NSResponder overrides: event handling ------------------------------
		Handling Mouse down, Mouse drag and Mouse up
*/
-(BOOL)acceptsFirstMouse:(NSEvent*)event
{
	return yes;
}

-(void)mouseDown:(NSEvent*)event
{
	XXLogIn("KbdView:MouseDown");
	NSPoint where = [event locationInWindow];			// window coordinates
	where = [self convertPoint:where fromView:nil];		// nil == my window
	mousekey = [self findKeyForPoint:where];
	if( mousekey!=NOKEY ) { XXLogLine("mousekey=$%02x",(int)mousekey); [(id)[self window] specciKeyDown:mousekey];	}
}

-(void)mouseDragged:(NSEvent*)event
{
	XXLogIn("KbdView:MouseDrag");
	NSPoint where = [event locationInWindow];			// window coordinates
	where = [self convertPoint:where fromView:nil];		// nil == my window
	if( mousekey!=NOKEY && [self mouse:where inRect:KeyRect(model,mousekey)] ) return;	// still over same key

// mouse left previous key
	if( mousekey!=NOKEY ) [(id)[self window] specciKeyUp:mousekey];	
	mousekey = [self findKeyForPoint:where];
	if( mousekey!=NOKEY ) [(id)[self window] specciKeyDown:mousekey];	
}

-(void)mouseUp:(NSEvent*)event
{
	XXLogIn("KbdView:MouseUp");	
	if( mousekey!=NOKEY ) [(id)[self window] specciKeyUp:mousekey];
	mousekey=NOKEY;
}


-(uchar&)getMousekey
{
	return mousekey;
}

-(uchar*)getKeymap
{
	return keymap;
}

-(void)dealloc
{
	XLogIn("KbdView:dealloc: %s",zx_info[model].name);
	[super dealloc];
}

@end







// ####################################################################################
// ####################################################################################






@implementation KbdPanel


- (id)initWithContentRect:(NSRect)contentRect styleMask:(uint)styleMask backing:(NSBackingStoreType)backingType defer:(BOOL)flag
{
	XLogIn("KbdPanel:initWithContentRect");

	kbd_panel = self = 
	[super initWithContentRect:contentRect styleMask:styleMask backing:backingType defer:flag];
	
	if(self)
	{
		model		= 0xFFFF;	// -> updatePanel legt kbd_view an
		mode		= kbdlog;
		memset(keymap,0xff,8);
		//stickykeys	= 0;
		//modifiers		= 0;
		//pressed_keys	= 0;
		kbd_view	= nil;		// wird in awakeFromNib angelegt
	}
	
	return self;
}

//-(void)close{}

-(void)awakeFromNib
{
	XLogIn("KbdPanel:awakeFromNib");

//	if([preferences getBoolForKey:@"startup_open_keyboard"])				<-- done in MenuController
	bool f = [preferences getBoolForKey:"runtime_leave_keyboard_open"];		[self setHidesOnDeactivate:!f];	
}


/*	Update Panel: 
	- call from main event loop
	- for regular panel update
	- called only if panel is visible
	- front machine m is locked for reading (RLOCK/RUNLOCK)
*/
-(void)updatePanel:(Machine*)m
{
	if( mode != m->GetKeyboardMode() ) 
	{
		mode = m->GetKeyboardMode();		// toggle mode
		XXTRAP(mode!=kbdlog&&mode!=kbdphys);
		stickykeys	 = 0;
		modifiers	 = 0;
		pressed_keys = 0;					// key specs können jetzt abweichen!
	}
	
	if( model!=m->GetModel() )
	{
		model = m->GetModel();

		[ self setTitle:ToNSString(zx_info[model].name) ];

		stickykeys	 = 0;
		modifiers	 = 0;
		pressed_keys = 0;

		kbd_view = [[KbdView alloc] initWithModel:model];
		[self setContentView:kbd_view];
		[kbd_view release];					// note: setContentView retaint den contentView
	}	
}


// reset keyboard: all keys up
-(void)allKeysUp
{
	stickykeys	 = 0;
	modifiers	 = 0;
	pressed_keys = 0;
	[self updateKeymap];
}



-(void)specciKeyDown:(uchar)zxkey
{
	XXLogIn("KbdPanel:specciKeyDown");
	if(zxkey==NOKEY) return;

	uint n = pressed_keys;		XTRAP( n>=NELEM(zx_keycodes) );
	zx_keycodes[n] = zxkey;
	os_keycodes[n] = zxkey;		// note: oskey == zxkey for virtual keyboard keys
	pressed_keys   = n+1;
	[self updateKeymap];
}


-(void)specciKeyUp:(uchar)zxkey 
{
	XXLogIn("KbdPanel:specciKeyUp");
//	if(zxkey==NOKEY) return;
	
	int n = pressed_keys;
	for(int i=0;i<n;i++) 
	{ 
		if(os_keycodes[i]==zxkey) { pressed_keys = --n; os_keycodes[i]=os_keycodes[n]; zx_keycodes[i]=zx_keycodes[n]; break; }
	}

	if( stickykeys != 0x0000 )	// any key sticky down ?
	{
		stickykeys = 0x0000;	// => release any sticky keys
	}
	else						// no key sticky down
	{
		switch(zxkey)			// => shift key going up becomes sticky
		{
		case CSH:	stickykeys = capsshiftsticky; break;		
		case CSH2:	stickykeys = capsshift2sticky; break;
		case SSH:	if(model>=jupiter /*not ZX80 or ZX81 which have DOT on that position*/) 
						stickykeys = symbolshiftsticky; break;	
		case SSH2:	stickykeys = symbolshift2sticky; break;
		}
	}

	[self updateKeymap];
}


-(uchar)specciKeyForRealworldKey:(ushort)keycode:(unichar)unicode
{
	XXLogIn("KbdPanel:specciKeyForRealworldKey($%02x,$%02x)",(int)keycode,(int)unicode);

	bool f80 = model==zx80, f81 = model==zx81;
	logtab const&  log2spec  = f80 ? log2spec_zx80 : f81 ? log2spec_zx81 : log2spec_zxsp;
	phystab const& phys2spec = f80||f81 ? phys2spec_zx80 : phys2spec_zxsp;

	if( mode==kbdlog )
	{		
		uchar lspec = unicode<sizeof(logtab) ? log2spec[unicode]  : NOKEY;	// logical
		if( lspec!=NOKEY ) return lspec;
		return keycode<sizeof(phystab) ? phys2spec[keycode] : NOKEY;		// fallback to physical
	}
	else
	{
		uchar pspec = keycode<sizeof(phystab) ? phys2spec[keycode] : NOKEY;	// physical
		if( pspec!=NOKEY ) return pspec;
		return unicode<sizeof(logtab) ? log2spec[unicode] : NOKEY;			// fallback to logical
	}
}

-(uchar)specciKeyForEvent:(NSEvent*)event
{
	ushort  keycode = [event keyCode];				
	NSString* chars = [event charactersIgnoringModifiers/*exceptForShift((scheiße!))*/];	
					  if( [chars length]==0 ) { XXLogLine("dead key"); return NOKEY; } /* dead keys return "" */
	unichar unicode = [chars characterAtIndex:0];		
	return [self specciKeyForRealworldKey:keycode:unicode];
}

// ----	replacement handlers ----
//		get all key events to the machine!
//
-(void)keyDown:(NSEvent*)event
{
	if([event isARepeat]) return;

	XXLogIn("KbdPanel:keyDown:$%02x",(int)[event keyCode]);

	uint  n		= pressed_keys;						XXASSERT( n<NELEM(zx_keycodes) );
	uchar zxkey = [self specciKeyForEvent:event];
	uchar oskey = 0xff - [event keyCode];
	uint  new_modifiers = modifier_keys = [event modifierFlags];

	if(zxkey==NOKEY) 
	{
		return;			// TODO: keyDown(Event) for ÄÖÜ etc.
	}

// Nachdem wir ja schon wissen, dass wir im Keyup-Event andere Character Codes mitgeteilt bekommen können, nun auch noch das:
// es kann vorkommen, dass wir für ein Keydown-Event überhaupt kein Keyup-Event kriegen. Billigpfurzer@Apple!
// und zwar wenn: +Taste +Command -Taste -Command. Dann wird '-Taste' nicht durchgestellt.
// dann muss es möglich sein, die Taste durch nochmaliges Drücken zu lösen
// dazu darf sie aber nicht zweifach eingetragen werden, sonst löst das Keyup nur eine der Dubletten.
	for(uint i=0;i<n;i++) if(os_keycodes[i]==oskey) return;	

	zx_keycodes[n] = zxkey;
	os_keycodes[n] = oskey;
	pressed_keys = n+1;
	
	modifiers = new_modifiers;
	[self updateKeymap];
}

-(void)keyUp:(NSEvent*)event
{
	XXLogIn("KbdPanel:keyUp:$%02x",(int)[event keyCode]);

//	uchar zxkey = [self specciKeyForEvent:event];
//	if(zxkey==NOKEY) return;			// TODO: keyUp(Event) for ÄÖÜ etc.

	uchar oskey = 0xff - [event keyCode];
	int n = pressed_keys;
	for(int i=0;i<n;i++) 
	{ 
		if(os_keycodes[i]==oskey) { pressed_keys = --n; os_keycodes[i]=os_keycodes[n]; zx_keycodes[i]=zx_keycodes[n]; break; }
	}

	modifier_keys = modifiers = [event modifierFlags];
	stickykeys = 0x0000;			// key up => release sticky keys
	[self updateKeymap];
}

-(void)flagsChanged:(NSEvent*)event
{
	XXLogIn( "KbdPanel:flagsChanged: $%02x", (int)[event keyCode] );
// note: man bekommt zwar einen Tastencode mit, der wohl die Taste bezeichnet, die rauf/runter geht.
// linke und rechte Modifiertasten werden dabei aber _nicht_ unterschieden!

	bool is_cmdkey = [event keyCode]==0x37/*Command Key*/;			// command key going up or down?
	bool cmdkey_up = is_cmdkey && (modifier_keys&NSCommandKeyMask);	// command key going up?

	uint new_modifiers = modifier_keys = [event modifierFlags];	
	if( cmdkey_up ) modifier_keys = new_modifiers & ~NSCommandKeyMask;
	// note: auch wenn die Cmd-Taste hochgeht, ist Bit NSCommandKeyMask in [event modifierFlags] gesetzt!

	if(XXLOG)
	{
		static cstr names[] = { /*16*/"AlphaShiftKey", /*17*/"ShiftKey", /*18*/"ControlKey", /*19*/"AlternateKey",
								/*20*/"CommandKey", /*21*/"NumericKeypadKey", /*22*/"HelpKey", /*23*/"FunctionKey" };
		uint up = modifiers & ~new_modifiers;
		uint dn = ~modifiers & new_modifiers;
		for( int i=16; i<=23; i++ )
		{
			if(up&(1<<i)) LogLine("%s up",names[i-16]);
			if(dn&(1<<i)) LogLine("%s down",names[i-16]);
		}
	}

	if( Mouse::IsGrabbed() && (new_modifiers&NSCommandKeyMask) ) Mouse::HandleCommandKey();

// 1. wenn Shifttasten toggeln während Command nieder ist kriegen wir ein Event => ignorieren
// 2. wenn Command nieder geht, kriegen wir ein Event => ignorieren
// 3. wenn Command rauf geht, kriegen wir ein Event => ignorieren (es könnt zB auch noch Alt nieder sein, 
//    wenn Cmd+Alt gedrückt wurden und Cmd zuerst hoch geht. Dann sollten wir jetzt Alt nicht erkennen. Also raus.)
	if( (new_modifiers&NSCommandKeyMask) || [event keyCode]==0x37/*Command Key*/ ) 
	{
		if( pressed_keys || modifiers || stickykeys )
		{
			pressed_keys = 0;			// CMD key down/up ?
			modifiers    = 0x0000;		// release all keys
			stickykeys	 = 0x0000;		// => no problems with 'stuck keys'
			[self updateKeymap];		
		}
		return;
	}

	if( (modifiers&~new_modifiers)&(capsmask|controlmask|metamask|capslockmask) )
	{ stickykeys = 0x0000; }			// key up => release sticky shift keys
	modifiers = new_modifiers;
	[self updateKeymap];
}



/* --------------------------------------------------------------
	recalculate state of keyboard maps

	kbdType=kbdPhysical	=> physical mapping is preferred (for games)
	kbdType=kbdLogical	=> logical  mapping is preferred (for basic)

	the table of the preferred method is tried first. if it does not
	contain an entry for the pressed/released key, the other table is tried.
	
	"alt" works as auxiliary "caps shift" keys in both mapping modes.
	they are essential in logical mapping mode to get the "caps shift" + "number key",
	e.g. "caps shift" + "1" to edit a basic line. if you simply press "shift" plus "1"
	this is "!" and this is decoded as "symbol shift" + "1" which results in a "!"
	on the specci. on the other hand, "alt" + "1" works. useful for breaking into 
	running basic programs too, since shift + space results in a simple space too.

	the state of control key is taken from ev->state in both modes, since it does
	not influence the character code returned by XLookupString().

	the state of shift and shift lock are taken from xev->state only in physical 
	mapping mode. they are not decoded separately in logical mode, since the key 
	translation table supplies the corresponding shift and control key state and an 
	additional modifier detected will disturb. e.g. if you press "!" (that is: "shift" + "1") 
	this results in "symbol shift" + "1" which is transposed to "!" by the specci
	keyboard routines just as expected. adding "caps shift" because "shift" is 
	pressed on the real keyboard (just to get the "!") will result in "1" and both
	shift keys pressed to the specci, decoding to nothing (3 keys pressed!)

---------------------------------------------------------------*/

-(void)updateKeymap
{
	XXLogIn("KbdPanel:updateKeymap");
	
// work on a keymap copy, so that running machines do not intermediately see no keys
	uchar newkeymap[8];
	memset( newkeymap,0xff,8 );
	
	bool ownmeta = false;		// set if keys supplied own caps/symbol flag
	 
	for(int k=0;k<pressed_keys;k++)
	{
		uchar key = zx_keycodes[k]; 
		//if(key==NOKEY) continue;
		if(key&c) { SET_KEY(newkeymap,CSH); ownmeta=true; key-=c; }
		if(key&s) { SET_KEY(newkeymap,SSH); ownmeta=true; key-=s; }
		if(key==NOKEY) continue;
		SET_KEY(newkeymap,key);
		if( (key&7)>=5 ) ownmeta=true;
	}

	bool hasSSH = model!=zx80 && model!=zx81;

	if( modifiers&capsmask && !ownmeta ) SET_KEY(newkeymap,CSH);
	if( modifiers&controlmask )			 SET_KEY(newkeymap,hasSSH?SSH:CSH);
	if( modifiers&metamask )			 SET_KEY(newkeymap,CSH);
	
	if(stickykeys!=0x0000)
	{
		if( stickykeys&symbolshiftsticky )  SET_KEY(newkeymap,SSH);
		if( stickykeys&symbolshift2sticky ) SET_KEY(newkeymap,SSH2);
		if( stickykeys&capsshiftsticky )    SET_KEY(newkeymap,CSH);
		if( stickykeys&capsshift2sticky )   SET_KEY(newkeymap,CSH2);
	}

// update specci and visual keyboard
	if( [self isVisible] ) [kbd_view updateKeymap:newkeymap];
	DecomposeKeys(zx48i2,newkeymap);
	AddGhostKeys(newkeymap);

// update keymap
	memcpy( keymap, newkeymap, 8 );
	//for(int i=0;i<8;i++) Log("%2hhx ",newkeymap[i]); LogNL();
}





-(uint)getPressedKeysCount
{
	return pressed_keys;
}

-(uchar*)getPressedOSKeycodes
{
	return os_keycodes;
}

-(uchar*)getKeymap
{
	return keymap;
}

-(uint) getMode
{
	return mode;
}


@end













