/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	LOG		1
#define	SAFE	3

#import <fcntl.h>
#import <unistd.h>

#import "unix/file_utilities.h"
#import "cstrings/cstrings.h"

#import "utilities.h"
#import "globals.h"
#import "Machine.h"
#import "MachinePanel.h"
#import "Preferences.h"
#import "Ula.h"
#import "Mmu.h"
#import "Z80.h"
#import "Z80 opcodes.h"
#import "KbdPanel.h"
#import "Ay.h"
#import "ICTester.h"
#import	"SinclairZxInterface1.h"
#import "SinclairZxPrinter.h"
#import	"SinclairZxInterface2.h"
#import	"ZxPlus2JoystickPorts.h"
#import	"ZxPlus2JoystickPortsPanel.h"
#import	"KempstonJoystickInterface.h"
#import	"KempstonMouseInterface.h"
#import "SinclairZxInterface1Panel.h"
#import "FullerBox.h"
#import "ZonxBox.h"
#import "Screen.h"
#import "TapeRecorder.h"
#import "ZxInfo.h"
#import "AlertPanel.h"
#import "MenuController.h"
#import "GifEncoder.h"
#import "GifReport.h"
#import "Mmu128k.h"
#import "MmuPlus3.h"



#define ForAllItems( WHAT ) { ZxItem*p=FirstItem(); do{ p->WHAT; } while( (p=p->Next())!=FirstItem() ); }


void Machine::MoveToFrontInMList ( )
{
	machine_list.MoveToFront(this); 
	[kbd_panel updatePanel:this];				// TODO!
}



/* ----	screenUpdate performance counter ----
*/
int frames_hit_count= 0;
static void add_frame_hit (int n)	
{ 
	static char map[100]; 
	static int i=0; ++i; i%=100; 
	frames_hit_count += n - map[i]; 
	map[i] = n; 
}



static cstr load_file ( cstr fpath, uptr& data, ulong& count )
{
	int fd = open( fpath, O_RDONLY, 0 );	
	if( fd!=-1 ) 
	{
		count = FileLength(fd);
		data  = (uptr)TempStr(count);
		ulong n = read(fd,data,count);
		close(fd);
		if(n==count) return NULL; /* ok */
	}
	data = NULL; count = 0;
	return Using("read file \"%s\" failed: %s",fpath,ErrorStr());
}


#import "Snapshots.mm"


void Machine::set_title()					
{
	XXXTRAP(!screen);
	cstr title = filepath ? FileNameFromPath(filepath) : model->name;
	[screen setTitle:ToNSString(title)]; 
}

void Machine::SetWindowZoom (int factor)
{
	[screen setZoom:factor];
}



// querries:
Sample	Machine::GetBeeperVolume ()		{ return ula->GetBeeperVolume(); }
long	Machine::GetCurrentCpuCycle()	{ return cpu->CpuCycle() - int( time_dsp_start * cpu_cycles_per_second ); }
bool	Machine::IsFullscreen()			{ return [screen isFullscreen]; }
ushort* Machine::GetVideoRam()			{ return ula->GetVideoRam(); }



void Machine::SavePreferencesNow()
{
	[preferences setBool:GetAy()!=NULL						forFile:filepath andKey:"newsnapshot_soundchip"];
	[preferences setBool:kempston_joystick_interface!=NULL	forFile:filepath andKey:"newsnapshot_joystick"];
	[preferences setBool:kempston_mouse_interface!=NULL		forFile:filepath andKey:"newsnapshot_mouse"];
	[preferences setInt: keyboard_mode						forFile:filepath andKey:"newsnapshot_keyboard_mode"];
	[preferences setBool:autostart_tape						forFile:filepath andKey:"newsnapshot_autostart_tape"];
}


/* ----	dealloc everything except screen ----
*/
void Machine::dealloc_items()
{
	XXLogIn("Machine:dealloc_items");

	SavePreferencesNow();
	
	while(Child()) delete Child();

	breakpoints = 0;
	delete[] filepath;					filepath = NULL; 
	if( instant_load_fd!=-1 ) close(instant_load_fd); instant_load_fd=-1;	// ignore error
	if( instant_save_fd!=-1 ) close(instant_save_fd); instant_save_fd=-1;	// ignore error
	delete[] instant_load_filename;		instant_load_filename = NULL;
	delete[] instant_save_filename;		instant_save_filename = NULL;
}
	

Machine::~Machine()
{
	XLogIn("delete Machine");

	XXXASSERT(MAIN_THREAD);	

	// lock.Lock();				// nicht sperren! Screen:updateThread könnte daran hängen bleiben!
	HaltCpu();					// -> runMachinesForSound will skip and not wait for this machine ((sicherheitshalber))
	machine_list.UnlinkMachine(this);	// -> runMachinesForSound will no longer find me ((mutexed via mlock))
	[screen closeAndDispose];	// -> updateThread won't access anything beyond screen
	dealloc_items();
}

Mmu* Machine::new_mmu_for_this_machine() 
{
	switch( model->model )
	{
	case zx128:
	case zx128_span:
	case zxplus2:
	case zxplus2_span:
	case zxplus2_french:	
			XXXASSERT( model->has_port_7ffd && !model->has_port_1ffd );
			return new Mmu128k(this);
	case zxplus3:
	case zxplus3_span:
	case zxplus2a:
	case zxplus2a_span:
			XXXASSERT( model->has_port_7ffd && model->has_port_1ffd );
			return new MmuPlus3(this);
	case tc2048:
	case ts2068:
			XXXASSERT( model->has_port_F4 && model->has_port_FF );
			TODO();
	default:
			XXXTRAP( model->has_port_F4 );
			XXXTRAP( model->has_port_FF );
			XXXTRAP( model->has_port_1ffd );
			XXXTRAP( model->has_port_7ffd );
			return new Mmu(this);
	}
}


Ay* Machine::new_ay_for_this_machine()
{
	cstr s,w,r;
	
	switch( model->model )
	{
	case zx80:
	case zx81:
	case jupiter:	// ?
		return new ZonxBox(this,Zonx81);
		
	case samcoupe:	// ?
	case tc2048:	// Anm.: der TC2048 hat keinen AY-Chip
	case ts2068:	// Anm.: 0xF4 ist auch benutzt (Paging)
		s = "----.----.----.0-01";		// ? select: üblicher Port: 0xF5
		w = "----.----.----.0-10";		// ? write:  üblicher Port: 0xF6
		r = "----.----.----.0-10";		// ? read:   üblicher Port: 0xF6
		return new InternalAy( this, s, w, r, model->psg_cycles_per_second );
		
	case pentagon:	// ?
	case scorpion:	// ?
	case zx128:		// Anm.: Bit 15 auch decodiert wg. Mmu
	case zx128_span:
	default:
		s = "11--.----.----.--0-";		// ? select: üblicher Port: 0xfffd
		w = "10--.----.----.--0-";		// ? write:  üblicher Port: 0xbffd
		r = "11--.----.----.--0-";		// ? read:   üblicher Port: 0xfffd
		return new InternalAy( this, s, w, r, model->psg_cycles_per_second );
	};
	
}


void Machine::load_preferences()
{
	XLogIn("Machine:load_preferences");

	if(filepath)	// snapshot file, use old settings if available and enabled in prefs, else use snapshot defaults
	{
		if( !GetAy() && [preferences getBoolForFile:filepath andKey:"newsnapshot_soundchip"] ) {
			Ay* ay = new_ay_for_this_machine();	
			ay->PowerOn( time_dsp_start, cc_dsp_start );
		}
		if( !kempston_joystick_interface && !zxplus2_js_ports && 
			[preferences getBoolForFile:filepath andKey:"newsnapshot_joystick"] ) 
					ToggleKempstonJoystickInterface(); 
		if( !kempston_mouse_interface && 
			[preferences getBoolForFile:filepath andKey:"newsnapshot_mouse"] ) 
					ToggleKempstonMouseInterface(); 
		SetAutostartTape( [preferences getBoolForFile:filepath andKey:"newsnapshot_autostart_tape"] );
		keyboard_mode = [preferences getIntForFile:filepath andKey:"newsnapshot_keyboard_mode"];
		[kbd_panel updatePanel:this];	// note: wenn invisible wird NXPanel::updatePanel nicht aufgerufen!
	}
	else			// new machine
	{
		if( !GetAy() && [preferences getBoolForKey:"newmachine_soundchip"] ) {
			Ay* ay = new_ay_for_this_machine();	
			ay->PowerOn( time_dsp_start, cc_dsp_start );
		}
		if( !kempston_joystick_interface && !zxplus2_js_ports && 
			[preferences getBoolForKey:"newmachine_joystick"] ) ToggleKempstonJoystickInterface();
		SetAutostartTape( [preferences getBoolForKey:"newmachine_autostart_tape"] );
		keyboard_mode = [preferences getIntForKey:"newmachine_keyboard_mode"];	// note: SetKeyboardMode tut evtl. nichts 
		[kbd_panel updatePanel:this];	// note: wenn invisible wird NXPanel::updatePanel nicht aufgerufen!
	}
}


/* --------------------------------------------------------
	designated initializer
	init with default model
	note: Machine.screen is preserved and must be cleared to NULL by caller if not yet allocated
*/
void Machine::init_with_model( uint modl )
{
	XLogIn( "Machine:init_with_model: %i = %s", (int)modl, modl<zxmodels?zx_info[modl].name:"(bummer!)" );
	if( modl>=zxmodels || !zx_info[modl].is_supported ) modl = zx48i3;	// force legal

// clear instance data:

	model		= zx_info+modl;
	filepath	= NULL;

	//screen	= NULL;			// screen is not reallocated when changing model
	cpu			= NULL;
	ula			= NULL;
	mmu			= NULL;	

	kempston_mouse_interface	= NULL;
	kempston_joystick_interface	= NULL;
	sinclair_zx_interface1		= NULL;
	sinclair_zx_interface2		= NULL;
	sinclair_zx_printer			= NULL;
	fuller_box					= NULL;
	zonx_box					= NULL;
	ic_tester					= NULL;
	zxplus2_js_ports			= NULL;

	rom_dis		= no;			// set from external item, e.g. IF2			
	rom.Purge();				
	ram.Purge();								

	breakpoints	= 0;
	break_ptr	= NULL;

	cpu_state	= 1 /*halted*/;
	tape_state	= 0;	
	cpu_options	= cpu_crtc;	if( model->has_contention() ) cpu_options |= cpu_waitmap;
	tape_options = 0;

	instant_load_fd			= -1;
	instant_load_filename	= NULL;
	instant_load_filetype	= unknown_file;	// see TapeData.h
	instant_save_fd			= -1;
	instant_save_filename	= NULL;
	instant_save_filetype	= unknown_file;

	keyboard_mode	= kbdlog;	
	[kbd_panel updatePanel:this];	// note: wenn invisible wird NXPanel::updatePanel nicht aufgerufen!
	
	cpu_cycles_per_second = model->cpu_cycles_per_second;

	time_dsp_start		= 0.0;
	cc_dsp_start		= 1000;			// +2A:  ok: 0k..2k, 7k..14k    boot error: 3k..6k, 15k..40k

	stop_tape_count		= 0;
	is_saving			= no;


// allocate stuff:

	uptr bu; 
	ulong size;
	cstr err;
	
	rom.Grow(model->rom_size);		
			err = load_file( CatStr(rsrc_path,model->rom_filename), bu, size ); 
			XTRAP( err );
			XTRAP( size!=model->rom_size );
			for( uint i=0; i<size; i++ ) { rom[i] = bu[i]; }		// also clears all flags
			XXXASSERT(model->contended_rompages==0);				// no model with contended rom pages

	ram.Grow(model->ram_size);
			//ram.Clear();											// flags nullen		
			long bits = model->contended_rampages;					// contended memory flaggen
			for( int i=0,e=0; i<model->ram_pages && (bits<<i)!=0; i++ )	
			{
				int j = e; e += model->page_size;
				if( (bits<<i) < 0 ) { while( j<e ) { ram[j++] |= cpu_waitmap; } }
			}
			
			for( uint i=0; i<NELEM(model->videoram_size); i++ )		// videoram flaggen
			{
				XXXTRAP( model->videoram_start[i]+model->videoram_size[i] > model->ram_size );
				ushort* v = &ram[model->videoram_start[i]];
				for( ulong j=0, e=model->videoram_size[i]; j<e; j++ ) v[j] |= cpu_crtc;
			}

// Items
	cpu = new Z80(this);
	ula = new Ula(this);
	mmu = new_mmu_for_this_machine();
	if( model->has_ay_soundchip ) { new_ay_for_this_machine(); }
	if( modl>=zxplus2 && modl<tc2048 )	
	{
		zxplus2_js_ports = new ZxPlus2JoystickPorts(this);
		if( [preferences getBoolForKey:"newmachine_open_builtin_js_panel"] ) [zxplus2_js_ports_panel showPanel];
	}
	if( modl==tc2048 )					  
	{
		kempston_joystick_interface = new KempstonJoystickInterface(this);
		if( [preferences getBoolForKey:"newmachine_open_builtin_js_panel"] ) [kempston_jsif_panel showPanel];
	}
	
	ForAllItems( PowerOn(time_dsp_start,cc_dsp_start) );

// attach screen / open window
	if(!screen) { screen = [[Screen alloc] initWithMachine:this]; ASSERT(screen); }
	set_title();

// show image of machine:					evtl.: preferences.new_machine_open_machine_image
//	if( [preferences getBoolForKey:"new_machine_open_machine_image"] ) [machine_panel showPanel];	
}


/* --------------------------------------------------------
	init with default model
	note: Machine.screen is preserved and must be cleared to NULL by caller if not yet allocated
*/
void Machine::init_with_default_model()
{
	XLogIn( "Machine:init_with_default_model" );
	init_with_model( [preferences getIntForKey:"newmachine_model"] );
}


/* --------------------------------------------------------
	init with snapshot
	note: Machine.screen is preserved and must be cleared to NULL by caller if not yet allocated
*/
cstr Machine::init_with_file( cstr file_path )
{
	XLogIn("Machine:init_with_file: \"%s\"", file_path);

	ulong file_ext = Peek4X( CatStr(LowerStr(ExtensionFromPath(file_path)),"  ") );
	model    = NULL;	// => detect whether init_with_model was already called by init_with_sna/init_with_z80
	cstr err = NULL;	

	switch(file_ext)
	{
	case '.o  ':
	case '.80 ':	err = init_with_80(file_path); break;
	case '.p  ':
	case '.81 ':	err = init_with_81(file_path); break;	
	case '.sna':	err = init_with_sna(file_path); break;
	case '.z80':	err = init_with_z80(file_path); break;
	case '.rom':	err = init_with_rom(file_path); break;
	case '.scr':	err = init_with_scr(file_path); break;
	case '.ass':	err = init_with_ass(file_path); break;
	case '.src':	err = init_with_ass(file_path); break;
	case '.tap':	{ init_with_model(zx48i3); OSStatus e = LoadFile(file_path); err=e?ErrorStr(e):NULL; break; }
	case '.tzx':	err = "tzx files must be loaded from the tape recorder"; break;
	default:		err = "unknown snapshot format"; break;
	}

	if (err) 
	{
		if(!model) init_with_default_model();
		return err;
	}
	else
	{
		filepath = NewCopy(file_path);	// durable copy
		set_title();
		return NULL; /*ok*/
	}
}


Machine::Machine()
:	HierObj("Machine"),
	lock(PLock::recursive)
{
	XLogIn("new Machine from preferences");

	Lock();
		screen = NULL;
		init_with_default_model();
		load_preferences();
		machine_list.LinkMachine(this);
	Unlock();
	ResumeCpu();	
}

Machine::Machine (uint model)
:	HierObj("Machine"),
	lock(PLock::recursive)
{
	XLogIn("new Machine: %i",(int)model);

	Lock();
		screen = NULL;
		init_with_model (model);
		load_preferences();
		machine_list.LinkMachine(this);
	Unlock();
	ResumeCpu();
}

Machine::Machine (cstr file_path)
:	HierObj("Machine"),
	lock(PLock::recursive)
{
	XLogIn("new Machine: \"%s\"",file_path);

	Lock();
		screen = NULL;
		cstr err = init_with_file(file_path); if(err) ShowWarning("Machine:Machine: error: ",err);
		load_preferences();
		machine_list.LinkMachine(this);
	Unlock();
	ResumeCpu();
}


void Machine::SetModel (uint modl)
{
	XLogIn("Machine:SetModel");

	HaltCpu();				// -> prevent runForSound blocking on run_lock
	Lock();					// -> prevent forced update in drawScreen
		dealloc_items();
		init_with_model(modl);
		load_preferences();
	Unlock();	
	ResumeCpu();
}


OSStatus Machine::LoadFile( cstr file_path )
{
	XLogIn("Machine:LoadFile: \"%s\"",file_path);
	XXXASSERT(cpu_state&7);		// cpu must be halted

	ulong file_ext = Peek4X( LowerStr(ExtensionFromPath(file_path)) );

	switch( file_ext )
	{
	case '.tap':
		{
		// load tapeloader snapshot, patch tape loader
			static cstr ld_name[zxmodels] =
			{	0,0,0,0,											// zx80, zx81, jupiter, sep1,
				"16k_i1","48k_i2","48k_i3","48k_plus","48k_inves",	// zx16k, zx48i2, zx48i3, zxplus, zxplus_span, 
				"128k","128k_sp",									// zx128, zx128_span, 
				"plus2","plus2_sp","plus2_fr",						// zxplus2, zxplus2_span, zxplus2_french,
				"plus3","plus3_sp",									// zxplus3, zxplus3_span, 
				"plus2A","plus2A_sp",								// zxplus2a, zxplus2a_span,
				0,"tc2048","ts2068","pentagon","scorpion",			// sep2, tc2048, ts2068, pentagon, scorpion,
				0,0,												// sep3, TODO:samcoupe,
			};
			
			uint modl   = model->model;
			cstr loader = ld_name[modl];
			if( !loader ) 
			{
				ShowAlert( Using("Machine:LoadFile:\n  Tape loader for %s is not yet implemented.",model->name) );
				return error;
			}
			loader = Using( "%sload_tape_%s.z80", rsrc_path, loader );
			
			HaltCpu();
			Lock();					// -> prevent forced update in drawScreen
			
				dealloc_items();
				cstr errmsg = init_with_z80(loader);
				if (errmsg) 
				{
					ShowAlert( Using("Machine:LoadFile:\n  Loading Tape Loader for %s failed:\n%s",model->name,errmsg) );
					init_with_model(modl); 
					load_preferences();
				}
				else
				{
					delete[] filepath;
					filepath = NewCopy(file_path);		// durable copy
					set_title();
					load_preferences();

					instant_load_fd = open( file_path, O_RDONLY, 0 ); 
					if( instant_load_fd==-1 ) 
					{
						errmsg = "open file failed"; 
						PowerOn();
						load_preferences();
					}
					else
					{
						SetInstantLoadFromTape(file_path);				// TODO: temporarily only, and don't set prefs...
						SetAutostartTape(instant_load_fd!=-1);	// TODO: temporarily only, and don't set prefs...
					}
				}

			Unlock();	
			ResumeCpu();
			
			return errmsg?error:ok;
		}

	case '.rom':
		{
			// try to load the rom into the running machine
			uchar bu[64 kB +1];
			int fd = open(file_path,O_RDONLY,0); 
			if(fd==-1) goto r;				// file does not exist => init_with_snapshot() wird den Fehler behandeln
			ulong size = read(fd,bu,64 kB +1);
			close(fd);	
			if(size==model->rom_size) 
			{
				Lock();
					delete[]filepath; filepath=NewCopy(file_path); set_title();
					for( uint i=0; i<size; i++ ) rom[i] = bu[i];		// also clears all flags
					Reset();						
				Unlock();
				return ok;
			}
			else
			{
				ShowWarning( Using("Machine:LoadFile \"%s\":\n  file size does not match current machine's rom\n",file_path) );
				return error;
			}
		}

	case '.scr':
		{					// load the screenshot into the running machine
			uchar bu[24*32*(8+1) +1];
			int fd = open(file_path,O_RDONLY,0);
			if(fd==-1) goto r;				// file does not exist => init_with_snapshot() wird den Fehler behandeln
			ulong size = read(fd,bu,24*32*(8+1) +1);
			close(fd);	
			if(size==24*32*(8+1)) 
			{
				Lock();
					ushort* vram = ula->GetVideoRam();
					for( uint i=0; i<size; i++ ) vram[i] = (vram[i]&0xff00) + bu[i];		// preserve flags
				Unlock();
				return ok;
			}
			else
			{
				ShowWarning( Using("Machine:loadFile \"%s\":\n  file size does not match screen$\n",file_path) );
				return error;
			}
		}
		
	case '.gif':
		{
			LogGifFile(file_path);
			return ok;
		}
	}
	
r:	HaltCpu();
	Lock();				// -> prevent forced update in drawScreen
		dealloc_items();
		cstr errmsg = init_with_file(file_path);
		load_preferences();
	Unlock();	
	ResumeCpu();

	if(errmsg) { ShowWarning( Using("Machine:LoadFile \"%s\":\n  %s\n",file_path,errmsg) ); return error; }
	else return ok;
}


void Machine::ReloadSnapshot()
{
	XLogIn("Machine:ReloadSnapshot");

	cstr path = filepath ? DupStr(filepath) : GetRecentFilePath();
	OSStatus err = error;

	if( path ) 
	{
		/* LBLOCK */ HaltCpu(); Lock();	
			err = LoadFile(path); 
		/* LUNBLOCK */ Unlock(); ResumeCpu();
	}
	else 
	{
		PowerOn();
	}
}


void Machine::SaveFile (cstr path)
{
	XLogIn("Machine:SaveFile: \"%s\"",path);
	
	OSStatus e = ok;
	cstr file_ext  = CatStr( LowerStr(ExtensionFromPath(path)), "   " );

	Lock();
		switch( Peek4X(file_ext) )
		{
		case '.80 ': 
		case '.o  ': e = Save_80(path); break;
		case '.81 ': 
		case '.p  ': e = Save_81(path); break;
		case '.sna': e = SaveSNA(path); break;
		case '.z80': e = SaveZ80(path); break;
		case '.rom': e = SaveROM(path); break;
		case '.scr': e = SaveSCR(path); break;
		case '.gif': e = [screen saveScreenshot:path]; break;
		default:     e = notyetimplemented; break;
		}
	Unlock();
	
	if (e) ShowWarning( (ErrorStr(e)) ); 
}


void Machine::SaveScreenshot( cstr path )
{
	XLogIn("Machine:SaveScreenshot: \"%s\"", path);
	if(!path) path = CatStr( FullPath("~/Desktop/",1),"Screen_",strchr(DateStr(CurrentTime()),' ')+1,".gif" );
	OSStatus e = [screen saveScreenshot:path];
	if (e) ShowWarning( (ErrorStr(e)) ); 
}


void Machine::RecordMovie( cstr path )
{
	if(!path) path = CatStr( FullPath("~/Desktop/",1),"Movie_",strchr(DateStr(CurrentTime()),' ')+1,".gif" );
	OSStatus e = [screen recordMovie:path]; 
	if (e) ShowWarning( (ErrorStr(e)) ); 
}



// ======================================================================
//					RUNNING THE MACHINE
// ======================================================================
 

/*	CPU callback: output instruction
*/
void Machine::OutputAtCycle (long cc, ushort addr, uchar byte)
{
	if( (addr&0xc001)==0x4001 ) cc = ula->AddSpecialWaitStates(cc);

	Time now = time_dsp_start + double(cc-cc_dsp_start) / cpu_cycles_per_second;	

	if( now >= seconds_per_dsp_buffer_plus_stitching )		// fake real time if overshooting too far
		now = seconds_per_dsp_buffer_plus_stitching-1e-6;	// e.g. if throttled too low (<300kHz) or single stepped

	XXXXASSERT( FirstItem() );

	ZxItem* p = FirstItem();
	for(;;) 
	{
		if( p->out_mask ) 
		{
			if ( (addr & p->out_mask) == p->out_bits ) p->Output( now, cc, addr, byte ); 
			p = p->Next(); continue;
		}
		if( p->in_mask ) { p = p->Next(); continue; }
		break;
	};
}


/*	CPU callback: input instruction
*/
uchar Machine::InputAtCycle (long cc, ushort addr)
{
	if( (addr&0xc001)==0x4001 ) cc = ula->AddSpecialWaitStates(cc);

	Time now = time_dsp_start + double(cc-cc_dsp_start) / cpu_cycles_per_second;	

	if( now >= seconds_per_dsp_buffer_plus_stitching )		// fake real time if overshooting too far
		now = seconds_per_dsp_buffer_plus_stitching-1e-6;	// e.g. if throttled too low (<300kHz) or single stepped

	uchar c = 0xff;		// input byte, 0-bits override colliding 1-bits
	uchar m = 0x00;		// bits actually set by extensions

	XXXXASSERT( FirstItem() );

	ZxItem* p = FirstItem();
	for(;;) 
	{
		if( p->in_mask ) 
		{
			if ( (addr & p->in_mask) == p->in_bits ) p->Input( now, cc, addr, c, m ); 
			p = p->Next(); continue;
		}
		if( p->out_mask ) { p = p->Next(); continue; }
		break;
	};

	return m==0xff ? c								// all bits set by external item
		: (c&m) | (ula->Input(now,cc,addr)&~m);		// else sneak missing bits from ula bus
}


	
/* ----	The Main Thing ----

	Run the machine until the end time of the dsp sample buffer is reached
	Unavoidable overshoot is allowed:
		DSP_SAMPLES_STITCHING provides the (fully handled) overshoot buffer
		DSP_BUFFER_GUARDZONE  provides an additional (discarded) overshoot zone for security only!

		standard values: 4 samples @ 44.1kHz  <=>  90.7 µs  <=> 317 T @ 3.5 MHz  <=>  90 T @ 1 MHz  <=>  1 T @ 11 kHz

	real-world and virtual time are 'connected' at:
		real-world time:	self->sampletime_dsp_start
			this is the exact time when the cpu finished last runForSound, including any unavoidable overshoot.
			it is not the start time for the current dsp buffer (in which case it would always be 0.0) 
			but is slightly ahead (-> overshoot).

		virtual time:		self->cc_dsp_start  
			taken from cpu->CpuCycle()
			this is the cpu cycle when the cpu finished last runForSound, including any unavoidable overshoot.
			it corresponds not with the start time for the current dsp buffer (in which case it would be fractional)
			but is slightly ahead (-> overshoot).
*/
void Machine::RunForSound()
{
	if( IsHalted() )				// don't lock the machine while halted: may be halted for longish work!
	{			
		if( stop_tape_count>0 )		// stop tape if auto-playing:
		{ 
			[tape_recorder pause];	// stop if playing, stop_recording if recording
			stop_tape_count = 0; 
		}
	}
	else  // if( IsRunning() )
	{		
		Lock();
		if( IsRunning() ) { run_for_sound(); }
		Unlock();
	}		
}

void Machine::run_for_sound()
{
	XXXLogIn("Machine:run_for_sound");


	long& cc = cpu->CpuCycle();
		  cc_dsp_start = cc;
	long  cc_dsp_end   = cc + 1 + long( (seconds_per_dsp_buffer - time_dsp_start) * cpu_cycles_per_second );
	long  cc_per_frame = ula->GetCpuCyclesPerFrame();
	long  cc_end;
	bool  preempted = no;

	do
	{
		cc_end = ula->GetWaitmapStart(); if( cc_end>cc_dsp_end ) cc_end=cc_dsp_end;
		if ( cc < cc_end ) if( preempted = cpu->Run(cc_end,cpu_options&~(cpu_waitmap|cpu_crtc)) ) goto c;
		cc_end = ula->GetWaitmapEnd(); if( cc_end>cc_dsp_end ) cc_end=cc_dsp_end;
		if ( cc < cc_end ) if( preempted = cpu->Run(cc_end,cpu_options) ) goto c;
		cc_end = cc_per_frame; if( cc_end>cc_dsp_end ) cc_end=cc_dsp_end;
		if ( cc < cc_end ) preempted = cpu->Run(cc_end,cpu_options&~(cpu_waitmap|cpu_crtc));		

	c:	if ( cc>=cc_per_frame )
		{
			ula->DoFrameFlyback();
			
			cc				-= cc_per_frame;
			cc_dsp_end		-= cc_per_frame;
			cc_dsp_start	-= cc_per_frame;
			
			if(this==front_machine()) add_frame_hit([screen isNotDirty]);	
			[screen nextFrame];			// trigger drawing
			AutoStartStopTape();
		}
	}
	while ( cc < cc_dsp_end && !preempted );
	
	if( cc>=cc_dsp_end /* && !preempted */ )
	{
		ForAllItems( DspBufferEnd(seconds_per_dsp_buffer,cc) );
		time_dsp_start += double(cc-cc_dsp_start) / cpu_cycles_per_second - seconds_per_dsp_buffer;
		XXXTRAP( time_dsp_start<0.0 );
	}
}
	


void Machine::AutoStartStopTape()
{
	/*	Anmerkungen zum Auto start/stop Feature:
		Das Band wird eingeschaltet, wenn sehr viele INs auf den ULA-Port gemacht werden. 
		Die Tape-Routine erzeugt typischerweise ~650 .. 850 pro 50Hz-Frame.
		Während sie auf den Pilotton wartet, sind es etwa 1150. 
		Danach kommt eine 1-sekündige Ausblende völlig ohne INs. :-(((((      
		INs während der Pilotphase sind etwa 950 pro Frame.
		Und danach erwartet sie, dass sie sofort Pilotton vorfindet um sich darauf zu synchronisieren.
		Wenn nicht, beginnt sie wieder ganz von vorne. 
		Das Band darf also keinesfalls sofort ausgeschaltet werden, wenn kein Input mehr erfolgt,
		sondern es muss die Ausblende durchlaufen! 
		stop_tape_count = 50 (1 sec) war etwas zu niedrig.
		Der TapeRecorder liefert 2 sec Stille + 2 sec Pilotton. 
		stop_tape_count = 75 (1.3 sec) scheint gut zu funktionieren.			
	*/

	double f = cpu_cycles_per_second / (ula->GetCpuCyclesPerFrame() * 50.0); 
	double ear_out_count = f * ula->EarOutCount();	ula->EarOutCount() = 0;
	double ear_in_count  = f * ula->EarInCount();	ula->EarInCount()  = 0; 

	if( stop_tape_count>0 )				// tape currently auto-running
	{
		if(is_saving)					// currently saving
		{
			if( ear_out_count<30 )		// end of saving
			{
				if( --stop_tape_count==0 && [tape_recorder isRecording] ) [tape_recorder pause];
			}
			else						// still saving
			{
				stop_tape_count = 50;
			}
		}
		else							// currently loading
		{
			if( ear_in_count<500 )		// end of loading
			{
				if( --stop_tape_count==0 && [tape_recorder isPlaying] ) [tape_recorder pause];
			}
			else						// still loading
			{
				stop_tape_count = 75;
			}
		}
	}
	
	else								// tape currently stopped (at least not auto-running)
	{
		if( ear_in_count>800 )			// started loading?
		{
			if( tape_options&autostart_tape )	// feature enabled?
			{
				[tape_recorder orderFront];		// unschön...
				if( [tape_recorder isLoaded] && [tape_recorder isStopped] ) [tape_recorder startPlaying]; 
				stop_tape_count = 75;
				is_saving = no;
			}
		}
		else if( ear_out_count>30 && ear_out_count<35 )	// started saving?  ((pilot == 32 .. 33))
		{
			if( tape_options&autostart_tape )	// feature enabled?
			{
				[tape_recorder orderFront]; 
				if( [tape_recorder isLoaded] && [tape_recorder isStoppedRecording] ) [tape_recorder startRecording];
				stop_tape_count = 50;
				is_saving = yes;
			}
		}
		
	}
}



void Machine::clear_break_ptr()
{
	break_ptr = cpu_options & cpu_break_x & *cpu->RdPtr(cpu->GetRegisters().rr.pc) 
			? cpu->RdPtr(cpu->GetRegisters().rr.pc) 
			: NULL; 
}


/* Laufe für (mindestens) cc cpu cycles.
*/
void Machine::RunCpuCycles (long cc)
{
	if(cc<=0) return;
	
	XXASSERT( IsHalted() );

	Lock();			// prevent forced update in draw screen thread

		cc_dsp_start = cpu->CpuCycle();
		
	// bei extrem niedergetakteter Maschine würde runForSound einige Aufrufe lang nichts tun
		while( time_dsp_start >= seconds_per_dsp_buffer ) 
		{
			ForAllItems( DspBufferEnd(seconds_per_dsp_buffer,cc_dsp_start) );
			time_dsp_start -= seconds_per_dsp_buffer;
		}

		Frequency c = cpu_cycles_per_second;
		
		time_dsp_start = 0.0;
		cpu_cycles_per_second = double(cc) / seconds_per_dsp_buffer;

		run_for_sound();
		time_dsp_start = 0.0;		// die kann z.B. bei cc=1 extrem in die Zukunft zeigen
		
		cpu_cycles_per_second = c;

	Unlock();
}

void Machine::StepOver()
{ 
	XLogIn("Machine:StepOver"); 

/*	Vorbedingung: state == cpu_halted
*/
	if( IsHalted() )
	{
		clear_break_ptr();

		ushort pc = cpu->GetRegisters().rr.pc;		// pc
		uchar  o  = cpu->Peek(pc);					// Opcode
		uchar  f  = cpu->GetRegisters().r.f;		// Flags
		
		switch(o)
		{
		case CALL_NZ: if(f&Z_FLAG) goto s; else goto c;
		case CALL_Z:  if(f&Z_FLAG) goto c; else goto s;
		case CALL_NC: if(f&C_FLAG) goto s; else goto c;
		case CALL_C:  if(f&C_FLAG) goto c; else goto s;
		case CALL_PO: if(f&P_FLAG) goto s; else goto c;
		case CALL_PE: if(f&P_FLAG) goto c; else goto s;
		case CALL_P:  if(f&S_FLAG) goto s; else goto c;
		case CALL_M:  if(f&S_FLAG) goto c; else goto s;
				
		case RST00:	case RST08: case RST10: case RST18: 
		case RST20: case RST28: case RST30: case RST38:
		case CALL:	
	c:		cpu->stack_breakpoint = cpu->GetRegisters().rr.sp;
			cpu_options |= cpu_break_sp; 
			ResumeCpu(); 
			break; 

		case HALT:
		// Bis Interrupt abarbeiten, halt bei Interrupt auch wenn Interrupts gesperrt
		// Halt wenn Breakpoint aktiviert wird ((kann nur ein cpu_break_x genau hier auf pc sein))
			if( cpu->CpuCycle()>=ula->GetCpuCycleOfIrptEnd() ) RunCpuCycles(ula->GetCpuCyclesPerFrame()-cpu->CpuCycle());
			if(break_ptr==NULL) RunCpuCycles (ula->GetCpuCycleOfInterrupt()-cpu->CpuCycle());
			if(break_ptr==NULL) goto s; else break;
				
		case PFX_ED:
		// Blockbefehle komplett abarbeiten:
		// Halt bei Interrupt, wenn der Befehl überschrieben oder ausgeblendet wird
		// Halt wenn Breakpoint aktiviert wird
			o = cpu->Peek(cpu->GetRegisters().rr.pc +1);	// ED Opcode
		//	if( (o&~0x0b) != LDIR ) goto s;					// kein Blockbefehl  ((Test nicht nötig))
			do RunCpuCycles(1); 
			while( break_ptr==NULL && cpu->GetRegisters().rr.pc==pc && cpu->Peek(pc)==PFX_ED && cpu->Peek(pc+1)==o );
			break;
			
		default:
	s:		RunCpuCycles(1);
			break;
		}
	}	
}


void Machine::StepOut()					
{ 
	XLogIn("Machine:StepOut"); 

	if( IsHalted() )
	{
		clear_break_ptr();

		cpu->stack_breakpoint = cpu->GetRegisters().rr.sp +2;
		cpu_options |= cpu_break_sp; 
		ResumeCpu(); 
	}
}


void Machine::StepIn()
{ 
	XLogIn("Machine:StepIn"); 

/*	Bei "Step In" wird immer exakt ein Befehl ausgeführt 
	(oder ausnahmsweise Interrupt/NMI-Behandlung gestartet)
	
	Dazu wird runForSound so aufgerufen, dass runForSound nur einen Befehl lang läuft.
	Das wird durch kurzzeitige Veränderung der Taktfrequenz erreicht.

	Die Maschine wird in den Audio-Out-Puffer schreiben. Da der DSP-Interrupt aber auf
	höherer Priorität läuft, wird er unsere Maschine unterbrechen, den Audio-Out-Puffer
	wie immer löschen, die restlichen Maschinen und den Kassettenrekorder abarbeiten 
	und den Puffer in den OS-Puffer kopieren. Von unserem Sound bleibt da nichts übrig. Hoffentlich.
	
	Das ist insbes. wichtig, weil bei extrem niedergetakteter Maschine für einen Schritt evtl.
	erst mehrmals DspBufferEnd() aufgerufen werden muss. 

	Der Single-Stepper funktioniert logischerweise nur, wenn die Cpu angehalten ist.
*/
	if (IsHalted())
	{
		clear_break_ptr();
		RunCpuCycles(1);
	}
}


// trigger NMI
void Machine::Nmi()
{
	XLogIn("Machine:Nmi"); 
	cpu->TriggerNmi();
}


// reset machine. does not change cpu state or ula settings
void Machine::Reset()					
{ 
	XLogIn("Machine:Reset"); 
	if(1)							// reset filepath & fenstertitel
	{								// => 'reload snapshot' unmöglich
		delete[] filepath; filepath = NULL;
		set_title();
	}
	Lock();							// == QBLOCK  ((quick modify machine from main thread))
		ForAllItems( Reset(0.0,cpu->CpuCycle()) );		// <-- TODO: die Zeiten sollten vielleicht besser stimmen ?!?
		cpu_options &= ~cpu_break_sp;
		clear_break_ptr();
	Unlock();
}


// power-on reset machine 
// reset items
// restore timings
// clear break points
//
void Machine::PowerOn()
{
	XLogIn("Machine:PowerOn");
	
	XTRAP( !model || model->model>=zxmodels );
	XTRAP( !cpu || ram.Count()!=model->ram_size || rom.Count()!=model->rom_size || !screen || !ula );	

	HaltCpu();
	Lock();

		delete[] filepath; filepath = NULL;
		set_title();

		stop_tape_count		= 0;
		is_saving			= no;
		break_ptr			= NULL;
		cpu_options		   &= ~cpu_break_sp;
		ClearAllBreakpoints();
		
	// reset timing:
		cc_dsp_start		= 1000;		// possible value for buggy +3 reset code		2007-05-01
		time_dsp_start		= 0.0;

		cpu_cycles_per_second = model->cpu_cycles_per_second;
		{ Ay* ay=GetAy(); if(ay) ay->SetClock(model->psg_cycles_per_second); }

	// power-cycle items:
		ForAllItems( PowerOn(time_dsp_start,cc_dsp_start) );
		
	Unlock();
	ResumeCpu();
}


void Machine::SetBreakpoint (ushort* brkPtr, int mode)	// TODO: ändern auf page+index
{
	mode &= cpu_break_rwx; if(!mode) return;

	XXTRAP( (brkPtr<ram.Data() || brkPtr>=ram.Data()+ram.Count()) && (brkPtr<rom.Data() || brkPtr>=rom.Data()+rom.Count()) );	

	Lock();

		cpu_options |= mode;						// enable appropriate breakpoint checking
		*brkPtr |= mode;
		int i; for( i=0; i<breakpoints; i++ ) if( breakpoint[i]==brkPtr ) goto x;
		if( (uint)i<NELEM(breakpoint) ) breakpoint[breakpoints++] = brkPtr;	

x:	Unlock();
}


void Machine::ClearBreakpoint (ushort* brkPtr, int mode)
{
	mode &= cpu_break_rwx; if(!mode) return;

	Lock();

		int o = 0;

		int i; for( i=0; i<breakpoints; i++ ) 
		{
			if( breakpoint[i]==brkPtr )				// implied: brkPtr valid
			{ 
				*brkPtr &= ~mode;
				if( (*brkPtr & cpu_break_rwx) == 0 ) { breakpoint[i] = breakpoint[--breakpoints]; }
				break;
			}
			o |= *breakpoint[i];
		}
		for( ; i<breakpoints; i++ )
		{
			o |= *breakpoint[i];
		} 

		cpu_options &= ~cpu_break_rwx;
		cpu_options |= o & cpu_break_rwx; 

	Unlock();
}


void Machine::ClearAllBreakpoints()
{
	Lock();

		cpu_options &= ~cpu_break_rwx;				// disable breakpoint checking
		while(breakpoints) *breakpoint[--breakpoints] &= ~cpu_break_rwx;
		
	Unlock();
}


void Machine::set_cpu_clock_and_predivider( Time new_clock/*cycles per second*/, int new_div )
{
	Limit( 1.0, new_clock, 2e7 );	

	Lock();
		ula->SetCpuClockPredivider(new_div);	// ula validates new cpu clock predivider
		
		cpu_cycles_per_second = new_clock;

	// bei extrem niedergetakteter Maschine würde runForSound einige Aufrufe lang nichts tun
		while( time_dsp_start >= seconds_per_dsp_buffer ) 
		{
			ForAllItems( DspBufferEnd(seconds_per_dsp_buffer,cc_dsp_start) );
			time_dsp_start -= seconds_per_dsp_buffer;
		}
	Unlock();
}


void Machine::SetCpuCyclesPerSecond( Time new_clock )
{
	set_cpu_clock_and_predivider( new_clock, ula->GetCpuClockPredivider() );
}


/*	Set machine overdrive
	• keep cpu clock predivider 
	• change ula clock and cpu clock
*/
void Machine::SetMachineOverdrive( double factor )
{
	SetCpuCyclesPerSecond( model->ula_cycles_per_second * factor / ula->GetCpuClockPredivider() );
}


/*	Set CPU clock predivider in Ula
	• keep ULA clock
	• change the resutling CPU clock
*/
void Machine::SetCpuClockPredivider( int new_div )
{
	set_cpu_clock_and_predivider( cpu_cycles_per_second * ula->GetCpuClockPredivider() / new_div, new_div );
}


void Machine::SetFullscreen( bool fs )
{
	XLogIn("Machine:SetFullscreen: %s",onoffstr(fs));
	
	bool old = [screen isFullscreen];
	if(old==fs) { XLogLine("no change!"); return; }
	
	[screen setFullscreen:fs];
}


void Machine::ToggleInternalAy()
{
	XLogIn("Machine:ToggleInternalAy");

	HaltCpu();
	Lock();
		Ay* ay = (Ay*)FindItem(isa_InternalAy);
		if( ay ) { delete ay; ay=NULL; }
		else 
		{	ay = new_ay_for_this_machine();
			ay->PowerOn( time_dsp_start, cc_dsp_start );
		}
		if( filepath ) [preferences setBool:ay!=NULL forFile:filepath andKey:"newsnapshot_soundchip"];
	Unlock();
	ResumeCpu();
}

void Machine::ToggleFullerBox()
{
	XLogIn("Machine:ToggleFullerBox");
	HaltCpu();
	Lock();
		if( fuller_box ) delete fuller_box; 
		else			
		{	fuller_box = new FullerBox(this); 
			fuller_box->PowerOn( time_dsp_start, cc_dsp_start ); 
		}
		if( filepath ) [preferences setBool:fuller_box!=NULL forFile:filepath andKey:"newsnapshot_fullerbox"];
	Unlock();
	ResumeCpu();
}

void Machine::ToggleZonxBox()
{
	XLogIn("Machine:ToggleZonxBox");
	HaltCpu();
	Lock();
		if( zonx_box ) delete zonx_box;  
		else			
		{
			uint zonz = model->model < zx16k ? Zonx81 : Zonx;
			zonx_box = new ZonxBox(this,zonz); 
			zonx_box->PowerOn( time_dsp_start, cc_dsp_start ); 
		}
		if( filepath ) [preferences setBool:zonx_box!=NULL forFile:filepath andKey:"newsnapshot_zonx"];
	Unlock();
	ResumeCpu();
}

void Machine::ToggleKempstonJoystickInterface()
{
	XLogIn("Machine:ToggleKempstonJoystickInterface");
	HaltCpu();
	Lock();
		if( kempston_joystick_interface ) delete kempston_joystick_interface;
		else
		{
			kempston_joystick_interface = new KempstonJoystickInterface(this);
			kempston_joystick_interface->PowerOn( time_dsp_start, cc_dsp_start ); 				
		}
		if( filepath ) [preferences setBool:kempston_joystick_interface!=NULL forFile:filepath andKey:"newsnapshot_joystick"];
	Unlock();
	ResumeCpu();
}

void Machine::ToggleKempstonMouseInterface()
{
	XLogIn("Machine:ToggleKempstonMouseInterface");
	HaltCpu();
	Lock();
		if( kempston_mouse_interface ) delete kempston_mouse_interface;
		else
		{
			kempston_mouse_interface = new KempstonMouseInterface(this);
			kempston_mouse_interface->PowerOn( time_dsp_start, cc_dsp_start ); 				
		}
		if( filepath ) [preferences setBool:kempston_mouse_interface!=NULL forFile:filepath andKey:"newsnapshot_mouse"];
	Unlock();
	ResumeCpu();
}

void Machine::ToggleSinclairZxInterface2()
{
	XLogIn("Machine:ToggleSinclairZxInterface2");		
	HaltCpu();
	Lock();
		if( sinclair_zx_interface2 ) delete sinclair_zx_interface2;
		else
		{
			sinclair_zx_interface2 = new SinclairZxInterface2(this);
			sinclair_zx_interface2->PowerOn( time_dsp_start, cc_dsp_start ); 				
		}
		if( filepath ) [preferences setBool:sinclair_zx_interface2!=NULL forFile:filepath andKey:"newsnapshot_interface2"];
	Unlock();
	ResumeCpu();
}

void Machine::ToggleSinclairZxPrinter()
{
	XLogIn("Machine:ToggleSinclairZxPrinter");
	HaltCpu();
	Lock();
		if( sinclair_zx_printer ) delete sinclair_zx_printer;
		else
		{
			sinclair_zx_printer = new SinclairZxPrinter(this);
			sinclair_zx_printer->PowerOn( time_dsp_start, cc_dsp_start ); 				
		}
		if( filepath ) [preferences setBool:sinclair_zx_printer!=NULL forFile:filepath andKey:"newsnapshot_zxprinter"];
	Unlock();
	ResumeCpu();
}

void Machine::ToggleSinclairZxInterface1()
{
	XLogIn("Machine:ToggleSinclairZxInterface1");
	HaltCpu();
	Lock();
		if( sinclair_zx_interface1 ) delete sinclair_zx_interface1; 
		else
		{
			sinclair_zx_interface1 = new SinclairZxInterface1(this);
			sinclair_zx_interface1->PowerOn( time_dsp_start, cc_dsp_start ); 				
		}
		if( filepath ) [preferences setBool:sinclair_zx_interface1!=NULL forFile:filepath andKey:"newsnapshot_interface1"];
	Unlock();
	ResumeCpu();
}




// ==============================================================================
//					CALL BACKS FROM THE CPU
// ==============================================================================


/*	Rom Patch:
	in:  current cc, pc and instruction read at pc
		 all registers saved into cpu->registers
	out: cpu running: instruction to execute instead 
		 cpu halted:  Z80.Run() will exit
*/
uchar Machine::PatchAtCycle( long cc, ushort pc, uchar instr )
{
	XLogIn( "Machine:PatchAtCycle: pc=$%04X",(uint)pc );

	ushort*  pc_ptr = cpu->RdPtr(pc);
	long rom_offset = pc_ptr - rom.Data();

// tape load routine entry in rom:
// TODO: the actual rom must be determined (e.g. ZX81 rom in ZX80 or vice versa)
	if( rom_offset == model->tape_load_routine )
	{
		if( instant_load_fd != -1 )
		{
			try	// catch file error
			{
			// load block:
				switch( model->model )
				{
				case zx80:	LoadZX80TapeBlock( instant_load_fd, instant_load_filetype ); break;
				case zx81:	LoadZX81TapeBlock( instant_load_fd, instant_load_filetype ); break;
				default:	LoadZxspTapeBlock( instant_load_fd, instant_load_filetype ); break;
				}

			// close file if it can only hold one block or if near eof:
				switch( instant_load_filetype )
				{
				default:
				{	off_t flen = FileLength(instant_load_fd);				// if(flen<0) throw(file_error(errno));
					off_t fpos = lseek( instant_load_fd, 0, SEEK_CUR	);	// if(fpos<0) throw(file_error(errno));	
					if( flen>=0 && fpos>=0 && flen-fpos > 2 ) break;
				}
				case zx80_file:	case o_file:
				case zx81_file:	case p_file:	
					close_file( instant_load_fd, nothrow ); instant_load_fd = -1;
					delete[] instant_load_filename;	instant_load_filename = NULL;
					tape_options &= ~instant_load_from_tape;
					break;
				}

			// return:
				if( tape_options & halt_cpu_after_block_loaded ) { HaltCpu(); return NOP; }
				Z80Regs& regs = cpu->GetRegisters();
				return cpu->Peek(regs.rr.pc);
			}
			catch( bad_alloc ) 
			{
				ShowAlert("run_for_sound:RomPatch:LoadTapeBlock:\n%s","out of memory");  				
				close( instant_load_fd ); instant_load_fd = -1;
				delete[] instant_load_filename; instant_load_filename = NULL;
				tape_options &= ~instant_load_from_tape;
				goto a;	// open tape recorder
			}
			catch( any_error e )
			{
				ShowAlert("run_for_sound:RomPatch:LoadTapeBlock:\n%s",e.Text() );
				close( instant_load_fd ); instant_load_fd = -1;
				delete[] instant_load_filename; instant_load_filename = NULL;
				tape_options &= ~instant_load_from_tape;
				goto a;	// open tape recorder
			}
		}
		else
		{
	a:		tape_state = tape_options & (autostart_tape | halt_cpu_after_block_loaded);
			if( tape_options & autostart_tape )
			{
				if( [tape_recorder isStopped] && [tape_recorder isLoaded] ) [tape_recorder startPlaying];
			}
			if( [tape_recorder isStopped] || ![tape_recorder isLoaded] )
			{
				[tape_recorder orderFront];
			}
		}
	}
	
// tape save routine entry in rom:
	else if( rom_offset == model->tape_save_routine )
	{
		if( instant_save_fd != -1 )
		{
			try
			{
				switch( model->model )
				{
				case zx80:	SaveZX80TapeBlock( instant_save_fd, instant_save_filetype ); break;
				case zx81:	SaveZX81TapeBlock( instant_save_fd, instant_save_filetype ); break;
				default:	SaveZxspTapeBlock( instant_save_fd, instant_save_filetype ); break;
				}

			// close files which can only hold one block:	
				switch( instant_save_filetype )
				{
				case zx80_file:	case o_file:
				case zx81_file:	case p_file:	
					clip_and_close( instant_save_fd, nothrow ); instant_save_fd   = -1;
					delete[] instant_save_filename; instant_save_filename = NULL;
					break;
				default:
					break;
				}
				
			// setup cpu to return from 'save tape' successfully:
				Z80Regs& regs = cpu->GetRegisters();
				switch( model->model )
				{
				case zx80:	regs.rr.pc = 0x0283 /*MAIN_EXEC*/; break;
				case zx81:	regs.rr.pc = 0x0207 /*SLOW_FAST*/; break;
				default:	regs.r.a = 0; regs.r.f |= Z_FLAG; regs.r.f &= ~C_FLAG;	// z && cy  =>  ok
						//	regs.ff.de = -1; regs.ff.ix += cnt-2;					// probably most times don't matter
							regs.r.iff1 = regs.r.iff2 = enabled; 
							regs.rr.pc = cpu->Pop2(); break;
				}

			// return:
				return cpu->Peek(regs.rr.pc);
			}
			catch( bad_alloc ) 
			{
				ShowAlert("run_for_sound:RomPatch:SaveTapeBlock:\n%s","out of memory"); 				
				close( instant_save_fd );   instant_save_fd   = -1;
				delete[] instant_save_filename; instant_save_filename = NULL;
				goto b;	// open tape recorder  ((too late..))
			}
			catch( any_error e )
			{
				ShowAlert( "run_for_sound:RomPatch:SaveTapeBlock:\n%s",e.Text() );
				close( instant_save_fd );   instant_save_fd   = -1;
				delete[] instant_save_filename; instant_save_filename = NULL;
				goto b;	// open tape recorder  ((too late..))
			}
		}
		else
		{
b:			tape_state = tape_options & autostart_tape;
			if( tape_options & autostart_tape )
			{
				if( [tape_recorder isStoppedRecording] ) [tape_recorder startRecording];
				else [tape_recorder orderFront];
			}
		}
	}
	
// common exit for tape load and tape save routine in rom:
	else if( rom_offset == model->tape_load_ret_addr )
	{
		if( tape_state&(autostart_tape) ) [tape_recorder pause];	
		if( tape_state&halt_cpu_after_block_loaded ) HaltCpu();		// => exit. instr is not executed! 
		tape_state = 0;
	}
	
// patch not handled:	
	else
	{
		ShowAlert( Using("unexpected patch at pc=$4hX",(ushort)pc) );
		if( (ulong)rom_offset<model->rom_size ) UnregisterRomPatch(rom_offset);
	}
	
	return instr;
}


void Machine::InfoFirstLocAtCycle (long cc, ushort pc)
{
}

bool Machine::IllegalInstructionAtCycle (long cc, ushort pc)
{
	XXLogIn("Machine:IllegalInstrAtCycle");
	return no;	// don't abort
}


uchar Machine::InterruptAcknowledgedAtCycle (long cc, ushort pc)
{
	return 0xFF;
}

void Machine::NmiAcknowledgedAtCycle (long cc, ushort pc)
{
}

void Machine::InstructionRetiAtCycle (long cc, ushort pc)
{
}

void Machine::InstructionRetnAtCycle (long cc, ushort pc)
{
}


void Machine::SetBeeperVolume (Sample v)
{ 
	XLogIn("Machine:SetKeyBeeperVolume");

	ula->SetBeeperVolume(v); 
}


void Machine::SetKeyboardMode (uint newmode)
{
	XLogIn("Machine:SetKeyboardMode");

	if(keyboard_mode==newmode) return;
	keyboard_mode = newmode;
	[kbd_panel updatePanel:this];	// note: wenn invisible wird NXPanel::updatePanel nicht aufgerufen!
	if(filepath) [preferences setInt:newmode forFile:filepath andKey:"newsnapshot_keyboard_mode"];
}


void Machine::SetUseWaitmap (bool onoff)
{
	XLogIn("Machine:SetUseWaitmap");

	bool use_waitmap = cpu_options & cpu_waitmap;
	if(use_waitmap==onoff) return;
	use_waitmap = onoff && model->has_contention();
	if( use_waitmap ) cpu_options |= cpu_waitmap; else cpu_options &= ~cpu_waitmap;
}


void Machine::SetAutostartTape( bool onoff )
{
	XLogIn( "Machine:SetAutostartTape: %s", onoffstr(onoff) );

	if( AutostartsTape() == onoff ) return;
	uint addr = model->tape_load_routine;
	if( addr==0 ) return;					// TODO: disable menu item in menu::update

	if( onoff )
	{
		tape_options |= autostart_tape;
		if( !InstantLoadsFromTape() ) RegisterRomPatch( addr );
	}
	else
	{
		if( !InstantLoadsFromTape() ) UnregisterRomPatch( addr );
		tape_options &= ~autostart_tape;
	}

	if(filepath) [preferences setBool:onoff forFile:filepath andKey:"newsnapshot_autostart_tape/*_for_loading*/"];
}


void Machine::SetInstantLoadFromTape( cstr newpath )
{
	XLogIn( "Machine:SetInstantLoadFromTape: \"%s\"", newpath?newpath:"off" );

	uint addr = model->tape_load_routine;
	if( addr==0 ) return;						// TODO: error alert and/or disable menu item "instant load from tape…"

	PLocker z(lock);

	if( InstantLoadsFromTape() && !AutostartsTape() ) UnregisterRomPatch( addr );
	tape_options &= ~instant_load_from_tape;
	if( instant_load_fd!=-1 ) close( instant_load_fd ); instant_load_fd=-1;
	delete[] instant_load_filename; instant_load_filename = NULL;

	if( newpath )
	{
		instant_load_fd = open( newpath, O_RDONLY, 0 );
		if( instant_load_fd==-1 ) { ShowAlert( "%s", ErrorStr(errno) ); return; }
		instant_load_filetype = (FileType)Peek4X( CatStr( LowerStr(ExtensionFromPath(newpath)), "   " ) );
		instant_load_filename = NewCopy( newpath );
		tape_options |= instant_load_from_tape;
		if( !AutostartsTape() ) RegisterRomPatch( addr );
	}
}





void Machine::SetInstantSaveToTape( cstr newpath )
{
	XLogIn( "Machine:SetInstantSaveToTape: \"%s\"", newpath?newpath:"off" );

	uint addr = model->tape_save_routine;
	if( addr==0 ) return;						// TODO: error alert and/or disable menu item "instant load from tape…"

	PLocker z(lock);

	if( InstantSavesToTape() && !AutostartsTape() ) UnregisterRomPatch( addr );
	tape_options &= ~instant_save_to_tape;
	if( instant_save_fd!=-1 ) close( instant_save_fd ); instant_save_fd=-1;
	delete[] instant_save_filename; instant_save_filename = NULL;

	if( newpath )
	{
		instant_save_fd = open( newpath, O_WRONLY|O_CREAT, 0644 );
		if( instant_save_fd==-1 ) { ShowAlert( "%s", ErrorStr(errno) ); return; }
		instant_save_filename = NewCopy( newpath );
		instant_save_filetype = (FileType)Peek4X( CatStr( LowerStr(ExtensionFromPath(newpath)), "   " ) );
		tape_options |= instant_save_to_tape;
		if( !AutostartsTape() ) RegisterRomPatch( addr );
	}
}


void Machine::SetHaltCpuAfterLoad( bool onoff )
{
	XLogIn("Machine:SetHaltCpuAfterLoad: %s", onoffstr(onoff));

	if( HaltsCpuAfterBlockLoaded() == onoff ) return;
	uint addr = model->tape_load_ret_addr;
	if( addr==0 ) return;					// TODO: disable menu item in menu::update
	XXASSERT( addr < model->rom_size );

	if(onoff)
	{
		tape_options |= halt_cpu_after_block_loaded;
		RegisterRomPatch( addr );
	}
	else
	{
		UnregisterRomPatch( addr );
		tape_options &= ~halt_cpu_after_block_loaded;
	}
}


/* ----------------------------------------------------------------------------------------
	Handle Patches
*/
void Machine::RegisterRomPatch( uint addr )		// TODO: external rom
{
	XXASSERT( addr < rom.Count() );

	cpu_options |= cpu_patch;
	rom[addr]   |= cpu_patch; 
	
	patches++;
}

void Machine::UnregisterRomPatch( uint addr )		// TODO: external rom
{
	XXASSERT( addr < rom.Count() );

	patches--;
	if(!patches) cpu_options &= ~cpu_patch;		
	rom[addr]   &= ~cpu_patch; 
}

void Machine::ClearAllPatches()			// TODO: external rom
{
	for( int i=rom.Count(); --i>=0; ) rom[i] &= ~cpu_patch;
	for( int i=ram.Count(); --i>=0; ) ram[i] &= ~cpu_patch;			

	cpu_options &= ~cpu_patch;
	patches = 0;
}







/* ----------------------------------------------------------------------------------------
	IC Tester:
*/

void Machine::toggle_ic_tester()
{
	XLogIn("Machine:toggle_ic_tester");

	Lock();
		if (ic_tester)	delete ic_tester;
		else			new ICTester(this);
	Unlock();
}

void Machine::toggle_test_ic_none()
{
	XLogIn("Machine:toggle_test_ic_none");
	XXTRAP(!ic_tester);
	ic_tester->SetIC(0);
}

void Machine::toggle_test_ic_7400()
{
	XLogIn("Machine:toggle_test_ic_7400");
	XXTRAP(!ic_tester);
	ic_tester->SetIC(7400);
}

void Machine::toggle_test_ic_7402()
{
	XLogIn("Machine:toggle_test_ic_7402");
	XXTRAP(!ic_tester);
	ic_tester->SetIC(7402);
}

void Machine::toggle_test_ic_7406()
{
	XLogIn("Machine:toggle_test_ic_7406");
	XXTRAP(!ic_tester);
	ic_tester->SetIC(7406);
}

void Machine::toggle_test_ic_27128()
{
	XLogIn("Machine:toggle_test_ic_27128");
	XXTRAP(!ic_tester);
	ic_tester->SetIC(27128);
}









