/*	Copyright  (c)	Günter Woigk 1994 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#if 0
#define SAFE	3
#define LOG		1

#import <fcntl.h>
#import <unistd.h>

#import "unix/file_utilities.h"
#import "cstrings/cstrings.h"

#import "utilities.h"
#import "globals.h"
#import "Machine.h"
#import "Z80 opcodes.h"
#import "Z80.h"
#import "Ay.h"
#import "Ula.h"
#import "Mmu.h"
#import "Port1ffd.h"
#endif

#define z80v1len 30
#define z80v2len 55
#define z80v3len 86



typedef struct
{ 
	uchar
		i,
		l2,h2,e2,d2,c2,b2,f2,a2,
		l,h,e,d,c,b,
		yl,yh,xl,xh,
		iff,r,f,a,spl,sph,im,
		brdr; 
} snahead;


struct z80head
{
	// ----- Z80 Version 1.45 Header
	uchar
		a,f,c,b,l,h,pcl,pch,spl,sph,
		i,r,data,e,d,c2,b2,e2,d2,l2,h2,a2,f2,
		yl,yh,xl,xh,iff1,iff2,im,

	// ----- Z80 version 2.01 enhanced header (if PC=0)
		h2lenl,h2lenh,	// size of header extension: 23 for version 2.01 files
		npcl,npch,		// PC (instead of pcl+pch, which is set to 0)
		model,			// 	Value:	Meaning in v2.01	Meaning in v3.0
						//	0		48k					48k
						//	1		48k + If.1			48k + If.1
						//	2		SamRam				48k + M.G.T.
						//	3		128k				SamRam
						//	4		128k + If.1			128k
						//	5		-					128k + If.1
						//	6		-					128k + M.G.T.
		port_7ffd,		// If in SamRam mode, bitwise state of 74ls259.
						// For example, bit 6=1 after an OUT 31,13 (=2*6+1)
						// If in 128 mode, contains last OUT to 7ffd (paging control)
						// if timex ts 2068: last out to port 244
		if1paged,		// !=0 means: interface1 rom is paged in
						// if timex ts 2068: last out to port 255
		rldiremu,		// Bit 0: 1 if R register emulation on
						// Bit 1: 1 if LDIR emulation on
						// Bit 2: AY sound in use, even on 48K machines
                        // Bit 6: (if bit 2 set) Fuller Audio Box emulation
                        // Bit 7: Modify hardware (see below)
		port_fffd,		// Last OUT to fffd (soundchip register number)
		soundreg[16],	// Contents of the sound chip registers

	// ----- Z80 version 3.0 enhanced header (if PC=0 & h2len=54)
		t_l,t_m,t_h,			// T state counter
		spectator,				// Flag byte used by Spectator (QL spec. emulator)
								// Ignored by Z80 when loading, zero when saving
		MGT_ROM,				// 0xFF if MGT Rom paged
		Multiface_ROM,			// 0fFF if Multiface Rom paged. Should always be 0.
		RAM0,RAM1,				// 0xFF if 0-8191 / 8192-16383 is RAM
		joy[10],				// 5* ascii word: keys for user defined joystick
		stick[10],				// 5* keyboard mappings corresponding to keys above
       	MGT_type,				// MGT type: 0=Disciple+Epson,1=Disciple+HP,16=Plus D
		disciple_inhibit_button_status,	// 0=out, 0ff=in
		disciple_inhibit_flag,	// Disciple inhibit flag: 0=rom pageable, 0ff=not

	// ----	warajewo/xzx extension ----
		port_1ffd;				// last out to $1ffd ((bank switching on +3))
};




/* ----	Compress .z80 file image -------------------------------------------
	compression scheme:  dc.b $ed, $ed, count, char
	returns 0 if size shrinked
	returns !0 if size growed
	For security: the compression buffer must be at least  (5/3*sourcesize)  bytes long
*/
static OSErr compress_z80 (ushort const* source, ulong sourcesize, uptr dest, ulong*destsize)
{
	uchar c,n;
	ushort const* q  = source;
	ushort const* qe = source + sourcesize;
	uptr  z = dest;
	
	XXXASSERT(source!=NULL);
	XXXASSERT(dest!=NULL);
	XXXASSERT(sourcesize>=1 kB);				// actual page size is always 0x4000 
	
	while (q<qe)
	{	c=*q++;
		if (q==qe||(uchar)*q!=c)							// single byte
		{
			*z++ = c;
			if ((c==0xed)&&(q+2<=qe)&&((uchar)*q==(uchar)*(q+1)))	// special care for compressible sequence after single 0xed !!!
				*z++=*q++;
			continue;
		}
		for (n=1; n<255 && (uchar)*q==c; n++) q++;			// TODO: <-- crashed here 2007-094-23
		if (q>qe) { n -= q-qe; q=qe; };
		if ( (n>=4) || (c==0xed) )			// compress ?
		{
			*z++ = 0xed;
			*z++ = 0xed;
			*z++ = n;
			*z++ = c;
		}
		else								// don't compress
		{
			for (;n;n--) *z++ = c;
		}
	}
	
	*destsize = z-dest;
	return (*destsize<=sourcesize) ? ok : error /* z80 compression increased data size */;
}


/* ----	Decompress .z80 file image -----------------------------------------
		returns NULL if ok
		else returns error message
		note: preserves flags in page
*/
static cstr decompress_z80 (cuptr source, long sourcesize, ushort* dest, long destsize)
{
	uchar c,n;
	cuptr   q = source; source+=sourcesize;
	ushort* z = dest;	  dest+=destsize;

	#define store(c)	((twobytes*)(z++))->lo = c
	
	while ( z<dest && q<source ) 
	{
		c=*q++;									// read next byte
		if (c!=0xed)   { store(c); continue; }	// single byte
		if (q==source) { store(c); continue; } 	// rare case: last source byte is 0xed

		c=*q++;									// read next byte
		if (c!=0xed) { store(0xed); q--; continue; };
		
		n=*q++; c=*q++;							// read count & char
		if ( (z+n)>dest ) return "decompression of a ram page failed";
		if (q>source) return "decompression of a ram page failed";
		while(n--) store(c);					// expand compressed bytes
	};
	if (q==source&&z==dest) return NULL;		// ok

	if (z==dest&&q+4<=source&&destsize==0xC000)	// version 1.45 file with end marker ?
	if (*q++==0 && *q++==0xed && *q++==0xed && *q++==0) return NULL;
	
	return "decompression of a ram page failed";
}


static ssize_t fd_write_uncompressed( int fd, ushort const* core, int size )
{
	uchar bu[size];
	for( int i=0; i<size; i++ ) bu[i] = core[i];
	return write( fd, bu, size );	
}




/* ----	load .sna snapshot file --------------------------
		returns NULL if all went ok
		else returns error message
*/
cstr Machine::init_with_sna (cstr path)
{
	XLog("Machine:init_with_sna\n");

	uptr bu; ulong size;
	cstr err = load_file(path,bu,size); if(err) return err;

	snahead& head = *(snahead*)bu; bu+=27; size-=27;
	uint pages    = size/0x4000;

	if( size!=pages*0x4000 || pages<1 || pages>3 ) return Using( "Bad ram size: $%4X bytes",size );

	ushort sp = head.sph*256+head.spl;
	if(sp<0x4000) ShowWarning( Using( "Stack within rom: $%4X", (uint)sp ) );
	if(pages<3&&sp>=pages*0x4000) ShowWarning( Using( "Stack beyond ram: $%4X", (uint)sp ) );
	if (head.im>2) return Using( "Illegal interrupt mode: %i", (int)head.im ); 
	
// now do it
	init_with_model(pages>1?zx48i3:zx16k);

	ula->SetBorderColor(head.brdr);

	for( uint i=0; i<size; i++ ) ((twobytes*)(ram+i))->lo = bu[i];	// copy data, preserve flags

	Z80Regs& reg = cpu->GetRegisters();
	reg.r.i 	= head.i;		
	reg.r.r 	= head.r;
	reg.r.a 	= head.a;		reg.r.f		= head.f;	
	reg.r.a2	= head.a2;		reg.r.f2	= head.f2;
	reg.r.b		= head.b;		reg.r.c		= head.c;
	reg.r.d		= head.d;		reg.r.e		= head.e;
	reg.r.h		= head.h;		reg.r.l		= head.l;
	reg.r.xh	= head.xh;		reg.r.xl	= head.xl;	
	reg.r.yh	= head.yh;		reg.r.yl	= head.yl;	
	reg.r.b2	= head.b2;		reg.r.c2	= head.c2;
	reg.r.d2	= head.d2;		reg.r.e2	= head.e2;
	reg.r.h2	= head.h2;		reg.r.l2	= head.l2;

	reg.r.im	= head.im;		
	reg.r.iff1	= reg.r.iff2	= (head.iff&0x04) ? enabled : disabled;
	reg.r.pcl	= cpu->Peek(sp++);
	reg.r.pch	= cpu->Peek(sp++);	
	reg.rr.sp	= sp;
	
	return NULL;	// ok
}



/* ----	load .z80 snapshot file --------------------------
*/
cstr Machine::init_with_z80 (cstr path)
{
	XLog("Machine:init_with_z80\n");

	uptr bu; ulong size;
	cstr err = load_file(path,bu,size); if(err) return err;

	int loaded		= 0;
	z80head head	= *(z80head*)bu;

	bool ay_used  = no;
	bool if1_used = no;
	bool mgt_used = no;
	bool paged_mem = no;

	if (head.data==0x0ff) head.data = 1;

// load version 1.45 file
	if (head.pch||head.pcl)
	{	
		XLog("  Version 1.45\n  Selecting 48k issue3\n");
		bu   += z80v1len;
		size -= z80v1len;
		
		init_with_model(zx48i3);

		if (head.data&0x20)		// compressed
		{	
			XLog("  $c000 bytes compressed\n");
			err = decompress_z80(bu,size,ram,0xC000);
			if (err) return err; 
		}
		else					// uncompressed
		{
			XLog("  $c000 bytes uncompressed\n");
			if ( size < 0xC000 ) return "file is too short";
			for( int i=0; i<0xc000; i++ ) ((twobytes*)(ram+i))->lo = bu[i];	// copy data, preserve flags
		};
	}

// versio 2.01 or above
	else	
	{
		XLog( head.h2lenl==23 ? "  Version 2.01\n" : "  Version 3.00 or higher\n" );

		if (head.h2lenh!=0) return "bogus header extension size";
		bu   += z80v1len+2+head.h2lenl;
		size -= z80v1len+2+head.h2lenl;

		head.pcl=head.npcl; head.pch=head.npch;

		if (head.h2lenl==23 && head.model>=3) head.model += 1;		// z80 v2.01 -> v3.00
		if (head.rldiremu&0x04) ay_used=yes;						// xzx

	// select required hardware
		XLog("  Model: %i: ",(int)head.model);	
		switch (head.model)	
		{
		case 1:	if1_used=yes;										// 48k + IF1
		case 0:	XLog("48k issue 3\n");
				init_with_model(zx48i3);							// 48k
				break;

		case 2: return "SamRam required - not supported";			// 48k + SamRam

		case 3:	XLog("48k issue 3\n");
				init_with_model(zx48i3);							// 48k + MGT
				mgt_used=yes;
				break;

		case 5:	if1_used=yes;										// 128k + IF1
		case 4:	XLog("128k\n");
				init_with_model(zx128);								// 128k
				mmu->SetState7ffd(head.port_7ffd);
				ay_used=yes;
				paged_mem=yes;
				break;

		case 6:	XLog("128k\n");
				init_with_model(zx128);								// 128k + MGT
				mmu->SetState7ffd(head.port_7ffd);
				mgt_used=yes; 
				ay_used=yes;
				paged_mem=yes;
				break;

		case 128:XLog("TS 2068\n");
				init_with_model(ts2068);							// Timex TS 2068  ((warajevo))
				OutputAtCycle(0,244,head.port_7ffd);
				OutputAtCycle(0,255,head.if1paged);
				ay_used=yes;
				break;
		
		case 8:															// zx +3  ((xzx))
		case 7: if(head.h2lenl<z80v3len+1) return "bogus +3 snapshot";	// zx +3  ((xzx))
				XLog("+3\n");
				init_with_model(zxplus3);
				mmu->SetState7ffd1ffd(head.port_7ffd,head.port_1ffd);
				ay_used=yes;
				paged_mem=yes;
				break;
		
		case 9:	XLog("Pentagon\n");
				init_with_model(pentagon);							// pentagon ((xzx))
				mmu->SetState7ffd(head.port_7ffd);
				ay_used=yes;	
				// PENTAGON: •••more settings stored in snapshot?•••		TODO
				break;
		
		case 10:XLog("Scorpion\n");
				init_with_model(scorpion);							// scorpion ((xzx))
				mmu->SetState7ffd(head.port_7ffd);
				ay_used=yes;
				paged_mem=yes;
				// SCORPION: •••more settings stored in snapshot!•••		TODO
				break;

		default:return "Unknown hardware configuration";
		}

		if (if1_used)
		{
			if (head.if1paged) return "Interface-1 ROM paged in (not yet supported)"; 
			ShowWarning( "Interface 1 may be required!" );
		}
		if (mgt_used) 
		{
			if (head.h2lenl<z80v3len) return "M.G.T. with v2.01 header"; 
			if (head.MGT_ROM==0xff) return "M.G.T. ROM paged in - not supported"; 
			ShowWarning( "M.G.T. interface may be required!" );
		}
		if (!ay_used)
		{
			uchar c=head.soundreg[0]; int i;
			for(i=15;i;i--) if(head.soundreg[i]!=c) break;
			ay_used = i;
		}

	// set T cycle counter
		if(head.h2lenl>=z80v2len+3)
		{
		//	The hi T state counter counts up modulo 4. Just after the ULA generates its once-in-every-20-ms interrupt, it is 3, 
		//	and is increased by one every 5 emulated milliseconds. 
		//	In these 1/200s intervals, the low T state counter counts down from 17471 to 0 (17726 in 128K modes), 
		//	which make a total of 69888 (70908) T states per frame.
			long n = (head.t_h-3+1) * model->cpu_cycles_per_frame/4 - ushort(head.t_m*256+head.t_l);
			cpu->CpuCycle() = n<0 ? 0 : n>(long)model->cpu_cycles_per_frame ? (long)model->cpu_cycles_per_frame : n;
		}

	// loop: load all pages
		while(size>3&&!err)
		{
			struct bh { uchar  lenl,lenh,nr; };
			int len;
			
		// read page header
			bh& blockhead = *(bh*)bu; bu+=3;size-=3;
				
		// read and decompress page
			len = blockhead.lenl+256*blockhead.lenh;
			if ( (long)(size-len)<0 ) break;
			cuptr q = bu;
			bu += len; size -= len;
	
			if (!paged_mem)		// zx spectrum 48k
			{
				loaded |= 0xf1;
				XLog("  decompressing page ID %i\n",(int)blockhead.nr); 
				switch (blockhead.nr)
				{
				default: return "page ID out of range";
				case 0:	err = decompress_z80(q,len,rom+0x0000,0x4000); XLog("  --> rom image!\n"); break;  /* The ROM */
				case 1:	 /* IF1 ROM */			break;
				case 11: /* Multiface ROM */	break;
				case 8:	err = decompress_z80(q,len,ram+0x0000,0x4000); loaded |= 2; break;
				case 4:	err = decompress_z80(q,len,ram+0x4000,0x4000); loaded |= 4; break;
				case 5:	err = decompress_z80(q,len,ram+0x8000,0x4000); loaded |= 8; break;
				};
			}
			else			// zx 128k
			{
				XLog("  decompressing page ID %i\n",(int)blockhead.nr); 
				switch (blockhead.nr)
				{
				default: return "page ID out of range";
				case 0:	err = decompress_z80(q,len,rom+0x4000,0x4000); XLog("  --> 48k Basic rom image!\n"); break;  /* Basic ROM */
				case 1:	 /* IF1 ROM */			break;
				case 2:	err = decompress_z80(q,len,rom+0x0000,0x4000); XLog("  --> Boot rom image!\n"); break;  /* Boot ROM */
				case 11: /* Multiface ROM */	break;
				case 3:
				case 4:
				case 5:
				case 6:
				case 7:	
				case 8:
				case 9:  
				case 10: err = decompress_z80(q,len,ram+0x4000*(blockhead.nr-3),0x4000);
						 loaded |= 1<<(blockhead.nr-3); break;
				};
			}
		};
		
		if (err) return err; 
		if (loaded!=0xff) return "Some RAM pages missing!"; 
		if (size!=0) XLog ( "Some extra bytes at end of file ignored\n" );
	}

// setup screen
	ula->SetBorderColor(head.data>>1);

// setup registers
	Z80Regs& reg = cpu->GetRegisters();
	reg.r.i 	= head.i;		
	reg.r.r 	= head.r&0x7f + head.data<<7;	// bit 7 von r steckt in data bit 0 ... tzz tzz
	reg.r.a 	= head.a;		reg.r.f		= head.f;	
	reg.r.a2	= head.a2;		reg.r.f2	= head.f2;
	reg.r.b		= head.b;		reg.r.c		= head.c;
	reg.r.d		= head.d;		reg.r.e		= head.e;
	reg.r.h		= head.h;		reg.r.l		= head.l;
	reg.r.xh	= head.xh;		reg.r.xl	= head.xl;	
	reg.r.yh	= head.yh;		reg.r.yl	= head.yl;	
	reg.r.b2	= head.b2;		reg.r.c2	= head.c2;
	reg.r.d2	= head.d2;		reg.r.e2	= head.e2;
	reg.r.h2	= head.h2;		reg.r.l2	= head.l2;
	reg.r.pch	= head.pch;		reg.r.pcl	= head.pcl;
	reg.r.sph	= head.sph;		reg.r.spl	= head.spl;
	reg.r.iff1	= head.iff1;	reg.r.iff2	= head.iff2;
	reg.r.im = head.im&3; if (reg.r.im>2) reg.r.im = 1;

	//Log("ix = $%04x  (init_with_z80)\n",(uint)reg.rr.ix);

// setup soundchip:
	if (ay_used)
	{
		if (!ay) ay = NewAyForThisMachine(); 
		if (ay)
		{
			for ( int i=0; i<16; i++ ) ay->SetRegister(i,head.soundreg[i] );
			ay->SelectRegister(head.port_fffd);
		}
	}
	
	return NULL;		// ok
}





/* ----	Save Core as .sna file ------------------------------------------------------------
		note: .sna files were originally saved from an NMI routine. 
		therefore the PC is stored on the Z80 stack and only IFF1 needs to be saved
*/
OSErr Machine::SaveSNA (cstr path)
{
	XLog("Machine:SaveSNA\n");

	int zx = model->model;
	if (zx!=zx16k && zx!=zx48i2 && zx!=zx48i3) 
	{
		ShowWarning
		(	"A .sna file can only save a ZX 16k or ZX 48k.\n"
			"When you load this snapshot back later,\n"
			"it will be running on a ZX Spectrum 16k or 48k\n"
			"potentially missing some resources.\n"
		);
	}
	
	snahead head; 
	TRAP(sizeof(head)!=27);
	
	Z80Regs& reg = cpu->GetRegisters();
	ushort sp = reg.rr.sp;
	cpu->Poke(--sp,reg.r.pch);		// push retn address
	cpu->Poke(--sp,reg.r.pcl);
	
	head.im	= reg.r.im;
	head.iff= reg.r.iff1 ? 0x04 : 0x00;
	head.i	= reg.r.i;	head.r	= reg.r.r;
	head.a	= reg.r.a;	head.a2	= reg.r.a2;
	head.f	= reg.r.f;	head.f2	= reg.r.f2;
	head.b	= reg.r.b;	head.c	= reg.r.c;
	head.d	= reg.r.d;	head.e	= reg.r.e;
	head.h	= reg.r.h;	head.l	= reg.r.l;
	head.xh	= reg.r.xh;	head.xl	= reg.r.xl;
	head.yh	= reg.r.yh;	head.yl	= reg.r.yl;
	head.b2	= reg.r.b2;	head.c2	= reg.r.c2;
	head.d2	= reg.r.d2;	head.e2	= reg.r.e2;
	head.h2	= reg.r.h2;	head.l2	= reg.r.l2;
	head.sph= sp>>8;	head.spl= sp;
	head.brdr = ula->GetBorderColor();

	int fd = open( path, O_WRONLY|O_CREAT, 0644 );
	if (fd<0) return Error();

	write( fd, &head, 27 );

	if( model->ram_size<=0xc000 )
	{
		fd_write_uncompressed( fd, ram, model->ram_size );	
	}
	else
	{
		fd_write_uncompressed( fd, cpu->RdPtr(0x4000), 0x4000 );
		fd_write_uncompressed( fd, cpu->RdPtr(0x8000), 0x4000 );
		fd_write_uncompressed( fd, cpu->RdPtr(0xC000), 0x4000 );
	}

	close( fd );

	return ok;
}



/* ----	save .z80 file; version 2.01 --------------------------------------------------
*/
OSErr Machine::SaveZ80 (cstr path)
{
	XLog("Machine:SaveZ80\n");

	int fd = open( path, O_WRONLY|O_CREAT, 0644 );
	if (fd<0) return Error();

// write z80 v2.01 head:
	{
		z80head head;
		memset(&head,0,sizeof(head));
	
	// z80 registers
		Z80Regs& reg = cpu->GetRegisters();

		head.a	= reg.r.a;	head.a2	= reg.r.a2;
		head.f	= reg.r.f;	head.f2	= reg.r.f2;
		head.b	= reg.r.b;	head.c	= reg.r.c;
		head.d	= reg.r.d;	head.e	= reg.r.e;
		head.h	= reg.r.h;	head.l	= reg.r.l;
		head.xh	= reg.r.xh;	head.xl	= reg.r.xl;
		head.yh	= reg.r.yh;	head.yl	= reg.r.yl;
		head.b2	= reg.r.b2;	head.c2	= reg.r.c2;
		head.d2	= reg.r.d2;	head.e2	= reg.r.e2;
		head.h2	= reg.r.h2;	head.l2	= reg.r.l2;

		head.npch = reg.r.pch;	head.npcl = reg.r.pcl;
		head.sph  = reg.r.sph;	head.spl  = reg.r.spl;
		head.iff1 = reg.r.iff1; head.iff2 = reg.r.iff2; 	
		head.i    = reg.r.i;	head.im	  = reg.r.im;

		head.data = ((reg.r.r>>7)&1) + (ula->GetBorderColor()<<1);
		head.r	  = reg.r.r & 0x7F;


	// version 2.01 header
		head.h2lenl	= 23;			// size of header extension: 23 bytes for version 2.01 files

	// model:
		if (model->model>=zxplus2)		// TODO: save z80 version 3
		ShowAlert
		(	"A .z80 file version 2.01 can't save the +2 or later models.\n"
			"When you load this snapshot later, "		
			"it will be running on a ZX Spectrum 128k " 
			"potentially missing some essential hardware."
		);

		head.model = model->model >=zx128 ? 3 : 0;
				//			IF1 currently not supported
				//			MGT, SamRam probably never supported
				// 	Value:	Meaning in v2.01	Meaning in v3.0
				//	0		48k					48k
				//	1		48k + If.1			48k + If.1
				//	2		SamRam				48k + M.G.T.
				//	3		128k				SamRam
				//	4		128k + If.1			128k
				//	5		-					128k + If.1
				//	6		-					128k + M.G.T.

	// zx128: page control register
	// SamRam mode: bitwise state of 74ls259; p.ex. bit 6=1 after an OUT 31,13 (=2*6+1)
	// Timex TS 2068: last out to port 244
		head.port_7ffd = mmu ? mmu->GetState7ffd() : 0;

	// !=0 means: interface1 rom is paged in
	// Timex TS 2068: last out to port 255
		head.if1paged = 0;

	// Bit 0: 1 if R register emulation on
	// Bit 1: 1 if LDIR emulation on
	// Bit 2: AY sound in use, even on 48K machines
	// Bit 6: (if bit 2 set) Fuller Audio Box emulation
	// Bit 7: Modify hardware (see below)
		head.rldiremu = 0x03;		// R register and LDIR emu: on

	// AY soundchip:
		if(ay)
		{
			head.rldiremu |= (1<<2);										// ay in use, even on 48k machine
			if(FindStr(ay->GetName(),"Fuller")) head.rldiremu |= (1<<6);	// AY chip for Fuller box
			head.port_fffd	= ay->RegNr();									// selected register
			memcpy ( head.soundreg, ay->Registers(), 16 );					// registers
		}

	// write data
		write( fd, &head, z80v2len );
	}
	
	
	
/*	Now the compressed ram pages follow. Each block has a 3 byte header:
		dc.w		length of data (without this header; low byte first)
		dc.b		page number of block
		dc.s		compressed data follows

	The pages are numbered, depending on the hardware mode, in the following way:

		PageID	48 mode				128 mode			SamRam mode
		0		48K rom				rom (basic)			48K rom
		1		Interface I, Disciple or Plus D rom, according to setting
		2		-					rom (reset)			samram rom (basic)
		3		-					page 0				samram rom (monitor,..)
		4		8000-bfff			page 1				Normal 8000-bfff
		5		c000-ffff			page 2				Normal c000-ffff
		6		-					page 3				Shadow 8000-bfff
		7		-					page 4				Shadow c000-ffff
		8		4000-7fff			page 5				4000-7fff
		9		-					page 6				-
		10		-					page 7				-
		11		Multiface rom		Multiface rom		-

	In 16k mode, only page 8 is saved.					kio 1999-05-02
	In 48k mode, pages 4,5 and 8 are saved.
	In SamRam mode, pages 4 to 8 must be saved.			((SamRam not supported))
	In 128 mode, all pages from 3 to 10 are saved.

	The 128 has a memory map like:   Rom [switchable];   Ram 5;   Ram 2;   Ram [switchable]
*/
	{
		ulong size;
		uchar b[3+0x4000*5/3+1];

	// Test for custom ROM:
		if( model->model>=zx16k && (model->rom_size==0x4000 || model->rom_size==0x8000) )	// supported?
		{
			uptr bu; int i;
			cstr err = load_file( CatStr(rsrc_path,model->rom_filename), bu, size ); 
			if(!err)
			{
				for( i=0;i<0x4000; i++ ) 
				{
					if( bu[i]!=(uchar)rom[i] ) 
					{ 
						XLog("  saving non-standard ROM #0\n");
						compress_z80(rom+0x0000,0x4000,b+3,&size); 
						Poke2Z(b,size); b[2] = model->rom_size==0x4000?0:2; write(fd,b,size+3); 
						break; 
					}
				}
				if(model->rom_size==0x8000) for( i=0x4000;i<0x8000; i++ ) 
				{
					if( bu[i]!=(uchar)rom[i] ) 
					{ 
						XLog("  saving non-standard ROM #1\n");
						compress_z80(rom+0x4000,0x4000,b+3,&size); 
						Poke2Z(b,size); b[2] = 0; write(fd,b,size+3); 
						break; 
					}
				}
			}
		}

	// RAM:
		if (!mmu/*no paging*/)		// TODO: models with paging but no port 7ffd
		{
			int n = model->ram_size / 0x4000;
					   compress_z80(ram+0x0000,0x4000,b+3,&size); Poke2Z(b,size); b[2] = 8; write(fd,b,size+3);
			if(n>=2) { compress_z80(ram+0x4000,0x4000,b+3,&size); Poke2Z(b,size); b[2] = 4; write(fd,b,size+3); }
			if(n>=2) { compress_z80(ram+0x8000,0x4000,b+3,&size); Poke2Z(b,size); b[2] = 5; write(fd,b,size+3); }
		}
		else	// with paged memory		note: this works for the +3 too
		{
			for (int i=0;i<=7;i++)
			{
				compress_z80(ram+i*0x4000,0x4000,b+3,&size); Poke2Z(b,size); b[2] = i+3; write(fd,b,size+3); 
			}
		}
	}
	
	close(fd);

	return ok;
}



/* ----	Save ROM to file -------------------------------------------------
*/
OSErr Machine::SaveROM (cstr path)
{
	XLog("Machine:SaveROM\n");

	int fd = open( path, O_WRONLY|O_CREAT, 0644 );
	if (fd<0) return Error();

	fd_write_uncompressed( fd, rom, model->rom_size );
	close( fd );

	return ok;
}


/* ----	Save .scr screenshot file ----------------------------------------
*/
OSErr Machine::SaveSCR (cstr path)
{
	XLog("Machine:SaveSCR\n");

	int fd = open( path, O_WRONLY|O_CREAT, 0644 );
	if (fd<0) return Error();

	fd_write_uncompressed( fd, ula->GetVideoRam(), 6912 );
	close( fd );

	return ok;
}


/* ----	Save .gif screenshot file ----------------------------------------
*/
OSErr Machine::SaveGIF ( cstr path )
{
	XLog( "Machine:SaveGIF\n" );

	int fd = open( path, O_WRONLY|O_CREAT, 0664 );
	if (fd<0) return Error();

// allocate buffers
	ushort*	attr_pixel	= ula->NewAttrPixelArray();	// ushort[32*24*8] high byte=attr, low byte=pixel
	uchar*  border		= ula->NewBorderArray();	// uchar[ MAX_ULA_CLOCKS_PER_LINE / 16 *
													//	(MAX_LINES_BEFORE_SCREEN+192+MAX_LINES_AFTER_SCREEN+1) ]
	
// fetch attr_pixel data from screen:
// these are the arrays which would be swapped back to the ula 
// when receiving new screen data. they should still contain 
// valid data, which was recently drawn by the screen.
	screen->lock->Lock();
		Swap( attr_pixel, screen->new_attr_pixel );
		Swap( border,     screen->new_border     );
	screen->lock->Unlock();

	ulong  ula_row_bytes = ula->GetOctetsPerLine();		// pixel octets per scan row
	ulong  ula_top_lines = ula->GetLinesBeforeScreen();
//	ulong  ula_bot_lines = ula->GetLinesAfterScreen();

	uint const v_border  = 3*8;							// vert. border padding (rows)
	uint const h_border  = 4*8;							// hor. border padding	(pixel)

	uchar* screen_start	 = border + ula_top_lines  * ula_row_bytes - h_border/8;
	uchar* frame_start	 = screen_start - v_border * ula_row_bytes;
	uchar* screen_end	 = screen_start + 192      * ula_row_bytes;
//	uchar* frame_end	 = screen_end + v_border   * ula_row_bytes;

	uchar  bgcolor		 = screen_start[0];				// current border color

// search for 'flashing' attributes
	int i; for( i=0; i<32*192; i++ ) { if( (short)attr_pixel[i]<0 ) break; }
	bool has_flashing = i<32*192;	


// Header: 3 bytes signature, 3 bytes version
	if(has_flashing) WriteGif89aHeader(fd);			// we need vers. 89a for flashing
	else			 WriteGif87aHeader(fd);
	

// Logical screen descriptor:
	uchar flags   = (1<<7)			// gif stream contains a global color table
				  + ((4-1)<<4)		// color resolution of orig. image: 4 bit ~ 16 colors
				  + (0<<3)			// colors are not ordered by significance
				  + ((4-1)<<0);		// color resolution of global color table: 4 bit ~ 16 colors

	write_ushort_z( fd, 256+2*h_border );	// screen width
	write_ushort_z( fd, 192+2*v_border );	// screen height
	write_uchar( fd, flags );
	write_uchar( fd, bgcolor );		// background color index
	write_uchar( fd, 0 );			// pixel aspect ratio: 1:1


// Global color table:
	write_bytes( fd, cmap16, sizeof(cmap16) );


// Write Netscape Extension for looping animations
	if(has_flashing)
	{
		XLog("image has flashing pixels\n");
		WriteGifNetscapeExtension(fd);
		WriteGifGraphicControlExtension( fd, 32 /* 1/100 sec delay */, no_disposal + no_transparency );
	}


// Write Images for Border:
	if(1) bgcolor=0xff;					// force border images
	WriteGifImageBlock ( fd, bgcolor,					// border above screen
			frame_start,				// pixels[]
			0,0,						// x, y
			(256+2*h_border)/8,			// w (unscaled)
			v_border,					// h (unscaled)
			ula_row_bytes,8,1);			// dy, xscale, yscale

	if( has_flashing ) WriteGifGraphicControlExtension( fd, 0 /* 1/100 sec delay */, no_disposal + no_transparency );

	WriteGifImageBlock ( fd, bgcolor,					// border left of screen
			screen_start,				// pixels[]
			0,v_border,					// x, y
			(h_border)/8,				// w (unscaled)
			192,						// h (unscaled)
			ula_row_bytes,8,1);			// dy, xscale, yscale

	WriteGifImageBlock ( fd, bgcolor,					// border right of screen
			screen_start+(h_border+256)/8, // pixels[]
			h_border+256,v_border,		// x, y
			(h_border)/8,				// w (unscaled)
			192,						// h (unscaled)
			ula_row_bytes,8,1);			// dy, xscale, yscale

	WriteGifImageBlock ( fd, bgcolor,					// bordere below screen
			screen_end,					// pixels[]
			0,v_border+192,				// x, y
			(256+2*h_border)/8,			// w (unscaled)
			v_border,					// h (unscaled)
			ula_row_bytes,8,1);			// dy, xscale, yscale


// init attr_pixel -> color_index conversion tables
	uchar a2b[256]; Setup_a2b(a2b);		// attr -> back color
	uchar a2f[256]; Setup_a2f(a2f);		// attr -> fore color
	for( int i=64; i<128; i++ )			// replace bright black with black
	{
		if(a2b[i]==8) a2b[i]=0; 
		if(a2f[i]==8) a2f[i]=0; 
	}	
	Blink_bxf(a2b,a2f,0);				// spread the replaced colors

// allocate buffers
	delete[] border; border = NULL;
	uchar*  screen	= new uchar[256*192];		// 1 byte/pixel, flashphase 1

// convert attr_pixel -> indexed_color screen
	uptr z = screen;
	for( i=0; i<32*192; i++ ) 
	{
		uchar p = attr_pixel[i];	// 8 pixel
		uchar a = attr_pixel[i]>>8;	// attr
		uchar f = a2f[a];			// foreground color
		uchar b = a2b[a];			// background color
		for( int x=0; x<8; x++ ) { *z++ = p&0x80?f:b; p+=p; }
	}

// Write 1st Screen Image
	WriteGifImageBlock ( fd, 0xff,	
			screen,					// pixels[]
			h_border,v_border,		// x, y
			256,					// w (unscaled)
			192,					// h (unscaled)
			256,1,1);				// dy, xscale, yscale



// Write 2nd Screen Image if screen has blinking colors:

	if( has_flashing )
	{
	// toggle flashing attr
		Blink_bxf(a2b,a2f,1);
		

	// calc affected box:
		int xmin,xmax,ymin,ymax;
	
		ushort* p = attr_pixel;
		short row[192], col[32]; memset(row,0,192*sizeof(short)); memset(col,0,32*sizeof(short));
		for( int y=0;y<192;y++ )
		for( int x=0;x<32;x++ ) { row[y] |= *p; col[x] |= *p; p++; }
		
		for( xmin=0;  col[xmin]>=0; xmin++ ) {} // must terminate because at least one attr byte flashes!
		for( ymin=0;  row[ymin]>=0; ymin++ ) {}
		for( xmax=31; col[xmax]>=0; xmax-- ) {} 
		for( ymax=191;row[ymax]>=0; ymax-- ) {}
		XLog("transp. box: [%i..x..%i] x [%i..y..%i]\n", xmin*8,xmax*8+7, ymin,ymax);

		z = screen;
		for( int y=ymin; y<=ymax; y++ ) 
		for( int x=xmin; x<=xmax; x++ )
		{
			i = y*32 + x;		
			uchar p = attr_pixel[i];	// 8 pixel
			uchar a = attr_pixel[i]>>8;	// attr
			uchar f = a2f[a];			// foreground color
			uchar b = a2b[a];			// background color
			for( i=0; i<8; i++ ) 
			{
				uchar c = p&0x80?f:b;	// new pixel color
				*z = c==*z ? 8 : c;		// new color or transparent color
				p+=p; z++; 
			}
		}

	// Write 2nd Image:
		WriteGifGraphicControlExtension( fd, 32 /* 1/100 sec delay */, no_disposal + no_transparency );
		WriteGifImageBlock ( fd, 0xff,
			screen+256*ymin+xmin*8,		// pixels[]
			h_border+xmin*8,v_border+ymin,// x, y
			(xmax-xmin+1)*8,			// w (unscaled)
			ymax-ymin+1,				// h (unscaled)
			256,1,1);					// dy, xscale, yscale
	}


// Write Trailer
	WriteGifTrailer( fd );
	close( fd );

// Purge Buffers:
	delete[] screen;
	delete[] attr_pixel;

	return ok;
}


/* ---- Load ROM file -----------------------------
		load standard machine with user-defined rom
*/
cstr Machine::init_with_rom (cstr path)
{
	XLog("Machine:LoadROM\n");

	uptr bu; ulong size; 
	cstr err = load_file(path,bu,size); if(err) return err;
	if(size>64 kB) return "file too long"; 

	uint my_model = [preferences getIntForKey:"newmachine_model"];	
	if( my_model>=zxmodels || !zx_info[my_model].IsSupported() ) my_model = zx48i3;	// force legal

	if( zx_info[my_model].rom_size != size )
	{
		int i; for( i=0; i<zxmodels; i++ )
		{
			if( size == zx_info[i].rom_size && zx_info[i].IsSupported() ) break;
		}
			
		if( i==zxmodels ) return "file size does not match rom size of any supported machine"; 
		my_model=i;		
	}
	
	init_with_model(my_model);
	//XXTRAP( ram-rom != size );			ram[] und rom[] werden mittlerweile separat alloziiert
	XXTRAP( model->rom_size != size );	 //	test instead
	
	for( ulong i=0; i<size; i++ ) rom[i] = bu[i];		// also clear flags
	return NULL;
}


/* ---- Load SCR screenshot ---------------------------
		load screenshot into new machine
		note: setting machine into stuck mode (halt with di)
		note: always selects model zx48ki3 for proper screen mode
*/
cstr Machine::init_with_scr (cstr path)
{
	XLog("Machine:init_with_scr\n");

	uptr bu; ulong size;
	cstr err = load_file(path,bu,size); if(err) return err;
	if(size!=24*(8+1)*32) return "wrong file length: this is no valid .scr screenshot";

	init_with_model(zx48i3);

	ushort* vram = ula->GetVideoRam();
	for( uint i=0; i<size; i++ ) vram[i] = (vram[i]&0xff00) + bu[i];		// preserve flags

	*cpu->WrPtr(0x8000) = HALT;
	cpu->GetRegisters().rr.pc = 0x8000;	

	return NULL; // ok
}

	
#include "AssZ80.h"

cstr Machine::init_with_ass( cstr file )
{
	XLog("Machine:init_with_ass: %s\n",file);

	cstr srcfile  = NewCopy(file);
	cptr ext	  = ExtensionFromPath(file);
	cstr destfile = NewCopy(CatStr(SubStr(file,ext),".$"));
	cstr listfile = NULL;
	cstr errmsg   = NULL;

	AssZ80 ass;
	ass.AssembleFile( srcfile, destfile, listfile, no/*list1*/, no/*list2*/, no/*listv*/, no/*listw*/, 'b'/*targetstyle*/ );

	if(ass.Errors()) 
	{
		listfile = NewCopy(CatStr(SubStr(file,ext),".log"));
		ass.AssembleFile (	srcfile, destfile, listfile, ass.Pass()==1/*list1*/, ass.Pass()==2/*list2*/, 
							yes/*listv*/, yes/*listw*/, 'b'/*targetstyle*/ );

		errmsg = Using( "Assembling failed. Logfile was created at \"%s\".", listfile );
	}
	else	// assembled OK
	{
		errmsg = init_with_file( ass.TargetName() );	// 2007-03-25
		if( !errmsg )
		{
			delete[] filepath;					// delete target file name
			filepath = NewCopy(srcfile);		// filepath := source file name  ((durable copy))
			set_title();						// title := source file name
		}
	}
	
	delete[] srcfile;
	delete[] destfile;
	delete[] listfile;

	return errmsg;		/* NULL = ok */
}


/*	read zx80 tape file: "*.80" or "*.o"

		returns NULL if all went ok
		else returns error message

	ZX80 Cassette File Structure
		x seconds    your voice, saying "filename" (optional)
		x seconds    video noise
		5 seconds    silence
		LEN bytes    data, loaded to address 4000h, LEN=(400Ah)-4000h.
		x seconds    silence / video noise

		ZX80 files do not have filenames
		ZX80 files cannot be autostarted. 
		The data is loaded to address $4000++
		The data contains the whole system area, basic program, VARS.
		Video memory is NOT included in ZX80 files.		
		the last byte of a (clean) file should be $80 (the last byte of VARS)
		The system area should contain proper data. 
		$400A	defines the end address (used to calculate the file length). 
		$4028++	may be misused for whatever purpose.

	.80 and .o files:	include only the data, loaded to $4000++
	.o files:			typically there is some garbage at the file end
*/
cstr Machine::init_with_80 ( cstr path )
{
	XLog("Machine:init_with_80\n");

	int fd = open( path, O_RDONLY, 0 );	
	if( fd==-1 ) return ErrorStr();

	try
	{
		uchar sysvars[0x28]; 
		read_data( fd, sysvars, 0x28 );
		int end = Peek2Z( sysvars+0x0A );	
		int len = end - 0x4028;				if(len<=0) throw(file_error("sysvar $400A: illegal load end address"));
		uchar bu[len];
		read_data( fd, bu, len );

	// init 
		init_with_model(zx80);				// da sollte man noch das ram wählen können...

	// ram
		for( int i=0; i<0x28; i++ ) { ram[i] = sysvars[i]; }
		for( int i=0; i<len;  i++ ) { ram[0x28+i] = bu[i]; }

	// register:
	// main execution loop entry
		Z80Regs& reg = cpu->GetRegisters();
		reg.rr.af = 0x3583;
		reg.rr.bc = 0x0035;
		reg.rr.de = 0xffff;
		reg.rr.hl = 0x12C3;

		reg.rr.ix = 0x0000;
		reg.rr.iy = 0x4000;
		reg.rr.pc = 0x0283;
		reg.rr.sp = 0xbffe; /*32k*/ ram[0xbffe -0x4000] = 0x82; ram[0xbfff-0x4000] = 0x3f; 

		reg.rr.af2 = 0x0000;
		reg.rr.bc2 = 0x0021;
		reg.rr.de2 = 0xd8f0;
		reg.rr.hl2 = 0xd8f0;
		
		reg.r.im = 1;
		reg.r.i  = 0x0e;
		reg.r.iff1 = reg.r.iff2 = disabled;
		

		return NULL; /* ok */
	}
	catch( file_error err )
	{		
		return err.Text();
	}
	catch( std::bad_alloc )
	{
		return "Out of memory";
	}
}


/*	read zx81 tape file: "*.81" or "*.p"

		returns NULL if all went ok
		else returns error message

	ZX81 Cassette File Structure
		x seconds    your voice, saying "filename" (optional)
		x seconds    video noise
		5 seconds    silence
		1-127 bytes  filename (bit7 set in last char)
		LEN bytes    data, loaded to address 4009h, LEN=(4014h)-4009h.
		1 pulse      video retrace signal (only if display was enabled)
		x seconds    silence / video noise

		The data contains system area, basic program, video memory, VARS.
		the last byte of a (clean) file should be $80 (the last byte of VARS)

		$4014	defines the end address (used to calculate the file length)
		$4029	points to the next executed (autostarted) BASIC line
		$403B	indicates if program runs in SLOW or FAST mode (bit 6)
		$403C++	may be misused for whatever purpose, 
		video memory must contain 25 HALT opcodes if the file was saved in SLOW mode.

		Files should usually not exceed 16 KBytes. 
		The memory detection procedure in both ZX80 and ZX81 BIOS stops after 16 KBytes (at $8000), 
		and initializes the stack pointer at that address, even if more memory is installed. 
		Thus loading files of 16K or more would destroy the stack area, 
		unless a separate loader has previously moved the stack area to another location.
		However, most ZXes don't have more than 16K RAM, so bigger files won't work on most computers anyways.

	.81 and .p files:	include only the data, loaded to $4009++
	.p files:			typically there is some garbage at the file end
	.p81 files:			start with the 1..127 bytes filename, last byte ORed with $80
*/
cstr Machine::init_with_81 ( cstr path )
{
	XLog("Machine:init_with_81\n");

	int fd = open( path, O_RDONLY, 0 );	
	if( fd==-1 ) return ErrorStr();

	try
	{
		uchar sysvars[0x3c]; 
		read_data( fd, sysvars+0x09, 0x3c-0x09 );
		int end = Peek2Z( sysvars+0x14 );	
		int len = end - 0x403c;				if(len<=0) throw(file_error("sysvar $4014: illegal load end address"));
		uchar bu[len];
		read_data( fd, bu, len );

	// init
		init_with_model(zx81);				// da sollte man noch das ram wählen können...

	// ram
		for( int i=0x09; i<0x3c; i++ ) { ram[i] = sysvars[i]; }
		for( int i=0;    i<len;  i++ ) { ram[0x3c+i] = bu[i]; }

	// register:
	// Aussprungstelle der "Alle Bytes geladen?" Testroutine:
		Z80Regs& reg = cpu->GetRegisters();
		reg.rr.af = 0x3583;	// o. $3783
		reg.rr.bc = 0x0080;
		reg.rr.de = 0xffff;
		reg.rr.hl = end;

		reg.rr.ix = 0x0281;
		reg.rr.iy = 0x4000;
		reg.rr.pc = 0x0206;
		reg.rr.sp = 0xbffa; /*32k*/ 
			ram[0xbffa - 0x4000] = 0x61; ram[0xbffb - 0x4000] = 0x33; 
			ram[0xbffc - 0x4000] = 0x76; ram[0xbffd - 0x4000] = 0x06; 
			ram[0xbffe - 0x4000] = 0x00; ram[0xbfff - 0x4000] = 0x3e; 

		reg.rr.af2 = 0x0000;
		reg.rr.bc2 = 0x8102;
		reg.rr.de2 = 0x002b;
		reg.rr.hl2 = 0xd7ff;
		
		reg.r.im = 1;
		reg.r.i  = 0x1e;
		reg.r.r  = 0xa6;
		reg.r.iff1 = reg.r.iff2 = disabled;

			ram[0x00] = 0xff; ram[0x01] = 0x80; ram[0x02] = 0xfc; 
			ram[0x03] = 0xbf; ram[0x04] = 0x00; ram[0x05] = 0xc0; 
			ram[0x06] = 0x00; ram[0x07] = 0xfe; ram[0x08] = 0xff; 
		
		return NULL; /* ok */
	}
	catch( file_error err )
	{		
		return err.Text();
	}
	catch( std::bad_alloc )
	{
		return "Out of memory";
	}
}















