/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define LOG		1

#import "globals.h"
#import "MachineList.h"
#import "Machine.h"



MachineList	machine_list;

/*	
################################################################

		list of all machines

		modifying is only allowed from the main event loop
		locking   is only required,
		- when modifying the machines list, or
		- when accessing from threads other than main event loop

################################################################ */


MachineList::MachineList ( )
:	HierObj("Machine List"),
	lock(PLock::recursive)
{
	XLogIn("new MachineList");
}


void MachineList::Unlink ( )	{ IERR(); }
void MachineList::Link ( int, HierObj* ) { IERR(); }
		
		
void MachineList::MoveToFront ( Machine* m  )
{
	XXXASSERT(MAIN_THREAD);
	PLocker z(lock); 

//	no action:
//	• not yet linked: !parent: happens, when @class screen creator ordersFront it's window => callback => this->MoveToFront
//	• already in front
	if( m->IsFirst() ) return;

	m->HierObj::MoveToFront(); 
} 


void MachineList::LinkMachine ( Machine* m  )
{
	XXXASSERT(MAIN_THREAD);
	PLocker z(lock); 
	XXXASSERT( m->Parent()==NULL );
	
	HierObj::LinkAsFirstChild(m); 
} 


void MachineList::UnlinkMachine ( Machine* m )
{
	XXXASSERT(MAIN_THREAD);
	PLocker z(lock); 
	XXXASSERT( m->Parent()==this );

	m->HierObj::Unlink();
}


/* ----	run all machines to read/write sound ----
		static member function
*/
void MachineList::RunMachinesForSound()
{	
	PLocker z(&lock);					// access lock for mlist / front_machine()
	Machine* m = FirstMachine();		// may be NULL at powerup or shutdown

	if(m) 
	{
		do
		{
			m->RunForSound();
			m = m->Next();
		}
		while( m != FirstMachine() );
	}
}





	
	