/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef MachineList_h
#define MachineList_h

#import "HierObj.h"
#import "zxsp_types.h"
#import "unix/pthreads.h"



class MachineList : private HierObj
{
	PLock			lock;										// access lock for machine_list / front_machine()

private:		
					MachineList		( MachineList const& );		// prohibit
	MachineList&	operator=		( MachineList const& );		// prohibit
	MachineList*	Copy			( ) const {return 0;};		// prohibit. ((copy HierObj))
	void			Unlink			( );						// prohibit
	void			Link			( int where, HierObj* );	// prohibit

public:

					MachineList		( );
					//~MachineList	( );

// Queries:

virtual bool		IsA				( int id ) const		{ return id==isa_MachineList || HierObj::IsA(id); }
	Machine*		FirstMachine	( ) const				{ return (Machine*)HierObj::FirstChild(); }

// Actions:
		
	void			MoveToFront			( Machine* );		// move to front in list
	void			LinkMachine			( Machine* );		// insert into list (as new front machine)
	void			UnlinkMachine		( Machine* );		// unlink from list
	void			RunMachinesForSound	( );
};


extern	MachineList machine_list;

inline Machine* front_machine() { return machine_list.FirstMachine(); };


#endif















