/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

Global Notes:

	compiler flag 'Accelerated Objective-C Dispatch'	führt zu Crash unter 10.3.9!

	Info.plist:											Groups&Files:Targets:zxsp:Inspector:Properties:
		Version:			0.6.6 pre66
		Principal Class:	ZxspApp  oder  NSApplication (default)
*/


#define	SAFE	3
#define	LOG 	1

#import "cstrings/cstrings.h"
#import "unix/tempmem.h"
#import "globals.h"
#import "ZxInfo.h"
#import "MainApp.h"
#import "Mouse.h"


cstr	appl_name			= "zxsp";
cstr	appl_path			= 0;	// set by main()
cstr	rsrc_path			= 0;	// set by main()
cstr	appl_version_str	= 0;	// set by main() from "Info.plist"		e.g.: "0.6.6"
cstr	appl_pre_version_str= 0;	// set by main() from "Info.plist"		e.g.: "" or "pre99"



/*	-----
	Notes
	-----
	
	1) in MainMenu.nib the File's Owner custom class is set to MainApp
	2) In Project's Application settings the Principal class is set to MainApp
*/

@implementation MainApp

/* Central Event Dispatcher:
*/
- (void) sendEvent: (NSEvent*) event
{
	XXLogIn("MainApp:sendEvent: %i",(int)[event type]);

//	special handling of events during mouse grabbed by mouse interface:
	if( Mouse::IsGrabbed() && Mouse::HandleEvent(event) ) return;

// Standard handling:
	[super sendEvent: event];
}

@end




// ==== MAIN ====

int main(int argc, const char *argv[])
{
	LogLine("\n*** Welcome to zxsp on Mac OS X ***");
	XLogLine( "    ---------------------------");
#if defined(DEBUG)
	Log(  "    Increased debugging level\n\n");
#elif defined(FINAL) 
	//		 Deployment version
#else
	Log("    Development version\n\n");
#endif

	extern double kCFCoreFoundationVersionNumber;	// CFBase.h
	{	// 299 ~ 10.3
		// 368 ~ 10.4
		// 476 ~ 10.5
		XLogLine("Core Foundation version number: %.2lf",kCFCoreFoundationVersionNumber);
	}
	
#if defined(_POWERPC)
	XLogLine("Architecture: PPC");
#elif defined(_I386)
	XLogLine("Architecture: i386");
#else
	LogLine("Architecture: ¿unknown?");
#endif

#if defined(_BIG_ENDIAN)
	XLogLine("  Byte order: *\"1234\" = '1234' (big endian, ppc)");
#elif defined(_LITTLE_ENDIAN)
	XLogLine("  Byte order: *\"1234\" = '4321' (little endian, i386)");
#else
	LogLine("  Byte order: ¿unknown?");
#endif

	XLogLine("  Alignment: %s", _ALIGNMENT_REQUIRED?"required":"not required" );
	XLogLine("    max.:    %i", int(_MAX_ALIGNMENT) );
	XLogLine("    short:   %i", int(_SHORT_ALIGNMENT) );
	XLogLine("    int:     %i", int(_INT_ALIGNMENT) );
	XLogLine("    int32:   %i", int(_LONG_ALIGNMENT) );
	XLogLine("    int64:   %i", int(_LONG_LONG_ALIGNMENT) );
	XLogLine("    double:  %i", int(_DOUBLE_ALIGNMENT) );
	XLogLine("    pointer: %i", int(_POINTER_ALIGNMENT) );

// validate byte order:
	char abcd[5]="abcd";
	TRAP( Peek4X(abcd)!='abcd' );
	TRAP( Peek4Z(abcd)!='dcba' );

#ifndef _STDINT_H_
	ABORT("_STDINT_H_ not defined <-> stdint.h not included");
#endif
	
// the following generates no code: (except if library functions are broken!)	
	TRAP( CountBits( (int   )0x01020304 ) != 5 );
	TRAP( CountBits( (uint  )0x01020304 ) != 5 );
	TRAP( CountBits( (long  )0x01020304 ) != 5 );
	TRAP( CountBits( (ulong )0x01020304 ) != 5 );
	TRAP( CountBits( (short )0x0304     ) != 3 );
	TRAP( CountBits( (ushort)0x0304     ) != 3 );
	TRAP( CountBits( (char  )0x3b       ) != 5 );
	TRAP( CountBits( (uchar )0x3b       ) != 5 );

	TRAP( (long )Log2(0x20000000) != 29 );
	TRAP( (long )Log2(0x3fffffff) != 29 );
	TRAP( (ulong)Log2(0x20000000) != 29 );
	TRAP( (ulong)Log2(0x3fffffff) != 29 );
	TRAP( (short)Log2(0x2000)     != 13 );
	TRAP( (short)Log2(0x3fff)     != 13 );
	TRAP( (uchar)Log2(0x20)       != 5 );
	TRAP( (char) Log2(0x3f)       != 5 );
	TRAP( Log2(1)!=0 );
	TRAP( Log2(0)!=0 );
	
	TRAP( Bits(0)!=0 );
	TRAP( Bits(1)!=0 );
	TRAP( Bits(2)!=1 );
	TRAP( Bits(3)!=2 );
	TRAP( Nibbles(0)!=0 );
	TRAP( Nibbles(1)!=0 );
	TRAP( Nibbles(2)!=1 );
	TRAP( Nibbles(16)!=1 );
	TRAP( Nibbles(17)!=2 );
	TRAP( Bytes(0)!=0 );
	TRAP( Bytes(1)!=0 );
	TRAP( Bytes(2)!=1 );
	TRAP( Bytes(256)!=1 );
	TRAP( Bytes(257)!=2 );
	

	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
	NSBundle* mainBundle = [NSBundle mainBundle];			// -> "zxsp.app/Info.plist"
		TRAP(!mainBundle);
	
// appl name:
		//appl_name = "zxsp";		oder aus mainBundle (Info.plist) auslesen...
		XLogLine("appl_name: %s",appl_name);

// invocation:
		appl_path = argv[0];
		XLogLine("appl_path: %s",appl_path);

// rsrc_path:
		NSString* bundlePath = [mainBundle bundlePath];			// "..../zxsp.app"
		TRAP(!bundlePath);
		XLogLine("bndl_path: %s",[bundlePath UTF8String]);	
		rsrc_path = NewCopy( CatStr( [bundlePath UTF8String], "/Contents/Resources/" ) );
		XLogLine("rsrc_path: %s",rsrc_path);

// version:
		NSString* versionString = [mainBundle objectForInfoDictionaryKey:@"CFBundleVersion"];	// "0.6.6 pre4"
		TRAP(!versionString);
		XXLogLine("CFBundleVersion: %s",[versionString UTF8String]);
		cstr s = [versionString UTF8String];
		cstr p = FindStr(s,"pre"); 
		if(p)	// preview version:
		{
			appl_pre_version_str = NewCopy(p);
			while(p[-1]<'0') p--;				// find last digit of version, discard optional '.' or ' '
			appl_version_str     = NewCopy(SubStr(s,p));
		}
		else
		{
			appl_pre_version_str = "";
			appl_version_str     = NewCopy(s);
		}
		XLogLine("version: %s %s",appl_version_str,appl_pre_version_str);

	[mainBundle release];
	[pool release];


// *DOIT*
    return NSApplicationMain( argc, argv );
}












