/*	Copyright  (c)	Günter Woigk 2002 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define	LOG		1

#import "cstrings/cstrings.h"
#import "utilities.h"
#import <Cocoa/Cocoa.h>		// how to include CGDirectDisplay.h ?




/*	Log CGRect bounds
*/
cstr BoundsStr( CGRect& box )
{
	return Using( "x=%.0f, y=%.0f, w=%.0f, h=%.0f", box.origin.x,box.origin.y,box.size.width,box.size.height );
}



/*	calculate a 2-byte crc checksum
	probably based on y-modem crc.
	the 'crc' argument allows incremental calculation.
*/
ushort Crc16 ( ptr p, int count, short crc /* == 0 */ )
{
    while( --count>=0 )
	{
        crc = crc ^ (short)*p++ << 8;
        for( int i=0; i<8; ++i ) { crc = crc<0 ? crc<<1 ^ 0x1021 : crc<<1; }
    }

    return crc;
}




/*	convert frequency to beautified text for display
	returned string is allocated in the cstr pool.
	note: frequency == 1/time
*/
cstr MHzStr ( Frequency frequency )
{
	str s = TempStr(16);
	if( frequency>=1000000 ) sprintf(s,"%.5g MHz",frequency/1000000.0); else
	if( frequency>=1000    ) sprintf(s,"%.5g kHz",frequency/1000.0); else
							 sprintf(s,"%.5g Hz", frequency); 
	return s;
}




/*	convert Port Address Specs, e.g.: "----.----.--01.----" 
	to Mask and Bits
*/
ushort BitsForSpec ( cstr s )
{
	for( ushort bits=0;;) 
		switch(*s++)
		{
		case '1': bits += bits +1; continue;
		case '0': 
		case '-': bits += bits; 
		case '.': continue;
		case  0 : return bits;
		default: IERR();
		}
}

ushort MaskForSpec ( cstr s )
{	
	for( ushort mask=0;;) 
		switch(*s++)
		{
		case '0': 
		case '1': mask += mask +1; continue;
		case '-': mask += mask; 
		case '.': continue;
		case  0 : return mask;
		default: IERR();
		}
}

cstr And( cstr a, cstr b )
{
	ptr z = DupStr(a);
	for( a=z; *z; z++,b++ ) if(*b!=*z) *z='-';
	return a;
}





 

 



 
 