/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#import "kio/kio.h"
#import "DspTime.h"

typedef double Time;		// new seconds-based time	[s]
typedef double Frequency;	// frequency	[1/s]

typedef float 	Sample;				typedef Sample const	cSample;
typedef uchar	uchar32[32];
typedef uchar	uchar64[64];
typedef uchar	uchar256[256];

enum	FileType	{ unknown_file=0, tap_file='.tap', tzx_file='.tzx', pzx_file='.pzx', 
					  rles_file='.rle', aiff_file='.aif', wav_file='.wav',
					  zx80_file='.80 ', o_file='.o  ', 
					  zx81_file='.81 ', p_file='.p  ' };

enum zx_isa_id {
	isa_MachineList = 1,		
	isa_Machine,
	isa_Item,
		isa_Mmu,
			isa_Mmu128k,
				isa_MmuPlus3,
		isa_Ay,
			isa_InternalAy,
		isa_Ula,
		isa_Z80,
		isa_FullerBox,
		isa_ZonxBox,
		isa_ICTester,
		isa_dktronicsDualJoystickInterface,
		isa_KempstonMouseInterface,
		isa_KempstonJoystickInterface,
		isa_SinclairZxInterface1,
		isa_SinclairZxInterface2,
		isa_SinclairZxPrinter,
		isa_ZxPlus2JoystickPorts,
		isa_ProtekJoystickInterface,
		isa_CursorJoystickInterface,
	isa_JoystickList,
	isa_Joystick,
		isa_KeyboardJoystick,
		isa_USBJoystick,
	isa_ZxPanel,
};

typedef class StereoSample;			typedef StereoSample const	cStereoSample;
typedef	class GifEncoder;
typedef	class Pixelmap;				typedef Pixelmap const cPixelmap;
typedef	class ZxPixelmap;
typedef uchar Comp;					typedef Comp const cComp;
typedef class Colormap;				typedef Colormap const cColormap;
typedef class ZxInfo; 
typedef class HierObj;
typedef class Sig;
typedef class ZxItem;
typedef class Machine;
typedef class MachineList;
typedef class Z80;
typedef class Ay;
typedef class Ula;
typedef class Mmu;
typedef class ICTester;
typedef class KempstonMouseInterface;
typedef class KempstonJoystickInterface;
typedef class SinclairZxInterface1;
typedef class SinclairZxInterface2;
typedef class ZxPlus2JoystickPorts;
typedef class SinclairZxPrinter;
typedef class FullerBox;
typedef class ZonxBox;
typedef class Display;				typedef Display* DisplayRef;
typedef	class Joystick;
typedef	class JoystickList;
typedef class DisplayList;
typedef class TempMemPool;


#ifdef __OBJC__
	@class  Screen;
	@class	NXPanel;
	@class	NXWindowMenu;
	@class	NXTextField;
	@class	Preferences;
	@class	MenuController;
	@class	TapeRecorder;
	@class	KeyboardJoystickPanel;
	@class	CheckUpdateController;
	@class	ZXPrinterPanel;
	@class	SinclairZxInterface2Panel;
	@class	SinclairZxInterface1Panel;
	@class	ZxPlus2JoystickPortsPanel;
	@class	KempstonJoystickInterfacePanel;
	@class	KempstonMouseInterfacePanel;
	@class	ICTesterPanel;
	@class	ZonxBoxPanel;
	@class	AyPanel;
	@class	CpuPanel;
	@class	KbdPanel;
	@class	MemoryPanel;
	@class	FdcPanel;
	@class	UlaPanel;
	@class	MachinePanel;

#else
	typedef	void Screen;
	typedef	void NXPanel;
	typedef	void NXWindowMenu;
	typedef	void NXTextField;
	typedef	void Preferences;
	typedef	void MenuController;
	typedef	void TapeRecorder;
	typedef	void KeyboardJoystickPanel;
	typedef	void CheckUpdateController;
	typedef	void ZXPrinterPanel;
	typedef	void SinclairZxInterface2Panel;
	typedef	void SinclairZxInterface1Panel;
	typedef	void ZxPlus2JoystickPortsPanel;
	typedef	void KempstonJoystickInterfacePanel;
	typedef	void KempstonMouseInterfacePanel;
	typedef	void ICTesterPanel;
	typedef	void ZonxBoxPanel;
	typedef	void AyPanel;
	typedef	void CpuPanel;
	typedef	void KbdPanel;
	typedef	void MemoryPanel;
	typedef	void FdcPanel;
	typedef	void UlaPanel;
	typedef	void MachinePanel;

	typedef signed long OSStatus;

#endif















