/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#import <Cocoa/Cocoa.h>
#include "globals.h"


extern NSThread* main_thread;
extern MenuController* menu_controller;
extern cstr GetRecentFilePath();



@interface MenuController : NSObject
{
    IBOutlet id menubar;

// zxsp menu:
	IBOutlet id about_panel;
	IBOutlet id view_version_number;
	IBOutlet id view_about_message;
	IBOutlet id view_pre_version;

// file menu:
    IBOutlet id item_reload;
    IBOutlet id menu_open_recent;
	IBOutlet id item_clear_recent;
    IBOutlet id item_screenshot;		// was: item_print;
    IBOutlet id item_record_movie;		// was: item_page_setup;

// edit menu:
    IBOutlet id item_cut;
    IBOutlet id item_copy;
    IBOutlet id item_paste;
    IBOutlet id item_clear;
    IBOutlet id item_select_all;

// model menu:
    IBOutlet id menu_model;
	// menu items are accessible via: [menu_model itemAtIndex:MODEL]

// items menu:
	IBOutlet id menu_items;
    IBOutlet id item_ay;
    IBOutlet id item_fuller_box;
    IBOutlet id item_zonx_box;
    IBOutlet id item_sinclair_zx_interface1;
    IBOutlet id item_kempston_joystick_interface;
    IBOutlet id item_kempston_mouse_interface;
    IBOutlet id item_builtin_joystick_ports;
    IBOutlet id item_sinclair_zx_interface2;
    IBOutlet id item_sinclair_zx_printer;
    IBOutlet id item_ic_tester;
    IBOutlet id item_ic_tester_submenu;
    IBOutlet id item_test_ic_none;
    IBOutlet id item_test_ic_7400;
    IBOutlet id item_test_ic_7402;
    IBOutlet id item_test_ic_7406;
    IBOutlet id item_test_ic_27128;

// options:
    IBOutlet id item_cpu_predivider_1;
    IBOutlet id item_cpu_predivider_2;
    IBOutlet id item_cpu_predivider_4;
    IBOutlet id item_cpu_predivider_8;

    IBOutlet id item_overdrive_50;
    IBOutlet id item_overdrive_75;
    IBOutlet id item_overdrive_90;
    IBOutlet id item_overdrive_100;
    IBOutlet id item_overdrive_120;
    IBOutlet id item_overdrive_150;
    IBOutlet id item_overdrive_200;
    IBOutlet id item_overdrive_400;
	
    IBOutlet id item_keyboard_basic_mode;
    IBOutlet id item_keyboard_game_mode;

    IBOutlet id item_halt_cpu_after_block_loaded;
    IBOutlet id item_instant_load_from_tape;
    IBOutlet id item_instant_save_to_tape;
    IBOutlet id item_autostart_tape;
    //IBOutlet id item_autostart_tape_for_load;
    //IBOutlet id item_autostart_tape_for_save;
    //IBOutlet id item_use_waitmap;
	
	//IBOutlet id item_audio_input_use_standard_device;
	//IBOutlet id item_audio_output_use_standard_device;
	IBOutlet id item_audio_input_enabled;
	IBOutlet id item_audio_input_from_file;
	IBOutlet id item_audio_output_enabled;
	IBOutlet id item_audio_output_to_file;
	NSSlider* audio_output_volume_slider;				// 10.5++

// control menu:
	// power_on, reset, nmi: ohne pointer (outlet)
    IBOutlet id item_halt_cpu;
    IBOutlet id item_step_in;
    IBOutlet id item_step_out;
    IBOutlet id item_step_over;
	
// window menu:
    IBOutlet id item_window_x1;
    IBOutlet id item_window_x2;
    IBOutlet id item_window_x3;
    IBOutlet id item_window_x4;
    IBOutlet id item_fullscreen;	
}

- (IBAction)doAbout:(id)sender;

- (IBAction)doCloseTopWindow:(id)sender;
- (IBAction)doNew:(id)sender;
- (IBAction)doOpen:(id)sender;
- (IBAction)doReload:(id)sender;
- (IBAction)doSaveAs:(id)sender;
- (IBAction)doOpenRecent:(id)sender;
- (IBAction)doClearRecent:(id)sender;
- (IBAction)doScreenshot:(id)sender;
- (IBAction)doRecordMovie:(id)sender;

- (IBAction)setZX80:(id)sender;
- (IBAction)setZX81:(id)sender;
- (IBAction)setJupiterAce:(id)sender;
- (IBAction)setZX16k:(id)sender;
- (IBAction)setZX48ki2:(id)sender;
- (IBAction)setZX48ki3:(id)sender;
- (IBAction)setZXplus:(id)sender;
- (IBAction)setZXplus_span:(id)sender;
- (IBAction)setZXplus128:(id)sender;
- (IBAction)setZXplus128span:(id)sender;
- (IBAction)setZXplus2:(id)sender;
- (IBAction)setZXplus2_span:(id)sender;
- (IBAction)setZXplus2_french:(id)sender;
- (IBAction)setZXplus3:(id)sender;
- (IBAction)setZXplus3_span:(id)sender;
- (IBAction)setZXplus2A:(id)sender;
- (IBAction)setZXplus2A_span:(id)sender;
- (IBAction)setPentagon:(id)sender;
- (IBAction)setSamCoupe:(id)sender;
- (IBAction)setScorpion:(id)sender;
- (IBAction)setTimex2068:(id)sender;
- (IBAction)setTimex2048:(id)sender;

- (IBAction)toggleAy:(id)sender;
- (IBAction)toggleFullerBox:(id)sender;
- (IBAction)toggleZonxBox:(id)sender;
- (IBAction)toggleSinclairZxInterface1:(id)sender;
- (IBAction)toggleKempstonJoystickInterface:(id)sender;
- (IBAction)toggleKempstonMouseInterface:(id)sender;
- (IBAction)toggleSinclairZxInterface2:(id)sender;
- (IBAction)toggleSinclairZxPrinter:(id)sender;
- (IBAction)toggle_ic_tester:(id)sender;
- (IBAction)toggle_test_ic_none:(id)sender;
- (IBAction)toggle_test_ic_7400:(id)sender;
- (IBAction)toggle_test_ic_7402:(id)sender;
- (IBAction)toggle_test_ic_7406:(id)sender;
- (IBAction)toggle_test_ic_27128:(id)sender;

- (IBAction)setCpuPredivider1:(id)sender;
- (IBAction)setCpuPredivider8:(id)sender;
- (IBAction)setCpuPredivider2:(id)sender;
- (IBAction)setCpuPredivider4:(id)sender;
- (IBAction)setMachineOverdrive50:(id)sender;
- (IBAction)setMachineOverdrive75:(id)sender;
- (IBAction)setMachineOverdrive90:(id)sender;
- (IBAction)setMachineOverdrive100:(id)sender;
- (IBAction)setMachineOverdrive120:(id)sender;
- (IBAction)setMachineOverdrive150:(id)sender;
- (IBAction)setMachineOverdrive200:(id)sender;
- (IBAction)setMachineOverdrive400:(id)sender;

- (IBAction)setKeyboardBasicMode:(id)sender;
- (IBAction)setKeyboardGameMode:(id)sender;

- (IBAction)toggleHaltCpuAfterLoad:(id)sender;
- (IBAction)toggleInstantLoadFromTape:(id)sender;
- (IBAction)toggleInstantSaveToTape:(id)sender;
//- (IBAction)toggleUseWaitmap:(id)sender;
- (IBAction)toggleAutostartTape:(id)sender;
//- (IBAction)toggleAutostartTapeForLoad:(id)sender;
//- (IBAction)toggleAutostartTapeForSave:(id)sender;

- (IBAction)audioInputEnabled:(id)sender;
- (IBAction)audioOutputEnabled:(id)sender;
- (IBAction)audioInputFromFile:(id)sender;
- (IBAction)audioOutputToFile:(id)sender;
- (void)setAudioOutputVolume:(id)sender;			// 10.5++ 

- (IBAction)doReset:(id)sender;
- (IBAction)doPowerOn:(id)sender;
- (IBAction)doNmi:(id)sender;
- (IBAction)toggleHaltCpu:(id)sender;
- (IBAction)doStepIn:(id)sender;
- (IBAction)doStepOver:(id)sender;
- (IBAction)doStepOut:(id)sender;

- (IBAction)setWindowX1:(id)sender;
- (IBAction)setWindowX2:(id)sender;
- (IBAction)setWindowX3:(id)sender;
- (IBAction)setWindowX4:(id)sender;
- (IBAction)toggleFullscreen:(id)sender;


-(cstr)queryTapefileForSaving;
-(cstr)queryTapefileForLoading;




@end











