/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	MenuController

	Der MenuController hat zwei Aufgaben:

1.	Der MenuController dient als eine zentrale Anlaufstelle für alle Menü-Events,
	weil die Events im IB nicht fest an eine Maschine gekoppelt werden können,
	da sich diese ja ändert, jenachdem welche vorne ist.
	Der MenuHandler leitet die Events zumeist an die vordere Maschine weiter und
	behandelt nur wenige selbst, zB. solche, die keiner Maschine zuzuordnen sind.
	Evtl. notwendige Aktualisierung von Menüs muss vom Empfänger der Nachrichten 
	veranlasst werden mittels updateXxxxMenu Nachrichten an den MenuController.

2.	Der MenuController stellt Funktionen bereit, um die angezeigten Menüs 
	entsprechend dem Zustand der vorderen Maschine zu aktualisieren. 
	Dies sind die updateXxxxMenu Nachrichten.

	Der MenüController dient nicht dazu, um über irgendwelche seiner Funktionen 
	Einstellungen in laufenden Maschinen zu ändern oder dort Aktionen anzustoßen.
	(Abgesehen davon, dass er die Menüaktivitäten des Benutzers durchleitet).
	Aufruf von IBActions um das zu erreichen ist verboten und kann ggf. zu 
	infiniter Rekursion MenuController->Machine->MenuController usw. führen.
*/

#define SAFE	3 
#define LOG		2


#import "unix/file_utilities.h"
#import "cstrings/cstrings.h"
#import "utilities.h"
#import "MenuController.h"
#import "CpuPanel.h"
#import "FdcPanel.h"
#import "SinclairZxInterface1Panel.h"
#import "KbdPanel.h"
#import "AyPanel.h"
#import "TapeRecorder.h"
#import "UlaPanel.h"
#import "Ula.h"
#import "ICTester.h"
#import "Machine.h"
#import "Screen.h"
#import "Z80.h"
#import	"Dsp.h"
#import	"ZxInfo.h"
#import "KempstonJoystickInterface.h"
#import "Preferences.h"
#import "AlertPanel.h"
#import "NXTextField.h"


@implementation MenuController


/* ---- recent file handling ----
*/
#define	MAX_RECENT_FILES	20

static NSMutableArray* recent_files = NULL;

cstr GetRecentFilePath() {
	if( [recent_files count]==0 ) return NULL;	
	NSString* filePath = [recent_files objectAtIndex:0];	
	return [filePath fileSystemRepresentation];
}

-(void)remove_recent_file_at_index:(int)i {
	[recent_files     removeObjectAtIndex:i];
	[menu_open_recent removeItemAtIndex:i];
}
-(void)remove_recent_file:(NSString*)filepath {
	for( int i=[recent_files count]; --i>=0; )							
	{
		NSString* s = [recent_files objectAtIndex:i];
		if( [filepath isEqualToString:s] ) [self remove_recent_file_at_index:i];
	}
}
-(void)disable_vanished_recent_files { // and enable_existing_recent_files

	for(uint i=0;i<[recent_files count]; i++)
	{
		NSString* file = [recent_files objectAtIndex:i];	
		BOOL exists = IsFile([file fileSystemRepresentation]);
		[[menu_open_recent itemAtIndex:i] setEnabled:exists];
	}
}
-(void)truncate_recent_files:(uint)n {
	if( [recent_files count]<=n ) return;

// 1.: remove vanished files first:
	for (int i=[recent_files count]-1;i>=0;i--)
	{			
		NSMenuItem* item = [menu_open_recent itemAtIndex:i];
		if ([item isEnabled]) continue;

		[self remove_recent_file_at_index:i]; 
		if([recent_files count]==n) return; 
	}

// 2.: remove files at list end:
	while( [recent_files count]>n ) 
	{
		[self remove_recent_file_at_index:n];
	}
}
-(void)add_recent_file:(NSString*)filepath {	
 
	NSString*   s = [filepath lastPathComponent];
	SEL         a = @selector(doOpenRecent:);
	NSMenuItem* m = [[NSMenuItem alloc] initWithTitle:s action:a keyEquivalent:@""];
	[recent_files     insertObject:filepath atIndex:0];
	[menu_open_recent insertItem:m          atIndex:0];
}
-(void)sync_recent_files {
	[preferences setStringArray:recent_files forKey:"recent_files"];
}
-(void)initRecentFiles {

	XLogIn("MenuController:initRecentFiles");

	recent_files = [[NSMutableArray alloc] init];
	NSArray* liste = [[preferences getStringArrayForKey:"recent_files"] retain];
	for( int i=[liste count]; --i>=0; ) [self add_recent_file:[liste objectAtIndex:i]];
	[item_clear_recent setEnabled:[recent_files count]!=0];
	[self disable_vanished_recent_files];
}
-(void)clearRecentFiles {

	XLogIn("MenuController:clearRecentFiles");
	
	[self truncate_recent_files:0];
	[item_clear_recent setEnabled:0];
	[self sync_recent_files];
}
-(void)addRecentFile:(NSString*)filepath {

	XLogIn("MenuController:addRecentFile");

	[self disable_vanished_recent_files];
	[self remove_recent_file:filepath];
	[self add_recent_file:filepath];
	[self truncate_recent_files:MAX_RECENT_FILES];
	[item_clear_recent setEnabled:1];
	[self sync_recent_files];
}
-(void)openRecentFile:(id)sender {
	/*
		open selected item in "Recent Files" menu
		in: sender = menu item
	*/
	XLogIn("MenuController:openRecentFile");
	
	NSMenu* mm = [sender menu]; if(!mm)return;			// owning menu
	int i = [mm indexOfItem:sender]; if(i<0)return;		// own index

	NSString* filePath = [recent_files objectAtIndex:i];	
	cstr file = [filePath fileSystemRepresentation];
	OSStatus err = error;
	LBLOCK; 
		err = m->LoadFile(file); 
		if(!err) [self addRecentFile:filePath];
	LUNBLOCK;
}


/* ----	UTILITIES ----
*/

#define CHECK(value,item)	{ static bool old = 0; if(old!=bool(value)) [item setState:  (old = bool(value))]; }
#define ENABL(value,item)	{ static bool old = 1; if(old!=bool(value)) [item setEnabled:(old = bool(value))]; }
#define ACTION(msg,action)	{ XLogIn("MenuController:%s",msg); QBLOCK; m->action; QUNBLOCK; }


-(void)openSnapshotWithDialog {

	XLogIn("MenuController:openSnapshotWithDialog");

	if( fullscreen_mode ) { RLOCK; m->SetFullscreen(off); RUNLOCK; }

	NSArray* types = [ NSArray arrayWithObjects:@"ass",@"ASS",@"src",@"SRC",@"scr",@"SCR",@"rom",@"ROM", 
								@"z80",@"Z80",@"sna",@"SNA",@"tap",@"tape",@"TAP",@"TAPE",
								@"80",@"81",@"o",@"O",@"p",@"P",
								/* @"gif",@"GIF", @"tzx",@"TZX", */ nil ];
	NSOpenPanel* dialog = [NSOpenPanel openPanel];
	int result = [dialog runModalForTypes:types];

	if( result==NSOKButton )
	{
		NSString* filePath = [[dialog filenames] objectAtIndex:0];
		cstr file = [filePath fileSystemRepresentation];
		OSStatus err = error;
		LBLOCK; 
			err = m->LoadFile(file); 
			if(!err) [self addRecentFile:filePath];
		LUNBLOCK;
	}
}
-(void)saveSnapshotWithDialog {

	XLogIn("MenuController:saveSnapshotWithDialog");

	if( fullscreen_mode ) { RLOCK; m->SetFullscreen(off); RUNLOCK; }
	
	LBLOCK; 

	  NSSavePanel* dialog = [NSSavePanel savePanel];
	  NSArray* ftlist = nil;
	  switch( m->GetModel() )
	  {
	  case zx80:	ftlist = [NSArray arrayWithObjects:@"z80",@"rom",@"scr",@"gif",@"80",@"o",nil]; break;
	  case zx81:	ftlist = [NSArray arrayWithObjects:@"z80",@"rom",@"scr",@"gif",@"81",@"p",nil]; break;
	  case zx16k:
	  case zx48i2:
	  case zx48i3:
	  case zxplus:	ftlist = [NSArray arrayWithObjects:@"z80",@"rom",@"scr",@"gif",@"sna",nil]; break;	
	  default:		ftlist = [NSArray arrayWithObjects:@"z80",@"rom",@"scr",@"gif",nil]; break;
	  };
	  
	  [dialog setAllowedFileTypes:ftlist];
	  int result = [dialog runModal];		// or: runModalForDirectory:file
	
		if( result==NSFileHandlingPanelOKButton )
		{
			NSString* filePath = [dialog filename];
			m->SaveFile ( [filePath fileSystemRepresentation] );
			[self addRecentFile:filePath];
		}

	LUNBLOCK; 
}
-(cstr)queryTapefileForLoading {
	/*
		ask for tape file to load.
		returns path or
		returns NULL if user canceled.
	*/
	XLogIn("MenuController:queryTapefileForLoading");

	if( fullscreen_mode ) { RLOCK; m->SetFullscreen(off); RUNLOCK; }
	cstr path = NULL;
	
	RLOCK; 

		NSOpenPanel* dialog = [NSOpenPanel openPanel];
		NSArray*     ftlist = nil;
		
		switch( m->GetModel() )
		{
		case zx80:	ftlist = [NSArray arrayWithObjects:@"80",@"o",nil,@"tzx",@"pzx",@"TZX",@"PZX",@"O",nil]; break;
		case zx81:	ftlist = [NSArray arrayWithObjects:@"81",@"p",nil,@"tzx",@"pzx",@"TZX",@"PZX",@"P",nil]; break;
		default:	ftlist = [NSArray arrayWithObjects:@"tap",@"TAP",nil,@"tzx",@"pzx",@"tape",@"TAPE",@"TZX",@"PZX",nil]; break;
		};

		int result = [dialog runModalForTypes:ftlist];
	
		if( result==NSFileHandlingPanelOKButton )
		{
			path = [[dialog filename] fileSystemRepresentation];
		}

	RUNLOCK; 
	
	return path;
}
-(cstr)queryTapefileForSaving {

	XLogIn("MenuController:queryTapefileForSaving");

	if( fullscreen_mode ) { RLOCK; m->SetFullscreen(off); RUNLOCK; }
	cstr path = NULL;
	
	RLOCK; 

		NSSavePanel* dialog = [NSSavePanel savePanel];
		NSArray*     ftlist = nil;
		
		switch( m->GetModel() )
		{
		case zx80:	ftlist = [NSArray arrayWithObjects:@"o",@"80",@"tzx",@"pzx",nil]; break;
		case zx81:	ftlist = [NSArray arrayWithObjects:@"p",@"81",@"tzx",@"pzx",nil]; break;
		default:	ftlist = [NSArray arrayWithObjects:@"tap",@"tape",@"tzx",@"pzx",nil]; break;
		};

		[dialog setAllowedFileTypes:ftlist];
		int result = [dialog runModal];		// or: runModalForDirectory:file
	
		if( result==NSFileHandlingPanelOKButton )
		{
			path = [[dialog filename] fileSystemRepresentation];
		}

	RUNLOCK; 
	
	return path;
}


/* ---- DELEGATE NOTIFICATIONS ------------------------------------------
*/

/*	startup:	
	applicationWillFinishLaunching:
	application:openFile:				<-- for drag'n'dropped documents
	applicationDidFinishLaunching:
*/
-(void)applicationWillBecomeActive:(NSNotification *)aNotification	{ XLogIn("MenuController:applicationWillBecomeActive"); }
-(void)applicationDidBecomeActive:(NSNotification *)aNotification	{ XLogIn("MenuController:applicationDidBecomeActive"); }
-(void)applicationWillResignActive:(NSNotification *)aNotification	{ XLogIn("MenuController:applicationWillResignActive"); }
-(void)applicationDidResignActive:(NSNotification *)aNotification	{ XLogIn("MenuController:applicationDidResignActive"); }
-(void)applicationWillHide:(NSNotification *)aNotification			{ XLogIn("MenuController:applicationWillHide"); }
-(void)applicationDidHide:(NSNotification *)aNotification			{ XLogIn("MenuController:applicationDidHide"); }
-(void)applicationWillUnhide:(NSNotification *)aNotification		{ XLogIn("MenuController:applicationWillUnhide"); }
-(void)applicationDidUnhide:(NSNotification *)aNotification			{ XLogIn("MenuController:applicationDidUnhide"); }
-(BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication*)theApplication	{ return YES; }
-(NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication*)sender			{ return NSTerminateNow; }

-(void)applicationWillTerminate:(NSNotification *)aNotification {	
	XLogIn("MenuController:applicationWillTerminate");
	XXXASSERT(MAIN_THREAD);

	RLOCK;
		//m->SavePreferencesNow();	Machine::~Machine() will do that
		[m->GetScreen() savePreferencesNow];	// use frontmost machine to save default screen position
	RUNLOCK;
	
	if( DSP::audio_out_filename!=NULL ) DSP::StopAudioOutputToFile();	// else the .aiff file remains unfinished / corrupted
	[tape_recorder ejectTape];		// save to tape file if modified

	if( audio_output_volume_slider ) {
		Sample v = [audio_output_volume_slider doubleValue];
		[preferences setFloat:v*v forKey:"audio_output_volume"];
	}
}



-(void)disableUnsupportedMenus {
	/*
		update menus
		enable/disable menus
		add/remove tick
	*/
	XLogIn("MenuController:disableUnsupportedMenus");
	
// file:
//	[item_print			setEnabled:0];		note: menu item is replaced with doScreenshot
//	[item_page_setup	setEnabled:0];		note: menu item is replaced with doRecordMovie
//	[item_open_recent	setEnabled:0];			
//	[item_clear_recent	setEnabled:0];			
//	[item_reload		setEnabled:0];

// edit:
	[item_cut			setEnabled:0];
	[item_copy			setEnabled:0];
	[item_paste			setEnabled:0];
	[item_clear			setEnabled:0];
	[item_select_all	setEnabled:0];

// model:
	[menu_model setAutoenablesItems:off];
	for( int i=0; i<zxmodels; i++ ) { [[menu_model itemAtIndex:i] setEnabled:zx_info[i].is_supported]; }

// items:
	[item_fuller_box					setEnabled:0];
//	[item_zonx_box						setEnabled:0];
//	[item_kempston_joystick_interface	setEnabled:0];
//	[item_kempston_mouse_interface		setEnabled:0];
//	[item_sinclair_zx_interface2		setEnabled:0];

	[item_sinclair_zx_printer			setEnabled:0];
	[item_sinclair_zx_interface1		setEnabled:0];

// disable IC Tester items:
#if 1
	[item_ic_tester			setEnabled:0];
	[item_ic_tester_submenu	setEnabled:0];	// das Item, nicht das Menu!	submenus haben kein disable...
#endif
//	[item_test_ic_none		setEnabled:0];	// nicht nötig
//	[item_test_ic_7400		setEnabled:0];	// "" ...
//	[item_test_ic_7402		setEnabled:0];
//	[item_test_ic_7406		setEnabled:0];
//	[item_test_ic_27128		setEnabled:0];

// truncate 'Items' menu:
// remove IC Tester items
	{	NSMenu* m = menu_items;
		int n = [m numberOfItems];
		while(	n-- && ![[m itemAtIndex:n] isEnabled] )			// note: separatorItems are disabled
		{
			[m removeItemAtIndex:n];
		}
	}

// options:	
//	[item_halt_cpu_after_block_loaded	setEnabled:0];
//	[item_autostart_tape_for_load		setEnabled:0];
//	[item_autostart_tape_for_save		setEnabled:0];
//	[item_enable_audio_input			setEnabled:AudioInputDevicePresent()];		<-- geht nicht, da Dsp noch nicht initialisiert.
extern double kCFCoreFoundationVersionNumber;	// CFBase.h
	{	bool pre_1_4 = kCFCoreFoundationVersionNumber < /*kCFCoreFoundationVersionNumber10_4*/ 368.00;
		if( pre_1_4 )
		{	[item_audio_input_from_file	setEnabled:0];
			[item_audio_output_to_file	setEnabled:0];
		}		
	}
	
// windows:
//	[item_fullscreen	setEnabled:0];
}
-(void)updateMenus {
	/*	update all menus
		called regularly
	*/
	RLOCK;
	
// Appl 
//	CHECK( [check_update_panel isVisible], item_check_update );

// File 
	ENABL( [recent_files count]!=0 || m->GetFilePath()!=NULL,item_reload );

// Edit 

// Model	
	int model = m->GetModel();			
	{	static int old_model = -1; 
		if( model!=old_model )
		{
			if(old_model!=-1)	[[menu_model itemAtIndex:old_model]	setState:off];
			old_model = model;	[[menu_model itemAtIndex:model]		setState:on ];				
		}
	}

// Items
	CHECK(m->kempston_joystick_interface,item_kempston_joystick_interface);
	CHECK(m->kempston_mouse_interface,item_kempston_mouse_interface);
	ENABL(model!=tc2048,item_kempston_joystick_interface);		
	CHECK(m->zxplus2_js_ports,item_builtin_joystick_ports);
	ENABL(off,item_builtin_joystick_ports);		// actually not user controlled
	CHECK(m->sinclair_zx_interface1,item_sinclair_zx_interface1);
	CHECK(m->sinclair_zx_interface2,item_sinclair_zx_interface2);
	CHECK(m->sinclair_zx_printer,item_sinclair_zx_printer);
	CHECK(m->fuller_box,item_fuller_box);

	// ay soundchip:
	CHECK( m->FindItem(isa_InternalAy), item_ay );
	ENABL( m->GetModel()>=zx16k && !m->GetInfo()->has_ay_soundchip, item_ay );
	CHECK( m->zonx_box, item_zonx_box );

	// IC Tester:	
	{	static bool state    = 0;	// not present
		if(state!=bool(m->ic_tester)) 
		{
			state = !state;
			CHECK(state,item_ic_tester); 
			[item_test_ic_none  setEnabled:state]; 
			[item_test_ic_7400  setEnabled:state]; 
			[item_test_ic_7402  setEnabled:state]; 
			[item_test_ic_7406  setEnabled:state]; 
			[item_test_ic_27128 setEnabled:state]; 
		}
		if(state)
		{
			static long old = -1; long selected = m->ic_tester->GetIC();
			if( selected!=old)
			{
				old = selected;
				[item_test_ic_none  setState:selected==0]; 
				[item_test_ic_7400  setState:selected==7400]; 
				[item_test_ic_7402  setState:selected==7402]; 
				[item_test_ic_7406  setState:selected==7406]; 
				[item_test_ic_27128 setState:selected==27128]; 
			}
		}
	}

// Options - Speed
	{	// cpu clock predivider: 
		static int old = -1; 
		int prediv = m->GetUla()->GetCpuClockPredivider();
		if( prediv!=old )
		{
			old = prediv;
			[item_cpu_predivider_1 setState:prediv==1];
			[item_cpu_predivider_2 setState:prediv==2];
			[item_cpu_predivider_4 setState:prediv==4];
			[item_cpu_predivider_8 setState:prediv==8];
		}
	}
	{	// machine overdrive <=> ula clock factor
		static double old = -1; 
		double fak = m->GetCpuCyclesPerSecond() * m->GetUla()->GetCpuClockPredivider() / m->GetInfo()->ula_cycles_per_second;
		if( fak!=old )
		{
			old = fak;
			[item_overdrive_50  setState:fak==0.5];
			[item_overdrive_75  setState:fak==0.75];
			[item_overdrive_90  setState:fak==0.9];
			[item_overdrive_100 setState:fak==1.0];
			[item_overdrive_120 setState:fak==1.2];
			[item_overdrive_150 setState:fak==1.5];
			[item_overdrive_200 setState:fak==2.0];
			[item_overdrive_400 setState:fak==4.0];
		}	
	}	
	
// Options
	{	static uint old = 0xfff; 	
		if(old!=m->GetKeyboardMode())
		{
			old = m->GetKeyboardMode();
			[item_keyboard_basic_mode setState:old==kbdlog];
			[item_keyboard_game_mode  setState:old==kbdphys];
		}
	}
	//CHECK(m->GetCpuOptions()&cpu_waitmap,item_use_waitmap);
	//ENABL(m->GetInfo()->has_contention(),item_use_waitmap);
	CHECK(m->HaltsCpuAfterBlockLoaded(),item_halt_cpu_after_block_loaded);
//	CHECK(m->InstantLoadsFromTape(),item_instant_load_from_tape);
	{	static cstr old = NULL; cstr nju = m->InstantLoadFilename(); 
		if(old!=nju) {
			old=nju; NSMenuItem* item = item_instant_load_from_tape; 
			cstr title = nju ? CatStr("Loading from \"",FileNameFromPath(nju),"\"") : "Instant Load from Tape…";
			[item setState:nju!=NULL]; 
			[item setTitle:[NSString stringWithUTF8String:title]]; 
		}	
	}
//	CHECK(m->InstantSavesToTape(),item_instant_save_to_tape);
	{	static cstr old = NULL; cstr nju = m->InstantSaveFilename(); 
		if(old!=nju) {
			old=nju; NSMenuItem* item = item_instant_save_to_tape; 
			cstr title = nju ? CatStr("Saving to \"",FileNameFromPath(nju),"\"") : "Instant Save to Tape…";
			[item setState:nju!=NULL]; 
			[item setTitle:[NSString stringWithUTF8String:title]]; 
		}	
	}
	CHECK(m->AutostartsTape(),item_autostart_tape);
	//CHECK(m->AutostartsTapeForLoad(),item_autostart_tape_for_load);
	//CHECK(m->AutostartsTapeForSave(),item_autostart_tape_for_save);

	//ENABL(DSP::AudioInputDevicePresent(),item_audio_input_use_standard_device);
	//CHECK(DSP::AudioInputDeviceEnabled(),item_audio_input_use_standard_device);
	//CHECK(DSP::AudioOutputDeviceEnabled(),item_audio_output_use_standard_device);

	ENABL(DSP::AudioInputDevicePresent(),item_audio_input_enabled);
	CHECK(DSP::AudioInputDeviceEnabled(),item_audio_input_enabled);

	//CHECK(DSP::AudioOutputDeviceEnabled(),item_audio_output_enabled);
	{	static bool old = 0; 
		bool nju = DSP::AudioOutputDeviceEnabled();
		if(old!=nju) {
			[item_audio_output_enabled setState:  (old = nju)]; 
			if( [item_audio_output_enabled hasSubmenu] ) {			// 10.5: volume slider
				[[[[item_audio_output_enabled submenu] itemAtIndex:0] performSelector:@selector(view)] setEnabled:nju];				
			}
		}
	}

//	CHECK(DSP::audio_in_filename!=NULL,item_audio_input_from_file);
	{	static cstr old = NULL; cstr nju = DSP::audio_in_filename; 
		if(old!=nju) {
			old=nju; NSMenuItem* item = item_audio_input_from_file; 
			cstr title = nju ? CatStr("From File: \"",FileNameFromPath(nju),"\"") : "From File…";
			[item setState:nju!=NULL]; 
			[item setTitle:[NSString stringWithUTF8String:title]]; 
		}
	}	
//	CHECK(DSP::audio_out_filename!=NULL,item_audio_output_to_file);
	{	static cstr old = NULL; cstr nju = DSP::audio_out_filename; 
		if(old!=nju) {
			old=nju; NSMenuItem* item = item_audio_output_to_file; 
			cstr title = nju ? CatStr("Record to File: \"",FileNameFromPath(nju),"\"") : "Record to File…";
			[item setState:nju!=NULL]; 
			[item setTitle:[NSString stringWithUTF8String:title]]; 
		}	
	}

// Control
	{	static int old = -1; int halted = m->IsHalted();
		if( halted!=old )
		{
			old = halted;
			[item_halt_cpu  setState:  halted];
			[item_step_in   setEnabled:halted];
			[item_step_over setEnabled:halted];
			[item_step_out  setEnabled:halted];	
		}
	}
	
// Window
	{	static int old = -1; int zoom = m->IsFullscreen() ? 0 : [m->GetScreen() getZoom];
		if(old!=zoom)
		{
			old = zoom;
			[item_fullscreen setState:zoom==0];		
			[item_window_x1  setState:zoom==1];
			[item_window_x2  setState:zoom==2];
			[item_window_x3  setState:zoom==3];
			[item_window_x4  setState:zoom==4];
		}
	}
	
	RUNLOCK;
}
-(void)attachVolumeControl {

	audio_output_volume_slider = NULL;
	
	NSMenuItem* my_menuitem = item_audio_output_enabled;
	SEL setview = @selector(setView:);

	if( [my_menuitem respondsToSelector:setview] )			// => 10.5 or higher
	{
		XLogLine("Adding audio output volume control");
	
		[my_menuitem setTitle:@"Audio Output Volume"];		// change from "Enabled" to "Volume"
		NSMenu* my_submenu = [[NSMenu alloc]initWithTitle:@""];
		[my_submenu setAutoenablesItems:NO];
		[my_menuitem setSubmenu:my_submenu];
		NSMenuItem* my_slideritem = [[NSMenuItem alloc]initWithTitle:@"" action:NULL keyEquivalent:@""];
		[my_submenu addItem:my_slideritem];
		NSSlider* slider = [[NSSlider alloc]init];
		[slider setMinValue:0];
		[slider setMaxValue:1];
		Sample volume = [preferences getFloatForKey:"audio_output_volume"];
		[slider setFloatValue:sqrtf(volume)];	[self setAudioOutputVolume:slider];
		[slider setFrameSize:NSMakeSize(21,150)];
		[slider setAction:@selector(setAudioOutputVolume:)];
		[slider setTarget:self];
		//[slider sendActionOn:];		during mouse tracking only
		[slider setContinuous:yes];		// => continuous events during mouse tracking
		[my_slideritem performSelector:setview withObject:slider];	// 10.5++ only, daher etwas umständlich wg. 10.4
		audio_output_volume_slider = slider;
	}
}
-(void)setAudioOutputVolume:(id)sender {

	Sample v = [sender doubleValue];	// should be in range [0.0 ... 1.0]
	DSP::SetOutputVolume( v*v );
}




/* ====	MENU HOOKS ==================================
	Der MenuController dient als eine zentrale Anlaufstelle für alle Menü-Events,
	weil die Events im IB nicht fest an eine Maschine gekoppelt werden können,
	da sich diese ja ändert, jenachdem welche vorne ist.
	Der MenuHandler leitet die Events zumeist an die vordere Maschine weiter und
	behandelt nur wenige selbst, z.B. solche, die keiner Maschine zuzuordnen sind.
*/

/*	appl menu:
*/
//- (IBAction)doCheckUpdate:(id)sender		{ XLogIn("MenuController:doCheckUpdate"); [check_update_panel orderFront:self]; }

/* file menu:
*/
- (IBAction)doReload:(id)sender				{ ACTION("doReload",ReloadSnapshot()); }
- (IBAction)doSaveAs:(id)sender				{ [self saveSnapshotWithDialog]; }
- (IBAction)doCloseTopWindow:(id)sender		{ XLogIn("MenuController:doCloseTopWindow"); LBLOCK; delete m; LNOBLOCK; }
- (IBAction)doOpen:(id)sender				{ [self openSnapshotWithDialog]; }
- (IBAction)doNew:(id)sender				{ XLogIn("MenuController:doNew"); if(!fullscreen_mode) new Machine(); }
- (IBAction)doOpenRecent:(id)sender			{ [self openRecentFile:sender]; }
- (IBAction)doClearRecent:(id)sender		{ [self clearRecentFiles]; }
- (IBAction)doScreenshot:(id)sender			{ ACTION("doScreenshot",SaveScreenshot()); }
- (IBAction)doRecordMovie:(id)sender		{ ACTION("doRecordMovie",RecordMovie()); }

/*	model menu:
*/
- (IBAction)setPentagon:(id)sender			{ ACTION("setPentagon",SetModel(pentagon)); }
- (IBAction)setSamCoupe:(id)sender			{ ACTION("setSamCoupe",SetModel(samcoupe)); }
- (IBAction)setScorpion:(id)sender			{ ACTION("setScorpion",SetModel(scorpion)); }
- (IBAction)setTimex2068:(id)sender			{ ACTION("setTimex2068",SetModel(ts2068)); }
- (IBAction)setTimex2048:(id)sender			{ ACTION("setTimex2048",SetModel(tc2048)); }
- (IBAction)setZX16k:(id)sender				{ ACTION("setZX16k",SetModel(zx16k)); }
- (IBAction)setZX48ki2:(id)sender			{ ACTION("setZX48ki2",SetModel(zx48i2)); }
- (IBAction)setZX48ki3:(id)sender			{ ACTION("setZX48ki3",SetModel(zx48i3)); }
- (IBAction)setZX80:(id)sender				{ ACTION("setZX80",SetModel(zx80)); }
- (IBAction)setZX81:(id)sender				{ ACTION("setZX81",SetModel(zx81)); }
- (IBAction)setZXplus:(id)sender			{ ACTION("setZXplus",SetModel(zxplus)); }
- (IBAction)setZXplus_span:(id)sender		{ ACTION("setZXplus_span",SetModel(zxplus_span)); }
- (IBAction)setZXplus128:(id)sender			{ ACTION("setZXplus128",SetModel(zx128)); }
- (IBAction)setZXplus128span:(id)sender		{ ACTION("setZXplus128span",SetModel(zx128_span)); }
- (IBAction)setZXplus2:(id)sender			{ ACTION("setZXplus2",SetModel(zxplus2)); }
- (IBAction)setZXplus2_span:(id)sender		{ ACTION("setZXplus2_span",SetModel(zxplus2_span)); }
- (IBAction)setZXplus2_french:(id)sender	{ ACTION("setZXplus2_french",SetModel(zxplus2_french)); }
- (IBAction)setZXplus3:(id)sender			{ ACTION("setZXplus3",SetModel(zxplus3)); }
- (IBAction)setZXplus3_span:(id)sender		{ ACTION("setZXplus3_span",SetModel(zxplus3_span)); }
- (IBAction)setZXplus2A:(id)sender			{ ACTION("setZXplus2A",SetModel(zxplus2a)); }
- (IBAction)setZXplus2A_span:(id)sender		{ ACTION("setZXplus2A_span",SetModel(zxplus2a_span)); }
- (IBAction)setJupiterAce:(id)sender		{ ACTION("setJupiterAce",SetModel(jupiter)); }

/*	items menu:
*/
- (IBAction)toggleAy:(id)sender						{ ACTION("toggleInternalAy",ToggleInternalAy()); }
- (IBAction)toggleFullerBox:(id)sender				{ ACTION("toggleFullerBox",ToggleFullerBox()); }
- (IBAction)toggleZonxBox:(id)sender				{ ACTION("toggleZonxBox",ToggleZonxBox()); }
- (IBAction)toggleSinclairZxInterface1:(id)sender	{ ACTION("toggleSinclairZxInterface1",ToggleSinclairZxInterface1()); }
- (IBAction)toggleKempstonJoystickInterface:(id)s	{ ACTION("toggleKempstonJoystickInterface",ToggleKempstonJoystickInterface()); }
- (IBAction)toggleKempstonMouseInterface:(id)s		{ ACTION("toggleKempstonMouseInterface",ToggleKempstonMouseInterface()); }
- (IBAction)toggleSinclairZxInterface2:(id)s		{ ACTION("toggleSinclairZxInterface2",ToggleSinclairZxInterface2()); }
- (IBAction)toggleSinclairZxPrinter:(id)s			{ ACTION("toggleSinclairZxPrinter",ToggleSinclairZxPrinter()); }

- (IBAction)toggle_ic_tester:(id)s			{ ACTION("toggle_ic_tester",toggle_ic_tester()); }
- (IBAction)toggle_test_ic_none:(id)s		{ ACTION("toggle_test_ic_none",toggle_test_ic_none()); }
- (IBAction)toggle_test_ic_7400:(id)s		{ ACTION("toggle_test_ic_7400",toggle_test_ic_7400()); }
- (IBAction)toggle_test_ic_7402:(id)s		{ ACTION("toggle_test_ic_7402",toggle_test_ic_7402()); }
- (IBAction)toggle_test_ic_7406:(id)s		{ ACTION("toggle_test_ic_7406",toggle_test_ic_7406()); }
- (IBAction)toggle_test_ic_27128:(id)s		{ ACTION("toggle_test_ic_27128",toggle_test_ic_27128()); }

/*	options menu:
*/
- (IBAction)setCpuPredivider1:(id)sender	{ ACTION("setCpuClockPredivider1",SetCpuClockPredivider(1)); }
- (IBAction)setCpuPredivider2:(id)sender	{ ACTION("setCpuClockPredivider2",SetCpuClockPredivider(2)); }
- (IBAction)setCpuPredivider4:(id)sender	{ ACTION("setCpuClockPredivider4",SetCpuClockPredivider(4)); }
- (IBAction)setCpuPredivider8:(id)sender	{ ACTION("setCpuClockPredivider8",SetCpuClockPredivider(8)); }

- (IBAction)setMachineOverdrive50:(id)s		{ ACTION("setMachineOverdrive50",SetMachineOverdrive(0.5)); }
- (IBAction)setMachineOverdrive75:(id)s		{ ACTION("setMachineOverdrive75",SetMachineOverdrive(0.75)); }
- (IBAction)setMachineOverdrive90:(id)s		{ ACTION("setMachineOverdrive90",SetMachineOverdrive(0.9)); }
- (IBAction)setMachineOverdrive100:(id)s	{ ACTION("setMachineOverdrive100",SetMachineOverdrive(1.0)); }
- (IBAction)setMachineOverdrive120:(id)s	{ ACTION("setMachineOverdrive120",SetMachineOverdrive(1.2)); }
- (IBAction)setMachineOverdrive150:(id)s	{ ACTION("setMachineOverdrive150",SetMachineOverdrive(1.5)); }
- (IBAction)setMachineOverdrive200:(id)s	{ ACTION("setMachineOverdrive200",SetMachineOverdrive(2.0)); }
- (IBAction)setMachineOverdrive400:(id)s	{ ACTION("setMachineOverdrive400",SetMachineOverdrive(4.0)); }

- (IBAction)setKeyboardBasicMode:(id)sender	{ ACTION("setKeyboardBasicMode",SetKeyboardMode(kbdlog)); }
- (IBAction)setKeyboardGameMode:(id)sender	{ ACTION("setKeyboardGameMode",SetKeyboardMode(kbdphys)); }

- (IBAction)toggleHaltCpuAfterLoad:(id)s	{ ACTION("toggleHaltCpuAfterLoad",SetHaltCpuAfterLoad(![s state])); }
- (IBAction)toggleInstantLoadFromTape:(id)s	{
	cstr path = [s state] ? NULL : [self queryTapefileForLoading];
	RLOCK;
		m->SetInstantLoadFromTape(path);	// path!=NULL => on / path==NULL => off
	RUNLOCK;
}
- (IBAction)toggleInstantSaveToTape:(id)s	{ 
	cstr path = [s state] ? NULL : [self queryTapefileForSaving];
	RLOCK;
		m->SetInstantSaveToTape(path);	// path!=NULL => on / path==NULL => off
	RUNLOCK;
}
- (IBAction)toggleAutostartTape:(id)s		{ ACTION("toggleAutostartTape",SetAutostartTape(![s state])); }
//- (IBAction)toggleAutostartTapeForLoad:(id)s{ ACTION("toggleAutostartTapeForLoad",SetAutostartTapeForLoad(![s state])); }
//- (IBAction)toggleAutostartTapeForSave:(id)s{ ACTION("toggleAutostartTapeForSave",SetAutostartTapeForSave(![s state])); }
//- (IBAction)toggleUseWaitmap:(id)s  		{ ACTION("toggleUseWaitmap",SetUseWaitmap(![s state])); }
-(IBAction)audioInputEnabled:(id)s			{
	bool f=![s state]; 
	XLogIn("MenuController:audioInputEnabled: %s",(onoffstr(f))); 
	DSP::EnableAudioInputDevice(f); 
}
-(IBAction)audioOutputEnabled:(id)s			{
	bool f=![s state]; 
	XLogIn("MenuController:audioOutputEnabled: %s",(onoffstr(f))); 
	DSP::EnableAudioOutputDevice(f); 
}
-(IBAction)audioInputFromFile:(id)s			{

	if( DSP::audio_in_filename )
	{
		DSP::StopAudioInputFromFile();
	}
	else
	{
		try	{ DSP::StartAudioInputFromFile(); }
		catch( any_error e ) { ShowWarning( "Audio Input From File failed: \n%s",e.Text() ); }
	}
}
-(IBAction)audioOutputToFile:(id)s			{

	if( DSP::audio_out_filename!=NULL )
	{
		DSP::StopAudioOutputToFile();
	}
	else
	{
		try	{ DSP::StartAudioOutputToFile(); }
		catch( any_error e ) { ShowWarning( "Audio Output To File failed: \n%s",e.Text() ); }
	}
}


/* control menu:
*/
-(IBAction)doPowerOn:(id)sender				{ ACTION("doPowerOn",PowerOn()); }		// also reset timings et.al.
-(IBAction)doNmi:(id)sender					{ ACTION("doNmi",Nmi()); }
-(IBAction)doReset:(id)sender				{ ACTION("doReset",Reset()); }
-(IBAction)toggleHaltCpu:(id)sender			{ ACTION("toggleHaltCpu",HaltCpu(![sender state])); }
-(IBAction)doStepOver:(id)sender			{ ACTION("doStepOver",StepOver()); }
-(IBAction)doStepOut:(id)sender				{ ACTION("doStepOut",StepOut()); }
-(IBAction)doStepIn:(id)sender				{ ACTION("doStepIn",StepIn()); }

/* window menu:
*/																				// note: setZoom needs no BLOCKing
-(IBAction)setWindowX1:(id)sender			{ XLogIn("MenuController:setWindowX1"); RLOCK; m->SetWindowZoom(1); RUNLOCK; }
-(IBAction)setWindowX2:(id)sender			{ XLogIn("MenuController:setWindowX2"); RLOCK; m->SetWindowZoom(2); RUNLOCK; }
-(IBAction)setWindowX3:(id)sender			{ XLogIn("MenuController:setWindowX3"); RLOCK; m->SetWindowZoom(3); RUNLOCK; }
-(IBAction)setWindowX4:(id)sender			{ XLogIn("MenuController:setWindowX4"); RLOCK; m->SetWindowZoom(4); RUNLOCK; }
-(IBAction)toggleFullscreen:(id)sender		{ ACTION("toggleFullscreen",SetFullscreen(![sender state])); }






/* ----	INITIALIZATION ------------------------------------------------
*/

- (id)init
{
	XLogIn("MenuController:init");			// note: wird nach App.-Start ausgeführt!

//	main_thread = [NSThread currentThread];	// remember main thread

	self = [super init];	
	if (self) [NSApp setDelegate:self];
	menu_controller = self;
	return self;
}

- (void)awakeFromNib
{
	XLogIn("MenuController:awakeFromNib");
	menu_controller = self;
	[self disableUnsupportedMenus];
	[self initRecentFiles];
	[self attachVolumeControl];

// einen ersten Thread starten, damit Cocoa in den Multi-Threaded-Mode geht:
	[NSThread detachNewThreadSelector:@selector(goMultiThreaded:) toTarget:self withObject:nil];
}


-(void)goMultiThreaded:(id)dummy
{
	XLogIn("MenuController:goMultiThreaded");	
	/* nothing to do */
}


- (void)applicationWillFinishLaunching:(NSNotification*)msg	
{ 
	XLogIn("MenuController:applicationWillFinishLaunching"); 
	DSP::StartCoreAudio();			// <-- also calculates samples_per_second	
}


/* ----	open file event ----
		open file with double click or drag&drop:
*/
- (BOOL)application:(NSApplication*)app openFile:(NSString*)filePath
{
	XLogIn("MenuController:application:openFile"); 
	if( !front_machine() || [preferences getBoolForKey:"newsnapshot_opens_in_new_window"] )
	{
		new Machine([filePath fileSystemRepresentation]);
		[self addRecentFile:filePath];
		return true;		// true == ok
	}
	else
	{
		OSStatus err = error;
		cstr file = [filePath fileSystemRepresentation];
		LBLOCK;
			err = m->LoadFile(file); 
			if(!err) [self addRecentFile:filePath];
		LUNBLOCK;
		return err==ok;	// true = ok
	}
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification	
{
	XLogIn("MenuController:applicationDidFinishLaunching"); 

// start menu update thread:
	[NSTimer scheduledTimerWithTimeInterval:0.5/*sec*/ target:self selector:@selector(updateMenus) userInfo:(id)NULL repeats:yes];

// if started without file then show splash screen and open first machine:
	if( !front_machine() )
	{
		[self doAbout:self];			// show splash screen for 2.5 sec
		[NSTimer scheduledTimerWithTimeInterval:2.5/*sec*/ target:about_panel selector:@selector(close) userInfo:(id)NULL repeats:no];
		new Machine();					// create first machine
	}

// open panels according to prefs:	
	if( [preferences getBoolForKey:"startup_open_keyboard"] )		[kbd_panel  orderFront:self];
	if( [preferences getBoolForKey:"startup_open_taperecorder"] )	[tape_recorder orderFront];
	if( [preferences getBoolForKey:"startup_open_machine_image"] )	[machine_panel showPanel];
}


- (IBAction)doAbout:(id)sender
{	
	XLogIn("MenuController:doAbout"); 

	[view_version_number setStrValue:appl_version_str];
	[view_pre_version    setStrValue:appl_pre_version_str];

	[view_about_message	 setStrValue:
		"Copyright © 1994-2008 Günter Woigk\n"
		"Erlangen, Germany, kio@little-bat.de\n"
		"Z80 emulation based on fMSX\n"
		"© 1994-1995 Marat Fayzullin\n"
		"ZX Spectrum keyboard images by Andrew Owen\n"
		"ZX80/ZX81/Jupiter Ace emulation by Paul Farrow" ];

	NSScreen* ms	= [NSScreen mainScreen];
	NSRect ms_frame	= [ms visibleFrame];

// main screen alert position:
	float x = ms_frame.origin.x + ms_frame.size.width/2;
	float y = ms_frame.origin.y + ms_frame.size.height*3/4;

// move there and show:	
	NSRect alert_box = [about_panel frame];
	x -= alert_box.size.width / 2;
	y -= alert_box.size.height * 3/4;
	[about_panel setFrameOrigin:NSMakePoint(int(x),int(y))];
	[about_panel orderFront:self];
}


@end








		
		
		