/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#import "kio/kio.h"


#ifdef __OBJC__
	@class NSEvent;
#endif

namespace Mouse
{

/*	Show / hide mouse pointer:
*/
extern	int 	hide_mouse_count;
inline	void	HideMouse()					{ hide_mouse_count++; CGDisplayHideCursor(kCGDirectMainDisplay); } 
inline	void	ShowMouse()					{ hide_mouse_count--; CGDisplayShowCursor(kCGDirectMainDisplay); } 
inline	bool	MouseIsVisible()			{ return hide_mouse_count==0; }
inline	bool	MouseIsHidden()				{ return hide_mouse_count>0; }


/*	Lock / unlock mous pointer position:
*/
extern	bool	mouse_is_nailed;
inline	CGError	NailMousePointer( bool f )	{ mouse_is_nailed=f; return CGAssociateMouseAndMouseCursorPosition(!f); } 
inline	bool	MouseIsNailed()				{ return mouse_is_nailed; }

inline	CGError	MoveMouseTo	( CGPoint globPos )	{ return CGWarpMouseCursorPosition(globPos); } // no mouseMoved event


/*	Grab / release mouse pointer for virtual machine:
*/
extern	bool	mouse_is_grabbed;
inline	bool	IsGrabbed()					{ return mouse_is_grabbed; }
extern	void	GrabMouse();
extern	void	ReleaseMouse();
extern	bool	HandleEvent( NSEvent* );	// return 'true' if handled
extern	void	HandleCommandKey();



// Buttons:
extern	bool	left_button;			// 1 if pressed
extern	bool	right_button;
extern	bool	middle_button;

inline	void	AllButtonsUp( )			{ left_button=right_button=middle_button = 0; }
inline	bool	LeftButton()			{ return left_button; }
inline	bool	RightButton()			{ return right_button; }
inline	bool	MiddleButton()			{ return middle_button; }


// X- and Y-axis:
// x_pos and y_pos are stored scaled
//
const	uint	x_max		= 256;
const	uint	y_max		= 256;

extern	uint	scaling;				// x and y scaling: 1:1, 1:2, 1:3 ...
extern	uint	scaled_x;				// scaled x-axis
extern	uint	scaled_y;				// scaled y-axis

inline	uint	XPos		( )			{ return scaled_x / scaling; }
inline	uint	YPos		( )			{ return scaled_y / scaling; }

inline	uint	Scaling		( )			{ return scaling; }
inline	void	SetScaling	( uint n )	{ scaled_x = scaled_x*n/scaling; scaled_y = scaled_y*n/scaling; scaling=n; }

}












