/*	Copyright  (c)	Günter Woigk 2006 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#import <Cocoa/Cocoa.h>
#import "globals.h"
#import "HierObj.h"


@interface NXPanel : NSPanel
{
	class ZxPanel*			cppPanel;	// my cpp counterpart

	NSTextField*			edit_view;	// current edited view (if any)
	bool					do_update;	// automatically cleared if (void)updatePanel:(Machine*)m is not used (reimplemented)
	IBOutlet NSMenuItem*	menu_item;	// associated window menu item for this panel
}

+(NXPanel*)panelForMenuItem:(id)item;

-(void)controlTextDidBeginEditing:(NSNotification*)note;
-(void)controlTextDidEndEditing:(NSNotification*)note;

-(void)updatePanel:(Machine*)m;			// *** this is to be re-implemented in sub classes ***

-(void)setWindowMenuItem:(NSMenuItem*)theItem;
-(void)createWindowMenuItem;
-(void)removeWindowMenuItem;

-(void)showPanel;						// orderFront
-(void)hidePanel;						// orderOut

-(void)enableSubviews:(bool)yesno;

@end



class ZxPanel : public HierObj
{
	NXPanel*	objcPanel;		// my Obj-C counterpart

private:
				ZxPanel			( ZxPanel const& );		// prohibit
		ZxPanel* operator =		( ZxPanel const& );		// prohibit

virtual ZxPanel* Copy			( ) const				{ IERR(); return new ZxPanel(*this); }	// prohibit
				
public:
				ZxPanel			( uint pins, cstr name, NXPanel* );
				~ZxPanel		( );

// IsA stuff:

virtual bool	IsA				( int id ) const		{ return id==isa_ZxPanel || HierObj::IsA(id); }


// Actions:

static ZxPanel*	PanelForMenuItem		( NSMenuItem* );

void			UpdatePanel				( Machine* );		// *** this is to be re-implemented in sub classes ***

void			SetWindowMenuItem		( NSMenuItem* );
void			CreateWindowMenuItem	( );
void			RemoveWindowMenuItem	( );

void			ShowPanel				( );				// orderFront
void			HidePanel				( );				// orderOut

void			EnableSubviews			( bool );
	
};














