/*	Copyright  (c)	Günter Woigk 2005 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define LOG 	1

#import "kio/kio.h"
#import "pthreads.h"
#import "NXTextField.h"
#import	"MenuController.h"




@implementation NXTextField


- (void)setStrValue:(cstr)text
{
	[super setStringValue:[NSString stringWithUTF8String:(text?text:"")]];
}


/*	Set display from speaking bits
	the displayed text is blended from visual strings 'Zeros' and 'Ones'
	each '0' bit in value selects the character from 'Zeroes' and
	each '1' bit in value selects the character from 'Ones'
	Between significant characters Beautification characters can be inserted,
	which are those characters which are the same in 'Zeroes' and 'Ones'.
	
	e.g.:	setBinaryValue:0x55 withZeroes:"%---0.0000" andOnes: "%---1.1111"
			results in: "%---1.0101"
*/
-(void)setBinaryValue:(uint)value withZeros:(cstr)s0 andOnes:(cstr)s1
{
	uint len = strlen(s0);	
	XXASSERT( len==strlen(s1) );
	char bu[len+1];
	memcpy( bu, s0, len+1 );

	cptr q=s1+len;
	ptr  z=bu+len;

	while( value && z>bu )
	{
		char c = *--q; 
		char a = *--z;
		if(a==c) continue;
		
		if(value&1) *z = c;
		value = value >> 1;
	}
	[self setStrValue:bu];
}


#ifndef FINAL

/* ==== setNeedsDisplayInRect ===================================

	Man kann zwar NSViews in separaten Threads neu zeichnen,
	man darf sie aber nur im Main Thread als dirty markieren.
	Deshalb wird das in "setNeedsDisplayInRect" geprüft.
	
	Note: Alle "setXXXValue" Methoden rufen "setNeedsDisplayInRect" auf!

============================================================== */


- (void) setNeedsDisplayInRect:(NSRect)theRect
{
	XXXASSERT( [NSThread currentThread] == main_thread );		
	[super setNeedsDisplayInRect:theRect];
}

- (void) setNeedsDisplay
{
	XXXASSERT( [NSThread currentThread] == main_thread );		
	[super setNeedsDisplay];
}


#endif


@end
