/*	Copyright  (c)	Günter Woigk 2002 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#import <Cocoa/Cocoa.h>
#import "kio/kio.h"
#import "NXPanel.h"

@interface Preferences : NXPanel
{
	IBOutlet id my_tab_view;

    IBOutlet id startup_open_keyboard;					
    IBOutlet id startup_open_taperecorder;					
    IBOutlet id startup_open_machine_image;		
    IBOutlet id startup_enable_audio_input;					
    IBOutlet id startup_fullscreen_mode;					

    IBOutlet id runtime_leave_keyboard_open;
    IBOutlet id runtime_leave_taperecorder_open;
    IBOutlet id runtime_audio_playthrough;

    IBOutlet id newmachine_model;					
    IBOutlet id newmachine_keyboard_mode;
    IBOutlet id newmachine_soundchip;
    IBOutlet id newmachine_joystick;
    IBOutlet id newmachine_autostart_tape;
//	IBOutlet id newmachine_autostart_tape_for_loading;
//	IBOutlet id newmachine_autostart_tape_for_saving;
    IBOutlet id newmachine_open_builtin_js_panel;

    IBOutlet id newsnapshot_keyboard_mode;
    IBOutlet id newsnapshot_soundchip;
    IBOutlet id newsnapshot_joystick;
//	IBOutlet id newsnapshot_mouse;		has no button in prefs panel
    IBOutlet id newsnapshot_autostart_tape;
//	IBOutlet id newsnapshot_autostart_tape_for_loading;
//	IBOutlet id newsnapshot_autostart_tape_for_saving;
    IBOutlet id newsnapshot_use_old_settings;
	IBOutlet id newsnapshot_opens_in_new_window;
	
	IBOutlet id fullscreen_mode_enabled;		IBOutlet id toggle_fullscreen_mode_menu_item; /*in window menu */
	IBOutlet id fullscreen_mode_display_id;
	IBOutlet id fullscreen_mode_resolution;
	IBOutlet id fullscreen_mode_color_depth;
	IBOutlet id fullscreen_mode_refresh_rate;
	//IBOutlet id fullscreen_mode_hardware_acceleration;
	IBOutlet id fullscreen_mode_synchronize_with_vbl;
	
	NSUserDefaults*	user_defaults;
	
	bool	use_file_settings;		// cached value for "newsnapshot_use_old_settings"
}

-(IBAction)fullscreenModeEnable:(id)sender;
-(IBAction)fullscreenModeDisplayID:(id)sender;
-(IBAction)fullscreenModeResolution:(id)sender;
-(IBAction)fullscreenModeColorDepth:(id)sender;
-(IBAction)fullscreenModeRefreshRate:(id)sender;
//-(IBAction)fullscreenModeHardwareAccel:(id)sender;
-(IBAction)fullscreenModeSyncWithVBL:(id)sender;

-(IBAction)startupFullscreenMode:(id)sender;
-(IBAction)startupOpenKeyboard:(id)sender;
-(IBAction)startupOpenTaperecorder:(id)sender;
-(IBAction)startupOpenMachineImage:(id)sender;	
-(IBAction)startupEnableAudioInput:(id)sender;

-(IBAction)runtimeLeaveKeyboardOpen:(id)sender;
-(IBAction)runtimeLeaveTaperecorderOpen:(id)sender;
-(IBAction)runtimeAudioPlaythrough:(id)sender;

-(IBAction)newMachineModel:(id)sender;
-(IBAction)newMachineKeyboardMode:(id)sender;
-(IBAction)newMachineSoundchip:(id)sender;
-(IBAction)newMachineJoystick:(id)sender;
-(IBAction)newMachineAutostartTape:(id)sender;
//-(IBAction)newMachineAutostartTapeForLoading:(id)sender;
//-(IBAction)newMachineAutostartTapeForSaving:(id)sender;
-(IBAction)newMachineOpenBuiltinJsPanel:(id)sender;	

-(IBAction)newSnapshotKeyboardMode:(id)sender;
-(IBAction)newSnapshotSoundchip:(id)sender;
-(IBAction)newSnapshotJoystick:(id)sender;
-(IBAction)newSnapshotAutostartTape:(id)sender;
//-(IBAction)newSnapshotAutostartTapeForLoading:(id)sender;
//-(IBAction)newSnapshotAutostartTapeForSaving:(id)sender;
-(IBAction)newSnapshotUseOldSettings:(id)sender;
-(IBAction)newSnapshotOpensInNewWindow:(id)sender;

// query fullscreen mode parameters, 
// which are a little complicated to evaluate
-(uint)getFullscreenModeColorDepth;
-(uint)getFullscreenModeWidth;
-(uint)getFullscreenModeHeight;
-(float)getFullscreenModeRefreshRate;


-(bool)getBoolForKey:(cstr)key;
-(int)getIntForKey:(cstr)key;
-(float)getFloatForKey:(cstr)key;
-(NSString*)getStringForKey:(cstr)key;				// set to autorelease!
-(NSArray*)getStringArrayForKey:(cstr)key;			// set to autorelease!

-(void)setObject:(id)object forKey:(cstr)key;
-(void)setBool:(bool)flag forKey:(cstr)key;
-(void)setInt:(int)wert forKey:(cstr)key;
-(void)setFloat:(float)wert forKey:(cstr)key;
-(void)setString:(NSString*)string forKey:(cstr)key;
-(void)setStringArray:(NSArray*)array forKey:(cstr)key;

-(bool)getBoolForFile:(cstr)file andKey:(cstr)key;	// if newsnapshot_use_old_settings and file::key exists; else return pref for key
-(int)getIntForFile:(cstr)file andKey:(cstr)key;	// if newsnapshot_use_old_settings and file::key exists; else return pref for key

-(void)setBool:(bool)flag forFile:(cstr)file andKey:(cstr)key;	// only if newsnapshot_use_old_settings; else ignored
-(void)setInt:(int)wert forFile:(cstr)file andKey:(cstr)key;	// only if newsnapshot_use_old_settings; else ignored


@end



extern Preferences* preferences;



