/*	Copyright  (c)	Günter Woigk 2004 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		1

#import	"cstrings/cstrings.h"
#import	"unix/file_utilities.h"

#import "utilities.h"
#import "Preferences.h"
#import "ZxInfo.h"
#import "KbdPanel.h"
#import "Dsp.h"
#import "DisplayList.h"


// note: NSUserDefaults are thread-safe





@implementation Preferences

// NXPanel:
//-(void)updatePanel:(Machine*)m {}


// %%%%%%%%%%%%%%%%%%%% INIT %%%%%%%%%%%%%%%%%%%

- (id)initWithContentRect:(NSRect)contentRect styleMask:(uint)styleMask backing:(NSBackingStoreType)backingType defer:(BOOL)flag
{
	XLogIn("Preferences:initWithContentRect");

	preferences = self = [super initWithContentRect:contentRect styleMask:styleMask backing:backingType defer:flag];

	if(self) 
	{
		do_update = no;		// no regular updatePanel callback, please.
		
		user_defaults = [NSUserDefaults standardUserDefaults];

		// ######################################
		// Hier werden die Default-Werte gesetzt!
		// ######################################

		NSDictionary* dflt_dict = 
			[NSDictionary dictionaryWithObjectsAndKeys:
				@"0",			@"startup_open_keyboard",
				@"0",			@"startup_open_taperecorder",
				@"1",			@"startup_open_machine_image",
				@"1",			@"startup_enable_audio_input",
				@"0",			@"startup_fullscreen_mode",

				@"\x31\x7e\x7d\x7b\x7c",	@"startup_keyboard_joystick_keycodes",
				@" UDLR",					@"startup_keyboard_joystick_unicodes",

				@"0",			@"startup_check_update",
				@"9999",		@"startup_check_update_next_date",
				@"1",			@"startup_check_update_intervall",
				
				@"0",			@"runtime_leave_keyboard_open",
				@"0",			@"runtime_leave_taperecorder_open",
				@"1",			@"runtime_audio_playthrough",

				[NSString stringWithFormat:@"%i",zx48i3],	@"newmachine_model",
				[NSString stringWithFormat:@"%i",kbdlog],	@"newmachine_keyboard_mode",
				@"0",			@"newmachine_soundchip",
				@"0",			@"newmachine_joystick",
				@"0",			@"newmachine_autostart_tape",
			//	@"0",			@"newmachine_autostart_tape_for_loading",
			//	@"0",			@"newmachine_autostart_tape_for_saving",
				@"1",			@"newmachine_open_builtin_js_panel",

				[NSString stringWithFormat:@"%i",kbdphys],	@"newsnapshot_keyboard_mode",
				@"0",			@"newsnapshot_soundchip",
				@"0",			@"newsnapshot_joystick",
				@"0",			@"newsnapshot_mouse",			// note: has currently no button in the prefs panel
				@"0",			@"newsnapshot_autostart_tape",
			//	@"0",			@"newsnapshot_autostart_tape_for_loading",
			//	@"0",			@"newsnapshot_autostart_tape_for_saving",
				@"0",			@"newsnapshot_use_old_settings",
				@"0",			@"newsnapshot_opens_in_new_window",
				
				[NSArray array],				@"recent_files",
				appl_is_preview ? @"1" : @"0",	@"check_update_includes_preview_version",

				@"1",			@"fullscreen_mode_enabled",
				//@"0",			@"fullscreen_mode_hardware_acceleration",
				@"0",			@"fullscreen_mode_synchronize_with_vbl",

				@"0",			@"fullscreen_mode_display_id",		// 0 = main screen
				@"0",			@"fullscreen_mode_resolution",		// 0 = don't change
				@"0",			@"fullscreen_mode_color_depth",		// 0 = don't change
				@"0",			@"fullscreen_mode_refresh_rate",	// 0 = don't change
				
				@"0.3",			@"audio_output_volume",				// roughly middle volume (logarithmic)
				
				nil			// end of list
			];

		[user_defaults registerDefaults:dflt_dict];
	}

	return self;
}


- (void)awakeFromNib 
{ 
	XLogIn("Preferences:awakeFromNib");

	[my_tab_view setDelegate:self];

	// #############################################
	// Hier werden die Views im Prefs Panel gesetzt!
	// #############################################
	
// prefs at startup:
	[startup_open_keyboard					setState:[user_defaults boolForKey:@"startup_open_keyboard"]];		
	[startup_open_taperecorder				setState:[user_defaults boolForKey:@"startup_open_taperecorder"]];	
	[startup_open_machine_image				setState:[user_defaults boolForKey:@"startup_open_machine_image"]];	
	[startup_enable_audio_input				setState:[user_defaults boolForKey:@"startup_enable_audio_input"]];	
	[startup_fullscreen_mode				setState:0/*TODO*/&&[user_defaults boolForKey:@"startup_fullscreen_mode"]];	

// prefs while running:
	[runtime_leave_keyboard_open			setState:[user_defaults boolForKey:@"runtime_leave_keyboard_open"]];
	[runtime_leave_taperecorder_open		setState:[user_defaults boolForKey:@"runtime_leave_taperecorder_open"]];
	[runtime_audio_playthrough				selectCellWithTag:[user_defaults integerForKey:@"runtime_audio_playthrough"]]; 

// prefs for new machines:
	[newmachine_model setAutoenablesItems:off];
	for( int i=0; i<zxmodels; i++ ) { [[newmachine_model itemAtIndex:i] setEnabled:zx_info[i].is_supported]; }

	[newmachine_model						selectItemAtIndex:[user_defaults integerForKey:@"newmachine_model"]];
	[newmachine_keyboard_mode				selectItemAtIndex:[user_defaults integerForKey:@"newmachine_keyboard_mode"]];
	[newmachine_soundchip					setState:[user_defaults boolForKey:@"newmachine_soundchip"]];
	[newmachine_joystick					setState:[user_defaults boolForKey:@"newmachine_joystick"]];
	[newmachine_autostart_tape				setState:[user_defaults boolForKey:@"newmachine_autostart_tape"]];
	//[newmachine_autostart_tape_for_loading	setState:[user_defaults boolForKey:@"newmachine_autostart_tape_for_loading"]];
	//[newmachine_autostart_tape_for_saving	setState:[user_defaults boolForKey:@"newmachine_autostart_tape_for_saving"]];
	[newmachine_open_builtin_js_panel		setState:[user_defaults boolForKey:@"newmachine_open_builtin_js_panel"]];

// prefs for new snapshots:
	[newsnapshot_opens_in_new_window		setState:[user_defaults boolForKey:@"newsnapshot_opens_in_new_window"]];
	[newsnapshot_keyboard_mode				selectItemAtIndex:[user_defaults integerForKey:@"newsnapshot_keyboard_mode"]];
	[newsnapshot_soundchip					setState:[user_defaults boolForKey:@"newsnapshot_soundchip"]];
	[newsnapshot_joystick					setState:[user_defaults boolForKey:@"newsnapshot_joystick"]];
	[newsnapshot_autostart_tape				setState:[user_defaults boolForKey:@"newsnapshot_autostart_tape"]];
	//[newsnapshot_autostart_tape_for_loading	setState:[user_defaults boolForKey:@"newsnapshot_autostart_tape_for_loading"]];
	//[newsnapshot_autostart_tape_for_saving	setState:[user_defaults boolForKey:@"newsnapshot_autostart_tape_for_saving"]];

	use_file_settings = [user_defaults boolForKey:@"newsnapshot_use_old_settings"];
	[newsnapshot_use_old_settings			selectCellWithTag:use_file_settings];
	
// prefs for fullscreen mode
	if( display_list.Count()==0 ) [user_defaults setBool:NO forKey:@"fullscreen_mode_enabled"]; // disable FS if DirectDisplay broken
	//[fullscreen_mode_hardware_acceleration	setState:[user_defaults boolForKey:@"fullscreen_mode_hardware_acceleration"]];
	[fullscreen_mode_synchronize_with_vbl	setState:[user_defaults boolForKey:@"fullscreen_mode_synchronize_with_vbl"]];
// die folgenden werden erst bei Öffnen des 'fullscreen mode' tabs verifiziert und speichern meist nicht ihren Index!:
	//	[fullscreen_mode_enabled				setState:[user_defaults boolForKey:@"fullscreen_mode_enabled"]];
	//	[fullscreen_mode_display_id				selectItemAtIndex:[user_defaults integerForKey:@"fullscreen_mode_display_id"]];
	//	[fullscreen_mode_resolution				selectItemAtIndex:[user_defaults integerForKey:@"fullscreen_mode_resolution"]];
	//	[fullscreen_mode_color_depth			selectItemAtIndex:[user_defaults integerForKey:@"fullscreen_mode_color_depth"]];
	//	[fullscreen_mode_refresh_rate			selectItemAtIndex:[user_defaults integerForKey:@"fullscreen_mode_refresh_rate"]];
	[my_tab_view selectFirstTabViewItem:nil];	// falls im Interface Builder zufällig was anderes, insbes. der Fullscreen-Tab ausgewählt ist
}


// %%%%%%%%%%%%%%%%%%%% FULLSCREEN MODE call back %%%%%%%%%%%%%%%%%%%

static bool  fs_values_valid = 0;
static uint  fs_color_depth;
static ulong fs_size;
static float fs_refresh_rate;

/*	UP: update all pupup menus in the fullscreen mode tab view:
*/
-(void)update_fullscreenmode_tabviews
{
	XLogIn("Preferences:update_fullscreenmode_tabviews");
	
// get number of displays:
	uint displays = display_list.Count();
	if(displays==0)	 // bummer: DirectDisplay doesn't work!
	{		
		[user_defaults setBool:NO forKey:@"fullscreen_mode_enabled"];
		[fullscreen_mode_enabled setEnabled:NO];
	}

// enable/disable all buttons acc. to fullscreen mode enabled/disabled in preferences:
	bool f = [user_defaults boolForKey:@"fullscreen_mode_enabled"];	
	[fullscreen_mode_enabled				setState:f];
	//[fullscreen_mode_hardware_acceleration	setEnabled:f];	
	[fullscreen_mode_synchronize_with_vbl	setEnabled:f];
	[fullscreen_mode_display_id				setEnabled:f];
	[fullscreen_mode_resolution				setEnabled:f];
	[fullscreen_mode_color_depth			setEnabled:f];
	[fullscreen_mode_refresh_rate			setEnabled:f];
	[startup_fullscreen_mode				setEnabled:0/*TODO*/];	// checkbox in first tabView "zxsp": options for startup
	[toggle_fullscreen_mode_menu_item		setEnabled:f];	// menu item in "Window" menu
	if(!f) return;	// if disabled bail out
	
// get & validate display_idx
	uint displays_in_menu = [fullscreen_mode_display_id numberOfItems];
	if( displays > displays_in_menu ) { displays = displays_in_menu; }	// should not happen
	while( displays_in_menu > displays ) { [fullscreen_mode_display_id removeItemAtIndex:--displays_in_menu]; }
	uint display_idx = [user_defaults integerForKey:@"fullscreen_mode_display_id"];
	if( display_idx>=displays ) display_idx=0;
	[fullscreen_mode_display_id selectItemAtIndex:display_idx];
	[fullscreen_mode_display_id setEnabled:displays>1];

// get mode_list[] and mode_list_count:	
	DisplayRef display = display_list[display_idx];
	CFArrayRef mode_list = display->GetDisplayModeList();
	uint modes = display->GetDisplayModeCount();

// populate mode_info
	ASSERT(sizeof(uint)>=4);
	struct { uint size; uint hz; ushort bpp; } mode_info[modes];
	for( uint i=0; i<modes; i++ )
	{
		CFDictionaryRef mode = (CFDictionaryRef)CFArrayGetValueAtIndex(mode_list,i);
		mode_info[i].bpp	= get_int(mode,kCGDisplayBitsPerPixel);
		mode_info[i].size	= get_int(mode,kCGDisplayWidth)*0x10000 + get_int(mode,kCGDisplayHeight)
											+ 0x8000 * (get_int(mode,kCGDisplayModeIsTelevisionOutput)==1);
		mode_info[i].hz		= uint(get_float(mode,kCGDisplayRefreshRate)+0.5);
	}

// clear all menus	
	[fullscreen_mode_resolution removeAllItems];
	[fullscreen_mode_resolution addItemWithTitle:@"Don't change"];
	//[[fullscreen_mode_resolution lastItem] setTag:0];
	[fullscreen_mode_refresh_rate removeAllItems];
	[fullscreen_mode_refresh_rate addItemWithTitle:@"Don't change"];
	//[[fullscreen_mode_refresh_rate lastItem] setTag:0];
	[fullscreen_mode_color_depth removeAllItems];
	[fullscreen_mode_color_depth addItemWithTitle:@"Don't change"];
	//[[fullscreen_mode_color_depth lastItem] setTag:0];

// init other default values:
	ulong resolution		= [user_defaults integerForKey:@"fullscreen_mode_resolution"];	// (w<<16)+h
	uint  resolution_idx	= 0;
	uint  refresh_rate		= uint([user_defaults floatForKey:@"fullscreen_mode_refresh_rate"]+0.5);
	uint  refresh_rate_idx	= 0;
	uint  color_depth		= [user_defaults integerForKey:@"fullscreen_mode_color_depth"];
	uint  color_depth_idx	= 0;

	uint min,max;

// determine list of distinct available refresh rates:
	if( mode_info[0].hz==0.0 )	// TFT
	{  
		[fullscreen_mode_refresh_rate setEnabled:NO];
		[[fullscreen_mode_refresh_rate lastItem] setTitle:@"Not for TFT"];
		[fullscreen_mode_synchronize_with_vbl setEnabled:off];
		if( [fullscreen_mode_synchronize_with_vbl state]==on ) [self setBool:off forKey:"fullscreen_mode_synchronize_with_vbl"];
	}
	else		// CRT
	{
		for(min=49;;min=max)	// find refresh rates from 50 to 120 Hz
		{	
			max=121;
			for( uint i=0; i<modes; i++ )
			{
				uint val = mode_info[i].hz;
				if( val<max && val>min ) max = val;
			}
			if( max==121 ) break; 
			if( refresh_rate==max ) refresh_rate_idx = [fullscreen_mode_refresh_rate numberOfItems];
			[fullscreen_mode_refresh_rate addItemWithTitle:ToNSString(Using("%i Hz",max))];
			[[fullscreen_mode_refresh_rate lastItem] setTag:max];
		}
	}
	if( refresh_rate_idx==0 ) refresh_rate = uint(display->GetRefreshRate()+0.5);	// not found =>don't change
	[fullscreen_mode_refresh_rate selectItemAtIndex:refresh_rate_idx];

// remove from mode_list[] all modes with wrong refresh rate:  
	for( uint i=0; i<modes; i++ )
	{ if( mode_info[i].hz != refresh_rate ) mode_info[i--]=mode_info[--modes]; }


// determine list of distinct available resolutions:
	uint current_size = (display->GetDisplayWidth()<<16)+display->GetDisplayHeight();
	for(min=640<<16;;min=max)	// find display sizes from width=640 upward
	{
		max=current_size+1;
		for( uint i=0; i<modes; i++ )
		{
			uint val = mode_info[i].size;
			if( val<max && val>min ) max = val;
		}
		if( max==current_size+1 ) break; 
		if( resolution==max ) resolution_idx = [fullscreen_mode_resolution numberOfItems];
		[fullscreen_mode_resolution addItemWithTitle:ToNSString(Using("%u x %u%s",(max>>16),(max&0x7fff),max&0x8000?" (TV)":""))];
		[[fullscreen_mode_resolution lastItem] setTag:max];
	}
	if( resolution_idx==0 ) resolution = current_size;	// not found => don't change
	[fullscreen_mode_resolution selectItemAtIndex:resolution_idx];

// remove from mode_list[] all modes with wrong resolution:  
	for( uint i=0; i<modes; i++ )
	{ if( mode_info[i].size&~0x8000u != resolution ) mode_info[i--]=mode_info[--modes]; }


// determine list of distinct available color depths:
#ifdef FINAL
	for(min=15;;min=max)	// find color depths from 16 to 32 bpp	(( 8 bit indexed colors seem not to work))
#else
	for(min=7;;min=max)	// find color depths from 8 to 32 bpp
#endif
	{		
		max=33;
		for( uint i=0; i<modes; i++ )
		{
			uint val = mode_info[i].bpp;
			if( val<max && val>min ) max = val;
		}
		if( max==33 ) break; 
		if( color_depth==max ) color_depth_idx = [fullscreen_mode_color_depth numberOfItems];
		[fullscreen_mode_color_depth addItemWithTitle:ToNSString(Using("%i Bit",max))];
		[[fullscreen_mode_color_depth lastItem] setTag:max];
	}
	if( color_depth_idx==0 ) color_depth = display->GetBitsPerPixel();		// not found =>don't change
	[fullscreen_mode_color_depth selectItemAtIndex:color_depth_idx];

	fs_color_depth	= color_depth;
	fs_refresh_rate = refresh_rate;
	fs_size			= resolution & ~0x8000;
	fs_values_valid = yes;
}


/*	Delegate method for my_tab_view:
*/
-(void)tabView:(NSTabView*)tabView didSelectTabViewItem:(NSTabViewItem*)tabViewItem
{
	XLogIn("Preferences:tabView:didSelectTabViewItem");

	XXXASSERT(tabView == my_tab_view);
	if( [tabView indexOfTabViewItem:tabViewItem] == 3 ) [self update_fullscreenmode_tabviews];
}


// %%%%%%%%%%%%%%%%%%%% Fullscreen Mode Queries %%%%%%%%%%%%%%%%%%%

-(uint)getFullscreenModeColorDepth
{
	if(!fs_values_valid) [self update_fullscreenmode_tabviews];
	return fs_color_depth;
}
-(uint)getFullscreenModeWidth
{
	if(!fs_values_valid) [self update_fullscreenmode_tabviews];
	return fs_size>>16;
}
-(uint)getFullscreenModeHeight
{
	if(!fs_values_valid) [self update_fullscreenmode_tabviews];
	return fs_size&0xffff;
}
-(float)getFullscreenModeRefreshRate
{
	if(!fs_values_valid) [self update_fullscreenmode_tabviews];
	return fs_refresh_rate;
}


// %%%%%%%%%%%%%%%%%%%% PREFS PANEL call backs %%%%%%%%%%%%%%%%%%%

-(IBAction)fullscreenModeEnable:(id)sender
{
	[self setBool:[sender state] forKey:"fullscreen_mode_enabled"]; 
	[self update_fullscreenmode_tabviews];
}

-(IBAction)fullscreenModeDisplayID:(id)sender
{ 
	[self setInt:[sender indexOfSelectedItem] forKey:"fullscreen_mode_display_id"]; 
	[self update_fullscreenmode_tabviews]; 
}

-(IBAction)fullscreenModeResolution:(id)sender			
{
	int tag = [[sender selectedItem] tag];
	[self setInt:tag forKey:"fullscreen_mode_resolution"]; 
	[self update_fullscreenmode_tabviews];
}

-(IBAction)fullscreenModeRefreshRate:(id)sender			
{
	int tag = [[sender selectedItem] tag];
	[self setInt:tag forKey:"fullscreen_mode_refresh_rate"];
	[self update_fullscreenmode_tabviews];
}

-(IBAction)fullscreenModeColorDepth:(id)sender
{
	int tag = [[sender selectedItem] tag];	
	[self setInt:tag forKey:"fullscreen_mode_color_depth"]; 
	[self update_fullscreenmode_tabviews];
}

//-(IBAction)fullscreenModeHardwareAccel:(id)sender		{ [self setBool:[sender state] forKey:"fullscreen_mode_hardware_acceleration"]; }
-(IBAction)fullscreenModeSyncWithVBL:(id)sender			{ [self setBool:[sender state] forKey:"fullscreen_mode_synchronize_with_vbl"]; }

-(IBAction)startupFullscreenMode:(id)sender				{ [self setBool:0/*TODO*/&&[sender state] forKey:"startup_fullscreen_mode"]; }
-(IBAction)startupOpenKeyboard:(id)sender				{ [self setBool:[sender state] forKey:"startup_open_keyboard"]; }
-(IBAction)startupOpenTaperecorder:(id)sender			{ [self setBool:[sender state] forKey:"startup_open_taperecorder"]; }
-(IBAction)startupOpenMachineImage:(id)sender			{ [self setBool:[sender state] forKey:"startup_open_machine_image"]; }
-(IBAction)startupEnableAudioInput:(id)sender			{ [self setBool:[sender state] forKey:"startup_enable_audio_input"]; }

-(IBAction)runtimeLeaveKeyboardOpen:(id)sender			{ bool f = [sender state]; [self setBool:f forKey:"runtime_leave_keyboard_open"]; [kbd_panel setHidesOnDeactivate:!f]; }
-(IBAction)runtimeLeaveTaperecorderOpen:(id)sender		{ bool f = [sender state]; [self setBool:f forKey:"runtime_leave_taperecorder_open"]; [tape_recorder setHidesOnDeactivate:!f]; }
-(IBAction)runtimeAudioPlaythrough:(id)sender			{ int  i = [[sender selectedCell] tag]; [self setInt:i forKey:"runtime_audio_playthrough"]; DSP::SetPlaythrough(i); }

-(IBAction)newMachineModel:(id)sender					{ [self setInt:[sender indexOfSelectedItem] forKey:"newmachine_model"]; }
-(IBAction)newMachineKeyboardMode:(id)sender			{ [self setInt:[sender indexOfSelectedItem] forKey:"newmachine_keyboard_mode"]; }
-(IBAction)newMachineSoundchip:(id)sender				{ [self setBool:[sender state] forKey:"newmachine_soundchip"]; }
-(IBAction)newMachineJoystick:(id)sender				{ [self setBool:[sender state] forKey:"newmachine_joystick"]; }
-(IBAction)newMachineAutostartTape:(id)sender			{ [self setBool:[sender state] forKey:"newmachine_autostart_tape"]; }
//-(IBAction)newMachineAutostartTapeForLoading:(id)sender	{ [self setBool:[sender state] forKey:"newmachine_autostart_tape_for_loading"]; }
//-(IBAction)newMachineAutostartTapeForSaving:(id)sender	{ [self setBool:[sender state] forKey:"newmachine_autostart_tape_for_saving"]; }
-(IBAction)newMachineOpenBuiltinJsPanel:(id)sender		{ [self setBool:[sender state] forKey:"newmachine_open_builtin_js_panel"]; }

-(IBAction)newSnapshotKeyboardMode:(id)sender			{ [self setInt:[sender indexOfSelectedItem] forKey:"newsnapshot_keyboard_mode"]; }
-(IBAction)newSnapshotSoundchip:(id)sender				{ [self setBool:[sender state] forKey:"newsnapshot_soundchip"]; }
-(IBAction)newSnapshotJoystick:(id)sender				{ [self setBool:[sender state] forKey:"newsnapshot_joystick"]; }
-(IBAction)newSnapshotAutostartTape:(id)sender			{ [self setBool:[sender state] forKey:"newsnapshot_autostart_tape"]; }
//-(IBAction)newSnapshotAutostartTapeForLoading:(id)sender{ [self setBool:[sender state] forKey:"newsnapshot_autostart_tape_for_loading"]; }
//-(IBAction)newSnapshotAutostartTapeForSaving:(id)sender	{ [self setBool:[sender state] forKey:"newsnapshot_autostart_tape_for_saving"]; }
-(IBAction)newSnapshotUseOldSettings:(id)sender			{ use_file_settings = [[sender selectedCell] tag]; [self setBool:use_file_settings forKey:"newsnapshot_use_old_settings"]; }
-(IBAction)newSnapshotOpensInNewWindow:(id)sender		{ [self setBool:[sender state] forKey:"newsnapshot_opens_in_new_window"]; }



// %%%%%%%%%%%%%%%%%%%%%% READ PREFS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



-(bool)getBoolForKey:(cstr)key
{
	XXLogIn("Preferences:getBoolForKey: %s",key);
	return [user_defaults boolForKey:ToNSString(key)];
}

-(int)getIntForKey:(cstr)key
{	
	XXLogIn("Preferences:getIntForKey: %s",key);
	return [user_defaults integerForKey:ToNSString(key)];
}

-(float)getFloatForKey:(cstr)key
{	
	XXLogIn("Preferences:getFloatForKey: %s",key);
	return [user_defaults floatForKey:ToNSString(key)];
}

-(NSString*)getStringForKey:(cstr)key		// set to autorelease!
{
	XXLogIn("Preferences:getStringForKey: %s",key);
	NSString* s = [user_defaults stringForKey:ToNSString(key)];
	XXTRAP(!s);		// <-- in init müssen alle defaults gestellt werden!
	return s;
}

-(NSArray*)getStringArrayForKey:(cstr)key	// set to autorelease!
{
	XXLogIn("Preferences:getStringArrayForKey: %s",key);
	NSArray* a = [user_defaults stringArrayForKey:ToNSString(key)];
	XXTRAP(!a);		// <-- in init müssen alle defaults gestellt werden!
	return a;
}

-(bool)getBoolForFile:(cstr)file andKey:(cstr)key 
{
	if(file) file = BasenameFromPath(file);
	XXLogIn("Preferences:getBoolForFile: %s andKey: %s",Quoted(file),key);

	if(file&&use_file_settings)
	{
		cstr filekey = CatStr(file,"::",key);
		if( [user_defaults stringForKey:ToNSString(filekey)] ) key = filekey;	
	}
	return [user_defaults boolForKey:ToNSString(key)]; 
}

-(int)getIntForFile:(cstr)file andKey:(cstr)key 
{
	if(file) file = BasenameFromPath(file);
	XXLogIn("Preferences:getIntForFile: %s andKey: %s",Quoted(file),key);

	if(file&&use_file_settings)
	{
		cstr filekey = CatStr(file,"::",key);
		if( [user_defaults stringForKey:ToNSString(filekey)] ) key = filekey;	
	}
	return [user_defaults integerForKey:ToNSString(key)]; 
}



// %%%%%%%%%%%%%%%%%%%%%% WRITE PREFS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



-(void)setBool:(bool)flag forKey:(cstr)key
{
	XXLogIn("Preferences:setBool: %i forKey: %s",(int)flag,key);
	[user_defaults setBool:flag forKey:ToNSString(key)];
}

-(void)setInt:(int)wert forKey:(cstr)key
{
	XXLogIn("Preferences:setInt: %i forKey: %s",(int)wert,key);
	[user_defaults setInteger:wert forKey:ToNSString(key)];
}

-(void)setFloat:(float)wert forKey:(cstr)key
{
	XXLogIn("Preferences:setFloat: %f forKey: %s",(float)wert, key);
	[user_defaults setFloat:wert forKey:ToNSString(key)];
}

-(void)setObject:(id)object forKey:(cstr)key
{
	XXLogIn("Preferences:setObjectForKey: %s",key);
	[user_defaults setObject:object forKey:ToNSString(key)];
}

-(void)setString:(NSString*)string forKey:(cstr)key
{
	XXLogIn("Preferences:setStringForKey: %s",key);
	[self setObject:string forKey:key];
}

-(void)setStringArray:(NSArray*)array forKey:(cstr)key
{
	XXLogIn("Preferences:setArrayForKey: %s",key);
	[self setObject:array forKey:key];
}

-(void)setBool:(bool)flag forFile:(cstr)file andKey:(cstr)key
{
	if(file) file = BasenameFromPath(file);
	XXLogIn("Preferences:setBoolForFile: %s andKey: %s",Quoted(file),key);
	if(file&&use_file_settings) [self setBool:flag forKey:CatStr(file,"::",key)];
}

-(void)setInt:(int)wert forFile:(cstr)file andKey:(cstr)key
{
	if(file) file = BasenameFromPath(file);
	XXLogIn("Preferences:setIntForFile: %s andKey: %s",Quoted(file),key);
	if(file&&use_file_settings) [self setInt:wert forKey:CatStr(file,"::",key)];
}



@end












