/*	Copyright  (c)	Günter Woigk 2007 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE. 
*/


#ifndef DisplayList_h
#define DisplayList_h

#import <ApplicationServices/ApplicationServices.h>
#import "AlertPanel.h"


#define MAX_DISPLAYS 4
class Display;

extern float get_float( CFDictionaryRef values, CFStringRef key );
extern int   get_int  ( CFDictionaryRef values, CFStringRef key );


class DisplayList
{
		CGDisplayCount		num_displays;
		CGDirectDisplayID	display_ids[MAX_DISPLAYS];
		Display*			displays[MAX_DISPLAYS];
	
public:
					DisplayList		( );		// constructor

		Display*	operator[]		( uint i )	{ XXXTRAP(i>=num_displays); return displays[i]; }

		uint		Count			( )			{ return num_displays; }

static	void		RestoreDisplayModes()		{ CGRestorePermanentDisplayConfiguration(); }
};


/*	the one and only:
*/
extern DisplayList  display_list;


// ==============================================================================


class Display
{
	CGDisplayCount		myDisplayIndex;
	CGDirectDisplayID	myDisplay;
	CFArrayRef			myDisplayModeList;
	uint				myDisplayModeCount;
	ulong				myOpenGLDisplayID;
	
public:
				Display				( uint idx, CGDirectDisplayID disp ); 

	void		DumpCurrentDisplayMode();		// dump w,h,bpp,hz of current display mode of display
	void		DumpAllDisplayModes	( );		// dump w,h,bpp,hz of all possible display modes of display
	
	ulong		GetOpenGLDisplayID	( )			{ return myOpenGLDisplayID; }
	CFArrayRef	GetDisplayModeList	( )			{ return myDisplayModeList; }
	uint		GetDisplayModeCount	( )			{ return myDisplayModeCount; }

	CGRect		GetDisplayBounds	( )			{ return CGDisplayBounds( myDisplay ); }	
	CGSize		GetDisplaySize		( )			{ return CGDisplayScreenSize( myDisplay ); }
	uint		GetDisplayWidth		( )			{ return CGDisplayPixelsWide( myDisplay ); }
	uint		GetDisplayHeight	( )			{ return CGDisplayPixelsHigh( myDisplay ); }
	uint		GetBitsPerPixel		( )			{ return CGDisplayBitsPerPixel( myDisplay ); }
	float		GetRefreshRate		( );
	bool		IsTV				( );

// Get dictionary of current display modes
// returned dictionary is owned by the system => do not release!
	CFDictionaryRef GetCurrentDisplayMode()		{ return CGDisplayCurrentMode( myDisplay ); }

// set display mode: 
// return TRUE for success:
	bool		SetDisplayMode( CFDictionaryRef v )	{ return CGDisplaySwitchToMode( myDisplay, v ) == CGDisplayNoErr; }
	bool		SetDisplayMode( float width, float height, int bpp, float hz );	// set w,h,bpp,hz for display

	bool		CaptureDisplay		( );		// capture display for exclusive direct use. return TRUE for success
	bool		ReleaseDisplay		( );		// release exclusive use for display. return TRUE for success

};



#endif






