/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#import <Cocoa/Cocoa.h>
#import "zxsp_types.h"
#import	"Ula.h"
#import	"Machine.h"


extern bool fullscreen_mode;

//@protocol Display
//-(void)nextFrame;
//@end


@interface Screen : NSWindow // <Display>
{
@protected
	Machine*		machine;				// *** protected by 'lock'
	
	bool			fullscreen;				// are we fullscreen?
	DisplayRef		fs_display;				// Display handle of captured display
	NSOpenGLContext*gl_context;				// openGL drawing context
	NSString*		fs_old_frame;			// save window frame during fullscreen
		
	bool			draw_inner_screen;		// setting: draw screen$ (not only border)
	bool			update_all;				// flag: redraw entire screen, not only modified pixels
	bool			termi;					// request updateThread to terminate
	PSema*			sema;					// 'work' semaphore for updateThread

	enum { gif_none=0, gif_screenshot, gif_record_movie, gif_recording, gif_close_file, gif_h_border=32, gif_v_border=24 };
	struct GifData {
		int			state;	
		bool		commandkey_up;			// cmd key up detected => next dn will stop recording
		int			frame_count;			// frame counter during movie recording
		GifEncoder*	encoder;
		ZxPixelmap*	pixelmap;
		ZxPixelmap*	othermap;
	} gif;
	
//@public

	PLock*			lock;					// data access protection lock

// data passed by nextFrame: 
	bool			new_data_valid;			// flag: new data valid
	ushort*			new_attr_pixel;			// *** protected by 'lock'
	uchar*			new_border;				// *** protected by 'lock'
	int				ula_top_lines;		
	int				ula_bot_lines;
	int				ula_row_bytes;
	bool			ula_flashphase;

// current display state:
	ushort*			old_attr_pixel;
	uchar*			old_border;
	uchar			pixels[192*256];
}

-(id)initWithMachine:(Machine*)machine;
-(void)closeAndDispose;
-(void)savePreferencesNow;
-(OSStatus)recordMovie:(cstr)path;		// gif
-(OSStatus)saveScreenshot:(cstr)path;	// gif

-(void)setFullscreen:(bool)onoff;
-(bool)isFullscreen;
-(DisplayRef)getFSDisplayRef;

-(void)setZoom:(int)factor;
-(int)getZoom;
-(NSSize)size;
-(int)width;
-(int)height;

-(void)setDrawInnerScreen:(bool)onoff;			// draw inner screen else border only
-(bool)drawsInnerScreen;

-(void)updateThread:(id)dummy;
-(void)nextFrame;
-(void)updateAll;
-(bool)isNotDirty;
-(bool)isDirty;

@end









