/*	Copyright  (c)	Günter Woigk 2000 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define LOG		1

#include "RleBuffer.h"


RleBuffer::RleBuffer( RleBuffer const& q ) throw()		// note: move(), not copy()
:	ucharBuffer(q)	
{
	rle_size = q.rle_size; 
	rle_pos  = q.rle_pos; 
}


RleBuffer& RleBuffer::operator= ( RleBuffer const& q ) throw()	// note: move(), not copy()
{
	ucharBuffer::operator=( q );
	rle_size = q.rle_size; 
	rle_pos  = q.rle_pos; 
	return *this; 
}


inline ulong RleBuffer::run_length( ulong a, ulong e ) throw()
{
	ulong n = 0;
	cuptr bu = Data();
	e = Min(e,Size());
	while( a<e ) { n += bu[a++]; }
	return n;
}


void RleBuffer::Grow( ulong newused ) throw(bad_alloc)
{
	ucharBuffer::Grow(newused); 
}


void RleBuffer::Resize ( ulong newsize ) throw(bad_alloc)
{ 
	if( newsize<Pos() )	 rle_pos  -= run_length(newsize,Pos()); 
	if( newsize<Size() ) rle_size -= run_length(newsize,Size());

	ucharBuffer::Resize( newsize );
}

#if 0
void RleBuffer::Resize ( ulong newmax, ulong newsize ) throw(bad_alloc)
{
	if( newmax<newsize ) newmax = newsize;	// force valid

	if( newsize<Pos() )	 rle_pos  -= run_length(newsize,Pos()); 
	if( newsize<Size() ) rle_size -= run_length(newsize,Size());

	ucharBuffer::Resize( newmax,newsize );
}
#endif

void RleBuffer::Seek ( ulong newpos ) throw()
{
	if( newpos<Pos() )	rle_pos -= run_length(newpos,Pos()); 
	else				rle_pos += run_length(Pos(),newpos);

	ucharBuffer::Pos() = newpos;
}


void RleBuffer::Write( cuptr bu, ulong n ) throw(bad_alloc)
{
	ulong p0 = Pos();
	ulong p1 = p0+n;

	ulong n0 = run_length(p0,p1);
	ucharBuffer::Write(bu,n);
	ulong n1 = run_length(p0,p1);
	
	rle_pos  += n1; 
	rle_size += n1-n0;
}


void RleBuffer::Write( uchar c, ulong n ) throw(bad_alloc)
{ 
	ulong p0 = Pos();
	ulong p1 = p0+n;

	ulong n0 = run_length(p0,p1);
	ucharBuffer::Write(c,n);
	ulong n1 = c*n; // run_length(p0,p1);
	
	rle_pos  += n1; 
	rle_size += n1-n0;
}


