/*	Copyright  (c)	Günter Woigk 2000 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	Buffer for rle bytes: runlength encoded data
	
	based on Buffer<uchar>
	additionally maintains the current runlength position of the read/write pointer Pos()
	and the total runlength of the buffer.
*/


#ifndef RleBuffer_h
#define	RleBuffer_h

#import "templates/Buffer.h"


typedef Buffer<uchar> ucharBuffer;


class RleBuffer : protected ucharBuffer
{
	ulong		rle_size;				// rle count matching buffer.Size()
	ulong		rle_pos;				// rle count matching buffer.Pos()

	ulong		run_length( ulong a, ulong e )	throw();

public:
				RleBuffer()				throw()				: rle_size(0), rle_pos(0) { }
				~RleBuffer()			throw()				{ }
				RleBuffer( RleBuffer const& ) throw();		// note: move(), not copy()
    RleBuffer&	operator=( RleBuffer const& ) throw();		// note: move(), not copy()

	ulong		RlePos()				const throw()		{ return rle_pos; }
	ulong		RleSize()				const throw()		{ return rle_size; }

	ulong		Pos()					const throw()		{ return ucharBuffer::Pos(); }
	ulong		Size()					const throw()		{ return ucharBuffer::Size(); }	// used size

	bool		AtEnd()					const throw()		{ return Pos() >= Size(); }
	bool		Avail( ulong n )		const throw()		{ return ucharBuffer::Avail(n); }
	void		Seek( ulong pos )		throw();
	void		Rewind( )				throw()				{ ucharBuffer::Pos() = rle_pos = 0; }

	uchar		operator[]( ulong idx )	throw(index_error)	{ return ucharBuffer::operator[](idx); }

	void		Grow( ulong newused )	throw(bad_alloc);
	void		Resize( ulong newused ) throw(bad_alloc);	// grow or shrink
	void		Allocate( ulong newmax )throw(bad_alloc)	{ ucharBuffer::Allocate(newmax); }
	void		ShrinkToFit	( )			throw(bad_alloc)	{ ucharBuffer::ShrinkToFit(); }

	void		Purge()					throw()				{ ucharBuffer::Purge(); rle_size = rle_pos = 0; }
	void		Write( uchar c )		throw(bad_alloc);
	void		Write( cuptr, ulong n )	throw(bad_alloc);
	void		Write( uchar, ulong n )	throw(bad_alloc);
	
	uchar		Peek( )					throw(index_error)	{ return ucharBuffer::Peek(); }
	uchar		Read( )					throw(index_error)	{ uchar c = ucharBuffer::Read(); rle_pos += c; return c;  }
	void		Skip( )					throw(index_error)	{ rle_pos += ucharBuffer::Read(); }
	void		rSkip( )				throw(index_error)	{ rle_pos -= ucharBuffer::rRead(); }

	cuptr		Data( )					throw()				{ return ucharBuffer::Data(); }
};



inline void RleBuffer::Write( uchar c ) throw(bad_alloc)
{
	uchar c0 = Pos()<Size() ? Peek() : 0;
	ucharBuffer::Write(c); 
	rle_pos += c; 
	rle_size += c - c0;
}


#endif

















