/*	Copyright  (c)	Günter Woigk 2000 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#ifndef RlesData_h
#define	RlesData_h

#import	<new>
#import "TapeData.h"
#import "RleBuffer.h"
 
 
typedef class std::bad_alloc bad_alloc;	// #<new>:	std::bad_alloc


class RlesData : public TapeData
{
// data buffer:
	RleBuffer		buffer;

	Frequency		samples_per_second;	// sample rate for rles data
	Time			time_akku;			// fractional byte in progress 
	Sample			last_sample;		// for 'best zero crossing detection'

	ulong			data_start;			// buffer.Pos() of data start  ((sync pulse +2))
	ulong			data_end;			// buffer.Pos() of end of data

	void			init				();
	void			init				( TapeData const& );	
	void			kill				();

	ulong			TotalSamples() 		const { return buffer.RleSize(); }
	ulong			CurrentSamplePos()	const { return buffer.RlePos(); }
	
	void			start_recording		();
	void			stop_recording		();
	
	void			store_byte			( uchar c )						throw(bad_alloc);
	void			store_bytes			( uchar c, ulong n )			throw(bad_alloc);
	void			store_bytes			( uchar* c, ulong n )			throw(bad_alloc);
	void			store_pair			( uchar c0, uchar c1 )			throw(bad_alloc);
	void			store_pairs			( uchar c0, uchar c1, ulong n )	throw(bad_alloc);
	void			store_samples		( float n )						throw(bad_alloc);
	void			store_cc			( ulong n )						throw(bad_alloc);
	void			store_time			( Time t )					throw(bad_alloc);

	void			store_akku			();	
	void			append_time			( Time t, bool );
	void			append_time_hi		( Time t )			{ append_time(t,1); }
	void			append_time_lo		( Time t )			{ append_time(t,0); }
	void			store_tzx_pause		( Time t );
	void			store_pure_tone		( ulong, Time );
	void			store_pure_data		( uptr, ulong, Time,Time );
	void			store_data_block	( uptr, ulong, Time,Time,Time,Time,Time,Time,Time );
	void			store_data_block	( uptr, ulong, Time );
	void			store_tzx19_symbol	( uptr, ulong );
	void			store_zx81_bit		( bool );
	void			store_zx81_byte		( uchar );
	void			store_zx81_data		( uchar bu[], uint len );

	void			read_akku			();

	void			read_tap_file		( int fd )	throw(file_error,bad_alloc);
	void			read_rles_file		( int fd )	throw(file_error,bad_alloc);
	void			read_tzx_file		( int fd, uchar delim=0 )	throw(file_error,bad_alloc);
	void			read_pzx_file		( int fd )	throw(file_error,bad_alloc);
	void			read_aiff_file		( int fd )	throw(file_error,bad_alloc);
	void			read_wav_file		( int fd )	throw(file_error,bad_alloc);
	void			read_zx80_file		( int fd )	throw(file_error,bad_alloc);
	void			read_zx81_file		( int fd )	throw(file_error,bad_alloc);

	void			write_tap_file		( int fd )	throw(file_error,bad_alloc);
	void			write_rles_file		( int fd )	throw(file_error,bad_alloc);
	void			write_tzx_file		( int fd )	throw(file_error,bad_alloc);
	void			write_pzx_file		( int fd )	throw(file_error,bad_alloc);
	void			write_aiff_file		( int fd )	throw(file_error,bad_alloc);
	void			write_wav_file		( int fd )	throw(file_error,bad_alloc);


public:				RlesData			()						: TapeData()  { init();  }
					RlesData			( TapeData const& q )	: TapeData(q) { init(q); }
	virtual			~RlesData			()									  { kill();  }
	RlesData&		operator=			( TapeData const& q );	

	virtual Time	CurrentPosition		() const;
	virtual Time	TotalPlaytime		() const;

	virtual Data_Type GetDataType		() const		{ return rles; }
	virtual cstr	MajorBlockInfo		() const;
	virtual cstr	MinorBlockInfo		() const;
	virtual void	CalcMajorBlockInfo	();

	static bool		CanRead			( FileType );
	static bool		CanWrite		( FileType );

// actions:

	virtual long	ReadFromFile		( int fd, FileType );
	virtual long	WriteToFile			( int fd, FileType );	

	virtual void	SeekStart			();
	virtual void	SeekEnd				();
	virtual void	SeekPosition		( Time t );

	virtual void	Record				( Sample const* bu, long n, Signal_Type );
	virtual void	Record				( cStereoSample* bu, long n, Signal_Type );
	virtual long	Play				( Sample* bu,		long n );
	virtual void	Stop				();
	virtual void	StartRecording		();
	virtual void	StartPlaying		();
	virtual void	Purge				();
};





#endif

















