/*	Copyright  (c)	Günter Woigk 2000 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	class TapeData
	base class for block data containers
	for use by class TapeFile
*/

#define SAFE	3
#define	LOG 	1

#import  "cstrings/cstrings.h"
#import  "TapeData.h"
#import  "ZxInfo.h"


/*	initialization
*/
void TapeData::init()
{
	modified			= no;
	recording			= no;
	major_block_info	= NULL;
	minor_block_info	= NULL;
}


/*	init from other block data
*/
void TapeData::init ( TapeData const& q )
{
	modified			= q.modified;
	recording			= q.recording;
	major_block_info	= NewCopy(q.major_block_info);
	minor_block_info	= NewCopy(q.minor_block_info);
}


/*	destructor
*/
void TapeData::kill()
{
	delete[] major_block_info;
	delete[] minor_block_info;
}


/*	set custom block info
*/
void TapeData::SetMajorBlockInfo ( cstr s )	
{
	s = NewCopy(s); 
	delete[] major_block_info; 
	major_block_info = s;
	modified = yes;
}

void TapeData::SetMinorBlockInfo ( cstr s )	
{
	s = NewCopy(s); 
	delete[] minor_block_info; 
	minor_block_info = s;
	modified = yes;
}



/*	calculate major and minor block info from tape data
	minor block info is only set for header blocks
*/
void TapeData::SetBlockInfoFromData( int blen, uchar const* data )
{
	cstr major = 0;
	cstr minor = 0;

	{
		if (blen<=0)	 { major = "No data"; goto x; }	// not decodable
		if (blen<2)		 { major = "Blips"; goto x; }	// minimum length is 3 = type + data[1] + crc

		uchar btyp = data[0];							// block type
		uchar htyp = data[1];							// header type (if block is a header)
		if( btyp==0xff ) { major = "Data block"; goto x; }
		if( btyp!=0x00 ) { major = CatStr( "Block with illegal type: ", NumStr(btyp) ); goto x; }
		if( blen!=19 )   { major = "Long header"; goto x; }
		if( htyp>3 )     { major = CatStr( "Header with illegal type: ", NumStr(htyp) ); goto x; }

	// standard header block:

		char name[101];																// long enough for 10 x "RANDOMIZE "
		ptr d = name;
		char const brt[] = "£";
		char const cpy[] = "©"; 

		for (int i=2;i<12;i++)														// 10 bytes name
		{
			uchar c = data[i]; 
			if(c<32) *d++ = '?'; else
			if(c==96) { *d++ = brt[0]; *d++ = brt[1]; } else						// brit. pound
			if(c<127) *d++ = c; else
			if(c==127) { *d++ = cpy[0]; *d++ = cpy[1]; } else						// (c)
			{ strcpy(d,basic_token[c-128]); d=strchr(d,0); if(c>=0xa3) *d++=' '; }	// graphic characters, basic token		
		}
		
		while( d>name && d[-1]==' ' ) { d--; } *d=0;

		cstr const typstr[] = { "Program: ", "Data: ", "Data: ", "Code: " };		
		major = CatStr( typstr[htyp], d==name?"((unnamed))":name ); 

		int blen = Peek2Z(data+12);				// length of next block
		int bpos = Peek2Z(data+14);				// source/dest position of next block

		if( htyp==0 /*prog*/ ) minor = Using( bpos>9999 ? "length: %i bytes" : "%i bytes, starts at line %i", blen, bpos ); else
		if( htyp==3 /*code*/ ) minor = Using( "%i bytes, address: %i", blen, bpos ); else
		/* 1,2 = data */	   minor = Using( "length: %i bytes", blen );
	}
	
x:	SetMajorBlockInfo(major);
	SetMinorBlockInfo(minor);
}




















