/*	Copyright  (c)	Günter Woigk 2000 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

 
#ifndef TapeData_h
#define	TapeData_h


#import  "zxsp_types.h"
#import  "DspTime.h"

											// from "DspTime.h" / "Dsp.mm":
extern	Frequency	samples_per_second;		// DSP-Konstante & Zeitbasis des Systems: samples/second
											// typically 44100.0

//enum	FileType	{ unknown_file=0, tap_file='.tap', tzx_file='.tzx', pzx_file='.pzx', 
//					  rles_file='.rle', aiff_file='.aif', wav_file='.wav',
//					  zx80_file='.80 ', o_file='.o  ', 
//					  zx81_file='.81 ', p_file='.p  ' };
enum	Data_Type	{ unknown_data, rles, bytes, shorts };
enum	Signal_Type	{ data_signal, noise, silence };



/*
	container for block data
	for use by class TapeFile
*/

class TapeData
{
protected:
	bool			modified;
	bool			recording;
	cstr			major_block_info;
	cstr			minor_block_info;
	
protected:
	void			init				();
	void			init				( TapeData const& q );
	void			kill				();

					TapeData			()						{ init();  }
					TapeData			( TapeData const& q )	{ init(q); }
	TapeData&		operator=			( TapeData const& q )	{ if(this!=&q) { kill(); init(q); } return *this; }

public:
	virtual			~TapeData			()			{ kill(); }

// queries:
 
	bool			IsRecording			() const	{ return recording; }
	bool			IsModified			() const	{ return modified; }
	virtual cstr	MajorBlockInfo		() const	{ return major_block_info; }
	virtual cstr	MinorBlockInfo		() const	{ return minor_block_info; }
	virtual void	CalcMajorBlockInfo	() = 0;

	virtual Data_Type GetDataType		() const	= 0;
	virtual Time	CurrentPosition		() const	= 0;
	virtual Time	TotalPlaytime		() const	= 0;

	bool			IsAtStart			() const	{ return CurrentPosition() <= 0.0; }
	bool			IsAtEnd				() const	{ return CurrentPosition() >= TotalPlaytime(); }
	bool			IsNearStart			( Time proximity ) const	{ return CurrentPosition() <= proximity; }
	bool			IsNearEnd			( Time proximity ) const	{ return CurrentPosition() >= TotalPlaytime() - proximity; }

// actions:

	void			SetMajorBlockInfo	( cstr );
	void			SetMinorBlockInfo	( cstr );
	void			SetBlockInfoFromData( int data_size_in_bytes, uchar const* first_20_bytes );
	void			ClearMajorBlockInfo	()			{ SetMajorBlockInfo(NULL); }
	void			ClearMinorBlockInfo	()			{ SetMinorBlockInfo(NULL); }

	virtual long	ReadFromFile		( int fd, FileType )	= 0;
	virtual long	WriteToFile			( int fd, FileType )	= 0;

	virtual void	SeekPosition		( Time t )	= 0;	// { Stop(); current_position = MinMax(0.0,t,total_playtime); }
	virtual void	SeekStart			()				= 0;	// { Stop(); current_position = 0.0; }
	virtual void	SeekEnd				()				= 0;	// { Stop(); current_position = total_playtime; }

	virtual void	Record				( Sample const* bu, long n, Signal_Type )	= 0;
	virtual void	Record				( cStereoSample* bu, long n, Signal_Type )	= 0;
	virtual long	Play				( Sample* bu,       long n )	= 0;

	virtual void	Stop				()			{ recording = no; }
	virtual void	StartRecording		()			{ recording = yes; }
	virtual void	StartPlaying		()			{ recording = no; }
	virtual void	Purge				()			{ kill(); init(); }
};





#endif

















