/*	Copyright  (c)	Günter Woigk 2000 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#import <Cocoa/Cocoa.h>
#import "globals.h"
#import "DSP.h"
#import "unix/pthreads.h"


class TapeFile;


@interface TapeRecorder : NSView
{
	PLock*		tapefile_lock;		// tapefile access lock
	bool		view_tape_loaded;
	TapeFile*	tapefile;
	cstr		filename;
 
	int			state;				enum { stopped=0,stopped_recording,playing,recording,rewinding,winding,seekprev,seeknext };
	int			view_state;

	char		current_blockinfo1[100];
	char		current_blockinfo2[100];

	Time		next_time_l;			// calculated time for next frame
	Time		next_time_r; 

	int			current_frame_l;		// index of currently displayed frame (0 ... 11)
	int			current_frame_r;
	
	int			current_dia_l;
	int			current_dia_r;

    IBOutlet id view_tapeinfo1;
    IBOutlet id view_tapeinfo2;
    IBOutlet id btn_eject;
    IBOutlet id btn_record;
    IBOutlet id btn_prev;
    IBOutlet id btn_back;
    IBOutlet id btn_run;
    IBOutlet id btn_fore;
    IBOutlet id btn_next;
}


- (IBAction)doEject:(id)sender;
- (IBAction)doRecord:(id)sender;
- (IBAction)doPrev:(id)sender;
- (IBAction)doBack:(id)sender;
- (IBAction)doRun:(id)sender;
- (IBAction)doFore:(id)sender;
- (IBAction)doNext:(id)sender;

- (void)ejectTape;
- (void)startPlaying;			// <-- intended for the auto start/stop tape feature: to be called from interrupt
- (void)startRecording;			//		""
- (void)stop;					//		""
- (void)pause;					//		""
- (void)writeAudioToTape:(StereoSample*)buffer count:(int)samples;  	// called from core audio interrupt
- (void)readAudioFromTape:(Sample*)buffer count:(int)samples; 	// called from core audio interrupt

- (bool)isRecording;
- (bool)isPlaying;
- (bool)isLoaded;
- (bool)isStopped;
- (bool)isStoppedRecording;

- (void)updatePanel;			// note: @class TapeRecorder is not an NXPanel!

- (BOOL)isVisible;
- (void)insertTape:(cstr)filepath;
- (void)orderFront;
- (void)setHidesOnDeactivate:(BOOL)flag;




@end







