/*	Copyright  (c)	Günter Woigk 2000 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	2000-02-03 kio	started work on tape window for zxsp-mac
	2004-11-12 kio	started work on version for Mac OSX
*/

#define		SAFE	3
#define		LOG 	1

#import	<math.h>
#import	"unix/file_utilities.h"
#import "cstrings/cstrings.h"

#import	"TapeRecorder.h"
#import	"Machine.h"
#import	"KbdPanel.h"
#import	"MenuController.h"
#import	"NXTextField.h"
#import	"TapeFile.h"
#import	"Preferences.h"
#import "DspTime.h"
#import "utilities.h"
#import "AlertPanel.h"



// desription of taperecorder.gif 
// and taperecorder_empty.gif:

int const	frames			= 12;		// animation frames per reel (120°)
int const	frames_per_rot	= 12 * 3;

int const	width		= 288;			// overall window dimensions
int const	height		= 200;

int const	win_x		= 58;			// look-through window
int const 	win_y		= 88;
int const	win_w		= 174;
int const 	win_h		= 56;

int const	axis_d		= 26;			// diameter of motor axis

int const	reel_d		= 96;			// diameter of reel ~ distance between reel centers
int const	reel_dx		= 100;			// offset in ur_image
int const	min_d		= 50;//43;		// min. diameter of tape on reel  (( real: ~43px))
int const	max_d		= 150;//96;		// max. diameter of tape on reel  (( real: ~96px => ReqDia(0h45m) == full reel ))
int const	min_vis_d	= 55;			// min. visible diameter of tape on reel
int const	max_vis_d	= 87;			// max. visible diameter of tape on reel
 
int const	axis_y		= win_y +23;	
int const	axis_x1		= win_x +37;
int const	axis_x2		= win_x +135;

int const	comp_w		= win_w/2;		// width of pre-composed reel
int const	comp_h		= win_h;

Frequency const	fullspeed_rot_per_sec		= 3;
#define			fullspeed_time_per_frame	( 1.0 / fullspeed_rot_per_sec / frames_per_rot )


static CGImageRef		ur_recorder_leer;
static CGImageRef		ur_recorder_voll;
static CGImageRef		ur_12_achsen;
static CGImageRef		ur_12_reels;

static CGImageRef		composed_reel_l[frames];
static CGImageRef		composed_reel_r[frames];
static CGImageRef		composed_axis[frames];
static int				composed_dia_l[frames];	
static int				composed_dia_r[frames];
static CGContextRef		composer_l[frames];
static CGContextRef		composer_r[frames];
static CGContextRef		composer_a[frames];

static CGColorSpaceRef	colorspace = CGColorSpaceCreateDeviceRGB();

/*	Utilities:
*/
inline int reel_diameter_for_seconds ( Time seconds ) /* diameter[pixel] of tape on reel */ {

/*		assumes 0h45m fit on a reel
		uses: min_d and max_d

		piv	= pi/4										pi/4
		v	= r^2*pi = (d/2)^2*pi = d^2 * piv			volume
		
	 
		45 min == 1500*60*45 bits = 4,050,000 bits = 506,250 bytes  <->  full reel <->  d=max_d
		 0 min														<-> empty reel <->  d=min_d

		b45 = 506,250									bytes per full reel  (45 minutes)

		ppb = (max_d^2-min_d^2) *piv / b45				pixels per bytes	((volume/bytes))

		iv	= min_d^2*piv								inner reel volume @ bytes=0
		tv	= iv + bytes*ppb							tape volume + inner reel volume
		tv  = d^2*piv
		<=> d^2*piv = iv + bytes*ppb
		<=> d^2 = ( iv + bytes*ppb ) /piv
				= iv/piv + bytes*ppb/piv
				= min_d^2*piv/piv + bytes*((max_d^2-min_d^2) *piv / b45)/piv
				= min_d^2         + bytes*((max_d^2-min_d^2)      / b45)
				= min_d^2 + bytes *(max_d^2-min_d^2) /b45
*/	
	return int ( 0.5+ sqrt( seconds * (max_d*max_d-min_d*min_d) / (45*60) + min_d*min_d ) );	
}
inline Time delay_to_next_frame_for_animation ( int dia ) /* duration[sec] until next frame of animation */ {
/*
	calculation of animation speed:

	vl	= 47.5 mm/s					linear tape speed
	dd	= 42.5 mm					distance between axes
	u	= d * pi					circumference
	vr	= vl/u						rotational speed
	ptm	= dd / (axis_x2-axis_x1)	scaling factor pixels -> mm

	fpr	= 36						frames per rotation
	tpr	= 1/vr						time/rotation
	tpf	= tpr/fpr					time/frame
		= 1/vr/36
		= 1/36/vl*vu
		= 1/36/47.5*d*pi					*s/mm
		= 1/36/47.5*pi*diam*ptm				*s/mm
		= 1/36/47.5*42.5*pi*diam/reel_d		*s
		= 1000/36/47.5*42.5*pi/reel_d*diam	*ms
*/
	return dia / 36.0 / 47.5 * 3.141592653589793 * 42.5 / (axis_x2-axis_x1); 
}

/*	managing this image crap:
*/
static CGImageRef load_png_image ( cstr filename ) /* load one img from rsrc */ {

	CFStringRef filepath = CFStringCreateWithCString     ( NULL/*alloc*/, CatStr(rsrc_path,filename), kCFStringEncodingUTF8 );
	CFURLRef    fileurl  = CFURLCreateWithFileSystemPath ( NULL/*alloc*/, filepath, kCFURLPOSIXPathStyle, no/*isDirectory*/ );
	CGDataProviderRef dataprovider = CGDataProviderCreateWithURL ( fileurl );
	CGImageRef  image    = CGImageCreateWithPNGDataProvider ( dataprovider, NULL/*decode[]*/, no/*shouldInterpolate*/,				
																kCGRenderingIntentDefault );
	CGDataProviderRelease ( dataprovider );
	CFRelease(filepath);
	CFRelease(fileurl);
	return image; 
}
static void const* get_image_data_ptr_for_CGContext ( void* /* CGContextRef */ info ) {
	CGContextRef gc = (CGContextRef)info;
	return CGBitmapContextGetData(gc);
}
static CGImageRef new_image_for_context ( CGContextRef gc ) /* create img with data providers and return ptr */ {

#if 1
//struct CGDataProviderDirectAccessCallbacks {
//	CGDataProviderGetBytePointerCallback getBytePointer;
//	CGDataProviderReleaseBytePointerCallback releaseBytePointer;
//	CGDataProviderGetBytesAtOffsetCallback getBytes;
//	CGDataProviderReleaseInfoCallback releaseProvider; };
//typedef struct CGDataProviderDirectAccessCallbacks CGDataProviderDirectAccessCallbacks;
	size_t w = CGBitmapContextGetWidth(gc);
	size_t h = CGBitmapContextGetHeight(gc);
	CGDataProviderDirectAccessCallbacks callbacks = { get_image_data_ptr_for_CGContext, NULL, NULL, NULL };
	CGDataProviderRef provider = CGDataProviderCreateDirectAccess ( gc /* info */, w*h*4 /* bytes */, &callbacks );

#else
//struct CGDataProviderDirectCallbacks {
//   CGDataProviderGetBytePointerCallback getBytePointer;
//   CGDataProviderReleaseBytePointerCallback releaseBytePointer;
//   CGDataProviderGetBytesAtPositionCallback getBytesAtPosition;
//   CGDataProviderReleaseInfoCallback releaseInfo; };
//typedef struct CGDataProviderDirectCallbacks  CGDataProviderDirectCallbacks;	
	off_t w = CGBitmapContextGetWidth(gc);
	off_t h = CGBitmapContextGetHeight(gc);
	CGDataProviderDirectCallbacks callbacks = { get_image_data_ptr_for_CGContext, NULL, NULL, NULL };
	CGDataProviderRef provider = CGDataProviderCreateDirect ( gc /* info */, w*h*4 /* bytes */, &callbacks );
#endif
	
	return CGImageCreate ( w, h, 8 /* bitsPerComponent */, 32 /* bitsPerPixel */, w*4 /* bytesPerRow */,
							colorspace, kCGImageAlphaNone/*SkipLast*/, provider, NULL /* decode[] */, 
							no /* shouldInterpolate */, kCGRenderingIntentDefault );
}
static void compose_axis ( int i ) /* create composed_axis[i] (eins von 12 Motorachsen-Einzelbildern) */ {

	CGContextRef gc = composer_a[i];

// Achse aus Image	
	CGRect box = { {-axis_d*i,0}, {312,27} };
	CGContextDrawImage ( gc, box, ur_12_achsen );

	composed_axis[i] = new_image_for_context(gc);
}
static void load_images() /* load ur_xxx images, create composer_xxx[12], clear composed_reel_x[], create composed_axis[12] */ {

	XLogIn("TapeRecorder:load_images");

	ur_recorder_leer = load_png_image("Taperecorder leer.png");			TRAP(!ur_recorder_leer);
	ur_recorder_voll = load_png_image("Taperecorder mit Kassette.png");	TRAP(!ur_recorder_voll);
	ur_12_achsen     = load_png_image("12 Achsen.png");					TRAP(!ur_12_achsen);
	ur_12_reels      = load_png_image("12 Reels.png");					TRAP(!ur_12_reels);

	for(int i=0;i<frames;i++)
	{
		composer_l[i] = CGBitmapContextCreate 
		( new char[comp_w*comp_h*4], comp_w, comp_h, 8, comp_w*4, colorspace, kCGImageAlphaNoneSkipLast );	
		composer_r[i] = CGBitmapContextCreate 
		( new char[comp_w*comp_h*4], comp_w, comp_h, 8, comp_w*4, colorspace, kCGImageAlphaNoneSkipLast );
		composer_a[i] = CGBitmapContextCreate 
		( new char[axis_d*axis_d*4], axis_d, axis_d, 8, axis_d*4, colorspace, kCGImageAlphaNoneSkipLast );
		
		composed_reel_l[i] = NULL;
		composed_reel_r[i] = NULL;
		composed_dia_l[i]  = -1;	TRAP(!composer_l[i]);
		composed_dia_r[i]  = -1;	TRAP(!composer_r[i]);
		compose_axis(i);
	}
}
static void compose_left_reel ( int i, int dia ) /* create composed_reel_l[i] for diameter dia. called during animation. */ {

	CGContextRef gc = composer_l[i];

// gelber Hintergrund
	{	CGRect box = {{0,0},{comp_w,comp_h}};
		CGContextSetRGBFillColor ( gc, 202.0/255, 203.0/255, 114.0/255, 1.0 );	// hellgelb
		CGContextFillRect ( gc, box );
	}
	
// brauner Bandwickel
	CGContextBeginPath(gc);
	CGContextAddArc(gc,axis_x1-win_x,axis_y-win_y,dia/2.0,0,2*pi,1);
	CGContextAddArc(gc,axis_x1-win_x,axis_y-win_y,(min_vis_d-1)/2.0,0,2*pi,0);
	CGContextClosePath(gc);
	CGContextSetRGBFillColor ( gc, 68.0/255, 54.0/255, 60.0/255, 1.0 );		// dunkelbraun
	CGContextFillPath(gc);

// Bandscheibe aus Image	
	{	CGRect box = { {axis_x1-win_x-reel_d/2-i*reel_dx,axis_y-win_y-reel_d/2}, {1196,96} };
		CGContextDrawImage ( gc, box, ur_12_reels );
	}
	
// Kassettenfenster überlagern
	{	CGRect box = { {-win_x,-win_y}, {width,height} };
		CGContextDrawImage ( gc, box, ur_recorder_voll );
	}
	
	CGImageRelease( composed_reel_l[i] );
	composed_reel_l[i] = new_image_for_context(gc);
	composed_dia_l[i]  = dia;
}
static void compose_right_reel ( int i, int dia ) /* create composed_reel_r[i] for diameter dia. called during animation. */ {

	CGContextRef gc = composer_r[i];

// gelber Hintergrund
	{	CGRect box = {{0,0},{comp_w,comp_h}};
		CGContextSetRGBFillColor ( gc, 202.0/255, 203.0/255, 114.0/255, 1.0 );	// hellgelb
		CGContextFillRect ( gc, box );
	}
	
// brauner Bandwickel
	CGContextBeginPath(gc);
	CGContextAddArc(gc,axis_x2-(win_x+win_w/2),axis_y-win_y,dia/2.0,0,2*pi,1);
	CGContextAddArc(gc,axis_x2-(win_x+win_w/2),axis_y-win_y,(min_vis_d-1)/2.0,0,2*pi,0);
	CGContextClosePath(gc);
	CGContextSetRGBFillColor ( gc, 68.0/255, 54.0/255, 60.0/255, 1.0 );		// dunkelbraun
	CGContextFillPath(gc);

// Bandscheibe aus Image	
	{	CGRect box = { {axis_x2-(win_x+win_w/2)-reel_d/2-i*reel_dx,axis_y-win_y-reel_d/2}, {1196,96} };
		CGContextDrawImage ( gc, box, ur_12_reels );
	}
	
// Kassettenfenster überlagern
	{	CGRect box = { {-win_x-win_w/2,-win_y}, {width,height} };
		CGContextDrawImage ( gc, box, ur_recorder_voll );
	}
	
	CGImageRelease( composed_reel_r[i] );
	composed_reel_r[i] = new_image_for_context(gc);
	composed_dia_r[i]  = dia;
}




// ######################################################################################
//							Animated Tape Recorder 
// ######################################################################################

@implementation TapeRecorder		//	: NSView

-(id)initWithFrame:(NSRect)box {

	XLogIn("TapeRecorder:initWithFrame");

	self = tape_recorder = [super initWithFrame:box];
	if(self)
	{
		tapefile_lock		= new PLock(PLock::recursive);

		view_tape_loaded	= no;
		view_state			= stopped;
		state				= stopped;

		next_time_l			= 1e8;			// calculated time for next frame
		next_time_r			= 1e8;

		current_frame_l		= 0;			// index of currently displayed frame (0 ... 11)
		current_frame_r		= 0;
		
		current_dia_l		= 0;
		current_dia_r		= 0;

		tapefile			= NULL;
		filename			= NULL;
		current_blockinfo1[0] = 0;
		current_blockinfo2[0] = 0;
	}
	return self;
}
-(void)awakeFromNib {

	XLogIn("TapeRecorder:awakeFromNib");

	//XXXASSERT(tape_recorder==self);
	if(!tapefile_lock) tapefile_lock = new PLock(PLock::recursive);
	tape_recorder = self;

	load_images();		

//	if([preferences getBoolForKey:@"startup_open_taperecorder"]) [<-- done in MenuController
	bool f = [preferences getBoolForKey:"runtime_leave_taperecorder_open"]; [self setHidesOnDeactivate:!f];	
	[ view_tapeinfo1  setStrValue:"" ]; current_blockinfo1[0]=0; 
	[ view_tapeinfo2  setStrValue:"" ]; current_blockinfo2[0]=0; 
//	[self setDelegate:self];

/*	start redraw timer:
		ein Timer für die langsamen Reaktionen (Info Texte, Tasten)
		und einer für die Animation.
		Außerdem wird noch regelmäßig NXPanel:updatePanel aufgerufen,
		was aber beim TapeRecorder nicht genutzt wird!
*/
	[NSTimer scheduledTimerWithTimeInterval:0.02/*sec*/ target:self selector:@selector(animateTapeView) userInfo:(id)NULL repeats:yes];
	[NSTimer scheduledTimerWithTimeInterval:0.10/*sec*/ target:self selector:@selector(updatePanel) userInfo:(id)NULL repeats:yes];		
}

-(void)dealloc 
{
	XLogIn("TapeRecorder:dealloc");
	delete tapefile_lock;	tapefile_lock=NULL;
	delete tapefile;		tapefile=NULL;
	delete[] filename;		filename=NULL;
	[super dealloc];
}

-(BOOL)isOpaque /* override [NSView isOpaque]*/ { return yes; }

-(NSMenu*)menuForEvent:(NSEvent*)e /* override [NSView menuForEvent] to create and return context menu */ 
{
	// called by the GUI to get the context menu via [NSView menu]
	// TODO: release?

	XLogIn("TapeRecorder:menuForEvent");
	
	NSMenu*	theMenu	  = [[[NSMenu alloc] initWithTitle:@"TapeRecorder context menu"] autorelease];
						[theMenu setAutoenablesItems:NO];

	if( tapefile)
	{
		NSMenuItem* 
		item  = [NSMenuItem alloc]; 
				[item initWithTitle:@"Eject tape" action:@selector(ejectTape) keyEquivalent:@""]; 
				[item setToolTip:@"Close and eject tape file."];
				[theMenu insertItem:[item autorelease] atIndex:0];
	}
	else
	{
		NSMenuItem* 
		item  = [NSMenuItem alloc]; 
				[item initWithTitle:@"Create new tape..." action:@selector(insertNewTapeWithDialog) keyEquivalent:@""]; 
				[item setToolTip:@"Create and insert a new and emtpy tape for recording."];
				[theMenu insertItem:[item autorelease] atIndex:0];
		item  =	[NSMenuItem alloc]; 
				[item initWithTitle:@"Insert tape..." action:@selector(insertTapefileWithDialog) keyEquivalent:@""]; 
				[item setToolTip:@"Insert existing tape for loading, appending or overwriting."];
				[theMenu insertItem:[item autorelease] atIndex:1];
	}

	return theMenu;
}

// keyboard forwarding:
//-(void)keyDown:(NSEvent*)event  { [kbd_panel keyDown:event];  }
//-(void)keyUp:(NSEvent*)event    { [kbd_panel keyUp:event];    }
//-(void)flagsChanged:(NSEvent*)e { [kbd_panel flagsChanged:e]; }

/*	animation:
*/
-(void)drawLeftReel:(CGContextRef)gc 
{
	XXLogIn("TapeRecorder:drawLeftReel");

	int i = current_frame_l;		// wird per dsp interrupt weitergestellt!
	if( current_dia_l!=composed_dia_l[i]) compose_left_reel(i,current_dia_l);

	XXTRAP(!composed_reel_l[i]);
	CGRect box = {{win_x,win_y},{comp_w,comp_h}};
	CGContextDrawImage ( gc, box, composed_reel_l[i] );
}

-(void)drawRightReel:(CGContextRef)gc 
{
	XXLogIn("TapeRecorder:drawRightReel");

	int i = current_frame_r;		// wird per dsp interrupt weitergestellt!
	if( current_dia_r!=composed_dia_r[i]) compose_right_reel(i,current_dia_r);

	XXTRAP(!composed_reel_r[i]);
	CGRect box = {{win_x+win_w/2,win_y},{comp_w,comp_h}};
	CGContextDrawImage ( gc, box, composed_reel_r[i] );
}

-(void)drawLeftAxis:(CGContextRef)gc 
{
	XXLogIn("TapeRecorder:drawLeftAxis");	

	XXTRAP(!composed_axis[current_frame_l]);
	CGRect box = {{axis_x1-axis_d/2,axis_y-axis_d/2},{axis_d,axis_d}};
	CGContextDrawImage ( gc, box, composed_axis[current_frame_l] );
}

-(void)drawRightAxis:(CGContextRef)gc 
{
	XXLogIn("TapeRecorder:drawRightAxis");

	XXTRAP(!composed_axis[current_frame_r]);
	CGRect box = {{axis_x2-axis_d/2,axis_y-axis_d/2},{axis_d,axis_d}};
	CGContextDrawImage ( gc, box, composed_axis[current_frame_r] );
}

-(void)drawRect:(NSRect)dirtybox 
{
	XXLogIn("TapeRecorder:drawRect");
	
	CGContext* gc = (CGContext*)[[NSGraphicsContext currentContext] graphicsPort];		// Quartz-2D drawing context

	CGRect box = {{0,0},{width,height}};

	if(view_tape_loaded)
	{
		XXXTRAP(!ur_recorder_voll);
		if( dirtybox.origin.x<win_x || dirtybox.origin.y<win_y || dirtybox.origin.x+dirtybox.size.width>win_x+win_w || 
			dirtybox.origin.y+dirtybox.size.height>win_y+win_h ) CGContextDrawImage ( gc, box, ur_recorder_voll );
		if( dirtybox.origin.x<win_x+win_w/2 ) [self drawLeftReel:gc];
		if( dirtybox.origin.x+dirtybox.size.width>win_x+win_w/2 ) [self drawRightReel:gc];
	}
	else
	{
		XXXTRAP(!ur_recorder_leer);
		if( 1 ) CGContextDrawImage ( gc, box, ur_recorder_leer );
		if( dirtybox.origin.x<win_x+win_w/2 ) [self drawLeftAxis:gc];
		if( dirtybox.origin.x+dirtybox.size.width>win_x+win_w/2 ) [self drawRightAxis:gc];
	}
}

-(void)animateTape /* animation if tape inserted */ 
{
	XXLogIn("TapeRecorder:animateTape");

	XXXASSERT(MAIN_THREAD);		// setNeedsDisplayInRect darf nur im Main Thread aufgerufen werden!

	{	PLocker nn(tapefile_lock);
		if(!tapefile) return;
		 
		current_dia_l = reel_diameter_for_seconds(tapefile->CurrentPosition());
		current_dia_r = reel_diameter_for_seconds(tapefile->TotalPlaytime() - tapefile->CurrentPosition());
	}	Limit( min_vis_d, current_dia_l, max_vis_d );
		Limit( min_vis_d, current_dia_r, max_vis_d );
	
	Time now = system_time;		// seconds-based time
	 
	if(!view_tape_loaded)
	{
		view_tape_loaded	= yes;
		view_state			= stopped;		// stopped
		next_time_l			= now;			// calculated time for next frame
		next_time_r			= now;
		[self setNeedsDisplay:true];
		return;
	}
	
	if( view_state==state ) { if( next_time_l>now && next_time_r>now ) return; }
	else					{ next_time_l = next_time_r = now; view_state=state; }

	int d = 1;
	int f = 1;

	switch( state )
	{
	case stopped:	
	case stopped_recording:	
		next_time_l = 
		next_time_r = now+1e8; 
		return;

	case playing:	
	case recording:	
		break;

	case winding:
	case seeknext:
		f = 2;
		d = 2;
		break;
		
	case seekprev:
	case rewinding:	
		f = 2;
		d = frames-2;
		break;
	}
	
	if( now >= next_time_r )
	{
		next_time_r += delay_to_next_frame_for_animation(current_dia_r) / f;  
		if( next_time_r<=now ) next_time_r = now + 2*fullspeed_time_per_frame;
		current_frame_r = (current_frame_r+d) % frames;
		NSRect dbox = {{win_x+win_w/2,win_y},{comp_w,comp_h}};
		[self setNeedsDisplayInRect:dbox];
	}	
	if( now >= next_time_l )
	{
		next_time_l += delay_to_next_frame_for_animation(current_dia_l) / f;
		if( next_time_l<=now ) next_time_l = now + 2*fullspeed_time_per_frame;
		current_frame_l = (current_frame_l+d) % frames;
		NSRect dbox = {{win_x,win_y},{comp_w,comp_h}};
		[self setNeedsDisplayInRect:dbox];
	}
}

-(void)animateNoTape /* animation if no tape inserted */ 
{
	XXLogIn("TapeRecorder:animateNoTape");

	XXXASSERT( [NSThread currentThread] == main_thread );		// setNeedsDisplayInRect darf nur im Main Thread aufgerufen werden!

	Time now = system_time;					// seconds-based time
	
	if(view_tape_loaded) 
	{
		view_tape_loaded	= no;
		view_state			= stopped;		// stopped
		next_time_l			= now;			// calculated time for next frame
		next_time_r			= now;
		[self setNeedsDisplay:true];
		return;
	}

	if( view_state==state ) { if( next_time_l>now && next_time_r>now ) return; }
	else					{ next_time_l = next_time_r = now; view_state=state; }

	switch( state )
	{
	case stopped_recording:
	case stopped:	
		next_time_l = 
		next_time_r = now+1e8; 
		return;

	case winding:
	case seeknext:
	case playing:	
	case recording:	
		if( next_time_r>now ) return;
	  {	next_time_l = now+1e8;
		next_time_r += 2*fullspeed_time_per_frame; if( next_time_r<=now ) next_time_r = now + 2*fullspeed_time_per_frame;
		current_frame_r = (current_frame_r+frames-2) % frames;
		NSRect dbox = {{axis_x2-axis_d/2,axis_y-axis_d/2},{axis_d,axis_d}};
		[self setNeedsDisplayInRect:dbox];
	  }	return;

	case seekprev:
	case rewinding:	
		if( next_time_l>now ) return;
	  {	next_time_r = now+1e8;
		next_time_l += 2*fullspeed_time_per_frame; if( next_time_l<=now ) next_time_l = now + 2*fullspeed_time_per_frame;
		current_frame_l = (current_frame_l+2) % frames;
		NSRect dbox = {{axis_x1-axis_d/2,axis_y-axis_d/2},{axis_d,axis_d}};
		[self setNeedsDisplayInRect:dbox];
	  }	return;
	}
}

-(void)animateTapeView /* animation */ 
{
	XXLogIn("TapeRecorder:animateTapeView");

	if ([[self window] isVisible]) 
	{	
		if(tapefile) [self animateTape];
		else		 [self animateNoTape];

	//	if( [self needsDisplay] ) { [self display]; }
	}
}

/*	window redirects:
*/
-(void)orderFront { [[self window] orderFront:self]; }
-(BOOL)isVisible { return [[self window] isVisible]; }
-(void)setHidesOnDeactivate:(BOOL)f { [[self window] setHidesOnDeactivate:f]; }

/* get tape recorder state:
*/
-(bool)isRecording	{ return state==recording; }
-(bool)isPlaying	{ return state==playing;   }
-(bool)isStopped	{ return state==stopped;   }
-(bool)isStoppedRecording	{ return state==stopped_recording;   }
-(bool)isLoaded		{ return tapefile!=NULL;   }


-(void)setTitle /* UP: set tape recorder window title acc. to filename and tape length */ {

	if( filename && tapefile )
	{
		int pt;
		{
			PLocker nn(tapefile_lock);
			pt = (int)tapefile->TotalPlaytime();
		}
		char s[32]; sprintf(s," - %i'%02i\"", pt/60, pt%60 );
		[[self window] setTitle:ToNSString(CatStr(FileNameFromPath(filename),s))];
	}
	else
	{
		[[self window] setTitle:@"Tape Recorder"]; 
	}
}

-(void)writeAudioToTape:(StereoSample*)buffer count:(int)n /* Core Audio Interrupt */ 
{ 
	// if recording: write buffer to tape file

	if( /*!tapefile ||*/ state!=recording ) return;

	XXLogIn("TapeRecorder:writeAudioToTape");
	PLocker nn(tapefile_lock);
	if( !tapefile || state!=recording ) return;
	tapefile->Record(buffer,n); 
}

-(void)readAudioFromTape:(Sample*)buffer count:(int)n /* Core Audio Interrupt */ 
{ 
	// if playing: read buffer from tape file
	// if winding: advance tape position

	if( !tapefile || state==recording || state==stopped || state==stopped_recording ) return;

	XXLogIn("TapeRecorder:readAudioFromTape");
	PLocker nn(tapefile_lock);
	if( !tapefile /* || state=recording || state==stopped || state==stopped_recording */ ) return;

	Time time_pos = tapefile->CurrentPosition(); 

	switch(state)
	{
	case playing:	
		tapefile->Play(buffer,n); 
		if( tapefile->IsAtEndOfTape() ) { state=stopped; tapefile->Stop(); }	
		return;

	default:		
		return;		// recording, stopped, stopped_recording

	case rewinding:	
		#define	SECS_PER_SAMPLE				( 1.0 / samples_per_second )
		#define	SECS_PER_INTERRUPT			( SECS_PER_SAMPLE * DSP_SAMPLES_PER_BUFFER )
		#define	SPOOLING_FACTOR				( 20.0 )
		#define	SPOOLING_BYTES_PER_IRPT		long( LOADSAVE_BYTES_PER_SECOND * SECS_PER_INTERRUPT * SPOOLING_FACTOR )
		#define	SPOOLING_SECONDS_PER_IRPT	Time( SECS_PER_INTERRUPT * SPOOLING_FACTOR )
		time_pos -= SPOOLING_SECONDS_PER_IRPT;
		if(time_pos<=0.0) { state=stopped; tapefile->Stop(); }
		break;
		
	case winding:	
		time_pos += SPOOLING_SECONDS_PER_IRPT;
		if(time_pos>=tapefile->TotalPlaytime()) { state=stopped; tapefile->Stop(); }
		break;

	case seekprev:	
		time_pos -= SPOOLING_SECONDS_PER_IRPT;
		if( time_pos <= tapefile->StartOfBlock() ) { tapefile->SeekStartOfBlock(); state=stopped; return; }
		break;

	case seeknext:	
		time_pos += SPOOLING_SECONDS_PER_IRPT;
		if( time_pos >= tapefile->EndOfBlock() ) { tapefile->SeekStartOfNextBlock(); state=stopped; return; }
		break;

	}

	tapefile->SeekPosition(time_pos);
}

-(void)updatePanel /* Timer Interrupt started in awakeFromNib: update title, info and buttons */ 
{
	// update buttons and info text
	// note: this is no NXPanel:updatePanel() !

	XXLogIn("TapeRecorder:updatePanel");

	XXXASSERT(MAIN_THREAD);

//	if ( ![self isVisible] ) return;			TODO: oder doch?

	[btn_eject  setState:off];
	[btn_record setState:state==recording||state==stopped_recording];
	[btn_prev   setState:state==seekprev];
	[btn_back   setState:state==rewinding];
	[btn_run    setState:state==playing || state==recording];
	[btn_fore   setState:state==winding];
	[btn_next   setState:state==seeknext];

// update label
	if(tapefile)
	{
		tapefile_lock->Lock();
		cstr info = tapefile->MajorBlockInfo();
		if( strcmp(current_blockinfo1,info) )
		{
			strcpy(current_blockinfo1,info);
			tapefile_lock->Unlock();
			[view_tapeinfo1 setStrValue:info];
			tapefile_lock->Lock();
		}
		info = tapefile->MinorBlockInfo();
		if( strcmp(current_blockinfo2,info) )
		{
			strcpy(current_blockinfo2,info);
			tapefile_lock->Unlock();
			[view_tapeinfo2 setStrValue:info];
			tapefile_lock->Lock();
		}
		tapefile_lock->Unlock();
		if( state==recording ) { [self setTitle]; }		// update total play time
	}
	else
	{
		if( current_blockinfo1[0]!=0 ) { current_blockinfo1[0] = 0; [ view_tapeinfo1  setStrValue:"" ]; }
		if( current_blockinfo2[0]!=0 ) { current_blockinfo2[0] = 0; [ view_tapeinfo2  setStrValue:"" ]; }
	}
}

/*	Auto-Start/Stop Feature:
*/
-(void)set_state:(int)new_state 
{	
	// set tape recorder to new state
	// tapefile is set to playing, recording or stopped
	// reentrant

	XXLogIn("TapeRecorder:set_state");

	if( new_state==state ) return;	// quick test
	PLocker nn(tapefile_lock);
	if( new_state==state ) return;	// may have changed

	if( new_state==recording || new_state==stopped_recording ) 
	{
		if( !tapefile || !tapefile->IsWritable() ) return;	// reject
	}

	state = new_state;

	if(tapefile)
	{
		if( new_state==playing )   tapefile->StartPlaying(); else
		if( new_state==recording ) tapefile->StartRecording(); else
		tapefile->Stop();
	}
}

-(void)startPlaying /* called from run_for_sound() for auto-start/stop feature */ 
{
	XLogIn("TapeRecorder:startPlaying");
	[self set_state:playing];
}

-(void)startRecording /* called from run_for_sound() for auto-start/stop feature */ 
{
	XLogIn("TapeRecorder:startRecording");
	[self set_state:recording];
}

-(void)stop /* called from run_for_sound() for auto-start/stop feature */ 
{
	XLogIn("TapeRecorder:stop");
	[self set_state:stopped];
}

-(void)pause /* called from run_for_sound() for auto-start/stop feature */ 
{	
	XLogIn("TapeRecorder:pause");
	if( state==recording ) [self set_state:stopped_recording];
	else if (state==playing) [self set_state:stopped];
}

-(void)ejectTape 
{
	XLogIn("TapeRecorder:ejectTape");
	XXXASSERT(MAIN_THREAD);

	if( !tapefile ) return;			// quick test
	PLocker nn(tapefile_lock);
	XXXASSERT(tapefile);			// must not have changed: only main thread is allowed to insert/eject 

	tapefile->Stop(); 
	state = stopped;	
	delete tapefile;   tapefile = NULL; 
	delete[] filename; filename = NULL;
	[self setTitle]; /* "Tape Recorder" */
}

-(void)insertTape:(cstr)path 
{
	XLogIn("TapeRecorder:insertTape");
	XXXASSERT(MAIN_THREAD);

	if( tapefile ) return;			// quick test
	PLocker nn(tapefile_lock);
	XXXASSERT(!tapefile);			// must not have changed: only main thread is allowed to insert/eject 

	state = stopped;

	cstr      fn = NewCopy(path);
	TapeFile* tf = new TapeFile( fn );
	int		 err = Error();

	if(err) 
	{
		//[self setTitle:@"Tape Recorder"]; 
		delete tf;
		delete[] fn;
		ShowAlert( "Insert tape failed: %s", ErrorStr(err) ); 
	}
	else    
	{	
		filename = fn;
		tapefile = tf;		
		
		[self setTitle];
		[self orderFront];
		//[view_tapeinfo1 setStrValue:FileNameFromPath(filename)];
	}
}

-(void)insertTapefileWithDialog 
{
	XLogIn("TapeRecorder:insertTapefileWithDialog");
	XXXASSERT(MAIN_THREAD);
	XXXASSERT(!tapefile);
	state = stopped;		// [self stop];

	NSArray*	 types  = [NSArray arrayWithObjects:@"tap",@"tape",@"TAP",@"TAPE",@"tzx",@"TZX",@"pzx",@"PZX",
													@"o",@"O",@"80",@"81",@"P",@"p",nil];
	NSOpenPanel* dialog = [NSOpenPanel openPanel];
	int			 result = [dialog runModalForTypes:types];

	if (result==NSOKButton) 
	{
		XXXASSERT(!tapefile);
		NSString* path = [[dialog filenames] objectAtIndex:0];
		[ self insertTape:[path fileSystemRepresentation] ];
	}
}

-(void)insertNewTapeWithDialog /* ask for filename and insert new, empty tape file */ 
{ 
	XLogIn("TapeRecorder:insertNewTapeWithDialog");
	XXXASSERT(MAIN_THREAD);

	if( tapefile ) return;
	
	[self stop];
	if( fullscreen_mode ) { RLOCK; m->SetFullscreen(off); RUNLOCK; }	
	NSSavePanel* dialog = [NSSavePanel savePanel];
	NSArray* ftlist = [NSArray arrayWithObjects:@"tap",@"TAP",@"tape",@"TAPE",nil]; 
	[dialog setAllowedFileTypes:ftlist];
	[dialog setTitle:@"Create a new and empty tape"];
//	[dialog setMessage:@"MESSAGE"];
//	[dialog setNameFieldLabel:@"File name:"];
	[dialog setPrompt:@"Create"];
	int result = [dialog runModal];		// or: runModalForDirectory:file

	if( result==NSFileHandlingPanelOKButton )
	{
		NSString* filePath = [dialog filename];
		[ self insertTape:[filePath fileSystemRepresentation] ];
	}
}


/*	Button Actions:
*/
-(IBAction)doEject:(id)sender 
{
	XLogIn("TapeRecorder:doEject");
	XXXASSERT(MAIN_THREAD);

	switch(state)
	{
	case stopped:	if( tapefile )	[self ejectTape]; 
					else			[self insertTapefileWithDialog];
					break;
	default:		[self stop]; 		// Aktuell gedrückte Taste wird gelöst
					break;
	}
}

-(IBAction)doRecord:(id)sender 
{
	XLogIn("TapeRecorder:doRecord");
	XXXASSERT(MAIN_THREAD);

	switch(state)
	{
	case stopped:	[self set_state:stopped_recording]; break;
	default:		[self stop]; break;				// aktuell gedrückte Taste lösen
	}
}

-(IBAction)doRun:(id)sender 
{
	XLogIn("TapeRecorder:doRun");
	XXXASSERT(MAIN_THREAD);

	switch(state)
	{
	case stopped_recording:	[self set_state:recording]; break;
	case stopped:			[self set_state:playing];   break;		// Taste wird normal gedrückt
	default:				[self stop]; break;				// Aktuell gedrückte Tasten lösen
	}
}

-(IBAction)doPrev:(id)sender 
{
	XLogIn("TapeRecorder:doPrev");
	XXXASSERT(MAIN_THREAD);

	switch(state)
	{
	case stopped:
		[self set_state:seekprev];				// Taste normal gedrückt
		if( tapefile && tapefile->IsNearStartOfBlock(0.1) && !tapefile->IsFirstBlock() ) 
		{ tapefile->SeekEndOfPrevBlock(); }		// take care that an x-short block is not skipped
		break;
	default:	
		[self stop]; break;						// Aktuell gedrückte Tasten lösen
	}
}

-(IBAction)doNext:(id)sender 
{
	XLogIn("TapeRecorder:doNext");
	XXXASSERT(MAIN_THREAD);

	switch(state)
	{
	case stopped:	[self set_state:seeknext]; break;		// Taste normal gedrückt
	default:		[self stop]; break;						// Aktuell gedrückte Tasten lösen
	}
}

-(IBAction)doBack:(id)sender 
{
	XLogIn("TapeRecorder:doBack");
	XXXASSERT(MAIN_THREAD);

	switch(state)
	{
	case stopped:	[self set_state:rewinding]; break;		// Taste normal gedrückt
	default:		[self stop]; break;						// Aktuell gedrückte Tasten lösen
	}
}

-(IBAction)doFore:(id)sender 
{
	XLogIn("TapeRecorder:doFore");
	XXXASSERT(MAIN_THREAD);

	switch(state)
	{
	case stopped:	[self set_state:winding]; break;		// Taste normal gedrückt
	default:		[self stop]; break;						// Aktuell gedrückte Tasten lösen
	}	
}



@end

























