/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef ZxInfo_h
#define ZxInfo_h

#import "kio/kio.h"
#import "globals.h"
#import "DspTime.h"


// list of known models
// enumeration must match 'Model' menu.
enum
{
	zx80, zx81, jupiter,
	sep1, zx16k, zx48i2, zx48i3, zxplus, zxplus_span, 
	zx128, zx128_span, 
	zxplus2, zxplus2_span, zxplus2_french,
	zxplus3, zxplus3_span, zxplus2a, zxplus2a_span,
	sep2, tc2048, ts2068, pentagon, scorpion,
	sep3, samcoupe,
	zxmodels
};


extern	cstr	basic_token[128];			// Sonderzeichen und Token CHR$( 128 .. 255 ) 
											// CHR$(0..31) = div. controls;  CHR$(96) = brit.pfund;  CHR$(127) = (c)


class ZxInfo
{
public:
	int			model;

	cstr		name;
	cstr		rom_filename;
	cstr		kbd_filename;
	cstr		image_filename;				// image of machine
	bool		is_supported;				// -> enabled in models menu

	bool		has_ay_soundchip;	
	bool		has_port_7ffd;				// 128k paging control
	bool		has_port_1ffd;				// +3/+2A paging control
	bool		has_port_F4;				// Timex screen modes (?) and mmu (?)
	bool		has_port_FF;				// Timex screen modes (?) and mmu (?)

	ulong		rom_size;
	ulong		ram_size;
	ulong		page_size;					
	int			rom_pages;					// page_size * rom_pages == rom_size
	int			ram_pages;					// page_size * ram_pages == ram_size

	ulong		videoram_start[2];			// biased to ram start
	ulong		videoram_size[2];			// size or 0
	
	ulong		contended_rampages;			// bits/page:  msb=first page to lsb=last page 
	uchar		contended_rompages;			// bits/page:  ((all roms not contended, as known so far))
	uchar		waitmap;					// bits/cc:    contention waitmap
	int			waitmap_offset;
	bool		has_contention () const		{ return waitmap!=0; }

	int			tape_load_ret_addr;			// index of the original or modified ZX Spectrum rom inside machine's rom  ((for halt_cpu_after_loading))
	int			tape_load_routine;			// index in rom[] of start of tape load routine
	int			tape_save_routine;			// index in rom[] of start of tape save routine

	ulong		lines_before_screen;
	ulong		lines_in_screen;
	ulong		lines_after_screen;

	ulong		cpu_cycles_before_screen;
	ulong		cpu_cycles_in_screen;
	ulong		cpu_cycles_after_screen;
	
	ulong		cpu_cycles_per_frame;
	ulong		cpu_cycles_per_line;

	ulong		ula_cycles_per_second;		// real cycles per real second for ula
	ulong		cpu_cycles_per_second;		// real cycles per real second for cpu
	ulong		psg_cycles_per_second;		// real cycles per real second for snd
	Frequency	frames_per_second;			// real frames per real second for ula

	//double	cpu_cycles_per_sample	() const	{ return cpu_cycles_per_second / samples_per_second; }
	double		psg_cycles_per_sample	() const	{ return psg_cycles_per_second / samples_per_second; }
	//ulong		cpu_clock_predivider	() const	{ return ula_cycles_per_second / cpu_cycles_per_second; }

	Sample		ula_earin_threshold[2];		// threshold level for ula_ear_in when ula_mic_out==0/1

	void		init		(int);
				ZxInfo		()						{ init(this-zx_info); }
};


#endif









										

