/*	Copyright  (c)	Günter Woigk 1994 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		2

#import "kio/kio.h"
#import "ZxInfo.h"

#define	use_paul_farrows_zx80_emulation	1
#define	use_paul_farrows_zx81_emulation	1
#define use_paul_farrows_jupiter_emulation 1

#define	zxsp_tape_load		0x0556


/*	0x90 "A" - 0xa4 "U" are UDGs. 
	"SPECTRUM" and "PLAY" are new tokens on 128++ replacing UDG "T" and "U"
*/
cstr basic_token[] =
{
/* 0x80 */ " ", "▝", "▘", "▀", "▗", "▐", "▚", "▜", "▖", "▞", "▌", "▛", "▄", "▟", "▙", "█", 
/* 0x90 */ "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P", 
/* 0xa0 */ "Q", "R", "S", "SPECTRUM", "PLAY", "RND", "INKEY$", "PI", "FN", "POINT", "SCREEN$", "ATTR", "AT", "TAB", "VAL$", "CODE", 
/* 0xb0 */ "VAL", "LEN", "SIN", "COS", "TAN", "ASN", "ACS", "ATN", "LN", "EXP", "INT", "SQR", "SGN", "ABS", "PEEK", "IN", 
/* 0xc0 */ "USR", "STR$", "CHR$", "NOT", "BIN", "OR", "AND", "<=", ">=", "<>", "LINE", "THEN", "TO", "STEP", "DEF FN", "CAT", 
/* 0xd0 */ "FORMAT", "MOVE", "ERASE", "OPEN #", "CLOSE #", "MERGE", "VERIFY", "BEEP", "CIRCLE", "INK", "PAPER", "FLASH", "BRIGHT", "INVERSE", "OVER", "OUT", 
/* 0xe0 */ "LPRINT", "LLIST", "STOP", "READ", "DATA", "RESTORE", "NEW", "BORDER", "CONTINUE", "DIM", "REM", "FOR", "GO TO", "GO SUB", "INPUT", "LOAD", 
/* 0xf0 */ "LIST", "LET", "PAUSE", "NEXT", "POKE", "PRINT", "PLOT", "RUN", "SAVE", "RANDOMIZE", "IF", "CLS", "DRAW", "CLEAR", "RETURN", "COPY" 
};




/* ---- model info database ----
*/
ZxInfo zx_info[zxmodels];

// Initialize zx model info database
// Initialization after main() entry
#if 0
void InitZxInfo()
{
	XLogIn( "InitZxInfo" );
	if(XXXSAFE) { static bool f=0; TRAP(f); f=1; }
	for ( uint i=0; i<zxmodels; i++ ) zx_info[i].init(i);
}
#endif


void ZxInfo::init( int n )
{
	XXLogIn("ZxInfo:init");		// init() for static data => before main(){} entry!
	
	switch( n )
	{ 
	default:
		IERR();

	case sep1:
	case sep2:
	case sep3:
        init(zx48i3);		// for the sake of valid data; e.g. no "division by zero"
		name				= "(separator)";
		is_supported		= no;
		break;


// -------------------------------------------------------------------
// zx80, zx81, jupiter ace

#if use_paul_farrows_zx81_emulation	// which runs in a modified zx spectrum
	case zx81:
		init(zx48i3);
		name				= "Sinclair ZX81";
		rom_filename		= "zx81_on_zx128.rom";
		kbd_filename		= "zx81_kbd.png";
		image_filename		= "zx81.jpg";
	//	is_supported		= yes;

	//	has_port_7ffd		= no;
	//	has_port_1ffd		= no;
	//	has_port_F4			= no;
	//	has_port_FF			= no;
	//	has_ay_soundchip	= no;

	//	rom_size			= 16 kB;
	//	ram_size			= 48 kB;
		tape_load_ret_addr	= 0x0206;			// used by SAVE and LOAD:
												// -> ret addr = 0x031C = SAVE
												// -> ret addr = 0x0383 = LOAD
		tape_load_routine	= 0x0340;			// index in rom[] of start of tape load routine
		tape_save_routine	= 0x02FC;			// index in rom[] of start of tape save routine ((-> after ex hl,de))
	//	page_size			= 16 kB;

		contended_rampages	= 0x00000000;		// not contended
	//	contended_rompages	= 0x00;				// not contended
	//	waitmap				= 0xFC;				// %11111100
	//	waitmap_offset		= +1;				// 
		videoram_start[0]	= 0x8000;			// ram page #3 $c000++
	//	videoram_size[0]	= 32*24*8 + 32*24;
	//	videoram_start[1]	= 0;
	//	videoram_size[1]	= 0;

		cpu_cycles_per_second	= 3250000;
		ula_cycles_per_second	= cpu_cycles_per_second*4;
		psg_cycles_per_second	= cpu_cycles_per_second/2;	// e.g. ZON X

	//	lines_before_screen	= 64;
	//	cpu_cycles_per_line	= 224;				//  128 T screen, 24 T right border, 48 T hor. retrace, 24 T left border.
	//	lines_in_screen		= 24*8;
	//	lines_after_screen	= 56;
	//	cpu_cycles_before_screen = cpu_cycles_per_line*lines_before_screen;		
	//	cpu_cycles_after_screen	= cpu_cycles_per_line*lines_after_screen;
	//	ula_earin_threshold[0]	= 0.05;			// ((earout=0&micout=0 => non-loading setting))
	//	ula_earin_threshold[1]	= 0.01;			// ((earout=0&micout=1 => loadFromTape setting))
		break;

#else	// use the own stuff (which is not yet written..)
	case zx81:
		name				= "Sinclair ZX81";
		rom_filename		= "zx81.rom";
		kbd_filename		= "zx81_kbd.png";
		image_filename		= "zx81.jpg";
		is_supported		= no;

		has_port_7ffd		= no;
		has_port_1ffd		= no;
		has_port_F4			= no;
		has_port_FF			= no;
		has_ay_soundchip	= no;

		rom_size			= 8 kB;
		ram_size			= 1 kB;
		page_size			= 1 kB;

		tape_load_ret_addr	= 0x0206;				// used by SAVE and LOAD:
													// -> ret addr = 0x031C = SAVE
													// -> ret addr = 0x0383 = LOAD
		tape_load_routine	= 0x0340;				// index in rom[] of start of tape load routine
		tape_save_routine	= 0x02FC;				// index in rom[] of start of tape save routine ((-> after ex hl,de))

		contended_rampages	= 0xFFFFFFFF;			// ? dummy
		contended_rompages	= 0xFF;					// ? dummy
		waitmap				= 0xFF;					// ? dummy		
		waitmap_offset		= 0;					// ? dummy		
		videoram_start[0]	= 0;					// ? video buffer		
		videoram_size[0]	= 24*32;				// ?
		videoram_start[1]	= 0;
		videoram_size[1]	= 0;

		cpu_cycles_per_second	= 3250000;
		ula_cycles_per_second	= cpu_cycles_per_second*4;
		psg_cycles_per_second	= cpu_cycles_per_second/2;	// e.g. ZON X

		cpu_cycles_per_line	= 224;		// ?
		lines_before_screen	= 64;		// ?
		lines_in_screen		= 24*8;		// ?
		lines_after_screen	= 56;		// ?
		cpu_cycles_before_screen	= cpu_cycles_per_line*lines_before_screen;		// ?
		cpu_cycles_after_screen		= cpu_cycles_per_line*lines_after_screen;		// ?
		ula_earin_threshold[0]		= 0.05;			// ? ((earout=0&micout=0 => non-loading setting))
		ula_earin_threshold[1]		= 0.01;			// ? ((earout=0&micout=1 => loadFromTape setting))
		break;
#endif
		
#if use_paul_farrows_zx80_emulation	// which runs in a modified zx spectrum
	case zx80:
		init(zx81);
		name				= "Sinclair ZX80";
		rom_filename		= "zx80_on_zx128.rom";
		kbd_filename		= "zx80_kbd.png";
		image_filename		= "zx80.jpg";
	//	is_supported		= yes;

	//	has_port_7ffd		= no;
	//	has_port_1ffd		= no;
	//	has_port_F4			= no;
	//	has_port_FF			= no;
	//	has_ay_soundchip	= no;

	//	rom_size			= 16 kB;
	//	ram_size			= 48 kB;
		tape_load_ret_addr	= 0x0202;			// used in LOAD and SAVE command! 
												// -> ret addr = 0x01F6 = SAVE
												// -> ret addr = 0x024B = LOAD
		tape_load_routine	= 0x0207;			// index in rom[] of start of tape load routine
		tape_save_routine	= 0x01B6;			// index in rom[] of start of tape save routine
	//	page_size			= 16 kB;

//		contended_rampages	= 0x00000000;		// not contended
	//	contended_rompages	= 0x00;				// not contended
	//	waitmap				= 0xFC;				// %11111100
	//	waitmap_offset		= +1;				// 
//		videoram_start[0]	= 0x8000;			// ram page #3 $c000++
	//	videoram_size[0]	= 32*24*8 + 32*24;
	//	videoram_start[1]	= 0;
	//	videoram_size[1]	= 0;

	//	ula_cycles_per_second	= 3250000*4;
	//	cpu_cycles_per_second	= ula_cycles_per_second/4;
	//	psg_cycles_per_second	= ula_cycles_per_second/8;
	//	lines_before_screen	= 64;
	//	cpu_cycles_per_line	= 224;				//  128 T screen, 24 T right border, 48 T hor. retrace, 24 T left border.
	//	lines_in_screen		= 24*8;
	//	lines_after_screen	= 56;
	//	cpu_cycles_before_screen = cpu_cycles_per_line*lines_before_screen;		
	//	cpu_cycles_after_screen	= cpu_cycles_per_line*lines_after_screen;
	//	ula_earin_threshold[0]	= 0.05;			// ((earout=0&micout=0 => non-loading setting))
	//	ula_earin_threshold[1]	= 0.01;			// ((earout=0&micout=1 => loadFromTape setting))
		break;

#else	// use the own stuff (which is not yet written..)
	case zx80:
		init(zx81);

		name				= "Sinclair ZX80";
		rom_filename		= "zx80.rom";
		kbd_filename		= "zx80_kbd.png";
		image_filename		= "zx80.jpg";
	//	is_supported		= no;

		rom_size			= 4 kB;
	//	ram_size			= 1 kB;
	//	page_size			= 1 kB;
		videoram_start[0]	= 0;					// ?		video buffer		
		videoram_size[0]	= 24*32;				// ?

		tape_load_ret_addr	= 0x0202;			// used in LOAD and SAVE command! 
												// -> ret addr = 0x01F6 = SAVE
												// -> ret addr = 0x024B = LOAD
		tape_load_routine	= 0x0207;			// index in rom[] of start of tape load routine
		tape_save_routine	= 0x01B6;			// index in rom[] of start of tape save routine
		break;
#endif


#if use_paul_farrows_jupiter_emulation	// which runs in a modified zx spectrum
	case jupiter:
		init(zx48i3);

		name				= "Jupiter Ace";
		rom_filename		= "jupiter_on_zx48k.rom";
		kbd_filename		= "jupiter.png";
		image_filename		= "jupiter.jpg";
		is_supported		= yes;
		tape_load_ret_addr	= 0;			// TODO used in LOAD and SAVE command! 
		tape_load_routine	= 0;			// TODO index in rom[] of start of tape load routine
		tape_save_routine	= 0;			// TODO index in rom[] of start of tape save routine
		break;

#else	// use the own stuff (which is not yet written..)
	case jupiter:
		init(zx48i3);

		name				= "Jupiter Ace";
		rom_filename		= "jupiter.rom";
		kbd_filename		= "jupiter.png";
		image_filename		= "jupiter.jpg";
		is_supported		= no;

		rom_size			= 4 kB;					// ?
		ram_size			= 2 kB;					// ?
		page_size			= 2 kB;
		videoram_start[0]	= 0;					// ?		video buffer		
		videoram_size[0]	= 24*32;				// ?

		tape_load_ret_addr	= 0;			// TODO used in LOAD and SAVE command! 
		tape_load_routine	= 0;			// TODO index in rom[] of start of tape load routine
		tape_save_routine	= 0;			// TODO index in rom[] of start of tape save routine
		break;
#endif
		
		
		
// -------------------------------------------------------------------
// ZX Spectrum 16k - 48k
		
	case zx48i3:
		name				= "ZX Spectrum 48k Issue 3";
		rom_filename		= "48.rom";
		kbd_filename		= "48k.png";
		image_filename		= "zxsp_issue3.jpg";
		is_supported		= yes;

		has_port_7ffd		= no;
		has_port_1ffd		= no;
		has_port_F4			= no;
		has_port_FF			= no;
		has_ay_soundchip	= no;

		rom_size			= 16 kB;
		ram_size			= 48 kB;
		tape_load_ret_addr	= 0x053F;
		tape_load_routine	= zxsp_tape_load;	// index in rom[] of start of tape load routine
		tape_save_routine	= 0x04C2;			// index in rom[] of start of tape save routine
		page_size			= 16 kB;

		contended_rampages	= 0x80000000;		// ram#0 ($4000-$7fff) contended
		contended_rompages	= 0x00;				// not contended
		waitmap				= 0xFC;				// %11111100
		waitmap_offset		= +1;				// warm ZX48k -kio-tested-  =>  contention start at 14336+1  (cold: 14336+2)
		videoram_start[0]	= 0x0000;
		videoram_size[0]	= 32*24*8 + 32*24;
		videoram_start[1]	= 0;
		videoram_size[1]	= 0;

		ula_cycles_per_second	= 14000000;
		cpu_cycles_per_second	= ula_cycles_per_second/4;
		psg_cycles_per_second	= ula_cycles_per_second/8;
		lines_before_screen	= 64;
		cpu_cycles_per_line	= 224;				//  128 T screen, 24 T right border, 48 T hor. retrace, 24 T left border.
		lines_in_screen		= 24*8;
		lines_after_screen	= 56;
		cpu_cycles_before_screen = cpu_cycles_per_line*lines_before_screen;		
		cpu_cycles_after_screen	= cpu_cycles_per_line*lines_after_screen;
		ula_earin_threshold[0]	= 0.05;			// ((earout=0&micout=0 => non-loading setting))
		ula_earin_threshold[1]	= 0.01;			// ((earout=0&micout=1 => loadFromTape setting))
	//	ula_earin_threshold[1]	= 0.00;			// TODO: Issue 4A: rnd noise when no signal ((earout=0&micout=1 => loadFromTape setting))
		break;
		
	case zx16k:
		init(zx48i3);
		name				= "ZX Spectrum 16k Issue 1";
		kbd_filename		= "16k.png";		// grey keys
		image_filename		= "zxsp_issue1.jpg";
		ram_size			= 16 kB;
		ula_earin_threshold[0]	=  0.05;	// ((earout=0&micout=0 => non-loading setting))
		ula_earin_threshold[1]	= -0.01;	// ((earout=0&micout=1 => loadFromTape setting))
	//	is_supported		= yes;
		break;
		
	case zx48i2:
		init(zx48i3);
		name				= "ZX Spectrum 48k Issue 2";
		ula_earin_threshold[0]	=  0.05;	// ((earout=0&micout=0 => non-loading setting))
		ula_earin_threshold[1]	= -0.01;	// ((earout=0&micout=1 => loadFromTape setting))
	//	image_filename		= "zxsp_issue3.jpg";
	//	is_supported		= yes;
		break;
		
	case zxplus:
		init(zx48i3);
		name				= "ZX Spectrum+";
		kbd_filename		= "128k.jpg";		// black plastic keys
		image_filename		= "zxplus.jpg";
	//	has_ay_soundchip	= no;				// ? (probably no)
	//	is_supported		= yes;
		break;
		
	case zxplus_span:		// TODO: has 64K RAM. mmu?
		init(zx48i3);
		rom_filename		= "zxplus_span.rom";		
		name				= "Inves Spectrum+ (Spanish)";
		kbd_filename		= "128k_span.jpg";		// black plastic keys
		image_filename		= "zxplus_span.jpg";
	//	has_ay_soundchip	= no;					// has no AY
	//	is_supported		= yes;
		ula_cycles_per_second	= 17734500/5*4;			// note: XTAL is 17.7345 = 5 * 3546900 Hz. TODO: get info!
		cpu_cycles_per_second	= ula_cycles_per_second/4;	// divide-by-5 is not supported by the class Ula. => pretend 14 MHz/4
		psg_cycles_per_second	= ula_cycles_per_second/8;	// probably the divide-by-5 is required for contendion-free VRam.

	//	note: cc/frame = 70908 [tested: 2007-06-04 kio]
	//		  70908 = (311 lines) * (57*4 cc_per_line)
	//		  311 = 63 + 192 + 56					// to be verified	((note: 63+192+56 is used by +128 .. +3))
		lines_before_screen	= 63;					// to be verified
		cpu_cycles_per_line	= 57*4;					// 228 T = 128 T screen, 24 T right border, 52 T hor.retrace, 24 T left border
		lines_in_screen		= 24*8;
		lines_after_screen	= 56;
		cpu_cycles_before_screen = cpu_cycles_per_line*lines_before_screen;		
		cpu_cycles_after_screen	= cpu_cycles_per_line*lines_after_screen;

		contended_rampages	= 0x00000000;		// no contended ram [tested: 2007-06-04 kio]
	//	contended_rompages	= 0x00;				// not contended
		waitmap				= 0x00;				// not used: all pages are marked as non-contended
		waitmap_offset		= 0;				// not used: ""
		break;



// -------------------------------------------------------------------
// ZX Spectrum 128k, +2 and +3

	case zx128_span:
		init(zx128);
		name				= "Sinclair ZX Spectrum+ 128K (Spanish)";
		rom_filename		= "128_span.rom";
		kbd_filename		= "128k_span.jpg";
		image_filename		= "zx128_span.jpg";
		break;
		
	case zx128:
		name				= "Sinclair ZX Spectrum+ 128K";
		rom_filename		= "128.rom";
		kbd_filename		= "128k.jpg";
		image_filename		= "zx128.jpg";
		is_supported		= yes;

		has_port_7ffd		= yes;
		has_port_1ffd		= no;
		has_port_F4			= no;
		has_port_FF			= no;
		has_ay_soundchip	= yes;

		rom_size			= 32 kB;
		ram_size			= 128 kB;
		tape_load_ret_addr	= 0x4000 + 0x053F;		
		tape_load_routine	= 0x4000 + zxsp_tape_load;	// index in rom[] of start of tape load routine
		tape_save_routine	= 0x4000 + 0x04C2;			// index in rom[] of start of tape save routine
		page_size			= 16 kB;

		contended_rampages	= 0x55000000;		// ram#1,3,5,7 is contended		(128/+2)
		contended_rompages	= 0x00;				// not contended
		waitmap				= 0xFC;				// %11111100
		waitmap_offset		= 0;				// -kio-tested- in 48k mode on cold ZX128. timings on warm machine unstable
		videoram_start[0]	= page_size*5;
		videoram_size[0]	= 32*24*8 + 32*24;
		videoram_start[1]	= page_size*7;
		videoram_size[1]	= 32*24*8 + 32*24;

		ula_cycles_per_second	= 3546900*4;
		cpu_cycles_per_second	= ula_cycles_per_second/4;
		psg_cycles_per_second	= ula_cycles_per_second/8;

		lines_before_screen	= 63;
		cpu_cycles_per_line	= 228;
		lines_in_screen		= 24*8;		
		lines_after_screen	= 56;		
		cpu_cycles_before_screen = cpu_cycles_per_line*lines_before_screen;		
		cpu_cycles_after_screen	 = cpu_cycles_per_line*lines_after_screen;	

		ula_earin_threshold[0] = 0.05;	// PCB 6U  ((earout=0&micout=0 => non-loading setting))
		ula_earin_threshold[1] = 0.01;	//         ((earout=0&micout=1 => loadFromTape setting))
		break;
		
	case zxplus2:
		init(zx128);
		name				= "ZX Spectrum +2";
		rom_filename		= "plus2.rom";
		kbd_filename		= "plus2.png";
		image_filename		= "zxplus2.jpg";
	//	is_supported		= yes;
	//	ula_earin_threshold[0] = 0.05;	// ? ((earout=0&micout=0 => non-loading setting))
	//	ula_earin_threshold[1] = 0.01;	// ? ((earout=0&micout=1 => loadFromTape setting))
		break;

	case zxplus2_span:
		init(zxplus2);
		name				= "ZX Spectrum +2 (Spanish)";
		rom_filename		= "plus2-spanish.rom";
		kbd_filename		= "plus2.png";				// TODO: spanish +2 keyboard
		image_filename		= "zxplus2.jpg";			// TODO: spanish +2 image
	//	is_supported		= yes;
		break;
	
	case zxplus2_french:
		init(zxplus2);
		name				= "ZX Spectrum +2 (French)";
		rom_filename		= "plus2-french.rom";
		kbd_filename		= "plus2.png";				// TODO: french +2 keyboard
		image_filename		= "zxplus2.jpg";			// TODO: french +2 image
	//	is_supported		= yes;
		break;
	
	case zxplus3:
		init(zx128);
		name				= "ZX Spectrum +3 v4.1";
		rom_filename		= "plus3_v41.rom";
		kbd_filename		= "plus3.png";
		image_filename		= "zxplus3.jpg";
		rom_size			= 64 kB;
		tape_load_ret_addr	= 3 * 0x4000 + 0x053F;		
		tape_load_routine	= 3 * 0x4000 + zxsp_tape_load;	// index in rom[] of start of tape load routine
		tape_save_routine	= 0x04C2;						// index in rom[] of start of tape save routine
		has_port_1ffd		= yes;
		contended_rampages	= 0x0F000000;		// ram#4,5,6,7 is contended		(+2A/+3)
		contended_rompages	= 0x00;				// not contended
		waitmap				= 0xFE;				// %11111110
	//	ula_earin_threshold[0] = 0.05;	// ? ((earout=0&micout=0 => non-loading setting))
	//	ula_earin_threshold[1] = 0.01;	// ? ((earout=0&micout=1 => loadFromTape setting))
		is_supported		= no;		// FDC support pending
		break;

	case zxplus3_span:
		init(zxplus3);
		name				= "ZX Spectrum +3 (Spanish)";
		rom_filename		= "plus3-spanish.rom";
		kbd_filename		= "plus3.png";					// TODO: spanish +3 keyboard
		image_filename		= "zxplus3-spanish.jpg";
		is_supported		= no;							// FDC support pending
		break;

	case zxplus2a:
		init(zxplus3);
		name				= "ZX Spectrum +2A";
	//	rom_filename		= "plus3_v41.rom";
	//	kbd_filename		= "plus3.png";
		image_filename		= "zxplus2B.jpg";
	//	rom_size			= 64 kB;
	//	tape_load_ret_addr	= 3 * 0x4000 + 0x05e2;		
	//	has_port_1ffd		= yes;
	//	contended_rampages	= 0x0F000000;		// ram#4,5,6,7 is contended		(+2A/+3)
	//	contended_rompages	= 0x00;				// not contended
	//	waitmap				= 0xFE;				// %11111110
	//	ula_earin_threshold[0] = 0.05;	// ? ((earout=0&micout=0 => non-loading setting))
	//	ula_earin_threshold[1] = 0.01;	// ? ((earout=0&micout=1 => loadFromTape setting))
		is_supported		= yes;		
		break;

	case zxplus2a_span:
		init(zxplus2a);
		name				= "ZX Spectrum +2A (Spanish)";
		rom_filename		= "plus3-spanish.rom";
	//	kbd_filename		= "plus3.png";					// TODO: spanish +3 keyboard
		image_filename		= "zxplus2a-spanish.jpg";
		is_supported		= yes;
		break;
		



// --------------------------------------------------------------
// Timex Computers
		
	case tc2048:			// built by Times Portugal
		init(zx48i3);
		name				= "Timex TC 2048";
		rom_filename		= "tc2048.rom";						// 2005-11-18
		kbd_filename		= "tc2048.png";
		image_filename		= "tc2048.jpg";
		is_supported		= yes;
		has_port_F4			= yes;
		has_port_FF			= yes;				// ? ((TS2068 hat ihn))
		page_size			= 8 kB;

		ula_cycles_per_second	= 3528000*4;
		cpu_cycles_per_second	= ula_cycles_per_second/4;
		psg_cycles_per_second	= ula_cycles_per_second/8;

		contended_rampages	= 0xC0000000;		// ram#0 ($4000-$7fff) contended	?		(caveat: 8k pages!)
	//	contended_rompages	= 0x00;				// not contended					?
	//	waitmap				= 0xFC;				// %11111100						?
	//	waitmap_offset		= +1;				// ?
		videoram_start[0]	= 0x0000;
		videoram_size[0]	= 32*24*8 + 32*24;
		videoram_start[1]	= 0x2000;
		videoram_size[1]	= 32*24*8 + 32*24;

	//	cpu_cycles_per_line	= 224;								// ?
	//	lines_before_screen	= 64;								// ?
	//	lines_in_screen		= 24*8;		
	//	lines_after_screen	= 56;								// ?
	//	cpu_cycles_before_screen = cpu_cycles_per_line*lines_before_screen;		// ?
	//	cpu_cycles_after_screen  = cpu_cycles_per_line*lines_after_screen;		// ?
	//	ula_earin_threshold[0] = 0.05;	// ? ((earout=0&micout=0 => non-loading setting))
	//	ula_earin_threshold[1] = 0.01;	// ? ((earout=0&micout=1 => loadFromTape setting))
		break;
		
	case ts2068:			// built by Timex USA
		init(zx48i3);
		name				= "Timex TS 2068";
		rom_filename		= "ts2068.rom";
		kbd_filename		= "ts2068.png";
		image_filename		= "ts2068.jpg";
		is_supported		= no;						// timex-mmu support pending
		
		has_ay_soundchip	= yes;						// but different port
		has_port_F4			= yes;
		has_port_FF			= yes;

		ram_size			= 48 kB;
		rom_size			= 24 kB;
		tape_load_ret_addr	= 0;				// TODO
		tape_load_routine	= 0;				// TODO: index in rom[] of start of tape load routine
		tape_save_routine	= 0;				// TODO: index in rom[] of start of tape save routine
		page_size			= 8 kB;
				
		contended_rampages	= 0xC0000000;		// ram#0 ($4000-$7fff) contended	?		(caveat: page_size = 8k!)
	//	contended_rompages	= 0x00;				// not contended					?
	//	waitmap				= 0xFC;				// %11111100						?
	//	waitmap_offset		= +1;				// ?
		videoram_start[0]	= 0x0000;
		videoram_size[0]	= 32*24*8 + 32*24;
		videoram_start[1]	= 0x2000;
		videoram_size[1]	= 32*24*8 + 32*24;

		ula_cycles_per_second	= 3528000*4;
		cpu_cycles_per_second	= ula_cycles_per_second/4;
		psg_cycles_per_second	= ula_cycles_per_second/8;
		
		cpu_cycles_per_line	= 224;						// ?
		lines_before_screen	= 34;						// ? ((60 Hz))
		lines_in_screen		= 24*8;	
		lines_after_screen	= 34;						// ?
		cpu_cycles_before_screen = cpu_cycles_per_line*lines_before_screen;		// ?
		cpu_cycles_after_screen  = cpu_cycles_per_line*lines_after_screen;		// ?

	//	ula_earin_threshold[0] = 0.05;	// ? ((earout=0&micout=0 => non-loading setting))
	//	ula_earin_threshold[1] = 0.01;	// ? ((earout=0&micout=1 => loadFromTape setting))

		break;



// --------------------------------------------------------------
// Russian Computers
				
	case pentagon:
		name				= "Pentagon (USSR)";
		rom_filename		= "pentagon.rom";
		kbd_filename		= "pentagon.png";
		image_filename		= "pentagon_provisorisch.jpg";				// TODO: image of Pentagon
		is_supported		= yes;

		has_port_7ffd		= yes;						
		has_port_1ffd		= no;
		has_port_F4			= no;
		has_port_FF			= no;
		has_ay_soundchip	= yes;

		rom_size			= 32 kB;
		tape_load_ret_addr	= 0x4000 + 0x053F;			// ?
		tape_load_routine	= 0x4000 + zxsp_tape_load;	// ? index in rom[] of start of tape load routine
		tape_save_routine	= 0x4000 + 0x04C2;			// ? index in rom[] of start of tape save routine
		ram_size			= 128 kB;					
		page_size			= 16 kB;					// ?	probably, if port 7ffd

		contended_rampages	= 0x00000000;				// not contended 
		contended_rompages	= 0x00;						// not contended
		waitmap				= 0x00;						// %00000000
		waitmap_offset		= 0;						// no contention => no offset
		videoram_start[0]	= 0x4000*5;					// ?	probably, if port 7ffd
		videoram_size[0]	= 32*24*8 + 32*24;			// ?
		videoram_start[1]	= 0x4000*7;					// ?	probably, if port 7ffd
		videoram_size[1]	= 32*24*8 + 32*24;			// ?

		ula_cycles_per_second	= 14000000;					// ?
		cpu_cycles_per_second	= ula_cycles_per_second/4;
		psg_cycles_per_second	= ula_cycles_per_second/8;

		cpu_cycles_per_line	= 224;						// ?
		lines_before_screen	= 64;						// ?
		lines_in_screen		= 24*8;		
		lines_after_screen	= 56;						// ?
		cpu_cycles_before_screen	= cpu_cycles_per_line*lines_before_screen;	// ?
		cpu_cycles_after_screen		= cpu_cycles_per_line*lines_after_screen;	// ?
	
		ula_earin_threshold[0] = 0.05;	// ? ((earout=0&micout=0 => non-loading setting))
		ula_earin_threshold[1] = 0.01;	// ? ((earout=0&micout=1 => loadFromTape setting))
		break;
		
	case scorpion:
		name				= "Scorpion (USSR)";
		rom_filename		= "scorpion.rom";
		kbd_filename		= "scorpion.png";
		image_filename		= "scorpion.jpg";			// TODO: image of Scorpion
		is_supported		= no;						// FDC support etc. pending

		has_port_7ffd		= yes;						// ?
		has_port_1ffd		= no;						// ?
		has_port_F4			= no;
		has_port_FF			= no;
		has_ay_soundchip	= yes;

		rom_size			= 64 kB;
		tape_load_ret_addr	= 0x4000 + 0x053F;			// ?
		tape_load_routine	= 0x4000 + zxsp_tape_load;	// ? index in rom[] of start of tape load routine
		tape_save_routine	= 0x4000 + 0x04C2;			// ? index in rom[] of start of tape save routine
		ram_size			= 256 kB;
		page_size			= 16 kB;					// probably, if port 7ffd

		contended_rampages	= 0x00000000;				// not contended 
		contended_rompages	= 0x00;						// not contended
		waitmap				= 0x00;						// %00000000
		waitmap_offset		= 0;						// no contention => no offset
		videoram_start[0]	= 0x4000*5;					// ?	probably, if port 7ffd
		videoram_size[0]	= 32*24*8 + 32*24;			// ?
		videoram_start[1]	= 0x4000*7;					// ?	probably, if port 7ffd
		videoram_size[1]	= 32*24*8 + 32*24;			// ?

		ula_cycles_per_second	= 14000000;							// ?
		cpu_cycles_per_second	= ula_cycles_per_second/4;
		psg_cycles_per_second	= ula_cycles_per_second/8;

		cpu_cycles_per_line	= 224;								// ?
		lines_before_screen	= 64;								// ?
		lines_in_screen		= 24*8;		
		lines_after_screen	= 56;								// ?
		cpu_cycles_before_screen	= cpu_cycles_per_line*lines_before_screen;		// ?
		cpu_cycles_after_screen= cpu_cycles_per_line*lines_after_screen;	// ?
		ula_earin_threshold[0] = 0.05;	// ? ((earout=0&micout=0 => non-loading setting))
		ula_earin_threshold[1] = 0.01;	// ? ((earout=0&micout=1 => loadFromTape setting))
		break;
		
	case samcoupe:
		name				= "Sam Coupé";
		rom_filename		= "coupe.rom";
		kbd_filename		= "coupe.png";		
		image_filename		= "samcoupe.jpg";			// TODO: image of Sam Coupé
		is_supported		= no;						// ula, fdc etc. missing

		has_port_7ffd		= yes;						// ?
		has_port_1ffd		= no;						// ?
		has_port_F4			= no;
		has_port_FF			= no;
		has_ay_soundchip	= yes;						// port ?

		rom_size			= 32 kB;
		tape_load_ret_addr	= 0;						// TODO
		tape_load_routine	= 0;						// TODO: index in rom[] of start of tape load routine
		tape_save_routine	= 0;						// TODO: index in rom[] of start of tape save routine
		ram_size			= 256 kB;
		page_size			= 16 kB;					// ?

		contended_rampages	= 0x00000000;				// ?
		contended_rompages	= 0x00;						// ?
		waitmap				= 0x00;						// ?
		waitmap_offset		= 0;						// ?
		videoram_start[0]	= 0x4000*5;					// ?
		videoram_size[0]	= 32*24*8 + 32*24;			// ?
		videoram_start[1]	= 0x4000*7;					// ?
		videoram_size[1]	= 32*24*8 + 32*24;			// ?

		ula_cycles_per_second	= 6000000*2;						// ? ? ? exact ?
		cpu_cycles_per_second	= ula_cycles_per_second/2;				// ? ? ?
		psg_cycles_per_second	= ula_cycles_per_second/4;				// ? ? ?

		cpu_cycles_per_line	= 224;								// ?
		lines_before_screen	= 64;								// ?
		lines_in_screen		= 24*8;						
		lines_after_screen	= 56;								// ?
		cpu_cycles_before_screen = cpu_cycles_per_line*lines_before_screen;		// ?
		cpu_cycles_after_screen  = cpu_cycles_per_line*lines_after_screen;		// ?
		ula_earin_threshold[0] = 0.05;	// ? ((earout=0&micout=0 => non-loading setting))
		ula_earin_threshold[1] = 0.01;	// ? ((earout=0&micout=1 => loadFromTape setting))
		break;
	}
	
	model = n;
	cpu_cycles_in_screen  = lines_in_screen * cpu_cycles_per_line;
	cpu_cycles_per_frame  = cpu_cycles_before_screen+cpu_cycles_in_screen+cpu_cycles_after_screen;
	frames_per_second	  = (Frequency)cpu_cycles_per_second / cpu_cycles_per_frame;
	rom_pages = rom_size/page_size; 
	ram_pages = ram_size/page_size; 
}






