/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define	LOG		1

#import "cstrings/cstrings.h"
#import "utilities.h"
#import "AyPanel.h"
#import "Ay.h"
#import "KbdPanel.h"
#import "MenuController.h"
#import "NXTextField.h"
#import "Machine.h"






// currently displayed values
#define XXX 0xffffffff
static long  frequ_a=XXX,frequ_b=XXX,frequ_c=XXX;
static long  frequ_n=XXX; 
static long  mixer=XXX;
static long  vol_a=XXX,vol_b=XXX,vol_c=XXX;
static long  frequ_env=XXX;
static long  shape_env=XXX;
static long  port_a=XXX,port_b=XXX;
static Frequency  xtal=XXX;
static long	 stereo=XXX;


// constant texts 
const  char io[]="io";
const  char ea[]="A-";
const  char eb[]="B-";
const  char ec[]="C-";
const  char na[]="a-";
const  char nb[]="b-";
const  char nc[]="c-";
const  char*es[]= {	"\\＿＿＿",     "\\＿＿＿","\\＿＿＿",   "\\＿＿＿",
					"/＿＿＿",      "/＿＿＿", "/＿＿＿",    "/＿＿＿",
					"\\\\\\\\\\\\","\\＿＿＿","\\/\\/\\/", "\\￣￣￣",
					"//////",      "/￣￣￣", "/\\/\\/\\", "/＿＿＿"   };


static ulong IntValue( cstr s )
{
	if (s[0]==0) return XXX;
	ulong v = (ulong)StrToDouble(s);
	return v & 0x7fffffff;
}


/*	Bestimme Hüllkurvennummer aus Buchstabengrafik
	Signifikant sind die ersten beiden Buchstaben
*/
static ulong IntValueForEnvelopeShape( cstr s )
{
	int n = strlen(s);
	
	if(n>=4)
	{
		switch(Peek4X(s))
		{
		case '\\＿': return 25;		// ＿ = 3 bytes in utf-8
		case '\\–':					// 1.5 long dash
		case '\\—':					// 2.0 long dash
		case '\\￣': return 27;		// ￣ = 3 bytes in utf-8
		case '/–':					// 1.5 long dash
		case '/—':					// 2.0 long dash
		case '/￣':	 return 29;
		case '/＿':	 return 31;
		}
	}
	
	if(n>=2)
	{
		switch(Peek2X(s))
		{
		case '\\\\':return 24;
		case '\\_':	return 25;
		case '\\/':	return 26;
		case '\\-':	return 27;
		case '//':	return 28;
		case '/-':	return 29;
		case '/\\':	return 30;
		case '/_':	return 31;
		}
	}

	return IntValue(s);
}


static ulong IntValueForVolume( cstr s )
{
	if ((s[0]|0x20)=='e') return 16;	// envelope
	return IntValue(s);
}


static ulong IntValue( id sender )
{
	cstr s = [[sender stringValue]UTF8String]; if (s[0]==0) return XXX;
	ulong v = (ulong)StrToDouble(s);
	return v & 0x7fffffff;
}


static double Value( id sender )
{
	cstr s = [[sender stringValue]UTF8String]; if (s[0]==0) return XXX;
	s = LowerStr(s);
	double f = 1.0;
	ptr p;
	if( (p=FindStr(s,"mhz")) ) { *p=0; f=1e6; } else		// 12.34 MHz
	if( (p=FindStr(s,"khz")) ) { *p=0; f=1e3; } else		// 12.34 kHz
	if( (p=FindStr(s,"hz")) )  { *p=0; }					// 12.34 Hz

	double v = StrToDouble(s,&s); while(*s==' ') s++; if(*s!=0) return XXX;
	return v * f;
}




@implementation AyPanel

// NSTextView delegate methods:
//	- (BOOL)control:(NSControl *)control textShouldEndEditing:(NSText *)fieldEditor{ return yes; }
//	- (BOOL)control:(NSControl *)control textShouldBeginEditing:(NSText*)fieldEditor{ return yes; }
//	- (void)controlTextDidChange:(NSNotification*)note { }

//-(void)close{}
//-(void)awakeFromNib {}


- (id)initWithContentRect:(NSRect)r styleMask:(uint)m backing:(NSBackingStoreType)b defer:(BOOL)f
{
	XLogIn("AyPanel:initWithContentRect");
	return ay_panel = [super initWithContentRect:r styleMask:m backing:b defer:f];
}


- (void)controlTextDidEndEditing:(NSNotification*)note
{

// wenn wir nur set_on_enter machen,
// kriegen wir evtl. keine set_irgendwas Message. 
// um sicher zu stellen, dass das Feld aktualisiert wird,
// löschen wir hier die Spiegelvariable.
	if(0) switch([edit_view tag])
	{
	case 1:	 xtal=XXX;		break;
	case 2:	 frequ_a=XXX;	break;
	case 3:	 frequ_b=XXX;	break;
	case 4:	 frequ_c=XXX;	break;
	case 5:	 frequ_n=XXX;	break;
	case 6:	 mixer=XXX;		break;
	case 7:	 vol_a=XXX;		break;
	case 8:	 vol_b=XXX;		break;
	case 9:	 vol_c=XXX;		break;
	case 10: frequ_env=XXX; break;
	case 11: shape_env=XXX; break;
	case 12: port_a=XXX;	break;
	case 13: port_b=XXX;	break;
	}

	[super controlTextDidEndEditing:note];
}


/*	Update Panel: 
	- call from main event loop
	- for regular panel update
	- called only if panel is visible
	- front machine m is locked for reading (RLOCK/RUNLOCK)
*/
-(void)updatePanel:(Machine*)m
{	
	Ay* ay = m->GetAy();
	
	if( ay==NULL )
	{
		if( xtal!=-1 )
		{
			cstr s = "---";
			[view_a_pitch setStrValue:s];	frequ_a=-1;
			[view_b_pitch setStrValue:s];	frequ_b=-1;
			[view_c_pitch setStrValue:s];	frequ_c=-1;
			[view_noise_pitch setStrValue:s];frequ_n=-1;
			[view_mixer setStrValue:s];		mixer=-1;
			[view_a_vol setStrValue:s];		vol_a=-1;
			[view_b_vol setStrValue:s];		vol_b=-1;
			[view_c_vol setStrValue:s];		vol_c=-1;
			[view_env_pitch setStrValue:s];	frequ_env=-1;
			[view_env_shape setStrValue:s];	shape_env=-1;
			[view_a_port setStrValue:s];	port_a=-1;
			[view_b_port setStrValue:s];	port_b=-1;
			[view_xtal setStrValue:s];		xtal = -1;
		} 
		return;
	}
	
	Frequency t = ay->GetClock();	// freq = 1/time
	if (t!=xtal) { xtal=t; [view_xtal setStrValue:MHzStr(t)];	}

	uchar const* r  = ay->Registers();
	int    n;		
	char   s[20];

	n=r[0]+256*r[1]; 	if (n!=frequ_a)  { frequ_a=n; [view_a_pitch setIntValue:n];		}
	n=r[2]+256*r[3]; 	if (n!=frequ_b)  { frequ_b=n; [view_b_pitch setIntValue:n];		}
	n=r[4]+256*r[5]; 	if (n!=frequ_c)  { frequ_c=n; [view_c_pitch setIntValue:n];		}
	n=r[6];			 	if (n!=frequ_n)  { frequ_n=n; [view_noise_pitch setIntValue:n];	}

	n=r[7];			 	if (n!=mixer)	 { mixer=n;	sprintf(s,"%c%c%c%c%c%c%c%c",io[(n>>7)&1],io[(n>>6)&1],nc[(n>>5)&1],nb[(n>>4)&1],na[(n>>3)&1],ec[(n>>2)&1],eb[(n>>1)&1],ea[(n>>0)&1]);
						[view_mixer setStrValue:s];	}

	n=r[8];				if (n!=vol_a)	 { vol_a=n; if(n&0x10) [view_a_vol setStrValue:"Envelope"]; else [view_a_vol setIntValue:n]; }
	n=r[9];				if (n!=vol_b)	 { vol_b=n; if(n&0x10) [view_b_vol setStrValue:"Envelope"]; else [view_b_vol setIntValue:n]; }
	n=r[10];			if (n!=vol_c)	 { vol_c=n; if(n&0x10) [view_c_vol setStrValue:"Envelope"]; else [view_c_vol setIntValue:n]; }
 
	n=r[11]+256*r[12];	if (n!=frequ_env){ frequ_env=n;	[view_env_pitch setIntValue:n];	}
	n=r[13]&0x0F;		if (n!=shape_env){ shape_env=n; [view_env_shape setStrValue:es[n]];	}
	n=r[14];			if (n!=port_a)	 { port_a=n;	sprintf(s,"$%02X",n); [view_a_port setStrValue:s];	}
	n=r[15];			if (n!=port_b)	 { port_b=n;	sprintf(s,"$%02X",n); [view_b_port setStrValue:s];	}
	
	n=ay->GetStereoMix();	if (n!=stereo)	 { stereo=n;    [menu_stereo selectItemAtIndex:n]; }
}


static void SetRR ( int n, long& rr, id sender )
{
	uint value = IntValue(sender);  
	if(value!=XXX) { QBLOCK; m->GetAy()->SetRegister(n+0,value); m->GetAy()->SetRegister(n+1,value>>8); QUNBLOCK; }
	rr = XXX; 
}

static void SetR ( int n, long& r, id sender )
{
	uint value = IntValue(sender);  
	if(value!=XXX) { QBLOCK; m->GetAy()->SetRegister(n,value); QUNBLOCK; }
	r = XXX; 
}

static void SetVol ( int n, long& r, id sender )
{
	cstr s = [[sender stringValue]UTF8String]; 
	uint value = IntValueForVolume(s);
	if(value!=XXX) { QBLOCK; m->GetAy()->SetRegister(n,value); QUNBLOCK; }
	r = XXX; 
}

static void SetEnvShape ( id sender )
{
	cstr s = [[sender stringValue]UTF8String]; 
	uint value = IntValueForEnvelopeShape(s);
	if(value!=XXX) { QBLOCK; m->GetAy()->SetRegister(13,value); QUNBLOCK; }
	shape_env = XXX; 
}

- (IBAction)set_a_pitch:(id)sender		{ SetRR(0,frequ_a,sender); }
- (IBAction)set_b_pitch:(id)sender		{ SetRR(2,frequ_b,sender); }
- (IBAction)set_c_pitch:(id)sender		{ SetRR(4,frequ_c,sender); }
- (IBAction)set_noise_pitch:(id)sender	{ SetR(6,frequ_n,sender); }
- (IBAction)set_a_volume:(id)sender		{ SetVol(8,vol_a,sender); }
- (IBAction)set_b_volume:(id)sender		{ SetVol(9,vol_b,sender); }
- (IBAction)set_c_volume:(id)sender		{ SetVol(10,vol_c,sender); }
- (IBAction)set_env_pitch:(id)sender	{ SetRR(11,frequ_env,sender); }
- (IBAction)set_env_shape:(id)sender	{ SetEnvShape(sender); }
- (IBAction)set_a_port:(id)sender		{ SetR(14,port_a,sender); }
- (IBAction)set_b_port:(id)sender		{ SetR(15,port_b,sender); }


- (IBAction)set_xtal:(id)sender
{
	Frequency freq = Value(sender);  xtal = XXX;	
	if( freq>=1000000 && freq<=4000000 ) { QBLOCK; Ay* ay = m->GetAy(); if(ay) ay->SetClock(freq); QUNBLOCK; }
}


- (IBAction)set_mixer:(id)sender	// "ooabcABC"
{	
	mixer		   = XXX;
	NSString* text = [sender stringValue]; if([text length]==0) return; 
	int      value = 0xff;
	int		   len = [text length]; 
	int       port = 0x40;

//  scan text for ooabcABC
//  abc = enable noise on channel
//  ABC = enable channel
//		if any of abcABC is not encountered, then that channel is disabled
//  i/o = set port to input/output. 
//		rightmost / first encountered 'i' or 'o' is for port A
//		leftmost / second encountered 'i' or 'o' is for port B
//		if no setting for port A or B is encountered, old state is preserved

	for ( int i=len-1; i>=0; i-- )
	{
		unichar c = [text characterAtIndex:i];
		if (c=='A') value &= ~0x01;			// '0' == enabled
		if (c=='B') value &= ~0x02;
		if (c=='C') value &= ~0x04;
		if (c=='a') value &= ~0x08;
		if (c=='b') value &= ~0x10;
		if (c=='c') value &= ~0x20;
		if (c=='i'||c=='I') { value &= ~port; port*=2; }	// '0' == input
		if (c=='o'||c=='O') { value |=  port; port*=2; } 
	}
	
	if (port==0x40) { value &= (mixer|~port); port*=2; }
	if (port==0x80) { value &= (mixer|~port); port*=2; }
		      
	QBLOCK; 
		Ay* ay = m->GetAy();
		if(ay) ay->SetRegister(7,value); 
	QUNBLOCK;
}


- (IBAction) set_stereo:(id)sender		// mono, abc_stereo, acb_stereo
{
	int n = [sender indexOfSelectedItem]; 
	if(n==stereo) return;
	stereo = n;
	
	QBLOCK;
		Ay* ay = m->GetAy();
		if(ay) { ay->SetStereoMix(n); }
	QUNBLOCK;
}




@end
