/*	Copyright  (c)	Günter Woigk 2006 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#define	SAFE	3
#define	LOG		2

#import "kio/kio.h"
#import "KeyboardJoystickPanel.h"
#import "KempstonJoystickInterfacePanel.h"
#import "KempstonJoystickInterface.h"
#import "MenuController.h"
#import "NXTextField.h"
#import "NXImage.h"
#import "Preferences.h"
#import "Joystick.h"
#import "Machine.h"





@implementation KempstonJoystickInterfacePanel

-(id)initWithContentRect:(NSRect)bbox styleMask:(uint)style 
				 backing:(NSBackingStoreType)backing defer:(BOOL)flag
{
	XLogIn("KempstonJoystickInterfacePanel:initWithContentRect");
	kempston_jsif_panel = self =
	[super initWithContentRect:bbox styleMask:style backing:backing defer:flag];
	
	return self;
}



//- (void)awakeFromNib { XLogIn("KempstonJoystickInterfacePanel:awakeFromNib"); }


- (IBAction)selectJoystick:(id)sender
{
	QBLOCK;
		KempstonJoystickInterface* kjsif = m->kempston_joystick_interface; 
		if(kjsif) { kjsif->SetJoystick((JoystickID)[sender indexOfSelectedItem]); }
	QUNBLOCK;
}


/*	Update Panel: 
	- call from main event loop
	- for regular panel update
	- called only if panel is visible
	- front machine m is locked for reading (RLOCK/RUNLOCK)
*/
-(void)updatePanel:(Machine*)m
{
	KempstonJoystickInterface* kjsif = m->kempston_joystick_interface; 

	static bool old = 1;
	if( old != (kjsif!=0) )
	{
		old = kjsif!=0;
		[self enableSubviews:old];
	}

	if( !kjsif ) return;

	static int old_model = zx48i3;
	int model = m->GetModel();
	if( model!=old_model )
	{	
		old_model = model;
		static NSImage* if_img		= nil;	
		static NSImage* tc2048_img	= nil;	
		if( !if_img )	  if_img	 = [imageView image];
		if( !tc2048_img ) tc2048_img = [[NXImage alloc] initWithContentsOfResourceFile:"tc2048_sideview.jpg"];
		[imageView setImage:model==tc2048?tc2048_img:if_img];
	}

	static uchar old_buttons = 0xff;
	uchar buttons = GetState( kjsif->GetJoystick() );
	if( buttons != old_buttons )
	{
		char s[]="%111FUDLR";
		for(uint i=1;i<=8;i++) { if( (buttons&(0x100>>i))==0 ) s[i]='-'; }
		[buttonsView setStrValue:s];
		old_buttons = buttons;
	}
		
	if( [joystickPopupMenu indexOfSelectedItem] != int(kjsif->GetJoystick()) )
	{
		[joystickPopupMenu selectItemAtIndex:kjsif->GetJoystick()];
	}
		
	static bool old_js1 = yes;
	if( IsPluggedIn(hid_joystick1) != old_js1 ) 
	{
		old_js1 = !old_js1;
		[[joystickPopupMenu itemAtIndex:0] setEnabled:old_js1];
	} 
		
	static bool old_js2 = yes;
	if( IsPluggedIn(hid_joystick2) != old_js2 ) 
	{
		old_js2 = !old_js2;
		[[joystickPopupMenu itemAtIndex:1] setEnabled:old_js2];
	} 
}


- (IBAction)rescanDevices:(id)sender
{
	XLogIn("KempstonJoystickInterfacePanel:rescanDevices");
	FindJoysticks ( 1 /*force*/ );
}


- (IBAction)setKeys:(id)sender		
{
	XLogIn("KempstonJoystickInterfacePanel:setKeys");
	[keyboard_joystick_panel makeKeyAndOrderFront:self];
}


@end












