/*	Copyright  (c)	Günter Woigk 2006 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define	LOG		1

#import "KempstonMouseInterface.h"
#import "Machine.h"


/*	acc. to schematics on k1.dyndns.org:

		$FBDF	%----.-011.--0-.----	x coord
		$FFDF	%----.-111.--0-.----	y coord
		$FADF	%----.--10.--0-.----	buttons: D0/D1 = right/left; 0 = pressed

	there are also other informations:

		middle button = D2 ((existed such a version?))
		address decoding A5 + full decoding of A8 to A11 for all: x, y, and buttons.
*/


#define o_addr	NULL
#define	i_addr	"----.--1-.--0-.----"



KempstonMouseInterface::KempstonMouseInterface ( Machine* m )
:	ZxItem( m, "Kempston Mouse Interface",o_addr,i_addr)
{
	m->kempston_mouse_interface = this;
	[kempston_mouse_if_panel showPanel];
	if(fullscreen_mode) Mouse::GrabMouse();
}


KempstonMouseInterface::~KempstonMouseInterface ( )
{
	GetMachine()->kempston_mouse_interface = NULL;
	if( Mouse::IsGrabbed() ) Mouse::ReleaseMouse();
}


void	 
KempstonMouseInterface::Input( Time, long, ushort address, uchar& byte, uchar& mask )
{
	switch( (address>>8)&7 )
	{
	case 6:	// read buttons:
	case 2:	// note: A10 not decoded for buttons 
		byte &= GetButtons();
		mask |= 3;				// 2-button version
		return;
	
	case 3:	// read x-axis:
		byte &= GetXAxis();
		mask  = 0xff;			// all bits driven
		return;

	case 7:	// read y-axis:
		byte &= GetYAxis();       
		mask  = 0xff;			// all bits driven
		return;

	default: 
		return;
	}
}






