/*	Copyright  (c)	Günter Woigk 1995 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef Mmu_h
#define Mmu_h

#import "templates/SharedArray.h"
#import "ZxItem.h"
#import "zxsp_types.h"



class Mmu : public ZxItem
{
protected:
	Z80*		cpu;
	Ula*		ula;
	SharedArray<ushort>	ram;
	SharedArray<ushort>	rom;
	uchar		port_7ffd;
	uchar		port_1ffd;
	uchar		port_F4;
	uchar		port_FF;
	
public: 
				Mmu					( Machine* );
				Mmu					( Machine*, cstr name, cstr o_addr, cstr i_addr );
virtual			~Mmu				();

virtual bool	IsA					( int id ) const		{ return id==isa_Mmu || ZxItem::IsA(id); }
virtual void	Reset				( Time, long cc );
virtual void	PowerOn				( Time, long cc );
//virtual void	Output				( Time, long cc, ushort addr, uchar byte );
//virtual uchar	Input				( Time, long cc, ushort addr );

virtual bool	HasPort7ffd			()		const throw()	{ return no; }
virtual bool	HasPort1ffd			()		const throw()	{ return no; }
virtual bool	HasPortF4			()		const throw()	{ return no; }
virtual bool	HasPortFF			()		const throw()	{ return no; }

	uchar		GetPort7ffd			()		const throw()	{ return port_7ffd; }
	uchar		GetPort1ffd			()		const throw()	{ return port_1ffd; }
	uchar		GetPortF4			()		const throw()	{ return port_F4; }
	uchar		GetPortFF			()		const throw()	{ return port_FF; }

//	set ports:
//	override locked_to_48k
//	• setup during initialisation
//	• perform selections in control panels
//	do not update screen up to current cc!

virtual void	SetPort7ffd			( uchar n )				{ /*nop*/ }
virtual void	SetPort1ffd			( uchar n )				{ /*nop*/ }
virtual void	SetPortF4			( uchar n )				{ /*nop*/ }
virtual void	SetPortFF			( uchar n )				{ /*nop*/ }

virtual void	SetLockedTo48k		( bool f )				{ /*nop*/ }
virtual bool	IsLockedTo48k		()		const throw()	{ return yes; }
};




#endif









