/*	Copyright  (c)	Günter Woigk 1995 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		1

#import "Mmu.h"
#import "Machine.h"
#import "Z80.h"
#import "Ula.h"


/*	creator for use by derived classes:
	for models with mmu
*/
Mmu::Mmu ( Machine* m, cstr name, cstr o_addr, cstr i_addr )
:	ZxItem( m, name, o_addr, i_addr )
{
	XXLogIn("new Mmu"); 
	cpu = NULL;		// m->GetCpu();
	ula = NULL;		// m->GetUla();
	port_7ffd = port_1ffd = port_F4 = port_FF = 0;
	m->mmu = this;
}


/*	creator for dummy mmu:
	for models without mmu
*/
Mmu::Mmu ( Machine* m )
:	ZxItem( m, "Mmu 48k" )
{
	XXLogIn("new Mmu"); 
	cpu = NULL;		// m->GetCpu();
	ula = NULL;		// m->GetUla();
	port_7ffd = port_1ffd = port_F4 = port_FF = 0;
	m->mmu = this;
}


Mmu::~Mmu() 
{
	XXLogIn("delete Mmu"); 
	if( GetMachine()->mmu==this ) GetMachine()->mmu = NULL;
}



void Mmu::PowerOn ( Time t, long cc )
{
	XLogIn("Mmu:PowerOn"); 

	cpu = (Z80*)GetMachine()->FindItem(isa_Z80); XXASSERT(cpu); // Connect( infoCpuDeleted, cpu, ZxItem::infoDeleted );
	ula = (Ula*)GetMachine()->FindItem(isa_Ula); XXASSERT(ula); // Connect( infoUlaDeleted, ula, ZxItem::infoDeleted );

	ram = GetMachine()->GetRam();	// => shared array
	rom = GetMachine()->GetRom();	// => shared array
	XXASSERT( ram.Count()>=16 kB && rom.Count()>=16 kB );

	Mmu::Reset( t, cc );
}


void Mmu::Reset ( Time, long cc )
{
	XLogIn("Mmu:Reset"); 
	
	XXASSERT(ula&&cpu);
	
	if( !GetMachine()->InternalRomDisabled() )
							 cpu->MapRom( 0x0000, 0x4000, &rom[0x0000], NULL, 0 /*no waitstates*/ );
							 cpu->MapRam( 0x4000, 0x4000, &ram[0x0000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	if( ram.Count()>0x4000 ) cpu->MapRam( 0x8000, 0x4000, &ram[0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	if( ram.Count()>0x8000 ) cpu->MapRam( 0xC000, 0x4000, &ram[0x8000], ula->GetWaitmap(), ula->GetWaitmapSize() );

	ula->SetVideoRam( &ram[GetMachine()->GetInfo()->videoram_start[0]] );
}









