/*	Copyright  (c)	Günter Woigk 1995 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
	
	Mmu for +2A and +3.

	based on Mmu128k
*/

#ifndef MmuPlus3_h
#define MmuPlus3_h

#import "Mmu128k.h"


class MmuPlus3 : public Mmu128k
{
	void		page_rom_plus3			();
	void		page_ram_plus3			();
	void		page_mem_plus3			();
	void		page_only_ram			();
	void		set_port_7ffd_and_1ffd	(uchar,uchar);
	void		send_printer_strobe		(bool)			{ /*if(machine->lprinter)machine->lprinter->SetStrobe(f);*/ }
	void		send_disc_motor_state	(bool)			{ /*if(machine->discdrive)machine->discdrive->SetMotor(f);*/ }
	
public:

				MmuPlus3			( Machine* );
//virtual		~MmuPlus3			(); 

virtual bool	IsA					( int id ) const	{ return id==isa_MmuPlus3 || Mmu128k::IsA(id); }
virtual void	Reset				( Time, long cc );
virtual void	PowerOn				( Time, long cc );
virtual void	Output				( Time, long cc, ushort addr, uchar byte );
//virtual uchar	Input				( Time, long cc, ushort addr );

//virtual bool	HasPort7ffd			()	const throw()	{ return yes; }
virtual bool	HasPort1ffd			()	const throw()	{ return yes; }

//	set ports:
//	override locked_to_48k
//	• setup during initialisation
//	• perform selections in control panels
//	do not update screen up to current cc!

virtual void	SetPort7ffd			( uchar n )			{ set_port_7ffd_and_1ffd(n,port_1ffd); }
virtual void	SetPort1ffd			( uchar n )			{ set_port_7ffd_and_1ffd(port_7ffd,n); }
virtual void	SetLockedTo48k		( bool f )			{ SetPort7ffd( f ? port_7ffd|0x20 : port_7ffd&~0x20 ); }
	void		SetDiscMotorState	( bool f )			{ SetPort1ffd( f ? port_1ffd|0x08 : port_1ffd&~0x08 ); }
	void		SetPrinterStrobe	( bool f )			{ SetPort1ffd( f ? port_1ffd|0x10 : port_1ffd&~0x10 ); }
	void		SetRamOnlyMode		( bool f )			{ SetPort1ffd( f ? port_1ffd|0x01 : port_1ffd&~0x01 ); }

//virtual bool	IsLockedTo48k		()	const throw()	{ return port_7ffd&0x20; }
	bool		GetDiscMotorState	()	const throw()	{ return port_1ffd&0x08; }
	bool		GetPrinterStrobe	()	const throw()	{ return port_1ffd&0x10; }
	bool		IsRamOnlyMode		()	const throw()	{ return port_1ffd&0x01; /* except if locked to 48k */ }
};






#endif









