/*	Copyright  (c)	Günter Woigk 1995 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		1

#import "utilities.h"
#import "MmuPlus3.h"
#import "Machine.h"
#import "Z80.h"
#import "Ula.h"


// $7ffd:	paging control register on 128 and +2, +3, +2A
// $1ffd:	on +3 and +2A the register $1ffd is also used.

#define	o_addr_7ffd	"01--.----.----.--0-"	// üblicher Port: 0x7ffd
#define	o_addr_1ffd	"0001.----.----.--0-"	// üblicher Port: 0x1ffd


MmuPlus3::MmuPlus3 ( Machine* m )
:	Mmu128k( m, "Mmu +2A/+3", And(o_addr_7ffd,o_addr_1ffd), NULL )
{
	XLogIn("new MmuPlus3"); 

	//TODO();
	//m->RegisterForOutput(this,o_addr_1ffd);		
}


void MmuPlus3::Reset ( Time, long cc ) 	
{
	XLogIn("MmuPlus3:Reset");	 
	if( cc > ula->GetCpuCycleOfNextCrtRead() ) ula->UpdateScreenUpToCycle(cc);
	set_port_7ffd_and_1ffd(0,0);
}


void MmuPlus3::PowerOn ( Time t, long cc )
{
	XLogIn("MmuPlus3:PowerOn");
	Mmu::PowerOn(t,cc);
	XXASSERT(ram.Count()==0x20000);
	XXASSERT(rom.Count()==0x10000);
	//if( cc > ula->GetCpuCycleOfNextCrtRead() ) ula->UpdateScreenUpToCycle(cc);
	set_port_7ffd_and_1ffd(0,0);
}


/*	page in selected rom 
	assumes: not locked to 48k and not ram-only mode
*/
void MmuPlus3::page_rom_plus3()
{
	if( GetMachine()->InternalRomDisabled() ) return;
	int c = ( (int(port_1ffd&0x04)) << 13 )		// hibit of rom_idx from bit 2 to bit 15
		  + ( (int(port_7ffd&0x10)) << 10 );	// lobit from bit 4 to bit 14
	cpu->MapMem( 0x0000, 0x4000, &rom[c], nowritepage, NULL, 0 );
}


/*	page in selected ram at $c000 in ROM+RAM mode
	assumes: not locked to 48k and not ram-only mode
*/
inline void MmuPlus3::page_ram_plus3()
{
	int i = port_7ffd & 0x07;
	cpu->MapRam( 3*0x4000, 0x4000, &ram[i*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
}


/*	page in rom and ram pages in ROM+RAM mode
	assumes: not locked to 48k and not ram-only mode
*/
void MmuPlus3::page_mem_plus3()
{
	page_ram_plus3();
	cpu->MapRam( 2*0x4000, 0x4000, &ram[2*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	cpu->MapRam( 1*0x4000, 0x4000, &ram[5*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	page_rom_plus3();
}


/*	page in 4 ram pages in ram-only mode
	assumes: ram-only mode
*/
void MmuPlus3::page_only_ram()
{
	int const  pg[] = { 0,1,2,3, 4,5,6,7, 4,5,6,3, 4,7,6,3 };
	int const* pp   = pg + 2*(port_1ffd&6);

//	TODO: does ROMDIS also disable RAM at $0000 in ram only mode?
//	Falls nicht, muss der ROMDIS-Setter auf only-ram mode prüfen!!
//	Oder die MMU bietet SetRomDis(rom*) und ResRomDis() o.Ä. an.

	int i = 0;
	int n = pp[i];
	
	if( GetMachine()->InternalRomDisabled() ) 
	{
		ushort* xrom = cpu->RdPtr(0);
		cpu->MapMem( i*0x4000, 0x4000, xrom, &ram[n*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	}

	for ( ;i<4;i++ )
	{
		n = pp[i];
		cpu->MapRam( i*0x4000, 0x4000, &ram[n*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	}
}	


/*	set port $7ffd and $1ffd
	override locked_to_48k
	• setup during initialisation
	• perform selections in control panels
	does not update screen up to current cc!
*/
void MmuPlus3::set_port_7ffd_and_1ffd ( uchar new_7ffd, uchar new_1ffd )
{
	XXLogIn("MmuPlus3:set_port_7ffd_and_1ffd:$%2x,$%2x",(int)new_7ffd,(int)new_1ffd);

	port_7ffd = new_7ffd;
	port_1ffd = new_1ffd;

	if( new_7ffd&0x20 )	// locked to 48k ?
	{
		page_mem_48k();
		ula->SetVideoRam( &ram[5*0x4000] );
	}
	else
	{
		if (new_1ffd&1)	page_only_ram();		// ram only mode
		else			page_mem_plus3();		// rom+ram mode
		ula->SetVideoRam( &ram[5*0x4000 + (new_7ffd&0x08)*0x1000] );
	}

	send_printer_strobe(new_1ffd&0x10);
	send_disc_motor_state(new_1ffd&0x08);
}


void MmuPlus3::Output ( Time, long cc, ushort addr, uchar byte )
{
	XXLogIn("MmuPlus3:Output($%4x,$%2x)",(int)addr,(int)byte);

	if(port_7ffd&0x20) return;					// mmu disabled. 
												// Note: some demos seem to assume that the rom/ram state is freezed
												// and not set to 48k compatible mode!
	if( (addr&0xC000)==0x4000 )					// write to port 7ffd
	{
		uchar toggled = port_7ffd ^ byte;
		port_7ffd = byte;

		if (toggled&0x08)						// screen changed?
		{										// (byte&0x08) ? ram+0x4000*7 : ram+0x4000*5;
			if( cc > ula->GetCpuCycleOfNextCrtRead() ) ula->UpdateScreenUpToCycle(cc);
			ula->SetVideoRam( &ram[5*0x4000 + (byte&0x08)*0x1000] );
		}
		
		if (port_1ffd&1) return;				// +3/+2A: special mode: all ram, all set by port $1ffd only
		if (toggled&0x07) page_ram_plus3();		// ram page at $c000 changed?
		if (toggled&0x10) page_rom_plus3();		// rom page changed?
	}
	else if( (addr&0xF000)==0x1000 )			// write to port 1ffd
	{
		uchar toggled = port_1ffd ^ byte;
		port_1ffd = byte;

		if(byte&1)			// only ram
		{
			if (toggled&7) page_only_ram();			// switch from rom+ram mode to ram-only mode
		}											// or switched ram config. in ram-only mode
		else				// rom+ram
		{
			if (toggled&1) page_mem_plus3();		// switch from ram only mode to rom+ram mode
			else if (toggled&4) page_rom_plus3();	// switch rom in rom+ram mode
		}
		
		if (toggled&0x08) send_disc_motor_state(byte&0x08);		// bit 3: Disc motor
		if (toggled&0x10) send_printer_strobe(byte&0x10);		// bit 4: Printer strobe
	}
}












