/*	Copyright  (c)	Günter Woigk  2006
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define	LOG		1

#include <unistd.h>

#import "cstrings/cstrings.h"
#import "unix/file_utilities.h"
#import	"SinclairZxInterface2.h"
#import	"SinclairZxInterface2Panel.h"
#import "Machine.h"
#import "Z80.h"
#import "Mmu.h"


/*
	sinclair1:
	data byte:	%xxxLRDUF active low
	address:	0xEFFE
				%---0.----.----.---0

	sinclair2:
	data byte:	%xxxFUDRL active low
	address:	0xF7FE
				%----.0---.----.---0
				
	Reading both joysticks simultaneously with addr = 0xE7FE is not possible.
	The upper data bits D5 to D7 are not set when reading the joystick port - 
	they are not connected to the MT62001 interface IC. 
	Data bit D6 is EAR_IN from the ULA, D5 and D7 are in most cases "1".
*/


#define o_addr	NULL
#define	i_addr	"----.----.----.---0"



SinclairZxInterface2::SinclairZxInterface2 ( Machine* m )
:	ZxItem(  m, "Sinclair ZX Interface 2",o_addr,i_addr)
{
	XLogIn("new SinclairZxInterface2");

	FindJoysticks ( );
	joystick1 = IsPluggedIn(hid_joystick1) ? hid_joystick1 : kbd_joystick;
	joystick2 = IsPluggedIn(hid_joystick2) ? hid_joystick2 : no_joystick;
	rom       = NULL;
	filepath  = NULL;

	m->sinclair_zx_interface2 = this;
	[sinclair_zx_interface2_panel showPanel];
}

SinclairZxInterface2::~SinclairZxInterface2 ( )
{
	UnloadRom();
	GetMachine()->sinclair_zx_interface2 = NULL;
}


void SinclairZxInterface2::Input ( Time/*t*/, long cc, ushort address, uchar& byte, uchar& mask )
{ 
	uchar rval  = 0; 

	if( (address&0x1801)==0x0800 )		// Sinclair 1: %000LRDUF active low
	{
		if( GetMachine() == front_machine() ) rval = GetState(joystick1);	//	%000FUDLR active high
		rval = ((rval&0x10)>>4) | ((rval&0x08)>>2) | ((rval&0x04)) | ((rval&0x03)<<3);
		mask |= rval;					// only pull down.
		byte &= ~rval;
	}
	else if( (address&0x1801)==0x1000 )	// Sinclair 2: %000FUDRL active low
	{
		if( GetMachine() == front_machine() ) rval = GetState(joystick2);	//	%000FUDLR active high 
		rval = (rval&0x1c) | ((rval&0x02)>>1) | ((rval&0x01)<<1);
		mask |= rval;					// only pull down.
		byte &= ~rval;
	}
}


void SinclairZxInterface2::UnloadRom ( )
{
	GetMachine()->Lock();

	//	if( rom && machine->GetRomDis() == rom && machine->ula && machine->cpu ) 
		if( rom && GetMachine()->InternalRomDisabled() && GetMachine()->cpu && GetMachine()->cpu->RdPtr(0) == rom && GetMachine()->ula ) 
		{
			uptr  wm  = GetMachine()->ula->GetWaitmap(); 
			ulong wmsz= GetMachine()->ula->GetWaitmapSize();
			GetMachine()->EnableInternalRom();	
			GetMachine()->cpu->MapRom( 0x0000, 0x4000, GetMachine()->GetRom().Data(), wm,wmsz );
			GetMachine()->Reset();
		}

	delete[] rom; rom = NULL;
	delete[] filepath; filepath = NULL;

	GetMachine()->Unlock();
}


void SinclairZxInterface2::LoadRom ( cstr file_path )
{ 
	UnloadRom();

	int fd = open(file_path,O_RDONLY,0); 
	if (fd==-1) return;						// file does not exist
	uchar bu[16 kB];
	ulong size = read(fd,bu,16 kB);
	close(fd);
	if (size!=16 kB) return;				// too short, read error, ...
	
	rom = new ushort[16 kB];
	for( uint i=0; i<size; i++ ) rom[i] = bu[i];	// also clears all flags
	filepath = NewCopy(FileNameFromPath(file_path));

	GetMachine()->Lock();

		uptr  wm  = GetMachine()->ula->GetWaitmap(); 
		ulong wmsz= GetMachine()->ula->GetWaitmapSize();
		GetMachine()->DisableInternalRom();
		GetMachine()->cpu->MapRom( 0x0000, 0x4000, rom, wm,wmsz );
		GetMachine()->Reset();

	GetMachine()->Unlock();
}










