
#define SAFE	3
#define LOG		1

#import "cstrings/cstrings.h"
#import "unix/file_utilities.h"
#import "KeyboardJoystickPanel.h"
#import "SinclairZxInterface2Panel.h"
#import "NXTextField.h"
#import "NXImage.h"
#import "Machine.h"
#import "SinclairZxInterface2.h"

static NSImage* if2_image		= nil;
static NSImage* if2_w_rom_image	= nil;



@implementation SinclairZxInterface2Panel


-(id)initWithContentRect:(NSRect)contentRect 
			   styleMask:(uint)styleMask 
			     backing:(NSBackingStoreType)backingType 
				   defer:(BOOL)flag
{
	XLogIn("SinclairZxInterface2Panel:initWithContentRect");
	
	return sinclair_zx_interface2_panel = self =
		[super initWithContentRect:contentRect styleMask:styleMask 
						backing:backingType defer:flag];
}



//- (void)awakeFromNib { XLogIn("SinclairZxInterface2Panel:awakeFromNib"); }


- (IBAction)selectJoystick1:(id)sender
{
	QBLOCK;
	SinclairZxInterface2* zxif2 = m->sinclair_zx_interface2; 
	if(zxif2) { zxif2->SetJoystick1((JoystickID)[sender indexOfSelectedItem]); }
	QUNBLOCK;
}

- (IBAction)selectJoystick2:(id)sender
{
	QBLOCK;
	SinclairZxInterface2* zxif2 = m->sinclair_zx_interface2; 
	if(zxif2) { zxif2->SetJoystick2((JoystickID)[sender indexOfSelectedItem]); }
	QUNBLOCK;
}


/*	Update Panel: 
	- call from main event loop
	- for regular panel update
	- called only if panel is visible
	- front machine m is locked for reading (RLOCK/RUNLOCK)
*/
-(void)updatePanel:(Machine*)m
{
	SinclairZxInterface2* zxif2 = m->sinclair_zx_interface2; 

	static bool old = 1;
	static cstr old_rom = "xxx";

	if( old != (zxif2!=0) )
	{
		old = zxif2!=0;
		[self enableSubviews:old];
	}

	if(zxif2)
	{
	// Insert/Eject button:
	// image:
	// loaded cartridge name:

		if( old_rom != zxif2->LoadedRom() )
		{
			old_rom = zxif2->LoadedRom();
			[selectedRomView setStrValue:old_rom?BasenameFromPath(old_rom):""];
			
			if( !if2_image ) 
			{
				if2_w_rom_image = [[NXImage alloc] initWithContentsOfResourceFile:"sinclair_if2_with_cart.jpg"];
				if2_image       = [imageView image];				
			}

			bool loaded = old_rom != NULL;
			[selectRomButton setTitle:loaded?@"Eject":@"Insert.."];
			[imageView setImage:loaded?if2_w_rom_image:if2_image];
	  }	
	  	  
	  {	static uchar old = 0xff;
		uchar rval  = GetState(zxif2->GetJoystick1());	//	%000FUDLR active high

		if( rval != old )
		{
			old = rval;
			rval = ((rval&0x10)>>4) | ((rval&0x08)>>2) | ((rval&0x04)) | ((rval&0x03)<<3);
			char s[]="%000LRDUF";	// sinclair1
			for(uint i=1;i<=8;i++) { if( (rval&(0x100>>i))==0 ) s[i]='-'; };
			[joystick1ButtonsView setStrValue:s];
		}
	  }
	  {	static uchar old = 0xff;
		uchar rval  = GetState(zxif2->GetJoystick2());	//	%000FUDLR active high 

		if( rval != old )
		{
			old = rval;
			rval = (rval&0x1c) | ((rval&0x02)>>1) | ((rval&0x01)<<1);
			[joystick2ButtonsView setBinaryValue:rval withZeros:"%000-----" andOnes:"%000FUDRL"];
			//char s[]="%000FUDRL";	// sinclair2
			//for(uint i=1;i<=8;i++) { if( (rval&(0x100>>i))==0 ) s[i]='-'; };
			//[joystick2ButtonsView setStrValue:s];
		}
	  }
	  	if( [joystick1PopupMenu indexOfSelectedItem] != int(zxif2->GetJoystick1()) )
		{
			[joystick1PopupMenu selectItemAtIndex:zxif2->GetJoystick1()];
		}
	  
	  	if( [joystick2PopupMenu indexOfSelectedItem] != int(zxif2->GetJoystick2()) )
		{
			[joystick2PopupMenu selectItemAtIndex:zxif2->GetJoystick2()];
		}

		static bool old_js1 = yes;
		if( IsPluggedIn(hid_joystick1) != old_js1 ) 
		{
			old_js1 = !old_js1;
			[[joystick1PopupMenu itemAtIndex:0] setEnabled:old_js1];
			[[joystick2PopupMenu itemAtIndex:0] setEnabled:old_js1];
		} 
			
		static bool old_js2 = yes;
		if( IsPluggedIn(hid_joystick2) != old_js2 ) 
		{
			old_js2 = !old_js2;
			[[joystick1PopupMenu itemAtIndex:1] setEnabled:old_js2];
			[[joystick2PopupMenu itemAtIndex:1] setEnabled:old_js2];
		} 	  
	}
}


/* Select / Eject Button:
*/
- (IBAction)selectRom:(id)sender
{
	XLogIn("SinclairZxInterface2Panel:selectRom");

	LBLOCK;
	
		SinclairZxInterface2* zxif2 = m->sinclair_zx_interface2;		
		if(zxif2)
		{
			cstr filepath = zxif2->LoadedRom();
			if( filepath )	// loaded => eject
			{
				zxif2->UnloadRom();
			}
			else	// no rom loaded => load
			{
				NSArray*	 types  = [NSArray arrayWithObjects:@"rom",@"ROM",nil];
				NSOpenPanel* dialog = [NSOpenPanel openPanel];
				int			 result = [dialog runModalForTypes:types];

				if (result==NSOKButton)
				{
					filepath = [[[dialog filenames] objectAtIndex:0] fileSystemRepresentation];		
					zxif2->LoadRom(filepath);
				}
			}
		}
		
	LUNBLOCK;
}


- (IBAction)rescanDevices:(id)sender
{
	XLogIn("SinclairZxInterface2Panel:RescanDevices");
	FindJoysticks ( 1 /*force*/ );
}


- (IBAction)setKeys:(id)sender
{
	XLogIn("SinclairZxInterface2Panel:setKeys");
	[keyboard_joystick_panel makeKeyAndOrderFront:self];
}





@end
