/*	Copyright  (c)	Günter Woigk 1995 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/
#ifndef Ula_h
#define Ula_h


#import "ZxItem.h"
#import "zxsp_types.h"

#define	MIN_LINES_BEFORE_SCREEN		25
#define	MAX_LINES_BEFORE_SCREEN		75

#define	MIN_LINES_AFTER_SCREEN		25
#define	MAX_LINES_AFTER_SCREEN		75

#define	MIN_ULA_CLOCKS_PER_LINE		((8+32+8)*16)
#define	MAX_ULA_CLOCKS_PER_LINE		((16+32+16)*16)


class Ula : public ZxItem
{
/*	current timing parameters:
	Die Ula weiß nichts von ihrem Eingangstakt.
	Sie wird indirekt über die CRT Backcalls der Cpu getaktet.	
*/
	long		lines_before_screen;	// 63 or 64 for 50 Hz
	long		lines_in_screen;		// 192
	long		lines_after_screen;		// 56 for 50 Hz		
	long		uc_per_row;				// ula cycles per video row
	long		uc_per_cc;				// ula cycles per cpu cycle		==  cpu clock predivider:		1, 2, 4, or 8
	long		uc_per_byte;			// const: ula cycles per linear screen byte address increment:	16

/*	Aus den Timing-Parametern resultierende Werte
*/
	long		uc_per_side_border;		// Zeit für Seitenborder+Strahlrücklauf

	long	 	cc_per_row;				// Total cpu cycles per line
	long		cc_waitmap_start;		// Ab wann die Waitmap benutzt werden muss
	long		cc_screen_start;		// Erster cc für einen CRT Backcall
	long		cc_waitmap_end;			// Ab wann nicht mehr
	long		cc_frame_end;			// Total cpu clocks per Frame

	uchar		waitmap[MAX_ULA_CLOCKS_PER_LINE];	// Aus den Timing-Parametern resultierende Waitmap. 

// Beeper:
	uchar		ula_out_byte;
	Time		ula_out_time;			// time [seconds] of last ula out == last sound output
	struct 
	{	Sample	volume;					// 0.0 ... 1.0
		Sample	current_sample;			// current beeper elongation
		Time	last_sample_time;
	}			beeper;
	int			ear_out_count;			// for autostart when saving
	int			ear_in_count;			// for autostart when loading

// CRTC:
	uchar		border_color;			// current border color
	ushort*		video_ram;				// current video ram
	long		current_frame;			// counter, used for flash phase
	long		ccx;					// next cc for reading from video_ram
	long		ccb;					// next cc for reading from border_color

// specci screen:						// [0,0] is top left corner
	ushort*		attr_pixel;				// specci screen attribute and pixel tupels
	uchar*		border;					// recorded border bytes

// validate settings and setup new timing
	void		setup_timing			( );
	void		setup_waitmap			( uchar wm_bits );


public:
				Ula						( Machine* );
				~Ula					( );

virtual bool	IsA						( int id ) const		{ return id==isa_Ula || ZxItem::IsA(id); }
virtual	void	Reset					( Time, long cc );
virtual	void	PowerOn					( Time, long cc );
virtual void	DspBufferEnd			( Time, long cc );		
virtual	void	Output					( Time, long cc, ushort addr, uchar byte );
virtual	void	Input					( Time t, long cc, ushort address, uchar& byte, uchar& mask );
		uchar	Input					( Time t, long cc, ushort address );
		long	AddSpecialWaitStates	( long cc );
	
	Sample		GetBeeperVolume			( )				{ return beeper.volume; }
	void		SetBeeperVolume			( Sample s )	{ if(s<0) s=0; else if(s>1) s=1; beeper.volume = s; }

	int&		EarOutCount				( )				{ return ear_out_count; }
	int&		EarInCount				( )				{ return ear_in_count; }

	void		SetCpuClockPredivider	( long n )		{ if(uc_per_cc!=n)			 { uc_per_cc=n;			  setup_timing(); } }
	void		SetLinesBeforeScreen	( long n )		{ if(lines_before_screen!=n) { lines_before_screen=n; setup_timing(); } }
	void		SetLinesInScreen		( long n )		{ if(lines_in_screen!=n)     { lines_in_screen=n;     setup_timing(); } }
	void		SetLinesAfterScreen		( long n )		{ if(lines_after_screen!=n)  { lines_after_screen=n;  setup_timing(); } }
	void		SetUlaCyclesPerLine		( long n )		{ if(uc_per_row!=n)			 { uc_per_row=n;		  setup_timing(); } }
	void		SetCpuCyclesPerLine		( long n )		{ SetUlaCyclesPerLine(n*uc_per_cc); }

	long		GetCpuClockPredivider	( )				{ return uc_per_cc; }
	long		GetUlaCyclesPerLine		( )				{ return uc_per_row; }
	long		GetCpuCyclesPerLine		( )				{ return cc_per_row; }
	long		GetOctetsPerLine		( )				{ return uc_per_row/uc_per_byte; }
	long		GetCpuCyclesPerFrame	( )				{ return cc_frame_end; }
	long		GetCpuCycleOfInterrupt	( )				{ return 0; }
	long		GetCpuCycleOfIrptEnd	( )				{ return 32; }
	long		GetUlaCyclesPerFrame	( )				{ return cc_frame_end*uc_per_cc; }
	long		GetOctetsPerFrame		( )				{ return cc_frame_end*uc_per_cc/uc_per_byte; }
	long		GetLinesBeforeScreen	( )				{ return lines_before_screen; }
	long		GetLinesInScreen		( )				{ return lines_in_screen; }
	long		GetLinesAfterScreen		( )				{ return lines_after_screen; }
	long		GetScreenStart			( )				{ return cc_screen_start; }
	long		GetWaitmapStart			( )				{ return cc_waitmap_start; }
	long		GetWaitmapEnd			( )				{ return cc_waitmap_end; }
	uptr		GetWaitmap				( )				{ return waitmap; }
	long		GetWaitmapSize			( )				{ return cc_per_row; }
	int			GetEarOutState			( )				{ return ula_out_byte & 0x10; }
	int			GetMicOutState			( )				{ return ula_out_byte & 0x08; }

// CRTC:
	long		UpdateScreenUpToCycle	( long cc );
	void		UpdateBorderUpToCycle	( long cc );
	void		AddVideoBeamIndicator	( long cc );
	void		AddVideoBeamIndicator	( uchar* my_border, ushort* my_attr_pixel );
	void		DoFrameFlyback			( );

	void		SetVideoRam				( ushort* p )	{ video_ram = p; }
	void		SetBorderColor			( uchar b );

	ushort*		GetVideoRam				( )				{ return video_ram; }
	uchar		GetBorderColor			( )				{ return border_color; }
	uchar*&		GetBorder				( )				{ return border; }
	ushort*&	GetAttrPixel			( )				{ return attr_pixel; }
	long		GetCurrentFrame			( )				{ return current_frame; }
	bool		GetFlashPhase			( )				{ return (current_frame>>4)&1; }
	long		GetCpuCycleOfNextCrtRead( )				{ return ccx/uc_per_cc; }

	long		GetCurrentBorderIndex	( )				{ return ccb; }
	long		GetCurrentScreenIndex	( )				{ return ccx==1<<30 ? 32*24*8 : (ccx/uc_per_row-lines_before_screen)*32 + ccx%uc_per_row/uc_per_byte; }
	
	long		GetCrtRowForCpuCycle	( long cc )		{ return cc / cc_per_row; }
	long		GetCrtColForCpuCycle	( long cc )		{ return cc*uc_per_cc % uc_per_row / uc_per_byte; }

	ushort*		NewAttrPixelArray		( )				{ return new ushort[32*24*8]; }
	uchar*		NewBorderArray			( )				{ return new uchar[MAX_ULA_CLOCKS_PER_LINE/16*(MAX_LINES_BEFORE_SCREEN+192+MAX_LINES_AFTER_SCREEN+1)]; }
};



#endif









