/*	Copyright  (c)	Günter Woigk 1995 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		1

#import "Ula.h"
#import "Machine.h"
#import "Z80.h"
#import "Screen.h"
#import "Dsp.h"
#import "ZxInfo.h"
#import "KbdPanel.h"
#import "AlertPanel.h"


//#define io_addr			"----.----.----.---0"		// übliche Adresse: $FE
#define	i_addr				NULL						// => kein automatischer Aufruf in machine->Input()
#define	o_addr				"----.----.----.---0"		// übliche Adresse: $FE

#define	KEYBOARD_IN_MASK	0x1F
#define	EAR_IN_MASK			0x40		

#define	BORDER_OUT_MASK		0x07
#define	MIC_OUT_MASK		0x08
#define	EAR_OUT_MASK		0x10		
#define	MIC_OUT_SHIFT		3		// 0x01<<3 = 0x08



const long	UC_PER_BYTE		= 16;			// ula cycles per pixel block  ((2 bytes == 16 pixel))


Ula::~Ula()
{
	GetMachine()->ula = NULL;
	delete[] attr_pixel;
	delete[] border;
}


Ula::Ula ( Machine* m )
:	ZxItem(m,"ULA",o_addr,i_addr)
{	
	beeper.volume	= 1.0;				// 0.0 ... 1.0
	uc_per_byte		= UC_PER_BYTE;
	
	attr_pixel  = NewAttrPixelArray();
	border = NewBorderArray();
	m->ula = this;
}


#define	uc_per_byte		UC_PER_BYTE



void Ula::PowerOn ( Time t, long cc )
{
    XLogIn("Ula:PowerOn");

// timing parameters:
	ZxInfo const* m = GetMachine()->GetInfo();
	uc_per_cc				= 4;
	lines_before_screen		= m->lines_before_screen;		// 63 or 64 for 50 Hz
	lines_in_screen			= m->lines_in_screen;			// 192
	lines_after_screen		= m->lines_after_screen;		// 56 for 50 Hz		
	cc_per_row				= m->cpu_cycles_per_line;		// Total cpu cycles per line
	uc_per_row				= cc_per_row * uc_per_cc;		// Total ula clock cycles per line
	
// ear/mic:
	beeper.current_sample	= 0.0;		// current beeper elongation
	beeper.last_sample_time	= t;
	ula_out_byte			= 0;		
	ula_out_time			= t;		
	ear_out_count			= 0;		// auto start/stop tape
	ear_in_count			= 0;	
	border_color			= 0;		// border := black

// Setup CRTC:
	current_frame			= 0;
	//video_ram				= &machine->GetRam()[m->videoram_start[0]];		wird von Mmu gesetzt!
	ccx						= 0;
	ccb						= 0;
	setup_timing();
}


void Ula::setup_waitmap ( uchar wm_bits )		// e.g. wm_bits = %11111100
{
//	uchar m8[4]  = {3,2,1,0};
//	uchar m4[8]  = {6,5,4,3,2,1,0,0};
//	uchar m2[16] = {12,11,10,9,8,7,6,5,4,3,2,1,0,0,0,0};
//	uchar m1[32] = {24,23,22,21,20,19,18,17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0,0,0,0,0,0,0,0};

	uchar bu[60]; memset(bu,0,60);
	uptr m8=bu, m4=bu+4, m2=bu+(4+8), m1=bu+(4+8+16);

	for( int d=0,i=7; i>=0; i-- ) 
	{
		if( char(wm_bits<<i)>=0 ) { d=0; continue; }
		
		d++;
		if(i&1) 
		m8[i/2]   = d/2 +1;

		m4[i*1]   = 1*d;

		m2[2*i+1] = 2*d -1;
		m2[2*i]   = 2*d;

		m1[4*i+3] = 4*d -3;
		m1[4*i+2] = 4*d -2;
		m1[4*i+1] = 4*d -1;
		m1[4*i]   = 4*d;
	}

	XXXTRAP( wm_bits==0xfc && m4[0]!=6 );
	XXXTRAP( wm_bits==0xfe && m4[0]!=7 );

	memset( waitmap,0,cc_per_row );
	XTRAP( cc_per_row < 16*32/uc_per_cc );
	int d = GetMachine()->GetInfo()->waitmap_offset;

	switch(uc_per_cc)
	{
	case 1:	d*=4; for ( int i=0; i<16; i++ ) memcpy( waitmap +d +i*32, m1, 24 ); break;
	case 2:	d*=2; for ( int i=0; i<16; i++ ) memcpy( waitmap +d +i*16, m2, 12 ); break;
	case 4:	d*=1; for ( int i=0; i<16; i++ ) memcpy( waitmap +d +i* 8, m4, 6  ); break;
	case 8:	d/=2; for ( int i=0; i<16; i++ ) memcpy( waitmap +d +i* 4, m8, 3  ); break;
	}
}



/*	setup timing parameters
	in:	uc_per_cc
		uc_per_row
		lines_[before|in|after]_screen
*/
void Ula::setup_timing()
{
// validate settings:
	if( uc_per_cc!=1 && uc_per_cc!=2 && uc_per_cc!=4 && uc_per_cc!=8 ) uc_per_cc = 4; /* default */

	Limit( (32L+16L)*uc_per_byte, uc_per_row, long(sizeof(waitmap)) ); uc_per_row &= -uc_per_byte;

	Limit( 25L,  lines_before_screen, 75L ); 
	Limit( 192L, lines_in_screen,    192L );
	Limit( 25L,  lines_after_screen,  75L );

// Abgeleitete Timing-Werte
	uc_per_side_border		= uc_per_row - 32 * uc_per_byte;
	cc_per_row				= uc_per_row / uc_per_cc;
	cc_frame_end			= (lines_before_screen+lines_in_screen+lines_after_screen) * cc_per_row;

  // von wann bis wann soll die cpu die waitmap benutzen?
	cc_waitmap_start		= lines_before_screen*cc_per_row -32/*etwas früher*/;
	cc_waitmap_end			= cc_waitmap_start + lines_in_screen*cc_per_row;

  // startcycle für den CRTC
	cc_screen_start			= lines_before_screen*cc_per_row;

// Waitmap vom Timing ableiten:
	uchar wm = GetMachine()->GetInfo()->waitmap;		// e.g. %11111100	
	setup_waitmap(wm);

// ccx validieren:
	{
		int row = ccx / uc_per_row - lines_before_screen;
		int col = ccx % uc_per_row / uc_per_byte /2 *2;		// in screen: 0*2 .. 15*2; else in side boder

		if( row<0 )  { row = col = 0; }
		if( col>30 ) { col=0; row+=1; }
		if( row>=lines_in_screen ) ccx = 1<<30;
		else ccx = (lines_before_screen+row) * uc_per_row + col * uc_per_byte;
	}
	
// ccb validieren:
	ccb = ccb;		// ccb ist der byte-index in border[]
	
// Info
	XLogLine("Ula:setup_timing:");
	XLogLine("+ lines_before_screen = %i",lines_before_screen);
	XLogLine("  lines_in_screen     = %i",lines_in_screen);
	XLogLine("  lines_after_screen  = %i",lines_after_screen);

	XLogLine("+ uc_per_row          = %i",uc_per_row);
	XLogLine("  uc_per_side_border  = %i",uc_per_side_border);

	XLogLine("+ cc predivider       = %i",uc_per_cc);
	XLogLine("  cc_per_row          = %i",cc_per_row);
	XLogLine("  cc_waitmap_start    = %i",cc_waitmap_start);
	XLogLine("  cc_screen_start     = %i",cc_screen_start);
	XLogLine("  cc_waitmap_end      = %i",cc_waitmap_end);
	XLogLine("  cc_frame_end        = %i",cc_frame_end);

// andere Komponenten aktualisieren	
	[GetMachine()->screen updateAll];

	for( int i=0; i<0x10000; i+=CPU_PAGESIZE )
	{
		PgInfo& pg = GetMachine()->cpu->Page(i);
		pg.waitmap_r = pg.waitmap_w = waitmap;
		pg.waitmap_r_size = pg.waitmap_w_size = cc_per_row;
	}
}


void Ula::Reset ( Time, long cc )
{ XLogIn("Ula::Reset"); return; }					// Ann.: Die Ula hat nur einen Power-On Reset


void Ula::DspBufferEnd ( Time t, long cc )
{
// pad sound to dsp_buffer: 
	if( t > beeper.last_sample_time )
	{
		DSP::OutputSamples( beeper.current_sample, beeper.current_sample, beeper.last_sample_time, t );
		beeper.last_sample_time = t;
	}
	beeper.last_sample_time -= seconds_per_dsp_buffer;
}

void Ula::SetBorderColor ( uchar byte )
{
	border_color = byte & BORDER_OUT_MASK;
}

long Ula::AddSpecialWaitStates ( long cc )
{
/*	contendion (waitstates)
*/
/*		High byte   |       | 
		in 40 - 7F? | bit 0 | Contention pattern  
		------------+-------+-------------------
			 No     |   0   | N:1, C:3
			Yes     |   0   | C:1, C:3
			 No     |   1   | N:4
			Yes     |   1   | C:1, C:1, C:1, C:1		
*/
	if( cc>=cc_waitmap_start && cc<cc_waitmap_start+cc_per_row*lines_in_screen )
	{
		cc += waitmap[(cc-1)%cc_per_row];
		cc += waitmap[(cc+0)%cc_per_row];
		cc += waitmap[(cc+1)%cc_per_row];
		cc += waitmap[(cc+2)%cc_per_row];
		GetMachine()->cpu->CpuCycle() = cc; 
	}
	return cc;
}


/* ---------------------------------------------------------------
				Output to ULA
--------------------------------------------------------------- */

void Ula::Output ( Time now, long cc, ushort addr, uchar byte )
{
/*	contendion (waitstates)
*/
	if( cc>=cc_screen_start && cc<cc_screen_start+cc_per_row*lines_in_screen )
	{
		if( (addr>>14)==1 ) cc += waitmap[(cc-1)%cc_per_row];
		cc += waitmap[cc%cc_per_row];
		GetMachine()->cpu->CpuCycle() = cc; 
	}
		
/*	cssfaq:	
	The lowest three bits specify the border colour; 
	a zero in bit 3 activates the MIC output, 
	whilst a one in bit 4 activates the EAR output and the internal speaker. 
	However, the EAR and MIC sockets are connected only by resistors, 
	so activating one activates the other; 
	the EAR is generally used for output as it produces a louder sound. 
	The upper two bits are unused.
*/	

//	--- BEEPER ---		
	if( (ula_out_byte^byte)&(MIC_OUT_MASK|EAR_OUT_MASK) ) 
	{
		DSP::OutputSamples( beeper.current_sample, beeper.current_sample, beeper.last_sample_time, now );
		beeper.last_sample_time = now;

		uchar bb = byte;
		bb ^= MIC_OUT_MASK;				// mic pin is low active
		beeper.current_sample = beeper.volume * -0.25 * ( (bb&EAR_OUT_MASK)*3/EAR_OUT_MASK + (bb&MIC_OUT_MASK)*1/MIC_OUT_MASK );
	}
	
// --- BORDER ---
	if( (border_color^byte)&BORDER_OUT_MASK ) 
	{
		UpdateBorderUpToCycle(cc);
		
		border_color = byte & BORDER_OUT_MASK;

	/*	Anmerkung zum Autostart-Tape Feature:
		Wenn nur die Border-Farbe wechselt, ist das wohl ein Hires-Bordereffekt.
		Wenn nur das EAR-Bit wechselt, ist das leise Tonausgabe.
		Wenn nur das MIC-Bit wechselt, ist das normal laute Tonausgabe. 
		Wenn MIC und Border gleichzeitig wechseln, ist das die Save-Routine.
		Tape-I/O am Spectrum wird traditionellerweise am Border visualisiert.
		Custom Taperoutinen, die speichern ohne Borderfeature, aktivieren den Autostart nicht.
		Tonausgabe, die synchron die Borderfarbe ändert, führt zu einem Fehlstart.	***GEFÄHRLICH***
	*/
		if( (ula_out_byte^byte)&(MIC_OUT_MASK) ) ear_out_count++;		// for autostart_tape feature
	}

	ula_out_byte = byte;
	ula_out_time = now;					// für die MIC_OUT -- EAR_IN Verknüpfelung		z.Zt. unbenutzt ((0.7.2.pre11))
}




/* ---------------------------------------------------------------
				Input from ULA
--------------------------------------------------------------- */

void Ula::Input ( Time now, long cc, ushort address, uchar& byte, uchar& mask )
{
	ABORT("Ula::Input called.\n");
}

uchar Ula::Input ( Time now, long cc, ushort addr )
{
	if(addr&1)		// Bit A0 = 1  => ULA not selected, ULA does not stop
	{
	// dummy read:
	// ULA won'now stop -> read mixture of 0xff (idle bus) and attribute byte
	// 2006-10-06 kio: issue 2: reads only attribute bytes, never pixel bytes.
	// TODO: at which clock cycle inside each 2-characters pixel packet 
	//	     is which attribute byte read, and when $ff

		cc -= cc_screen_start;
		if( cc<0 ) return 0xff;							// above screen	

		int row = cc / cc_per_row;	
		if( row>=lines_in_screen ) return 0xff;			// below screen

		int col = (cc*uc_per_cc % uc_per_row) / uc_per_byte;
		if( col>=32 ) return 0xff;						// in side border

	// return pixel byte:
	//	row = (row&0xc0) + ((row>>3)&7) + ((row&7)<<3);	// adapt fancy line ordering
	//	return video_ram[32*row+col];					// return screen byte

	// return attribute byte:
		return video_ram[24*8*32+32*(row/8)+col];
	}

	else			// Bit A0 = 0  => ULA selected
	{
	//	waitstates:
		if( cc>=cc_screen_start && cc<cc_screen_start+cc_per_row*lines_in_screen )
		{
			if( (addr>>14)==1 ) cc += waitmap[(cc-1)%cc_per_row];
			cc += waitmap[cc%cc_per_row];
			GetMachine()->cpu->CpuCycle() = cc; 
		}
			
	// bits 0-4 come from the keyboard
	// bits 5+7 are always 1
	// bit  6	is the signal from the ear socket messed up with ear+mic outputs

	// merge in the keys
		uchar  c = 0xff;
		if(GetMachine()==front_machine())			// 2006-11-18 kio
		{
			uchar* p = [kbd_panel getKeymap];	// 2006-11-18 kio
			uchar minor = (~addr)>>8;
			while ( minor ) { if (minor&1) c &= *p; minor>>=1; p++; }
		}
		
	// insert bit from mic socket
		if ( ula_out_byte & EAR_OUT_MASK )	// EAR_OUT==1
		{
			// ULA_EAR_IN is **probably** always "1"  =>  nothing to do
			// verified 2006-06-29 for issue 2, 4A, and 6U (128K)
		}
		
		else								// EAR_OUT==0
		{
			// signal from EAR input is read into bit 6.
			// threshold level depends on state of EAR output bit and PCB issue/model!
			// EMULATION INFO: some games start loading (jump into the rom loader) with MIC_OUT=0, though it must be 1. The loading routine switches MIC_OUT to 1 as soon as it finds the first transition... but it must find it! Therefore the threshold of EAR_IN can'now be too much away from zero else these games could have never loaded.
			// TODO: simulate capacitor

			Sample threshold = GetMachine()->GetInfo()->ula_earin_threshold[ /* 0?1: TEST!*/ (ula_out_byte>>MIC_OUT_SHIFT)&1 ];
			ulong a = ulong(now*samples_per_second); 		
			if( a>=DSP_SAMPLES_PER_BUFFER+DSP_SAMPLES_STITCHING )
			{
				XXTRAP( (long)a<0 );
				ShowAlert( "Sample input beyond dsp buffer: +%i\n", int(a-DSP_SAMPLES_PER_BUFFER) );
				if(0.0<threshold) c &= ~EAR_IN_MASK;
			}
			else
			{
				if(DSP::audio_in_buffer[a]<threshold) c &= ~EAR_IN_MASK;
			}
			
			ear_in_count++;		// for autostart_tape feature
		}

		return c;
	}
}


void Ula::UpdateBorderUpToCycle ( long cc )
{
	cc--;												// ZX128: test --> weird_science
	int n = ccb;
	int b = border_color;
	int e = (cc*uc_per_cc) / uc_per_byte;				// ZX128: test --> weird_science
	while(n<e) border[n++]=b;
	ccb = n;
}


long Ula::UpdateScreenUpToCycle ( long cc )
{
	cc *= uc_per_cc;
	
	int row = ccx / uc_per_row - lines_before_screen;
	int col = ccx % uc_per_row / uc_per_byte;				// in screen: 0*2 .. 15*2; else in side boder

	XXXTRAP( row<0 );
	XXXTRAP( row>=lines_in_screen && ccx<(1<<30) );
	XXXTRAP( col>30				  && ccx<(1<<30) );
	XXXTRAP( col&1                && ccx<(1<<30) );

	ushort* zap = attr_pixel		+ ( 32 * row + col );

b:	ushort* qp = video_ram	+ ( 32 * ((row&0xc0)+((row>>3)&0x7)+((row&7)<<3)) + col );
	ushort* qa = video_ram	+ ( 32 * (192+row/8) + col );
	
a:	if( ccx>cc ) return (ccx)/uc_per_cc;
		
	*zap++ = (*qa++<<8) + (uchar)*qp++;
	*zap++ = (*qa++<<8) + (uchar)*qp++;
	
	ccx += 2*uc_per_byte;
	col += 2;
	if( col<32 ) goto a;

	ccx += uc_per_side_border;
	col  = 0;
	row += 1;
	if( row<lines_in_screen ) goto b;

	return ccx = 1<<30;
}


void Ula::DoFrameFlyback ( )
{
	UpdateBorderUpToCycle(cc_frame_end);	// border
	ccb = 0;								// update_border_cc
	UpdateScreenUpToCycle(cc_frame_end);	// screen
	ccx = lines_before_screen*uc_per_row;	// update_screen_cc
	current_frame++;						// flash phase
	GetMachine()->cpu->RaiseInterrupt(32);		// Interrupt. TODO: Pentagon "better" interrupt?
}


void Ula::AddVideoBeamIndicator ( uchar* border, ushort* attr_pixel )
{
	uchar color  = 2;				// yellow
	uchar bright = 1;				// fast blinking bright
	if( int(system_time*6)&1 ) { color=6; bright=1; }
	
	border[ ccb ] = color+bright*8;

	int row = ccx / uc_per_row - lines_before_screen;
	if( (uint)row >= (uint)lines_in_screen ) return;	// above/below screen

	if( row>ccb*uc_per_byte/uc_per_row-lines_before_screen ) return;	// side border

	int col = ccx % uc_per_row / uc_per_byte;	
	XXXTRAP( col>=32 || col&1 );

	attr_pixel[32*row+col] = 0x4000*bright + 0x800 * color; 	
}


void Ula::AddVideoBeamIndicator ( long cc )
{
	UpdateBorderUpToCycle(cc+uc_per_byte/uc_per_cc);		// border
	UpdateScreenUpToCycle(cc);								// screen
	AddVideoBeamIndicator(border,attr_pixel);				// add marker in my own buffers
}


















