/*	Copyright  (c)	Günter Woigk 2004 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		1

#import "cstrings/cstrings.h"
#import "globals.h"
#import "utilities.h"
#import "UlaPanel.h"
#import "Machine.h"
#import "Z80.h"
#import "NXTextField.h"
#import "Screen.h"
#import "ZxInfo.h"
#import "Mmu.h"
#import "MmuPlus3.h"




// Displayed state:

#define XXX	0xFFFFFFFF
static Time	ula_clock_percent	= XXX;
static Time	ula_clock			= XXX;
static int		cpu_predivider		= XXX;
static Time	cpu_clock			= XXX;

static int		lines_before_screen	= XXX;
static int		lines_in_screen		= XXX;
static int		lines_after_screen	= XXX;
static int		pixel_per_screen	= XXX;
static int		bytes_per_line		= XXX;
static int		cpu_clocks_per_line	= XXX;
static double	frames_per_sec		= XXX;
static int		cpu_cycles_per_frame= XXX;
static int		crt_row				= XXX;
static int		crt_col				= XXX;
static int		crt_in_screen		= XXX;
static int		waitstates_on		= XXX;
static int		show_inner_screen	= XXX;
static int		waitstates_offset	= XXX;
static char		waitstates_map[8]	= {9,9,9,9,9,9,9,9};
static int		border_color		= XXX;
static int		mic_out_state		= XXX;
static int		ear_out_state		= XXX;
static int		frames_hit			= XXX;

// port 7ffd & port 1ffd views:
static int		old_mmu				= XXX;	enum { mmu_48k, mmu_plus2, mmu_plus2_48k, mmu_plus3, mmu_plus3_48k };
static int		old_7ffd			= XXX;
static int		old_1ffd			= XXX;
static char		old_page_c000		= XXX;//,	old_page_c000_enabled = XXX;
static char		old_page_8000		= XXX;//,	old_page_8000_enabled = XXX;
static char		old_page_4000		= XXX;//,	old_page_4000_enabled = XXX;
static char		old_page_0000		= XXX;//,	old_page_0000_enabled = XXX;
static char		old_videopage		= XXX;//,	old_videopage_enabled = XXX;
static char		old_locked_to_48k	= XXX;
static char		old_ram_only_mode	= XXX;//,	old_ram_only_mode_enabled	= XXX;
static char		old_disc_motor		= XXX;//,	old_disc_motor_enabled		= XXX;
static char		old_printer_strobe	= XXX;//,	old_printer_strobe_enabled	= XXX;
//static cstr		old_textfield_0000	= NULL;	cstr const text_rom = "ROM page", text_ram = "RAM page";
//static int		old_window_height	= XXX;	// note:  window_height_48k < window_height_plus2 < window_height_plus3


// Constants:

// window heights:
int const	window_height_48k		= 340;
int const	window_height_plus2		= 457;
int const	window_height_plus3		= 479;
int const	window_width			= 250;

// port $1ffd bits:
int const	ram_only_mode_mask		= 1<<0;		
int const	ram_only_bits_mask		= 3<<1;
int const	high_rom_bit			= 2;	
int const	high_rom_bit_mask		= 1<<2;	
int const	disc_motor_mask			= 1<<3;	
int const	printer_strobe_mask		= 1<<4;	

// port $7ffd bits:
int const	videoram_mask			= 1<<3;
int const	low_rom_bit				= 4;	
int const	low_rom_bit_mask		= 1<<4;	
int const	lock_to_48k_mask		= 1<<5;	



/* Calculate Textfield Value:
		page 1		-> 1
		12345		-> 12345
		12345 Hz	-> 12345
		12345 kHz	-> 12345000
		12345 MHz	-> 12345000000
		123.45%		-> 123.45
		1/123		-> 123
*/
static double Value( id sender )
{
	cstr s = [[sender stringValue]UTF8String]; if (s[0]==0) return XXX;
	s = LowerStr(s);
	double f = 1.0;
	ptr p;
	if( (p=FindStr(s,"page"))) { s=p+4; } else				// page 1
	if( (p=FindStr(s,"mhz")) ) { *p=0; f=1e6; } else		// 12.34 MHz
	if( (p=FindStr(s,"khz")) ) { *p=0; f=1e3; } else		// 12.34 kHz
	if( (p=FindStr(s,"hz")) )  { *p=0; } else				// 12.34 Hz
	if( (p=FindStr(s,"%")) )   { *p=0; } else				// 12.34 %	--> 12.34 (Ann.: ula clock percent: immer %)
	if( (p=FindStr(s,"/")) )   { s=p+1; }					// 1/4		--> 4	  (for predivider)

	double v = StrToDouble(s,&s); while(*s==' ') s++; if(*s!=0) return XXX;
	return v * f;
}


// UP:	determine best fitting new ram-only mode ($1ffd byte)  
//		for current ram-only mode ($1ffd byte) 
//		and selected new ram page for selected address
//
//	in: old 1ffd state			old1ffd		(ram only mode)
//		requested new page		page		0 .. 7
//		for address range		address		$0000 .. $c000
//
//  wenn neuer Mode nicht eindeutig, dann bevorzuge: 
//		1.: keep current mode
//		2.: mode 0 <-> mode 2  or  mode 4 <-> mode 6
//
//	mode bits are bits 1,2 in $1ffd
//		mode 0:	0,1,2,3
//		mode 2:	4,5,6,7
//		mode 4:	4,5,6,3
//		mode 6:	4,7,6,3
//
static uchar best_ram_only_mode_for ( uchar old1ffd, int page, ushort address )
{
	uchar oldbits  = old1ffd &  ram_only_bits_mask;		// old ram state bits
	uchar oldother = old1ffd & ~ram_only_bits_mask;		// other bits from old1ffd

	switch(address>>14)
	{
	case 0:		// $0000: pages(mode): 0,4,4,4
		if (page==0)	return oldother | 0;
		if (page==4)	return oldbits!=0 ? old1ffd : oldother | 2;
		break;
	case 1:		// $4000: pages(mode): 1,5,5,7
		if (page==1)	return oldother | 0;
		if (page==7)	return oldother | 6;
		if (page==5)	return oldbits&4 ? oldother | 6 : oldother | 0;
		break;
	case 2:		// $8000: pages(mode): 2,6,6,6
		if(page==2)		return oldother | 0;
		if(page==6)		return oldbits!=0 ? old1ffd : oldother | 2;
		break;
	case 3:		// $C000: pages(mode): 3,7,3,3
		if (page==7)	return oldother | 2;
		if (page==3)	return oldbits!=2 ? old1ffd : oldother | 0;
		break;
	}
	return old1ffd;
}





@implementation UlaPanel

// NSTextView delegate methods:
//-(BOOL)control:(NSControl *)control textShouldEndEditing:(NSText *)fieldEditor{ return yes; }
//-(BOOL)control:(NSControl *)control textShouldBeginEditing:(NSText*)fieldEditor{ return yes; }
//-(void)controlTextDidChange:(NSNotification*)note { }
//-(void)close{}
//-(void)windowWillClose:(NSNotification *)aNotification	{}
//-(void)awakeFromNib {}


- (id)initWithContentRect:(NSRect)contentRect styleMask:(uint)styleMask backing:(NSBackingStoreType)backingType defer:(BOOL)flag
{
	XLogIn("UlaPanel:initWithContentRect");
	return ula_panel = self = [super initWithContentRect:contentRect styleMask:styleMask backing:backingType defer:flag];
}


-(IBAction)setPort7ffd:(id)sender
{
	long n = long(Value(sender));
	if (n>=-128 && n<=255) 
	{
		QBLOCK; 
			Mmu* p7 = m->GetMmu(); 
			if (p7) p7->SetPort7ffd(n); 
		QUNBLOCK; 
	}
	old_7ffd = XXX;
}

-(IBAction)setPort1ffd:(id)sender
{
	long n = long(Value(sender));
	if (n>=-128 && n<=255) 
	{
		QBLOCK; 
			Mmu* p7 = m->GetMmu(); 
			if (p7) p7->SetPort1ffd(n); 
		QUNBLOCK; 
	}
	old_1ffd = XXX;
}

-(IBAction)setDiscMotor:(id)sender
{
	ulong n = long(Value(sender));
	if(n<=1) 
	{
		QBLOCK; 
			Mmu* p7 = m->GetMmu(); 
			if (p7&&p7->HasPort1ffd()) ((MmuPlus3*)p7)->SetDiscMotorState(n);
		QUNBLOCK;
	}
	old_disc_motor = XXX;
}

-(IBAction)setPrinterStrobe:(id)sender
{
	ulong n = long(Value(sender));
	if(n<=1) 
	{
		QBLOCK; 
			Mmu* p7 = m->GetMmu(); 
			if (p7&&p7->HasPort1ffd()) ((MmuPlus3*)p7)->SetPrinterStrobe(n);
		QUNBLOCK;
	}
	old_printer_strobe = XXX;
}

-(IBAction)setRamOnlyMode:(id)sender
{
	bool f = [sender state];
	QBLOCK; 
		Mmu* p7 = m->GetMmu(); 
		if (p7&&p7->HasPort1ffd()) ((MmuPlus3*)p7)->SetRamOnlyMode(f);
	QUNBLOCK;
	old_ram_only_mode = XXX;
}

-(IBAction)setLockedTo48k:(id)sender
{
	bool f = [sender state];
	QBLOCK; 
		Mmu* p7 = m->GetMmu(); 
		if (p7) p7->SetLockedTo48k(f);
	QUNBLOCK;
	old_locked_to_48k = XXX;
}

-(IBAction)setVideoram:(id)sender
{
	ulong n = long(Value(sender));
	if (n==5 || n==7)
	{
		QBLOCK; 
			//Mmu* p7 = m->GetMmu(); 
			//if (p7) p7->SetVideoPage(n);
			Ula* ula = m->GetUla();
			Mem& ram = m->GetRam();
			if(ula&&ram.Count()>n*0x4000) ula->SetVideoRam(&ram[n*0x4000]);
		QUNBLOCK;
	}
	old_videopage = XXX;
}


// UP:
-(void) set_page:(int)address :(id)sender
{
	ulong n = long(Value(sender));
	if (n<=7)
	{
		QBLOCK;
			Mmu* p7 = m->GetMmu(); if(!p7) goto x;
			uchar old7ffd = p7->GetPort7ffd();

			if(p7->HasPort1ffd())	// +3 / +2A:	pageable: all ram and rom
			{
				uchar old1ffd = p7->GetPort1ffd();
				if( old1ffd&ram_only_mode_mask )	// ram only mode
				{
					p7->SetPort1ffd( best_ram_only_mode_for(old1ffd,n,address) );
				}
				else		// rom + ram mode: $0000:0-3 $c000:0-7
				{	
					if(address==0xc000)		// page ram
					{
			pram:		p7->SetPort7ffd( (old7ffd&~7)|n );
					}
					else if(address==0x0000 && n<=3) 	// page rom
					{
						p7->SetPort1ffd( n&2 ? old1ffd|high_rom_bit_mask : old1ffd&~high_rom_bit_mask );
			prom:		p7->SetPort7ffd( n&1 ? old7ffd|low_rom_bit_mask  : old7ffd&~low_rom_bit_mask  );
					}
					else {}
				}
			}
			else	// 128k / +2:	pageable: only $C000 0..7  and  $0000 0..1
			{	
				if (address==0xc000) goto pram;					// page ram
				else if (address==0x0000 && n<=1) goto prom;	// page rom
				else {}
			}
x:		QUNBLOCK;
	}
	// page_XXXX = XXX;			must be done by caller
}

-(IBAction)setPageC000:(id)sender	{ [self set_page:0xc000:sender]; old_page_c000=XXX; }
-(IBAction)setPage8000:(id)sender	{ [self set_page:0x8000:sender]; old_page_8000=XXX; }
-(IBAction)setPage4000:(id)sender	{ [self set_page:0x4000:sender]; old_page_4000=XXX; }
-(IBAction)setPage0000:(id)sender	{ [self set_page:0x0000:sender]; old_page_0000=XXX; }







-(IBAction)setUlaClock:(id)sender
{
	Time v = Value(sender); 
	if( v!=XXX ) { QBLOCK; Ula* ula=m->GetUla(); if(ula) m->SetCpuCyclesPerSecond(v/ula->GetCpuClockPredivider()); QUNBLOCK; }
	ula_clock = XXX;	// invalidate displayed value => Textfield will be redrawn with normalized representation of value
}

-(IBAction)setUlaClockPercent:(id)sender
{
	double v = Value(sender); 
	if( v!=XXX ) { QBLOCK; m->SetMachineOverdrive(v*0.01); QUNBLOCK; }
	ula_clock_percent = XXX;
}

-(IBAction)setCpuClock:(id)sender
{
	Time v = Value(sender); 
	if( v!=XXX && v>0 ) { QBLOCK; m->SetCpuCyclesPerSecond(v); QUNBLOCK; }
	cpu_clock = XXX;
}

-(IBAction)setCpuClockDivider:(id)sender
{
	double v = Value(sender); 
	if( v!=XXX ) { QBLOCK; if(v<1.0) v = 1.0/v; m->SetCpuClockPredivider(int(v+0.1)); QUNBLOCK; }
	cpu_predivider = XXX;
}

-(IBAction)setTopBorderLines:(id)sender
{
	ulong n = long(Value(sender));
	if( n>0&&n<100 ) { QBLOCK; m->GetUla()->SetLinesBeforeScreen(n); QUNBLOCK; }	// note: was BLOCK/UNBLOCK, but an IBAction can only me called from the main event loop
	lines_before_screen = XXX;
}

-(IBAction)setBottomBorderLines:(id)sender
{
	ulong n = long(Value(sender));
	if( n>0&&n<100 ) { QBLOCK; m->GetUla()->SetLinesAfterScreen(n); QUNBLOCK; }		// note: was BLOCK/UNBLOCK, but an IBAction can only me called from the main event loop
	lines_after_screen = XXX;
}

-(IBAction)setBytesPerLine:(id)sender
{
	ulong n = long(Value(sender));
	if( n>=32&&n<=128 ) { QBLOCK; m->GetUla()->SetUlaCyclesPerLine(n*16); QUNBLOCK; }	// note: was BLOCK/UNBLOCK, but an IBAction can only me called from the main event loop
	bytes_per_line = XXX;
}

-(IBAction)setCpuClocksPerLine:(id)sender
{
	ulong n = long(Value(sender));
	QBLOCK;												// note: was BLOCK/UNBLOCK, but an IBAction can only me called from the main event loop
		n *= m->GetUla()->GetCpuClockPredivider();
		if( n>=32*16 && n<=128*16 ) m->GetUla()->SetUlaCyclesPerLine(n); 
	QUNBLOCK; 
	cpu_clocks_per_line = XXX;
}

-(IBAction)setCpuCyclesPerFrame:(id)sender
{
	ulong n = long(Value(sender));
	QBLOCK;											// note: was BLOCK/UNBLOCK, but an IBAction can only be called from the main event loop
		Ula* ula = m->GetUla();
		n /= ula->GetCpuCyclesPerLine();
		n -= ula->GetLinesBeforeScreen();
		n -= ula->GetLinesInScreen();
		if( n>0 ) ula->SetLinesAfterScreen(n);
	QUNBLOCK; 
	cpu_cycles_per_frame = XXX;
}

-(IBAction)setShowInnerScreen:(id)sender
{
	QBLOCK; [m->GetScreen() setDrawInnerScreen:[sender state]]; QUNBLOCK;
}


-(IBAction)setFramesPerSecond:(id)sender
{
	double t = Value(sender); 
	if( t>0&&t<400 ) { QBLOCK; m->SetCpuCyclesPerSecond(t*m->GetCpuCyclesPerFrame()); QUNBLOCK; }
	frames_per_sec = XXX;
}

-(IBAction)setBorder:(id)sender
{
	ulong n = ulong(Value(sender));
	if( n>=0 && n<=15 ) { QBLOCK; m->GetUla()->SetBorderColor(n); QUNBLOCK; }
	border_color = XXX;
}

-(IBAction)setUseWaitmap:(id)sender
{
	QBLOCK; m->SetUseWaitmap ([sender state]); QUNBLOCK;
}

-(IBAction)setWaitmapOffset:(id)sender
{
	waitstates_offset = XXX;			// TODO: set waitmap offset in ula panel
}

-(IBAction)setWaitmap:(id)sender
{
	memset(waitstates_map,0xff,sizeof(waitstates_map));			// TODO: set waitmap in ula panel
}

-(IBAction)doRestoreDefaults:(id)sender
{
	LBLOCK;
		const ZxInfo* zx = m->GetInfo();	
		m->SetMachineOverdrive(1.0);
		m->SetCpuClockPredivider (zx->ula_cycles_per_second/zx->cpu_cycles_per_second);
		m->SetUseWaitmap (zx->has_contention());
		
		m->GetUla()->SetLinesBeforeScreen(zx->lines_before_screen);
		m->GetUla()->SetLinesAfterScreen(zx->lines_after_screen);
		m->GetUla()->SetCpuCyclesPerLine(zx->cpu_cycles_per_line);
	LUNBLOCK;
}

#define SET_LABEL(VIEW,OLD,NEW)		if (OLD!=(NEW)					 ) { OLD=NEW; [VIEW setStrValue:NEW]; }
#define SHOW_BYTE(VIEW,OLD,NEW)		if (OLD!=(NEW) && edit_view!=VIEW) { OLD=NEW; sprintf(s,"$%02x",int(OLD));   [VIEW setStrValue:s]; }
#define SHOW_PAGE(VIEW,OLD,NEW)		if (OLD!=(NEW) && edit_view!=VIEW) { OLD=NEW; sprintf(s,"page %i",int(OLD)); [VIEW setStrValue:s]; }
#define SHOW_BOOL(VIEW,OLD,NEW)		if (OLD!=(NEW) && edit_view!=VIEW) { OLD=NEW; [VIEW setIntValue:(NEW)!=0]; }
#define SET_CHECKBOX(VIEW,OLD,NEW)	if (OLD!=(NEW)                   ) { OLD=NEW; [VIEW setState:NEW]; }
#define SET_ENABLED(VIEW,OLD,NEW)	if (OLD!=(NEW)                   ) { OLD=NEW; [VIEW setEnabled:NEW]; }


-(void)resize_window_to_height:(int)newheight
{
	NSRect box = [self frame];		// frame rect, glob. ccord.
	box = [self contentRectForFrameRect:box];
	box.origin.y = box.origin.y + box.size.height - newheight;
	box.size.height = newheight;
	box = [self frameRectForContentRect:box];
	[self setFrame:box display:YES animate:YES];
}


/* change display for new mmu state:
	- old_mmu = new_mmu = mmu_48k, mmu_plus2, mmu_plus2_48k, mmu_plus3, mmu_plus3_48k 
	- resize window
	- show/hide items
	- enable/disable items
	- set constant values
*/

-(void) set_mmu_to: (int)new_mmu
{
	char s[20];
	BOOL locked = NO;		// 48k lock

	switch(new_mmu)
	{
	case mmu_48k:
		[self resize_window_to_height:window_height_48k];
		[separator_7ffd setHidden:YES];
		break;
	
	case mmu_plus2_48k:		// 128/+2 locked to 48k
		locked = YES;

	case mmu_plus2:
		if(old_mmu>mmu_plus2_48k) [self resize_window_to_height:window_height_plus2];
		
		[separator_7ffd	setHidden:NO];

		[view_1ffd				setHidden:YES];
		[textfield_1ffd			setHidden:YES];
		SET_CHECKBOX( checkbox_locked_to_48k, old_locked_to_48k, locked );
		[checkbox_ram_only_mode	setHidden:YES];

		[view_page_c000 setEnabled:!locked]; if(locked) SHOW_PAGE( view_page_c000, old_page_c000, 0 );	
		[view_page_8000 setEnabled:NO];					SHOW_PAGE( view_page_8000, old_page_8000, 2 );
		[view_page_4000 setEnabled:NO];					SHOW_PAGE( view_page_4000, old_page_4000, 5 );
		[view_page_0000 setEnabled:!locked]; if(locked) SHOW_PAGE( view_page_0000, old_page_0000, 1 );	
		[view_videopage setEnabled:!locked]; if(locked) SHOW_PAGE( view_videopage, old_videopage, 5 );
		[textfield_rom_0000 setStrValue:"ROM page"];

	//	[textfield_disc_motor		setHidden:YES];
	//	[view_disc_motor			setHidden:YES];
	//	[textfield_printer_strobe	setHidden:YES];
	//	[view_printer_strobe		setHidden:YES];

		if(old_mmu<mmu_plus2) [self resize_window_to_height:window_height_plus2];
		break;

	case mmu_plus3_48k:		// +3/+2A locked to 48k
		locked = YES;

	case mmu_plus3:
		[separator_7ffd			setHidden:NO];
		
		[view_1ffd				setHidden:NO];
		[textfield_1ffd			setHidden:NO];
		SET_CHECKBOX( checkbox_locked_to_48k, old_locked_to_48k, locked );
		[checkbox_ram_only_mode	setHidden:NO];
		[checkbox_ram_only_mode	setEnabled:!locked]; if(locked) SET_CHECKBOX( checkbox_ram_only_mode, old_ram_only_mode, NO );

		[view_page_c000			setEnabled:!locked]; if(locked) SHOW_PAGE( view_page_c000, old_page_c000, 0 );
		[view_page_8000			setEnabled:!locked]; if(locked) SHOW_PAGE( view_page_8000, old_page_8000, 2 );
		[view_page_4000			setEnabled:!locked]; if(locked) SHOW_PAGE( view_page_4000, old_page_4000, 5 );
		[view_page_0000			setEnabled:!locked]; if(locked) SHOW_PAGE( view_page_0000, old_page_0000, 3 );
		[view_videopage			setEnabled:!locked]; if(locked) SHOW_PAGE( view_videopage, old_videopage, 5 );
		[textfield_rom_0000		setStrValue:old_ram_only_mode?"RAM page":"ROM page"];

		[view_disc_motor		setEnabled:!locked]; if(locked) SHOW_BOOL( view_disc_motor,	old_disc_motor, 0 );
	//	[textfield_disc_motor	setHidden:NO];
	//	[view_disc_motor		setHidden:NO];

		[view_printer_strobe	setEnabled:!locked]; if(locked) SHOW_BOOL( view_printer_strobe, old_printer_strobe, 0 );
	//	[textfield_printer_strobe	setHidden:NO];
	//	[view_printer_strobe		setHidden:NO];
	
		if(old_mmu<mmu_plus3) [self resize_window_to_height:window_height_plus3];	
		break;
	}
		
	old_mmu = new_mmu;
	return;
	
}



/*	Update Panel: 
	- call from main event loop
	- for regular panel update
	- called only if panel is visible
	- front machine m is locked for reading (RLOCK/RUNLOCK)
*/
-(void)updatePanel:(Machine*)m
{
	Ula* ula = m->GetUla(); if(!ula) return;

	char s[30];
	Mmu* p7 = m->GetMmu();

	if( !p7 || !p7->HasPort7ffd() )	// 48k:
	{
		if(old_mmu!=mmu_48k) [self set_mmu_to:mmu_48k]; 
	}
	
	else if(!p7->HasPort1ffd())	// 128k, +2:
	{
		uchar new_7ffd	= p7->GetPort7ffd();
		bool  locked	= new_7ffd&lock_to_48k_mask;	
		uchar new_mmu	= locked ? mmu_plus2_48k : mmu_plus2;

		if(old_mmu!=new_mmu) [self set_mmu_to:new_mmu];
	
		SHOW_BYTE( view_7ffd, old_7ffd, new_7ffd );

		if(!locked) // not locked to 48k:
		{
			SHOW_PAGE( view_videopage, old_videopage, new_7ffd&videoram_mask?7:5 );		
			SHOW_PAGE( view_page_c000, old_page_c000, new_7ffd&7 );
			SHOW_PAGE( view_page_0000, old_page_0000, (new_7ffd>>low_rom_bit)&1 );
		}
	}

	else		// +3, +2A:
	{
		uchar new_1ffd	= p7->GetPort1ffd();
		uchar new_7ffd	= p7->GetPort7ffd();
		bool  locked	= new_7ffd&lock_to_48k_mask;	
		uchar new_mmu	= locked ? mmu_plus3_48k : mmu_plus3;

		if(old_mmu!=new_mmu) [self set_mmu_to:new_mmu];
	
		SHOW_BYTE( view_1ffd, old_1ffd, new_1ffd );
		SHOW_BYTE( view_7ffd, old_7ffd, new_7ffd );

		if(!locked) // not locked to 48k:
		{
			SHOW_BOOL( view_disc_motor,			old_disc_motor,		BOOL(new_1ffd&disc_motor_mask)		);
			SHOW_BOOL( view_printer_strobe,		old_printer_strobe, BOOL(new_1ffd&printer_strobe_mask)	);
			SHOW_PAGE( view_videopage,			old_videopage,		new_7ffd&videoram_mask?7:5			);
		
			if(new_1ffd&ram_only_mode_mask)	// ram only mode?
			{
				if(!old_ram_only_mode)
				{
					SET_CHECKBOX(checkbox_ram_only_mode,old_ram_only_mode,YES);
					[textfield_rom_0000 setStrValue:"RAM page"];
				}
				int n = (new_1ffd>>1)&3;
				char const pageC[]={3,7,3,3}; SHOW_PAGE( view_page_c000, old_page_c000, pageC[n] );
				char const page8[]={2,6,6,6}; SHOW_PAGE( view_page_8000, old_page_8000, page8[n] );
				char const page4[]={1,5,5,7}; SHOW_PAGE( view_page_4000, old_page_4000, page4[n] );
				char const page0[]={0,4,4,4}; SHOW_PAGE( view_page_0000, old_page_0000, page0[n] );
			}
			else	// not ram only mode:
			{
				if(old_ram_only_mode)
				{
					SET_CHECKBOX(checkbox_ram_only_mode,old_ram_only_mode,NO);
					[textfield_rom_0000 setStrValue:"ROM page"];
				}
			
				SHOW_PAGE( view_page_c000, old_page_c000, new_7ffd&7 );
				SHOW_PAGE( view_page_8000, old_page_8000, 2 );
				SHOW_PAGE( view_page_4000, old_page_4000, 5 );
				SHOW_PAGE( view_page_0000, old_page_0000, ((new_7ffd>>low_rom_bit)&1) + ((new_1ffd>>(high_rom_bit-1))&2) );
			}
		}
	}
	
	Time current_ula_clock = m->GetCpuCyclesPerSecond() * m->GetUla()->GetCpuClockPredivider();
	if( ula_clock != current_ula_clock && edit_view!=view_ula_clock ) 
	{
		ula_clock  = current_ula_clock;
		[view_ula_clock setStrValue:MHzStr(ula_clock)];	
	}

	if( ula_clock_percent != current_ula_clock && edit_view!=view_ula_clock_percent ) 
	{ 
		ula_clock_percent  = current_ula_clock;			// actually stores ula clock
		sprintf( s, "%.3g%%", current_ula_clock / m->GetInfo()->ula_cycles_per_second * 100.0 );
		[view_ula_clock_percent setStrValue:s];	
	}

	if( cpu_predivider != ula->GetCpuClockPredivider() && edit_view!=view_cpu_clock_divider ) 
	{
		cpu_predivider  = ula->GetCpuClockPredivider(); 
		sprintf(s,"1/%i",cpu_predivider);
		[view_cpu_clock_divider   setStrValue:s];	
	} 
	
	if( cpu_clock != m->GetCpuCyclesPerSample() && edit_view!=view_cpu_clock ) 
	{
		cpu_clock  = m->GetCpuCyclesPerSample(); 
		[view_cpu_clock setStrValue:MHzStr(cpu_clock*samples_per_second)];	
	}
	 
	if( lines_before_screen != ula->GetLinesBeforeScreen() && edit_view!=view_lines_before_screen ) 
	{
		lines_before_screen  = ula->GetLinesBeforeScreen();
		[view_lines_before_screen setIntValue:lines_before_screen];	
	}

	if( lines_in_screen != ula->GetLinesInScreen() && edit_view!=view_lines_in_screen ) 
	{
		lines_in_screen  = ula->GetLinesInScreen();
		[view_lines_in_screen setIntValue:lines_in_screen];	
	}

	if( lines_after_screen != ula->GetLinesAfterScreen() && edit_view!=view_lines_after_screen ) 
	{
		lines_after_screen  = ula->GetLinesAfterScreen();
		[view_lines_after_screen setIntValue:lines_after_screen];	
	}

	if( pixel_per_screen != 256 && edit_view!=view_screen_pixels_per_line ) 
	{
		pixel_per_screen  = 256;
		[view_screen_pixels_per_line setIntValue:pixel_per_screen];	
	}

	if( cpu_clocks_per_line != ula->GetCpuCyclesPerLine() && edit_view!=view_cpu_clocks_per_line )
	{
		cpu_clocks_per_line  = ula->GetCpuCyclesPerLine();
		[view_cpu_clocks_per_line setIntValue:cpu_clocks_per_line];	
	}

	if( cpu_cycles_per_frame != ula->GetCpuCyclesPerFrame() && edit_view!=view_cpu_cycles_per_frame )
	{
		cpu_cycles_per_frame  = ula->GetCpuCyclesPerFrame();
		[view_cpu_cycles_per_frame setIntValue:cpu_cycles_per_frame];	
	}

	if( bytes_per_line != ula->GetUlaCyclesPerLine()/16 )
	{
		bytes_per_line = ula->GetUlaCyclesPerLine()/16;
		[view_bytes_per_line setIntValue:bytes_per_line];	
	}

	if( frames_per_sec != m->GetCpuCyclesPerSample() * samples_per_second / ula->GetCpuCyclesPerFrame() && edit_view!=view_frames_per_second )
	{
		frames_per_sec  = m->GetCpuCyclesPerSample() * samples_per_second / ula->GetCpuCyclesPerFrame();
		sprintf(s,"%.4g",frames_per_sec);
		[view_frames_per_second setStrValue:s];	
	}

	if( waitstates_on != (m->GetCpuOptions()&cpu_waitmap) )
	{
		waitstates_on = (m->GetCpuOptions()&cpu_waitmap);
		[checkbox_use_waitmap setState:waitstates_on];
	}
	
	if( show_inner_screen != [m->GetScreen() drawsInnerScreen] )
	{
		show_inner_screen = !show_inner_screen;
		[checkbox_show_inner_screen setState:show_inner_screen];
	}
	
	if( waitstates_offset != 0 && edit_view!=view_waitmap_offset )
	{
		waitstates_offset = 0;
		sprintf(s,"%i",waitstates_offset);
		[view_waitmap_offset setStrValue:s];	
	}

	if( border_color != ula->GetBorderColor() && edit_view!=view_border )
	{
		if( border_color==(int)XXX ) [view_border setStrValue:""];
		border_color  = ula->GetBorderColor();
		[view_border setBackgroundColor:[NSColor colorWithDeviceRed:(border_color&2)*0.8/2 green:(border_color&4)*0.8/4 blue:(border_color&1)*0.8/1 alpha:1.0]];
	}
	
	if( mic_out_state != ula->GetMicOutState() && edit_view!=view_mic ) 
	{
		mic_out_state  = ula->GetMicOutState();
		[view_mic setStrValue:mic_out_state?"1":"0"];	
	}
	
	if( ear_out_state != ula->GetEarOutState() && edit_view!=view_ear ) 
	{
		ear_out_state  = ula->GetEarOutState();
		[view_ear setStrValue:ear_out_state?"1":"0"];	
	}

	if( m->IsHalted() || frames_per_sec <= 1.0 )		
	{
		if( crt_row != ula->GetCrtRowForCpuCycle(m->GetCpu()->CpuCycle()) && edit_view!=view_crt_current_row ) 
		{
			crt_row  = ula->GetCrtRowForCpuCycle(m->GetCpu()->CpuCycle());
			[view_crt_current_row setIntValue:crt_row];	
		}

		if( crt_col != ula->GetCrtColForCpuCycle(m->GetCpu()->CpuCycle()) && edit_view!=view_crt_current_col ) 
		{
			crt_col  = ula->GetCrtColForCpuCycle(m->GetCpu()->CpuCycle());
			[view_crt_current_col setIntValue:crt_col];	
		}
		if( crt_in_screen != ( ulong(crt_row-lines_before_screen) < ulong(lines_in_screen) && crt_col < 32 ) ) 
		{
			crt_in_screen  = ( ulong(crt_row-lines_before_screen) < ulong(lines_in_screen) && crt_col < 32 );
			[view_crt_is_in_screen setStrValue:crt_in_screen?"✔":"-"];	
		}
	}
	else
	{
		if( crt_row!=99999 )
		{
			if( edit_view!=view_crt_current_row ) [view_crt_current_row  setStrValue:"---"]; crt_row = 99999;
			if( edit_view!=view_crt_current_col ) [view_crt_current_col  setStrValue:"---"]; crt_col = 99999;
												  [view_crt_is_in_screen setStrValue:"-"  ]; crt_in_screen = 99999;
		}
	}
	
	if(frames_hit!=frames_hit_count)
	{
		frames_hit = frames_hit_count;
		sprintf(s,"%i%%",frames_hit);
		[view_frames_hit setStrValue:s];	
	}
	
//	char	waitstates_map[8]		= {9,9,9,9,9,9,9,9};		TODO
}



@end
