/*	Copyright  (c)	Günter Woigk 2006 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define	LOG		1

#include <unistd.h>

#import "cstrings/cstrings.h"
#import "unix/file_utilities.h"
#import	"ZxPlus2JoystickPorts.h"
#import	"ZxPlus2JoystickPortsPanel.h"
#import "Machine.h"
#import "Z80.h"
#import "Mmu.h"


/*
	sinclair1:
	data byte:	%000LRDUF active low
	address:	0xEFFE
				%---0.----.----.---0

	sinclair2:
	data byte:	%000FUDRL active low
	address:	0xF7FE
				%----.0---.----.---0
				
	Gemischte Abfrage mit addr = 0xE7FE ?		TODO
	Wert der oberen 3 Datenbits noch nicht geklärt.		TODO
*/


#define o_addr	NULL
#define	i_addr	"----.----.----.---0"



ZxPlus2JoystickPorts::ZxPlus2JoystickPorts ( Machine* m )
:	ZxItem( m, "built-in joystick ports",o_addr,i_addr)
{
	XLogIn("new ZxPlus2JoystickPorts");

	FindJoysticks ( );
	// kbd_joystick nur explizit einstecken, wg. "some keys do not react" desaster!
	joystick1 = IsPluggedIn(hid_joystick1) ? hid_joystick1 : no_joystick; // not kbd_joystick!
	joystick2 = IsPluggedIn(hid_joystick2) ? hid_joystick2 : no_joystick;

	m->zxplus2_js_ports = this;

	[zxplus2_js_ports_panel createWindowMenuItem];	// internal item: just create window menu entry
} 


ZxPlus2JoystickPorts::~ZxPlus2JoystickPorts ( )
{
	GetMachine()->zxplus2_js_ports = NULL;
}


void ZxPlus2JoystickPorts::Input ( Time/*t*/, long cc, ushort address, uchar& byte, uchar& mask )
{
	uchar rval  = 0;

	if( (address&0x1801)==0x0800 )		// Sinclair 1: %000LRDUF active low
	{
		if( GetMachine() == front_machine() ) rval = GetState(joystick1);	//	%000FUDLR active high
		rval = ((rval&0x10)>>4) | ((rval&0x08)>>2) | ((rval&0x04)) | ((rval&0x03)<<3);
		mask |= rval;					// only pull down.
		byte &= ~rval;
	}
	else if( (address&0x1801)==0x1000 )	// Sinclair 2: %000FUDRL active low
	{
		if( GetMachine() == front_machine() ) rval = GetState(joystick2);	//	%000FUDLR active high 
		rval = (rval&0x1c) | ((rval&0x02)>>1) | ((rval&0x01)<<1);
		mask |= rval;					// only pull down.
		byte &= ~rval;
	}
}










