/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define LOG 	1

#import "cstrings/cstrings.h"
#import "KeyboardJoystickPanel.h"
#import "ZxPlus2JoystickPortsPanel.h"
#import "NXTextField.h"
#import "NXImage.h"
#import "Machine.h"
#import "ZxPlus2JoystickPorts.h"





@implementation ZxPlus2JoystickPortsPanel


-(id)initWithContentRect:(NSRect)contentRect 
			   styleMask:(uint)styleMask 
			     backing:(NSBackingStoreType)backingType 
				   defer:(BOOL)flag
{
	XLogIn("ZxPlus2JoystickPortsPanel:initWithContentRect");
	
	return zxplus2_js_ports_panel = self =
		[super initWithContentRect:contentRect styleMask:styleMask 
						backing:backingType defer:flag];
}



//- (void)awakeFromNib { XLogIn("ZxPlus2JoystickPortsPanel:awakeFromNib"); }


- (IBAction)selectJoystick1:(id)sender
{
	QBLOCK;
		ZxPlus2JoystickPorts* js_ports = m->zxplus2_js_ports; 
		if( js_ports ) { js_ports->SetJoystick1((JoystickID)[sender indexOfSelectedItem]); }
	QUNBLOCK;
}


- (IBAction)selectJoystick2:(id)sender
{
	QBLOCK;
		ZxPlus2JoystickPorts* js_ports = m->zxplus2_js_ports; 
		if( js_ports ) { js_ports->SetJoystick2((JoystickID)[sender indexOfSelectedItem]); }
	QUNBLOCK;
}


- (IBAction)rescanDevices:(id)sender
{
	XLogIn("ZxPlus2JoystickPortsPanel:rescanDevices");
	FindJoysticks( 1 /*force*/ );
}


- (IBAction)setKeys:(id)sender
{
	XLogIn("ZxPlus2JoystickPortsPanel:setKeys");
	[keyboard_joystick_panel makeKeyAndOrderFront:self];
}


- (IBAction)doReset:(id)sender
{
	XLogIn("ZxPlus2JoystickPortsPanel:doReset");
	QBLOCK;
		m->Reset();
	QUNBLOCK;
}


/* ----------------------------------------------------------
	Update Panel: 
	Called by NXPanel:updateAllPanels() from main event loop:
	- for regular panel update
	- called only if panel is visible
	- front machine m is locked for reading (RLOCK/RUNLOCK)
*/
-(void)updatePanel:(Machine*)m
{
	ZxPlus2JoystickPorts* js_ports = m->zxplus2_js_ports; 

	static void* old_js_ports = ptr(999/*force redraw after showPanel*/);
	if( old_js_ports != js_ports )
	{
		old_js_ports = js_ports;
		[self enableSubviews:js_ports!=NULL];
	}

	if( !js_ports ) return;
	
	static int old_model = zxplus2;		// image in panel is for zx +2 in NIB file
	int model = m->GetModel();
		
	if( model!=old_model )
	{
		old_model = model;

		static NSImage* zxplus2_image  = nil;	if( !zxplus2_image ) zxplus2_image = [imageView image]; 
		static NSImage* zxplus3_image  = nil;
		static NSImage* zxplus2B_image = nil;
		cstr path;
		NSImage** image;

		switch(model)
		{
		case zxplus2_span:			// TODO: side view image of spanish zx +2
		case zxplus2_french:		// TODO: side view image of french zx +2
		case zxplus2:	path = "zxplus2_sideview.jpg";	image = &zxplus2_image;  break;
		case zxplus3_span:			// TODO: side view image of spanish zx +3
		case zxplus3:	path = "zxplus3_sideview.jpg";	image = &zxplus3_image;  break;
		case zxplus2a_span:			// TODO: side view image of spanish zx +2A
		case zxplus2a:	path = "zxplus2B_sideview.jpg";	image = &zxplus2B_image; break;
		default:		IERR();
		}
				
		if( !*image ) { *image = [[NXImage alloc] initWithContentsOfResourceFile:path]; }
		if( *image )  { [imageView setImage:*image]; }
	}

	static uchar old_state1 = 0xff;
	uchar state = GetState(js_ports->GetJoystick1());	//	%000FUDLR active high

	if( state != old_state1 )
	{
		old_state1 = state;
		state = ((state&0x10)>>4) | ((state&0x08)>>2) | ((state&0x04)) | ((state&0x03)<<3);
		char s[]="%000LRDUF";	// sinclair1
		for(uint i=1;i<=8;i++) { if( (state&(0x100>>i))==0 ) s[i]='-'; };
		[joystick1ButtonsView setStrValue:s];
	}
  
	static uchar old_state2 = 0xff;
	state = GetState(js_ports->GetJoystick2());			//	%000FUDLR active high 

	if( state != old_state2 )
	{
		old_state2 = state;
		state = (state&0x1c) | ((state&0x02)>>1) | ((state&0x01)<<1);
		char s[]="%000FUDRL";	// sinclair2
		for(uint i=1;i<=8;i++) { if( (state&(0x100>>i))==0 ) s[i]='-'; };
		[joystick2ButtonsView setStrValue:s];
	}

	if( [joystick1PopupMenu indexOfSelectedItem] != int(js_ports->GetJoystick1()) )
	{
		[joystick1PopupMenu selectItemAtIndex:js_ports->GetJoystick1()];
	}

	if( [joystick2PopupMenu indexOfSelectedItem] != int(js_ports->GetJoystick2()) )
	{
		[joystick2PopupMenu selectItemAtIndex:js_ports->GetJoystick2()];
	}

	static bool old_js1 = yes;
	if( IsPluggedIn(hid_joystick1) != old_js1 ) 
	{
		old_js1 = !old_js1;
		[[joystick1PopupMenu itemAtIndex:0] setEnabled:old_js1];
		[[joystick2PopupMenu itemAtIndex:0] setEnabled:old_js1];
	} 
		
	static bool old_js2 = yes;
	if( IsPluggedIn(hid_joystick2) != old_js2 ) 
	{
		old_js2 = !old_js2;
		[[joystick1PopupMenu itemAtIndex:1] setEnabled:old_js2];
		[[joystick2PopupMenu itemAtIndex:1] setEnabled:old_js2];
	} 	  
}




@end
