/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define LOG		1

#include "cstrings/cstrings.h"

#import "CpuPanel.h"
#import "KbdPanel.h"
#import "MenuController.h"
#import "NXTextField.h"
#import "Machine.h"
#import "Z80.h"
#import "Z80 opcodes.h"
#import "utilities.h"
#import "ZxInfo.h"





// displayed values:
#define XXX 0xffffffff
static Time		XTAL=XXX; 
static ulong 	CC=XXX;
static ulong	bc=XXX,de=XXX,hl=XXX,bc2=XXX,de2=XXX,hl2=XXX,ix=XXX,iy=XXX,pc=XXX,sp=XXX;
static ushort 	a=XXX,f=XXX,a2=XXX,f2=XXX,f_vis=XXX,i=XXX,IFF1=XXX,IM=XXX,NMI=XXX,IRPT=XXX,r=XXX;


// breakpoints:
static int breakpoints = XXX;
typedef char c20 [20];
static c20 break_text[100]; 
//static char break_delim;			?
static ushort* break_ptr;


// memory view:
#define mview_rows 21
#define mview_cols 24
typedef char	c25 [25];
static ushort	mview_on=XXX; 
static struct 
{
	char	display_mode;						// 0/1/2 == z80/dec/hex
	char	reg_nr;								// z80.registers.rr.nn[reg_nr]
	ushort	reg_val;							// value of register (-> show current line and force inside view if changed)
	ushort	data_base;							// base address of displayed data
	ushort	data_count;							// bytes imposed in display
	uchar	data[mview_rows*16];				// data. as displayed:	20 lines à 16 bytes max.
	c25		text[mview_rows];					// displayed text:		20 lines à 24+1 bytes max.
	char	delim;								// 0
} mview;
static char marker[2] = {' ','>'};


static ushort HexDumpLine( Z80*, int line, ushort addr, ushort reg )
{
//	1234  12345678.12345678 \n	
	long a = Peek4X(mview.data+line*8), b = Peek4X(mview.data+line*8+4);
	sprintf( mview.text[line], "%04hX%c %08X.%08X \n", addr, marker[addr==reg], a,b );
	return addr+8;
}

static ushort ByteDumpLine( Z80*, int line, ushort addr, ushort reg )
{
//	1234  123, 123, 123, 123\n
	uptr p = mview.data+line*4;
	sprintf( mview.text[line], "%04hX%c % 3hhu, % 3hhu, % 3hhu, % 3hhu\n", addr, marker[addr==reg], p[0],p[1],p[2],p[3] );
	return addr+4;
}

static ushort WordDumpLine( Z80*, int line, ushort addr, ushort reg )
{
//	1234  -12345, -12345    \n
	ushort a = Peek2Z(mview.data+line*4), b = Peek2Z(mview.data+line*4+2);
	sprintf( mview.text[line], "%04hX%c % 6hi, % 6hi    \n", addr, marker[addr==reg], a,b );
	return addr+4;
}

static ushort DisassLine( Z80* cpu, int line, ushort addr, ushort reg )
{
//	1234  ld   a,-123(ix)   \n
	char* p = mview.text[line];
	int n = sprintf( p, "%04hX%c ", addr, marker[addr==reg] );
	n += sprintf( p+n, "%s", cpu->Disassemble(addr) );
	strcpy( mview.text[line]+n, "                \n"+(16-24+n) );
	return addr;
}

static ushort TextDumpLine( Z80*, int line, ushort addr, ushort reg )
{
//	1234  "abcd"+$80...
	char* p = mview.text[line];
	int n = sprintf( p, "%04hX%c ", addr, marker[addr==reg] );
	uchar* data = mview.data;

	for(;;)
	{
		uchar c = data[addr-mview.data_base];
		if( ((c+1)&0x7f)<=' ' )		// c<=32 || c==127 || c<=32+128 || c==127+128
		{
			if(mview_cols-n<3) break;
			n += sprintf( p+n, "$%02hhX,", c );				// $12,
			addr++;
		}
		else						// printable oder printable+128
		{
			if( (mview_cols-n<3) || ((c>>7)&&(mview_cols-n<7)) ) break;		// "a",  oder  "a"+$80,
			p[n++] = '"';
			for(;;)
			{
				p[n++] = c&0x7f; addr++; if(c>>7) break;	// c hatte bit7 gesetzt -> string beenden weil ''+$80''
				if(addr==reg) break;						// raus damit wir den marker vor die nächste zeile setzen können
				uchar d = data[addr+mview.data_base];		// d = nächstes Zeichen
				if( ((d+1)&0x7f)<=' ' ) break;				// non-ascii
				if( (mview_cols-n<2) || ((d>>7)&&(mview_cols-n<6)) ) break;	// platz langt nicht
				c = d;										// ok -> anhängen
			}
			p[n++]='"';  if(c>>7) { strcpy(p+n,"+$80"); n+=4; } 
			p[n++]=',';
		}
	}
	n--; while(n<mview_cols) p[n++]=' '; p[n]='\n';	
	return addr;
}



static str FlagsStr( uchar f )
{
	static char s[9]; strcpy(s,"SZ1H1VNC");
	for( int i=0;i<8;i++ ) if( (signed char)(f<<i)>=0 ) s[i]='-'; 
	return s;
}


static ulong IntValue( id sender )
{
	cstr s = [[sender stringValue]UTF8String]; if (s[0]==0) return XXX;
	ulong v = (ulong)StrToDouble(s);
	return v & 0x7fffffff;
}


static double Value( id sender )
{
	cstr s = [[sender stringValue]UTF8String]; if (s[0]==0) return XXX;
	s = LowerStr(s);
	double f = 1.0;
	ptr p;
	if( (p=FindStr(s,"mhz")) ) { *p=0; f=1e6; } else		// 12.34 MHz
	if( (p=FindStr(s,"khz")) ) { *p=0; f=1e3; } else		// 12.34 kHz
	if( (p=FindStr(s,"hz")) )  { *p=0; }
	
	double v = StrToDouble(s,&s); while(*s==' ') s++; if(*s!=0) return XXX;
//	XLogLine("value = %g",v*f);
	return v * f;
}



@implementation CpuPanel


// NSTextView delegate methods:
//-(BOOL)control:(NSControl *)control textShouldEndEditing:(NSText *)fieldEditor{ return yes; }
//-(BOOL)control:(NSControl *)control textShouldBeginEditing:(NSText*)fieldEditor{ return yes; }
//-(void)controlTextDidChange:(NSNotification*)note { }

// standard stuff:
//-(void)windowWillClose:(NSNotification *)aNotification {}
//-(void)close{}
//-(void)awakeFromNib {}

-(id)initWithContentRect:(NSRect)contentRect styleMask:(uint)styleMask backing:(NSBackingStoreType)backingType defer:(BOOL)flag
{
	XLogIn("CpuPanel:initWithContentRect");
	return cpu_panel = [super initWithContentRect:contentRect styleMask:styleMask backing:backingType defer:flag];
}






#define SHOW_R(R,V)		if (z80.r.R!=R    && edit_view!=V) { R=z80.r.R;    sprintf(s,"$%02hx",R); [V setStrValue:s]; }
#define SHOW_RR(RR,V)	if (z80.rr.RR!=RR && edit_view!=V) { RR=z80.rr.RR; sprintf(s,"$%04x",RR); [V setStrValue:s]; }

#define	SET_RR(RR)	ulong v = IntValue(sender); if(v!=XXX) { QBLOCK; m->GetCpu()->GetRegisters().rr.RR = v; QUNBLOCK; } RR=XXX;	
// note: was BLOCK/UNBLOCK, but an IBAction can only me called from the main event loop
#define	SET_R(R)	ulong v = IntValue(sender); if(v!=XXX) { QBLOCK; m->GetCpu()->GetRegisters().r.R   = v; QUNBLOCK; } R=XXX;	
// note: was BLOCK/UNBLOCK, but an IBAction can only me called from the main event loop


-(void)updateBreakpointView:(Machine*)m
{
	XLogIn("CpuPanel:updateBreakpoints");
	
	TRAP( NELEM(m->breakpoint) != NELEM(break_text) );
	
	breakpoints = m->breakpoints;
	break_ptr   = m->break_ptr;
	break_text[0][0]=0;

	for( int i=0; i<breakpoints; i++ )
	{
		ushort* bp = m->breakpoint[i];
		ushort flags = *bp;

		char rwx[] = "---  "; 
		if(flags&cpu_break_r) rwx[0]='r';			// break for reading
		if(flags&cpu_break_w) rwx[1]='w';			// break for writing
		if(flags&cpu_break_x) rwx[2]='x';			// break for execute
		if(bp==break_ptr)     rwx[4]='*';			// 'bingo!'

		//bool rom = bp < m->GetRam();
		Mem& ram = m->GetRam();											// MEM
		bool rom = bp < ram.Data() || bp >= ram.Data()+ram.Count();		// MEM		TODO: external rom
		int  pgsz = m->GetInfo()->page_size;
		char bank = '0' + (bp-(rom?m->GetRom().Data()/*MEM*/:m->GetRam().Data()/*MEM*/)) / pgsz; if(bank>'9') bank+=7;	// page within rom or page within ram
		int  addr = (bp-(rom?m->GetRom().Data()/*MEM*/:m->GetRam().Data()/*MEM*/)) & (pgsz-1);		// address within page

		addr += 0x10000;	// indicate 'not found'
		if( flags&(cpu_break_r|cpu_break_x) ) 
		{
			for( addr-=0x10000; addr<0x10000; addr+=pgsz ) { if( m->GetCpu()->RdPtr(addr)==bp ) break; }
		}
		if( addr>=0x10000 && (flags&cpu_break_w) ) 
		{
			for( addr-=0x10000; addr<0x10000; addr+=pgsz ) { if( m->GetCpu()->WrPtr(addr)==bp ) break; }
		}
		if( addr>=0x10000 )
		{
			if(rom) addr -= 0x10000;	// dflt addr base: rom: 0x0000
			else    addr -= pgsz;		// dflt addr base: ram: 0xc000 ((last page))
		}		

		sprintf( break_text[i], "%s#%c: $%04x: %s\n", rom?"rom":"ram", bank, addr, rwx );
	}
	
	[view_breakpoints setStrValue:break_text[0]];
}


/*	Update Panel: 
	- call from main event loop
	- for regular panel update
	- called only if panel is visible
	- front machine m is locked for reading (RLOCK/RUNLOCK)
*/
-(void)updatePanel:(Machine*)m
{	
	Z80* cpu = m->GetCpu();	if(!cpu) return;
	Z80Regs& z80 = cpu->GetRegisters();

	{	if( m->IsHalted() != [btn_halt state] ) [btn_halt setNextState];
		static bool old = 1;
		if( m->IsHalted() != old ) 
		{ 
			old = !old; 
			[btn_step_in setEnabled:old]; 
			[btn_step setEnabled:old]; 
			[btn_step_out setEnabled:old]; 
			[view_cc setEnabled:old];
		}
	}

	char s[12];

	SHOW_R(a, view_a);
	SHOW_R(a2,view_a2);
	SHOW_R(i, view_i);
	SHOW_R(r, view_r);
	SHOW_R(f, view_f);
	SHOW_R(f2, view_f2);

	if( z80.r.f!=f_vis && edit_view!=view_f_visual ) { f_vis=z80.r.f; [view_f_visual setStrValue:FlagsStr(f_vis)]; }
	
	SHOW_RR(bc, view_bc);
	SHOW_RR(de, view_de);
	SHOW_RR(hl, view_hl);
	SHOW_RR(bc2, view_bc2);
	SHOW_RR(de2, view_de2);
	SHOW_RR(hl2, view_hl2);
	SHOW_RR(ix, view_ix);
	SHOW_RR(iy, view_iy);
	SHOW_RR(pc, view_pc);
	SHOW_RR(sp, view_sp);

	if (z80.r.im!=IM && edit_view!=view_im)					{ IM   = z80.r.im; sprintf(s,"%hi",IM); [view_im setStrValue:s]; }	
	if (z80.r.iff1!=IFF1)									{ IFF1 = z80.r.iff1; [btn_iff1 setState:IFF1]; }
	if (z80.r.nmi!=NMI)										{ NMI  = z80.r.nmi;  [btn_nmi  highlight:NMI]; }
	if ( (cpu->InterruptSignal()>cpu->CpuCycle()) != IRPT )	{ IRPT = cpu->InterruptSignal()>cpu->CpuCycle(); [btn_irpt highlight:IRPT]; }
	if (cpu->CpuCycle()!=(long)CC && edit_view!=view_cc)			{ CC   = cpu->CpuCycle(); [view_cc setIntValue:CC]; }

	if (XTAL!=m->GetCpuCyclesPerSample() && edit_view!=view_xtal)					
	{
		XTAL = m->GetCpuCyclesPerSample();    
		[view_xtal setStrValue:MHzStr(XTAL*samples_per_second)]; 
	}

// breakpoint view:
	if( ( breakpoints!=m->breakpoints || m->break_ptr!=break_ptr ) && edit_view!=view_breakpoints )
	{		
		[self updateBreakpointView:m];
	}

// memory view:
	if( m->IsHalted() || m->GetCpuCyclesPerSample()*samples_per_second < 100 )
	{
		bool f = mview_on!=1; mview_on = 1; 

		if( mview.display_mode != [sel_display_mode indexOfSelectedItem] )		// darstellungsmodus wurde geändert
		{ 
			f=1; mview.display_mode = [sel_display_mode indexOfSelectedItem]; 
		} 

		if( mview.reg_nr != [[sel_register selectedCell] tag] )				// ausgewähltes register wurde geändert
		{ 
			f=1; if(mview.reg_nr==10/*pc*/&&mview.display_mode==0/*disass*/) mview.display_mode=1/*hex*/; 
			mview.reg_nr = [[sel_register selectedCell] tag]; 
			if(mview.reg_nr==10/*pc*/) mview.display_mode=0/*disass*/; 
			mview.data_base = mview.reg_val = z80.nn[mview.reg_nr];
			[sel_display_mode selectItemAtIndex:mview.display_mode];
		} 
		ushort& reg = z80.nn[mview.reg_nr];

		if( mview.reg_val != reg )											// wert des registers hat sich geändert
		{
			f=1; mview.reg_val = reg;
			const short min_free[]={4,16,4,4,16}; /*disass,hex,bytes,words,ascii*/
			if( short(reg-mview.data_base)<0 || short(mview.data_base+mview.data_count-reg)<=min_free[mview.display_mode] ) 
				mview.data_base = reg; 
		}

		// copy current memory data to mview.data[] 
		// take care for paged memory
		// last chance to set 'dirty' flag f
		{
			const short max_size[]={4*mview_rows,8*mview_rows,4*mview_rows,4*mview_rows,16*mview_rows}; /*disass,hex,bytes,words,ascii*/
			short max = max_size[mview.display_mode]; if(mview.data_count>max) mview.data_count=max;

			ushort* q = cpu->RdPtr(mview.data_base);
			uptr    z = mview.data;

			short i = 0;
			short e = mview.data_base+max;
				  if( (e>>10) != (mview.data_base>>10) ) e &= -1<<10;	// Potentielle Page-Grenze ((min. page size = 1 kB assumed))
				  e = e - mview.data_base - 1;							// innere Grenze damit 0x0000 wie 0x10000 funktioniert

			if(!f) { for(;i<=e;i++) { if( (uchar)q[i]!=z[i] ) break; } f = i<=e && i<mview.data_count; }			
			while(i<=e) { z[i] = q[i]; i++; } 

			q = cpu->RdPtr(mview.data_base+i) -i;
			e = max-1;
			
			if(!f) { for(;i<=e;i++) { if( (uchar)q[i]!=z[i] ) break; } f = i<=e && i<mview.data_count; }
			while(i<=e) { z[i] = q[i]; i++; }
		}
		
		// if required redraw view
		if(f)
		{
			ushort(*fu)(Z80*,int,ushort,ushort);
			switch(mview.display_mode)
			{
			case 0:	fu=DisassLine; break;
			case 1:	fu=HexDumpLine;  break;
			case 2:	fu=ByteDumpLine; break;
			case 3:	fu=WordDumpLine; break;
			case 4:	fu=TextDumpLine; break;
			}
			ushort addr = mview.data_base, reg = mview.reg_val; 
			for( int i=0; i<mview_rows; i++ ) addr = (*fu)( cpu, i, addr, reg ); 
			mview.data_count = addr-mview.data_base; 
			mview.text[mview_rows-1][mview_cols] = 0;

			[view_memory setStrValue:mview.text[0]]; 			
		}
	}
	else
	{
		if(mview_on!=0) { mview_on = 0; [view_memory setStrValue:""]; }
	}
}




// ===============================================================================
//								ACTIONS
// ===============================================================================


-(IBAction)doStepIn:(id)sender				{ QBLOCK; m->StepIn(); QUNBLOCK; }
-(IBAction)doStepOver:(id)sender			{ QBLOCK; m->StepOver(); QUNBLOCK; }
-(IBAction)doStepOut:(id)sender				{ QBLOCK; m->StepOut(); QUNBLOCK; }
-(IBAction)toggleHaltCpu:(id)sender			{ QBLOCK; m->HaltCpu([sender state]); QUNBLOCK; }


// set registers

// note: r, cc, af, pc, sp are stored in registers while the Z80 is running.
// But because the Z80 is running on an interrupt (sound output)
// it is not running, when we come here.
// it may interrupt our execution, but setting of the register is atomic.
// TODO: on a dual cpu machine the Z80 may actually run while we set the register
// and here setting the register in the Z80 instance body won't have any effect,
// because they would not set the register variable. 
// But even this can only happen, while the cpu is running. 
// And setting registers is only real useful, when it is halted. ;-)

- (IBAction)set_bc2:(id)sender	{ SET_RR(bc2); }
- (IBAction)set_bc:(id)sender	{ SET_RR(bc); }
- (IBAction)set_de2:(id)sender	{ SET_RR(de2); }
- (IBAction)set_de:(id)sender	{ SET_RR(de); }
- (IBAction)set_hl2:(id)sender	{ SET_RR(hl2); }
- (IBAction)set_hl:(id)sender	{ SET_RR(hl); }
- (IBAction)set_iy:(id)sender	{ SET_RR(iy); }
- (IBAction)set_ix:(id)sender	{ SET_RR(ix); }
- (IBAction)set_sp:(id)sender	{ SET_RR(sp); }
- (IBAction)set_pc:(id)sender	{ SET_RR(pc); }

- (IBAction)set_r:(id)sender	{ SET_R(r); }
- (IBAction)set_i:(id)sender	{ SET_R(i); }
- (IBAction)set_f:(id)sender	{ SET_R(f); }
- (IBAction)set_f2:(id)sender	{ SET_R(f2); }
- (IBAction)set_a:(id)sender	{ SET_R(a); }
- (IBAction)set_a2:(id)sender	{ SET_R(a2); }


- (IBAction)set_f_visual:(id)sender		// TODO: set f visual
{
	char* s = UpperStr([[sender stringValue]UTF8String]); 
	f=XXX; if(s[0]==0) return;
	uchar new_flags = 0;

	while( *s )
	{
		switch(*s++)
		{
		case 'V':	new_flags|=V_FLAG; break;		// parity / overflow
		case 'S':	new_flags|=S_FLAG; break;
		case 'Z':	new_flags|=Z_FLAG; break;
		case 'H':	new_flags|=H_FLAG; break;
		case 'N':	new_flags|=N_FLAG; break;
		case 'C':	new_flags|=C_FLAG; break;
		}
	}

	QBLOCK;		// note: was BLOCK/UNBLOCK, but an IBAction can only me called from the main event loop
	m->GetCpu()->GetRegisters().r.f = new_flags;
	QUNBLOCK;
}


- (IBAction)set_cc:(id)sender
{
	ulong v = IntValue(sender); 
	CC=XXX; if(v==XXX) return;
	
	QBLOCK;
		if(m->IsHalted())
		{			
			ulong cc_ffb = m->GetUla()->GetCpuCyclesPerFrame();
			if(v>cc_ffb) v = cc_ffb;
			v = ( v - m->GetCpu()->CpuCycle() + cc_ffb ) % cc_ffb;
			if(v) m->RunCpuCycles(v);
		}
		CC=XXX; 
	QUNBLOCK;
}


- (IBAction)set_xtal:(id)sender
{
	Time v = Value(sender); 
	if(v!=XXX&&v>0)
	{
		QBLOCK;
			m->SetCpuCyclesPerSecond(v);
		QUNBLOCK;
	}
	XTAL = XXX;
}


- (IBAction)toggleIff1:(id)sender
{
	QBLOCK;
	m->GetCpu()->GetRegisters().rr.iff = [sender state] ? 0x0101 : 0x0000;
	QUNBLOCK;
	IFF1=XXX;
}


- (IBAction)toggleNmi:(id)sender
{
	QBLOCK;
	m->GetCpu()->GetRegisters().r.nmi ^= 0x01; 
	QUNBLOCK;
}


- (IBAction)toggleIrpt:(id)sender
{
	QBLOCK;			// note: was BLOCK/UNBLOCK, but an IBAction can only me called from the main event loop
	Z80* cpu = m->GetCpu(); cpu->InterruptSignal() = cpu->InterruptSignal()>cpu->CpuCycle() ? 0 : cpu->CpuCycle()+4;
	QUNBLOCK;
}


- (IBAction)set_im:(id)sender
{
	int v = IntValue(sender); 
	if(v>=0&&v<=2) 
	{
		QBLOCK;
		m->GetCpu()->GetRegisters().r.im = v;
		QUNBLOCK;
	}
	IM=XXX;
}


- (IBAction)sel_displaymode:(id)sender
{
	XLogIn("CpuPanel: sel_displaymode");	// wird in updatePanel direkt vom popup erfragt
}


- (IBAction)sel_register:(id)sender
{
	XLogIn("CpuPanel: sel_register");		// wird in updatePanel direkt vom radio button array erfragt
}


-(IBAction)setBreakpoints:(id)sender
{ 
	XLogIn("CpuPanel:setBreakpoints"); 
	
//	"rom#n nnnn rwx"

	char* s = UpperStr([[sender stringValue]UTF8String]); 
	breakpoints=XXX; if(s[0]==0) return;

	QBLOCK;			// note: was BLOCK/UNBLOCK, but an IBAction can only me called from the main event loop
		m->ClearAllBreakpoints();

		for(;;)
		{
			int addr=-1, bank=-1, flags = 0;
			bool rom=0,ram=0;

		a:	char c = *s++;
		b:	switch(c)
			{
			case 0:		
			case '\n':	break;

			case 'A':	ram=1; goto a;
			case 'O':	rom=1; goto a;
			case 'R':	if(addr>=0) flags |= cpu_break_r; goto a;
			case 'W':	if(addr>=0) flags |= cpu_break_w; goto a;
			case 'X':	if(addr>=0) flags |= cpu_break_x; goto a;

			case '#':	
				bank = _digit_value(*s++);
				if( bank<=9 && _is_dec_digit(*s) ) { bank=bank*10+_digit_val(*s++); }
				goto a;
				
			case '$':
				addr=0; c=*s++; 
				while(_is_hex_digit(c)) { addr=addr*16+_digit_value(c); c=*s++; }
				goto b;

			default:	
				if(!_is_dec_digit(c)) goto a;		// ignore anything else
				addr=0; while(_is_dec_digit(c)) { addr=addr*10+_digit_val(c); c=*s++; }
				goto b;
			}

			if( !flags ) flags = cpu_break_x;

			if( addr>=0 )
			{
				if( bank>=0 )
				{
					if(rom!=ram)
					{
						addr &= (m->GetInfo()->page_size-1); 
						addr += bank * m->GetInfo()->page_size; 

						if(rom) { if( bank<m->GetInfo()->rom_pages ) m->SetBreakpoint ( &m->GetRom()[addr], flags); }
						else	{ if( bank<m->GetInfo()->ram_pages ) m->SetBreakpoint ( &m->GetRam()[addr], flags); }		
					}
				} 
				else 				
				{
					if( !(rom||ram) )
					{
						if( flags&cpu_break_w ) { m->SetBreakpoint ( m->GetCpu()->WrPtr(addr), cpu_break_w); flags &= ~cpu_break_w; }
						if( flags )				{ m->SetBreakpoint ( m->GetCpu()->RdPtr(addr), flags); }
					}
				}
			}
			if(c) continue; else break;
		}
		breakpoints=XXX; 
		
	QUNBLOCK;
}





-(IBAction)scrollMemoryView:(id)sender		{ LogLine("TODO: CpuPanel:scrollMemoryView"); }
-(IBAction)setMemory:(id)sender				{ LogLine("TODO: CpuPanel:setMemory"); }




@end


