/*	Copyright  (c)	Günter Woigk  1996-2005
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.


	Z80 Emulator
	originally based on fMSX; Copyright (C) Marat Fayzullin 1994,1995

	macro definitions
*/


// access z80 registers:

#define	IFF1	registers.r.iff1		// actually used irpt flip flop
#define	IFF2	registers.r.iff2		// copy of iff1 during nmi processing
#define	RR		registers.r.r			// 7 bit DRAM refresh counter
#define	RI		registers.r.i			// hi byte of interrupt vector: i register

#define	RA		registers.r.a
#define	RF		registers.r.f
#define	RB		registers.r.b
#define	RC		registers.r.c
#define	RD		registers.r.d
#define	RE		registers.r.e
#define	RH		registers.r.h
#define	RL		registers.r.l

#define	RA2		registers.r.a2
#define	RF2		registers.r.f2

#define	XH		registers.r.xh
#define	XL		registers.r.xl
#define	YH		registers.r.yh
#define	YL		registers.r.yl
#define	PCH		registers.r.pch
#define	PCL		registers.r.pcl
#define	SPH		registers.r.sph
#define	SPL		registers.r.spl

#define	BC		registers.rr.bc
#define	DE		registers.rr.de
#define	HL		registers.rr.hl

#define	BC2		registers.rr.bc2
#define	DE2		registers.rr.de2
#define	HL2		registers.rr.hl2

#define	IX		registers.rr.ix
#define	IY		registers.rr.iy
#define	PC		registers.rr.pc
#define	SP		registers.rr.sp


// read/write data:

#ifndef		PEEK
#define		PEEK(R,A)		{ cc+=3; R=Peek(A); }
#endif
#ifndef		POKE
#define		POKE(A,R)		{ cc+=3; Poke(A,R); }
#endif

#ifndef		GET_INSTR
#define		GET_INSTR(R)	{ cc+=4; r+=1; R=Peek(pc++); }
#endif
#ifndef		GET_CB_OP
#define		GET_CB_OP(R)	{ cc+=4; r+=1; R=Peek(pc++); }
#endif
#ifndef		GET_ED_OP
#define		GET_ED_OP(R)	{ cc+=4; r+=1; R=Peek(pc++); }
#endif
#ifndef		GET_XY_OP
#define		GET_XY_OP(R)	{ cc+=4; r+=1; R=Peek(pc++); }
#endif
#ifndef		GET_XYCB_OP
#define		GET_XYCB_OP(R)	{ cc+=5; R=Peek(pc++); }
#endif
#ifndef		GET_N
#define		GET_N(R)		{ cc+=3; R=Peek(pc++); }
#endif
#ifndef		SKIP_N
#define		SKIP_N()		{ cc+=3; Peek(pc++); }
#endif

#ifndef OUTPUT
#define	OUTPUT(A,B)			printf("cpu::out(%$04hx,$%02hhx)\n",ushort(A),uchar(B)); cc+=4; 
#endif
#ifndef INPUT				
#define	INPUT(A,B)			printf("cpu::in(%$04hx)\n",ushort(A)); B=0xFF; cc+=4; 
#endif


#define	GET_NN(RR)			{ GET_N(RR); ushort w_nn; GET_N(w_nn); RR += 256*w_nn; }
#define	POP(R)				{ PEEK(R,SP); SP++; }
#define	PUSH(R)				{ --SP; POKE(SP,R); }



// profiling opcodes and executed locations:

#ifndef CPU_INFO_IRPT
	#define CPU_INFO_IRPT			/* nop */
#endif
#ifndef CPU_INFO_NMI
    #define	CPU_INFO_NMI			/* nop */
#endif
#ifndef CPU_INFO_RETI
    #define	CPU_INFO_RETI(PC)		/* nop */
#endif
#ifndef CPU_INFO_RETN
    #define	CPU_INFO_RETN(PC)		/* nop */
#endif
#ifndef CPU_INFO_HALT
    #define	CPU_INFO_HALT			/* nop */
#endif

#ifndef CPU_INFO_ILLEGAL
	#define CPU_INFO_ILLEGAL(PC)	/* nop */
#endif

#ifndef CPU_INFO_POP
	#define CPU_INFO_POP			/* nop */
#endif
#ifndef CPU_INFO_RET
	#define CPU_INFO_RET			/* nop */
#endif
#ifndef CPU_INFO_EX_HL_xSP
	#define CPU_INFO_EX_HL_xSP		/* nop */
#endif

#ifndef CPU_INFO_RST00
	#define CPU_INFO_RST00(PC)		/* nop */
#endif
#ifndef CPU_INFO_RST08
	#define CPU_INFO_RST08(PC)		/* nop */
#endif
#ifndef CPU_INFO_RST10
	#define CPU_INFO_RST10(PC)		/* nop */
#endif
#ifndef CPU_INFO_RST18
	#define CPU_INFO_RST18(PC)		/* nop */
#endif
#ifndef CPU_INFO_RST20
	#define CPU_INFO_RST20(PC)		/* nop */
#endif
#ifndef CPU_INFO_RST28
	#define CPU_INFO_RST28(PC)		/* nop */
#endif
#ifndef CPU_INFO_RST30
	#define CPU_INFO_RST30(PC)		/* nop */
#endif
#ifndef CPU_INFO_RST38
	#define CPU_INFO_RST38(PC)		/* nop */
#endif








// --------------------------------------------------------------------
// ----	INSTRUCTION MACROS --------------------------------------------
// --------------------------------------------------------------------


/*	RLC ... SRL:	set/clr C, Z, P, S;
			clear	N=0, H=0
			pres.	none
*/
#define M_RLC(R)      				\
	rf  = R>>7;						\
	R   = (R<<1)+rf; 				\
	rf |= zlog_flags[R]

#define M_RRC(R)      				\
	rf  = R&0x01;					\
	R   = (R>>1)+(rf<<7);		 	\
	rf |= zlog_flags[R]

#define M_RL(R)						\
	if (R&0x80)						\
	{	R 	= (R<<1)+(rf&0x01);		\
		rf	= zlog_flags[R]+C_FLAG;	\
	} else							\
	{	R 	= (R<<1)+(rf&0x01);		\
		rf	= zlog_flags[R];		\
	}

#define M_RR(R)						\
	if (R&0x01)						\
	{	R 	= (R>>1)+(rf<<7);		\
		rf	= zlog_flags[R]+C_FLAG;	\
	} else							\
	{	R 	= (R>>1)+(rf<<7);		\
		rf	= zlog_flags[R];		\
	}
  
#define M_SLA(R)					\
	rf	= R>>7;						\
	R <<= 1;						\
	rf |= zlog_flags[R]

#define M_SRA(R)					\
	rf	= R&0x01;					\
	R	= (R&0x80)+(R>>1);			\
	rf |= zlog_flags[R]

#define M_SLL(R)					\
	rf	= R>>7;						\
	R	= (R<<1)+1;					\
	rf |= zlog_flags[R]

#define M_SRL(R)					\
	rf	= R&0x01;					\
	R >>= 1;						\
	rf |= zlog_flags[R]


/*	BIT:	set/clr	Z
			clear	N=0, H=1
			pres	C
			takes other flags from corresponding bits in tested byte!
*/
#define M_BIT(N,R)								\
	rf	= (rf&C_FLAG) + 						\
		  (R&(S_FLAG+P_FLAG)) + 				\
		   H_FLAG + 							\
		  ((R&N)?0:Z_FLAG)


/*	ADD ... CP:	set/clr	Z, S, V, C, N, H
				pres	none
*/
#define M_ADD(R)								\
	wm	= ra+R;									\
	rf	= wmh + (wml?0:Z_FLAG) + (wml&S_FLAG)	\
			 + (~(ra^R)&(wml^ra)&0x80?V_FLAG:0)	\
			 + ((ra^R^wml)&H_FLAG);				\
	ra	= wml

#define M_SUB(R)								\
	wm	= ra-R;									\
	rf	= -wmh + (wml?0:Z_FLAG) + (wml&S_FLAG)	\
			  + ((ra^R)&(wml^ra)&0x80?V_FLAG:0)	\
			  + ((ra^R^wml)&H_FLAG) + N_FLAG;	\
	ra	= wml

#define M_ADC(R)								\
	wm	= ra+R+(rf&C_FLAG);						\
	rf	= wmh + (wml?0:Z_FLAG) + (wml&S_FLAG)	\
			 + (~(ra^R)&(wml^ra)&0x80?V_FLAG:0)	\
			 + ((ra^R^wml)&H_FLAG);				\
	ra	= wml

#define M_SBC(R)								\
	wm	= ra-R-(rf&C_FLAG);						\
	rf	= -wmh + (wml?0:Z_FLAG) + (wml&S_FLAG)	\
			  + ((ra^R)&(wml^ra)&0x80?V_FLAG:0)	\
			  + ((ra^R^wml)&H_FLAG) + N_FLAG;	\
	ra	= wml

#define M_CP(R)									\
	wm	= ra-R;									\
	rf	= -wmh + (wml?0:Z_FLAG) + (wml&S_FLAG)	\
			  + ((ra^R)&(wml^ra)&0x80?V_FLAG:0)	\
			  + ((ra^R^wml)&H_FLAG) + N_FLAG;


/*	AND ... XOR:	set/clr	Z, P, S
					clear	C=0, N=0, H=0/1 (OR,XOR/AND)
					pres	none
*/
#define M_AND(R)								\
	ra &= R;									\
	rf	= H_FLAG|zlog_flags[ra]

#define M_OR(R)									\
	ra |= R;									\
	rf	= zlog_flags[ra]

#define M_XOR(R)								\
	ra ^= R;									\
	rf	= zlog_flags[ra]


/*	INC ... DEC:	set/clr	Z,P,S,H
					clear	N=0/1 (INC/DEC)
					pres	C
*/
#define M_INC(R)								\
	R++;										\
	rf	= (rf&C_FLAG) + 						\
		  (R?0:Z_FLAG) + 						\
		  (R&S_FLAG) + 							\
		  (R==0x80?V_FLAG:0) + 					\
		  (R&0x0F?0:H_FLAG)

#define M_DEC(R)								\
	R--;										\
	rf	= (rf&C_FLAG) + 						\
		  (R?0:Z_FLAG) + 						\
		  (R&S_FLAG) + 							\
		  (R==0x7F?V_FLAG:0) + 					\
		  (((R+1)&0x0F)?0:H_FLAG) +				\
		   N_FLAG


/*	ADDW:	set/clr	C
			clear	N=0
			pres	Z, P, S
			unkn	H
*/
#define M_ADDW(R1,R2)							\
	rf &= ~(N_FLAG+C_FLAG);						\
	rf |= ((ulong)R1+(ulong)R2)>>16;			\
	R1 += R2;


/*	ADCW, SBCW:	set/clr	C,Z,V,S
			clear	N=0/1 (ADC/SBC)
			unkn	H
			pres	none
*/
#define M_ADCW(R)								\
	wm	= HL+R+(rf&C_FLAG);						\
	rf	= (((ulong)HL+(ulong)R+(rf&C_FLAG))>>16)\
			+ (wm?0:Z_FLAG) + (wmh&S_FLAG)		\
			+ (~(HL^R)&(wm^HL)&0x8000?V_FLAG:0);\
	HL	= wm
   
#define M_SBCW(R)								\
	wm	= HL-R-(rf&C_FLAG);						\
	rf	= (((ulong)HL-(ulong)R-(rf&C_FLAG))>>31)\
			+ (wm?0:Z_FLAG) + (wmh&S_FLAG)		\
			+ ((HL^R)&(wm^HL)&0x8000?V_FLAG:0)	\
			+ N_FLAG;							\
	HL	= wm


/*	IN	set/clr	Z, P, S, H
		clear	N=0
		pres	C
*/
#define M_IN(R)									\
	INPUT(BC,R);								\
	rf	= (rf&C_FLAG) + zlog_flags[R]










