/*	Copyright  (c)	Günter Woigk 1996 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
 	
 
	Z80 Emulator
	originally based on fMSX; Copyright (C) Marat Fayzullin 1994,1995

	1995-03-28 kio	Started work on C version
	1995-04-01 kio	Finished, yee it works!												
	1995-04-06 kio	Removed "final bug": im2 interrupt handling							
	1995-06-12 kio	revised EI_WITH_DELAY handling										
	1995-06-12 kio	all Info-Calls and profiling added									
	1995-06-12 kio	exact r register emulation completed								
	1995-06-13 kio	All opcodes covered, no unsupported opcodes!						
	1995-11-01 kio	Paged memory supported												
	1995-11-26 kio	Engine now completely hardware independent!							
	1996-03-28 kio	Core2Core() works even if src and dest block overlap at both ends	
	1996-03-28 kio	removed option to switch off latency of 'EI' instruction			
	1996-05-02 kio	LOAD_CC after IN and OUT added to make extra wait states possible	
	1998-12-24 kio	started port to linux, rework for c++								
	2000-02-12 kio	removed bug in LDDR emulation										
	2002-02-06 kio	started work on cocoa version										
	2004-11-08 kio	cleaned up source. removed 'class cpu' overhead and disfunct stuff
	2004-11-11 kio	wait cycle management and crt video callback finished
	2005-01-16 kio	wait testing cpu cycles validated against the real machine
	2006-10-06 kio	removed the 'class Cpu' layer, as it contained a lot of Z80 stuff...
*/

#define SAFE 2
#define LOG  1

#define		CPU_SOURCE
#include	"Z80.h"
#include	"Z80 macros.h"
#include	"Z80 opcodes.h"


static uchar	zlog_flags[256] =				// convert:   A -> Z80-flags with S, Z, V=parity and C=0
{
	#define FLAGS0(A)	( A&0x80 )  +  ( (A==0)<<6 )  +  (  ( (~A+(A>>1)+(A>>2)+(A>>3)+(A>>4)+(A>>5)+(A>>6)+(A>>7))&1 ) << 2  )
	#define FLAGS2(A)	FLAGS0(A), FLAGS0((A+1 )), FLAGS0((A+2 )), FLAGS0((A+3))
	#define FLAGS4(A)	FLAGS2(A), FLAGS2((A+4 )), FLAGS2((A+8 )), FLAGS2((A+12))
	#define FLAGS6(A)	FLAGS4(A), FLAGS4((A+16)), FLAGS4((A+32)), FLAGS4((A+48))
	FLAGS6(0), FLAGS6(64), FLAGS6(128), FLAGS6(192)
};


// ==================================================================================


ushort	noreadpage[CPU_PAGESIZE];		// dummy page for unmapped memory
ushort	nowritepage[CPU_PAGESIZE];		// dito
#if CPU_STATISTICS
static ulong	nocountmap[CPU_PAGESIZE];		// dummy count map
#endif




Z80::Z80( Machine* m ) 
:	ZxItem(m,"Z80 CPU")	
{ 
	m->cpu = this;

//	crt_cycle = 0; 
	cpu_cycle = 0;
	crtc	  = NULL; 

#if CPU_STATISTICS
	memset( cnt_instr,0,sizeof(cnt_instr) );
	memset( cnt_cb,0,   sizeof(cnt_cb) );
	memset( cnt_ed,0,   sizeof(cnt_ed) );
	memset( cnt_xy,0,   sizeof(cnt_xy) );
	memset( cnt_xycb,0, sizeof(cnt_xycb) );
	MapCounts( addr, 0x10000, NULL,NULL,NULL );
#endif

// init global dummy read page for non-mapped ram: 
// note: no need to init nowritepage[]: 0x0000 is ok 
	TRAP(sizeof(twobytes)!=2);
	for( int i=0; i<CPU_PAGESIZE; i++ ) noreadpage[i] = 0x00FF;		// flags=0x00, data=0xFF

// setup memory to no_memory && no_breakpoints && no_waitmap
	for( int i=0; i<CPU_PAGES; i++ )
	{
		PgInfo& p = page[i];
		p.core_w		 = nowritepage - i*CPU_PAGESIZE;
		p.core_r		 = noreadpage  - i*CPU_PAGESIZE;
		p.waitmap_w		 = p.waitmap_r 	    = NULL; 
		p.waitmap_w_size = p.waitmap_r_size = 0; 
	}
}


Z80::~Z80 ( )
{
	GetMachine()->cpu = NULL;
}


void Z80::PowerOn ( Time t, long cc )
{ 
//	crt_cycle = cc; 
	cpu_cycle = cc; 
	crtc	  = GetMachine()->ula;	
	
	XXTRAP(!crtc); 

	interrupt_signal = 0; 
	Reset( t, cc ); 
}


void Z80::Reset ( Time, long cc )
{
	cpu_cycle = cc;				// note: eigentlich müssten wir noch bis cc laufen lassen...

	BC=DE=HL=IX=IY=				// clear registers
	PC=SP=BC2=DE2=HL2=0;
	RA=RA2=RF=RF2=RR=RI=0;
	IFF1=IFF2 = disabled;		// disable interrupts
	registers.r.im  = 1;		// interrupt mode := 1
	registers.r.nmi = off;
}


ushort Z80::Peek2 ( ushort addr )
{
	return (ushort)Peek(addr) + (ushort)((ushort)Peek(addr+1)<<8);
}

void Z80::Poke2 ( ushort addr, ushort n )
{
	Poke( addr,   (uchar)n );
	Poke( addr+1, (uchar)(n>>8) );
}

ushort Z80::Pop2 ( )
{
	registers.rr.sp += 2;
	return Peek2( registers.rr.sp - 2 );
}

void Z80::Push2 ( ushort n )
{
	registers.rr.sp -= 2; 
	Poke2( registers.rr.sp, n );
}





/* ----	attach ram/rom to address space --------------
		Attach real memory to cpu address space
		
		For best performance set CPU_PAGESIZE to the page size of the emulated machine.
		If you emulate multiple machines, set it to the lowest page size of any machine.
	
		addr	= cpu start address of page
		size	= size of page 
					0x0000 is assumed to mean 0x10000 (entire address range)
					addr and size must be multiples of CPU_PAGESIZE
					addr+size must not extend beyond address space end 0x10000 

		page	= pointer to page data

		waitmap	= wait cycles map for this page				
		wmsize	= size of waitmap
					NULL  => no waitstates map
					access to waitmap: cc += waitmap[cc%wmsize]
		
		data and flags pointers in PgInfo struct point to the virtual location of address $0000
		so that you can access data using page.data_r[addr>>CPU_PAGEBITS][addr]
*/
void Z80::MapMem ( ushort addr, ushort size, ushort* r_data, ushort* w_data, uptr waitmap, long wm_size )
{
	size -= CPU_PAGESIZE;			// note: size=0  =>  size=0x10000

	XXXTRAP(!this);
	XXTRAP(!r_data);
	XXTRAP(!w_data);
	XXTRAP(addr&CPU_PAGEMASK);		// addr muss in's 'CPU_PAGESIZE' raster passen
	XXTRAP(size&CPU_PAGEMASK);		// addr+size ebenfass wieder
	XXTRAP(size>0xffff-addr);		// wrap around address space end
	XXTRAP(!waitmap&&r_data[0]&cpu_waitmap);
	XXTRAP(!waitmap&&w_data[0]&cpu_waitmap);
	//XXTRAP( core < machine->core );
	//XXTRAP( core+size+CPU_PAGESIZE > machine->ram+machine->model->ram_size);

	r_data -= addr;
	w_data -= addr;
	
	PgInfo* a = page + (addr>>CPU_PAGEBITS);
	PgInfo* e = a    + (size>>CPU_PAGEBITS);
	
	for( ;a<=e; a++ ) 
	{
		a->core_r		  = r_data;
		a->core_w		  = w_data;
		a->waitmap_w	  = a->waitmap_r	  = waitmap; 
		a->waitmap_w_size = a->waitmap_r_size = wm_size; 
	}
}

void Z80::MapRam ( ushort addr, ushort size, ushort* core, uptr waitmap, long wm_size )
{
	size -= CPU_PAGESIZE;			// note: size=0  =>  size=0x10000

	XXXTRAP(!this);
	XXTRAP(!core);
	XXTRAP(addr&CPU_PAGEMASK);		// addr muss in's 'CPU_PAGESIZE' raster passen
	XXTRAP(size&CPU_PAGEMASK);		// addr+size ebenfalls wieder
	XXTRAP(size>0xffff-addr);		// wrap around address space end
	XXTRAP(!waitmap&&core[0]&cpu_waitmap);	// note: wartezyklen werden nur addiert, wo auch core[x]&cpu_waitmap gesetzt ist!
//	XXTRAP( core < machine->core );
//	XXTRAP( core+size+CPU_PAGESIZE > machine->ram+machine->model->ram_size);

	PgInfo* a = page + (addr>>CPU_PAGEBITS);
	PgInfo* e = a    + (size>>CPU_PAGEBITS);
	
	core -= addr;
	
	for( ;a<=e; a++ ) 
	{
		a->core_w		  = a->core_r		  = core;
		a->waitmap_w	  = a->waitmap_r	  = waitmap; 
		a->waitmap_w_size = a->waitmap_r_size = wm_size; 
	}
}

void Z80::MapWom ( ushort addr, ushort size, ushort* core, uptr waitmap, long wm_size )
{
	size -= CPU_PAGESIZE;			// note: size=0  =>  size=0x10000

	XXXTRAP(!this);
	XXTRAP(!core);
	XXTRAP(addr&CPU_PAGEMASK);		// addr muss in's 'CPU_PAGESIZE' raster passen
	XXTRAP(size&CPU_PAGEMASK);		// addr+size ebenfass wieder
	XXTRAP(size>0xffff-addr);		// wrap around address space end
	XXTRAP(!waitmap&&core[0]&cpu_waitmap);
//	XXTRAP( core < machine->core );
//	XXTRAP( core+size+CPU_PAGESIZE > machine->ram+machine->model->ram_size);

	PgInfo* a = page + (addr>>CPU_PAGEBITS);
	PgInfo* e = a    + (size>>CPU_PAGEBITS);
	
	core -= addr;
	
	for( ;a<=e; a++ ) 
	{
		a->core_w		  = core;
		a->waitmap_w	  = waitmap; 
		a->waitmap_w_size = wm_size; 
	}
}

void Z80::MapRom ( ushort addr, ushort size, ushort* core, uptr waitmap, long wm_size )
{
	size -= CPU_PAGESIZE;			// note: size=0  =>  size=0x10000

	XXXTRAP(!this);
	XXTRAP(!core);
	XXTRAP(addr&CPU_PAGEMASK);		// addr muss in's 'CPU_PAGESIZE' raster passen
	XXTRAP(size&CPU_PAGEMASK);		// addr+size ebenfalls wieder
	XXTRAP(size>0xffff-addr);		// wrap around address space end
	XXTRAP(!waitmap&&core[0]&cpu_waitmap);

//	veraltete Tests: Externe ROMs liegen nicht in machine->core
//	XXTRAP( core < machine->core );
//	XXTRAP( core+size+CPU_PAGESIZE > machine->ram+machine->model->ram_size);

	PgInfo* a = page + (addr>>CPU_PAGEBITS);
	PgInfo* e = a    + (size>>CPU_PAGEBITS);
	
	core -= addr;
	
	for( ;a<=e; a++ ) 
	{
		a->core_r		  = core;
		a->waitmap_r	  = waitmap; 
		a->waitmap_r_size = wm_size; 
	}
}



#ifdef CPU_STATISTICS
/* ----	Map memory access count pages -------------------
		The following count arrays count memory access for each byte
		M1Counts counts opcode reading (program execution)
		RdCounts counts read access 
		WrCounts counts write access 
*/
void Z80::MapM1Counts ( ushort addr, ushort size, ulong* map )
{
	XXTRAP(addr&CPU_PAGEMASK);		// addr muss in's 'CPU_PAGESIZE' raster passen
	XXTRAP(size&CPU_PAGEMASK);		// addr+size ebenfass wieder

	pInfo* a = this->page  +  (addr>>CPU_PAGEBITS);
	pInfo* p = a           +  (size>>CPU_PAGEBITS);
	
	if( size==0 ) { p += CPU_PAGES; }				// 0 == 0x10000
	XXTRAP( p > this->page+CPU_PAGES );				// wrap around address space end

	if(map) { map  -= addr; while(--p>=a) p->mcount = map; }
	else	{ addr += size; while(--p>=a) { addr -= CPU_PAGESIZE; p->mcount = nocountmap-addr; } }
}

void Z80::MapRdCounts ( ushort addr, ushort sz, ulong* map )
{
	XXTRAP(addr&CPU_PAGEMASK);		// addr muss in's 'CPU_PAGESIZE' raster passen
	XXTRAP(size&CPU_PAGEMASK);		// addr+size ebenfass wieder

	pInfo* a = this->page  +  (addr>>CPU_PAGEBITS);
	pInfo* p = a           +  (size>>CPU_PAGEBITS);
	
	if( size==0 ) { p += CPU_PAGES; }				// 0 == 0x10000
	XXTRAP( p > this->page+CPU_PAGES );				// wrap around address space end

	if(map) { map  -= addr; while(--p>=a) p->rcount = map; }
	else	{ addr += size; while(--p>=a) { addr -= CPU_PAGESIZE; p->rcount = nocountmap-addr; } }
}

void Z80::MapWrCounts ( ushort addr, ushort sz, ulong* map )
{
	XXTRAP(addr&CPU_PAGEMASK);		// addr muss in's 'CPU_PAGESIZE' raster passen
	XXTRAP(size&CPU_PAGEMASK);		// addr+size ebenfass wieder

	pInfo* a = this->page  +  (addr>>CPU_PAGEBITS);
	pInfo* p = a           +  (size>>CPU_PAGEBITS);
	
	if( size==0 ) { p += CPU_PAGES; }				// 0 == 0x10000
	XXTRAP( p > this->page+CPU_PAGES );				// wrap around address space end

	if(map) { map  -= addr; while(--p>=a) p->wcount = map; }
	else	{ addr += size; while(--p>=a) { addr -= CPU_PAGESIZE; p->wcount = nocountmap-addr; } }
}

void Z80::MapCounts ( ushort addr, ushort sz, ulong* rmap, ulong* wmap, ulong* m1map )
{
	XXTRAP(addr&CPU_PAGEMASK);		// addr muss in's 'CPU_PAGESIZE' raster passen
	XXTRAP(size&CPU_PAGEMASK);		// addr+size ebenfass wieder

	pInfo* a = this->page  +  (addr>>CPU_PAGEBITS);
	pInfo* p = a           +  (size>>CPU_PAGEBITS);
	
	if( size==0 ) { p += CPU_PAGES; }				// 0 == 0x10000
	XXTRAP( p > this->page+CPU_PAGES );				// wrap around address space end

	if(rmap)  rmap -= addr; 
	if(wmap)  wmap -= addr; 
	if(m1map) m1map -= addr; 
	ulong* nomap = nocountmap + addr + size;
	
	while(--p>=a) 
	{ 
		nomap -= CPU_PAGESIZE; 
		p->mcount = m1map?m1map:nomap; 
		p->rcount = rmap?rmap:nomap; 
		p->wcount = wmap?wmap:nomap; 
	}
}
#endif







// ===========================================================================
// ====	The Z80 Engine =======================================================
// ===========================================================================


// ----	jumping -----------------------------------------------------------
#define	LOOP				goto nxtcmnd						// LOOP to next instruction
#define POKE_AND_LOOP(W,C)	{ w=W; c=C; goto poke_and_nxtcmd; }	// POKE(w,c) and goto next instr.
#define EXIT				  goto x							// exit from cpu

// ----	load and store local variables from/to global registers ------------
#define	LOAD_REGISTERS										\
	r		= registers.r.r;	/* refresh counter R	 */	\
	cc		= cpu_cycle;		/* cpu execution counter */	\
	pc		= registers.rr.pc;	/* program counter PC	 */	\
	ra		= registers.r.a;	/* register A			 */	\
	rf		= registers.r.f;	/* register F			 */	\

#define	SAVE_REGISTERS															\
	registers.r.r	= (registers.r.r&0x80)|(r&0x7f);/* refresh counter R	 */	\
 	cpu_cycle		= cc;							/* cpu execution counter */	\
 	registers.rr.pc = pc;							/* program counter PC	 */	\
	registers.r.a   = ra;							/* register A			 */	\
	registers.r.f   = rf;							/* register F			 */	\


/* ====	The Z80 ENGINE ====================================================
*/
bool Z80::Run ( long ccx, int options )
{		
	XXTRAP((uchar)options!=0);
	
	long	cc;				// cpu execution counter

	ushort	pc;				// z80 program counter
	uchar	ra;				// z80 a register
	uchar	rf;				// z80 flags
	uchar	r;				// z80 r register bit 0...6

	ushort*rzp;				// pointer to IX or IY after 0xDD or 0xFD
	#define			rz		(*rzp)			// IX or IY
	#ifdef _BIG_ENDIAN
	#define			rzh		(((uptr)rzp)[0])		// XH or YH
	#define			rzl		(((uptr)rzp)[1])		// XL or YL
	#else
	#define			rzh		(((uptr)rzp)[1])		// XH or YH
	#define			rzl		(((uptr)rzp)[0])		// XL or YL
	#endif

	uchar	c=0;							// general purpose byte register
		
	ushort	w;								// general purpose word register
	#define			wl		(uchar)w		// access low byte of w
	#define			wh		(w>>8)			// access high byte of w

	ushort	wm;								// scratch for macro internal use
	#define			wml		(uchar)wm		// access low byte of wm
	#define			wmh		(wm>>8)			// access high byte of wm


// load local variables from global registers:
	LOAD_REGISTERS;
	CPU_INFO_INIT;


// check for interrupts:
// interrupts are only tested once at start of z80::Run()
// and only tested again after EI processing!

// non-maskable interrupt
// the NMI is edge-triggered and automatically cleared
	if (registers.r.nmi)
	{								// 11 T cycles, probably: M1:5T, M2:3T, M3:3T
		if (Peek(pc)==HALT) pc++;		// TODO: HALT and NMI
		registers.r.nmi = no;			// clear nmi
	//	IFF2 = IFF1;					// save current irpt enable bit
		IFF1 = disabled;				// disable irpt, preserve IFF2
		cc+=1;
		CPU_INFO_NMI;
		cc+=4;						// M1: 5 T: interrupt acknowledge cycle
		w = 0x0066;						// M2: 3 T: push pch
		goto irpt_xw;					// M3: 3 T: push pcl
	}

// maskable interrupt
	if(interrupt_signal<=cc) { interrupt_signal=0; LOOP; }	// no interrupt or interrupt expired
	if(IFF1==disabled) LOOP;								// interrupts disabled

// handle interrupt
// the interrupt is not automatically cleared
	if (Peek(pc)==HALT) pc++;		// TODO: HALT and Interrupt
	IFF1 = IFF2 = disabled;			// disable iff1 & iff2
	
	// im irpt ack zyklus wird halt in Takt 4 getestet, also 2 Takte später als im normalen M1 Zyklus
	// entsprechend wird das busbyte am ende von takt 4 gelesen, also auch 2 takte später
	cc+=2;
	c = 0xFF;
	CPU_INFO_IRPT;					// get byte from bus. TODO: does the cpu a dummy read in IM1?
	cc+=4;						// at least ZX Spectrum videoram waitcycles is no issue because irpt is off-screen.
	
	switch (registers.r.im)
	{
	case 0:				
	/*	mode 0: read instruction from bus: 
		normal timing; M1 is preceeded by 2 additional cycles (s.o.)
		NOTE: docs say that M1 is always 6 T, but that is not conclusive:
		the 2 additional cycles are before opcode fetch, and thus they must always add to instruction execution
		the timing from the moment of instruction fetch (cycle 2 normally, 4 in interrupt ackn.) 
		can't be shortended. This RSTxx opcodes must take 7+3+3 cycles.				
		to be tested somehow.	kio 2004-11-12
	*/
		switch(c)
		{
		case RST00: case RST08: case RST10: 
		case RST18: case RST20: case RST28:
		case RST30:		
		case RST38:		// RSTxx: timing: 2+5+3+3
			cc+=1;		// M1: 7 T: acknowledge interrupt, read RSTxx instruction from bus
			w = c-RST00;	// M2: 3 T: push pch
			goto irpt_xw;	// M3: 3 T: push pcl
			
		default:		//	currently only 1-byte instructions are supported in irpt ackn cycle
			IERR();		//	thereoff RST only useful. others, like CALL NN, may be added someday
		};

	case 1:				
	/*	mode 1:  RST38; 13 T: 7+3+3
		NOTE: docs say, timing is 13T for implicit RST38 in mode 1 and 12T for fetched RST38 in mode 0.
		maybe it is just vice versa? in mode 1 the implicitely known RST38 can be executed with the start of the M1 cycle
		and finish after 5T, prolonged to the irpt ackn M1 cycle min. length of 6T. Currently i'll stick to 13T as doc'd.
		to be tested somehow.	kio 2004-11-12
	*/
		 cc+=1;			// M1: 7 T: acknowledge interrupt 
		 w = 0x0038;
irpt_xw: PUSH(pc>>8);		// M2: 3 T: push pch 
		 PUSH(pc);			// M3: 3 T: push pcl and jump to 0x0038							
		 pc = w;
		 LOOP;

	case 2:	
	/*	mode 2:  jmp via table; 19 T: 7+3+3+3+3
	*/
		cc+=1;			// M1: 7 T: acknowledge interrupt and read interrupt vector
		PUSH(pc>>8);		// M2: 3 T: push pch
		PUSH(pc);			// M3: 3 T: push pcl
		w = registers.r.i*256+c;
		PEEK(PCL,w);		// M4: 3 T: read low byte from table
		PEEK(PCH,w+1);		// M5: 3 T: read high byte and jump 
		pc=PC;
		LOOP;

	default:		
		IERR();			// bogus irpt mode
	};
	

// ----	THE MAIN INSTRUCTION DISPATCHER --------------------------------

poke_and_nxtcmd:
	POKE(w,c);				// --> CPU_WAITCYCLES, CPU_READSCREEN, cc+=3, Poke(w,c)

nxtcmnd:

	if( cc<ccx )			
	{
loop_ei:
		#include "Z80 codes.h"

	}


// ---- EXIT ----

x:	CPU_INFO_EXIT;
	SAVE_REGISTERS;
	
	return (ulong)cc<(ulong)ccx;		// 0: regular exit;  1: preempted
}






































