/*	Copyright  (c)	Günter Woigk   1995-2006
  					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

	1995-06-29 kio	started work on pascal string lib
	2000-01-28 kio	started work on c string lib
	2002-01-19 kio	started unix port
	2002-01-26 kio	simplified Using(): now using vsnprintf()
	2002-01-26 kio	added xpool[] for unlimited string length support
	2003-05-27 kio	StrToDouble()
	2003-06-07 kio	fixed problem with VaUsing(va_list); fixed bugs in StrToDouble()
	2003-11-28 kio	fixed problem with VaUsing(va_list); again.
	2006-10-24 kio	renamed QString() to TempStr(), because QString collides with Qt's class QString
	2006-11-01 kio	TempStr() version for Carbon MP threads
*/

#define	SAFE	3
#define	LOG		2

#include <Carbon/Carbon.h>
#include "cstrings.h"

#ifdef HAVE_NAN_H
#include <nan.h>
#endif
#ifndef NAN
const double NAN = 0.0/0.0;
#endif



/* ----------------------------------------------------------
			THREAD-SAFE POOL HANDLING
---------------------------------------------------------- */


#define	QSPOOLSIZE	5000		// string pool size per thread
#define	QSMAXLEN	255

struct QSPool 
{
	uint	idx;
	char	pool[QSPOOLSIZE];
			QSPool(){idx=0;};
};

static TaskStorageIndex 
			QSPoolIndex = 0;	// index of per-thread pool handle



/* ----	allocate and return a string from the pool ---------------------------------
		allocate semi-durable strings
		use them quickly, for the buffer will be re-used after some more calls ...
		intended to prepare output of numbers etc.
		the string's last byte is preset with 0		
*/
char* QuickStr ( int len )
{
	QSPool* qspool;
	
	if ( !QSPoolIndex )
	{
		OSStatus err = MPAllocateTaskStorageIndex ( &QSPoolIndex );	TRAP(err);
		XXLog( "QuickStr: QSPoolIndex = %i\n", (int)QSPoolIndex );
		XXXTRAP(QSPoolIndex==0);
		XXXTRAP(MPGetTaskStorageValue(QSPoolIndex)!=0);	
	}

	qspool = (QSPool*) MPGetTaskStorageValue ( QSPoolIndex );	
	if ( !qspool )
	{
		XXLog( "QuickStr: allocating pool for new thread\n" );
		qspool = new QSPool; ASSERT(qspool);
		OSStatus err = MPSetTaskStorageValue ( QSPoolIndex, (TaskStorageValue)qspool ); TRAP(err);
	}

	XXXTRAP( qspool->idx>QSPOOLSIZE );
	XXTRAP ( len>QSMAXLEN );

	len += 1;	// for final 0-byte
	if ( qspool->idx+len>QSPOOLSIZE ) qspool->idx = 0;
	qspool->idx += len;
	qspool->pool[qspool->idx-1] = 0;

	return qspool->pool + qspool->idx - len;
}


/* ----------------------------------------------------------
			STRING FUNCTIONS
---------------------------------------------------------- */

#include "cstrings_functions.h"








