/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#define	SAFE	3
#define	LOG		1

#import "Colormap.h"
#import "cstrings/cstrings.h"


int
Colormap::AddColor ( cComp q[3] ) throw(bad_alloc,limit_error) 
{ 
	int n=used_colors++; 
	if(n==CmapSize()) GrowCmap(n+1); 
	memcpy(array+n*3,q,3); 
	return n; 
}


int
Colormap::CondAddColor ( cComp q[3] ) throw(bad_alloc,limit_error)
{ 
	int i = FindColor(q);     
	return i==not_found ? AddColor(q) : i; 
}
	

int
Colormap::CondAddColor ( Comp r, Comp g, Comp b ) throw(bad_alloc,limit_error)
{
	int i = FindColor(r,g,b);
	return i==not_found ? AddColor(r,g,b) : i; 
}


void
Colormap::ShrinkCmap ( int n ) throw(limit_error)
{
	n=valid_count(n); 
	Cmap::Shrink(n*3); 
	if( used_colors>n ) used_colors = n; 
	if( transp>=used_colors ) transp = unset; 
}


void
Colormap::ShrinkColors ( int n ) throw(limit_error)
{	
	int m=valid_count(n); 
	Cmap::Shrink(m*3); 
	if( used_colors>n ) used_colors=n; 
	if( transp>=used_colors ) transp=unset; 
}


int 
Colormap::FindColor ( Comp r, Comp g, Comp b ) const
{	
	cComp* p0 = GetCmap();
	cComp* p  = p0 + CmapByteSize();

	while( p>p0 ) 
	{
		p-=3;
		if( p[0]==r && p[1]==g && p[2]==b ) { int i = (p-p0)/3; if( i!=transp ) return i; }
	}

	return not_found;
}


cstr 
Colormap::CmapStr() const
{
	str s = TempStr( UsedColors()*12+30+2 );
	ptr p = s;
	cComp* q = GetCmap();

	p += sprintf(p,"max/used=%i(%i)/%i(%i), cmap: ", CmapSize(),CmapBits(),UsedColors(),UsedBits() );
	
	for( int i=0; i<UsedColors(); i++ )
	{
		p += sprintf(p,"%hhu,%hhu,%hhu%s ",q[0],q[1],q[2],i==transp?"*":""); 
		q += 3;
	}
	*--p = 0;
	return s;
}























