

#define SAFE	3
#define LOG		1

#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#import "unix/file_utilities.h"
#import "GifReport.h"



static void LogColorTable( int fd, int size )
{
	uchar rgb[256*3];
	read_bytes(fd,rgb,size*3);
	Log("  Color Table");
	for( int i=0; i<size; i++ )
	{
		if(i%8==0) Log("\n    ");
		uchar* p = rgb+3*i;
		Log("%02hhX,%02hhX,%02hhX    ",p[0],p[1],p[2]);
	}
	LogNL();
}


static void LogBlock( int fd, uptr bu, int n, int flags=0 /* 'X' for pure Hex */ )
{
	for( int i=0;i<n;i++ ) 
	{ 
		if(i%64==0) Log("\n    ");
		uchar c=bu[i]; 
		if( flags=='X' ) Log("%02hhX",c);
		else if( c>=0x20&&c<0x7f ) Log("%c",c);
		else Log(":%02hhX",c);
	}
}

static void LogSubBlocks( int fd, int flags=0 /* 'X' for pure Hex */ )
{
	Log("  Sub Blocks:");
	while(1)
	{
		int n = read_uchar(fd); if(n==0) { LogNL(); return; }
		uchar bu[256];
		read_bytes(fd,bu,n);
		LogBlock(fd,bu,n,flags);
	}
}

static void LogImageBlock(int fd)
{
	ushort w,h,x,y; uchar flags;
	x=read_ushort_z(fd);
	y=read_ushort_z(fd);
	w=read_ushort_z(fd);
	h=read_ushort_z(fd);
	flags=read_uchar(fd);	
	bool has_local_cmap	= flags&0x80;
	bool is_interlaced	= flags&0x40;
	bool cmap_is_sorted	= flags&0x20;
	int cmap_bits		= (flags&7)+1;
	int cmap_size		= 2<<(flags&7);

	Log("Image Block\n");
	Log("  x=%hu, y=%hu, w=%hu, h=%hu\n", x,y,w,h);
	Log("  flags: $%02hhx\n",flags);
	Log("    Local color table: %s\n",has_local_cmap?"yes":"no");
	Log("    Image interlaced: %s\n",is_interlaced?"yes":"no");

	if(has_local_cmap) 
	{
		Log("    Color table is sorted: %s\n",cmap_is_sorted?"yes":"no");
		Log("    Local color table: %i bit / %i colors\n", cmap_bits, cmap_size );
		LogColorTable(fd,cmap_size); 
	}

	uchar min_bits = read_uchar(fd); 
	Log("  LZW initial symbol size: %hhu bits\n",min_bits);

	// dump sub blocks	
	LogSubBlocks(fd,'X');
}

static void LogGraphicControlExtension(int fd)
{
	int   block_size = read_uchar(fd);
	uchar flags = read_uchar(fd);
	uint  delay_time = read_ushort_z(fd);
	uchar transp_index = read_uchar(fd);
	uchar block_end = read_uchar(fd);
	int disposal = (flags>>2)&7;	
	cstr disp[8] = 
	{ "0: don't care", "1: keep new pixels", "2: restore bgcolor", "3: restore pixels", "4", "5", "6", "7" };
	bool user_input = flags&2;
	bool transparency = flags&1;
	
	Log("Graphic Control Extension\n");
	if(block_size!=4) Log("  wrong field size: %i\n",block_size);
	if(block_end!=0)  Log("  wrong block delimiter: %i\n",block_end);
	Log("  delay time: %u msec\n", delay_time*10u);
	Log("  flags: $%02hhx\n",flags);		
	Log("    disposal: %s\n",disp[disposal] );
	Log("    wait for user input: %s\n", user_input?"yes":"no");
	Log("    use transparency: %s\n", transparency?"yes":"no");
	Log("  transparent color index: %hhu\n",transp_index);
}

static void LogCommentBlock(int fd)
{
	Log("Comment Block\n");
	LogSubBlocks(fd);
}

static void LogPlainTextBlock(int fd)
{
	Log("Plain Text Block\n");
	TODO();
}

static void LogApplicationExtension(int fd)
{
	Log("Application Extension\n");
	int block_size = read_uchar(fd);
	if(block_size!=11) Log("  wrong field size: %i\n",block_size);
	uchar bu[11];
	read_bytes(fd,bu,11);
	Log("  application:");
	LogBlock(fd,bu,11,0);
	LogNL();
	LogSubBlocks(fd);	
}


void LogGifFile( cstr filepath )
{
	Log( "Gif File Report for \"%s\"\n",filepath );
	
	int fd = open( filepath, O_RDONLY, 0660 );
	if (fd<0) { Log("open() failed\n"); return; }

	try
	{
		char header[7]; header[6]=0;
		read_bytes(fd,header,6); 
		if( strncmp(header,"GIF",3)!=0 ) throw(any_error("header!=\"GIF\""));
		Log( "Header: %s\n",header );

		ushort w,h; uchar flags, bgcolor, aspectratio;
		w=read_ushort_z(fd); 
		h=read_ushort_z(fd); 
		flags=read_uchar(fd); 
		bgcolor=read_uchar(fd); 
		aspectratio=read_uchar(fd); 
		bool has_global_cmap	= flags&0x80;
		int  source_color_bits	= ((flags>>4)&0x07)+1;
		int  source_colors		= 2<<((flags>>4)&0x07);
		bool colors_sorted		= flags&0x08;
		int	 global_cmap_bits	= (flags&7)+1;
		int	 global_cmap_colors	= 2<<(flags&7);
		Log("Logical Screen Descriptor\n");
		Log("  screen width: %hu\n",w);
		Log("  screen height: %hu\n",h);
		Log("  flags: $%02hhx\n",flags);		
		Log("    Global color table: %s\n", has_global_cmap?"yes":"no" );
		Log("    Source color richness: %i bit / %i colors\n", source_color_bits, source_colors );
		Log("    Color table is sorted: %s\n", colors_sorted?"yes":"no" );
		Log("    Global color table: %i bit / %i colors\n", global_cmap_bits, global_cmap_colors );
		Log("  background color index: %hhu\n",bgcolor);
		Log("  aspect ratio: %s\n",aspectratio?(Using("%i/64",aspectratio+15)):"square");
		
		if( has_global_cmap ) LogColorTable(fd,global_cmap_colors);

		while(1)
		{
			uchar image_separator = read_uchar(fd);
		
			if( image_separator==0x2c )		// image separator
			{
				LogImageBlock(fd);
				continue;
			}
			if( image_separator==0x21 )		// extension
			{
				uchar label = read_uchar(fd);
				if( label==0xf9 )			// graphic control 
				{
					LogGraphicControlExtension(fd);
					continue;					
				}
				if( label==0xfe )			// comment
				{
					LogCommentBlock(fd);
					continue;
				}
				if( label==0x01 )			// plain text
				{
					LogPlainTextBlock(fd);
					continue;
				}
				if( label==0xff )			// application extension
				{
					LogApplicationExtension(fd);
					continue;
				}
				Log("Extension with unknown label: $%02hhx\n", label);
				break;
			}
			if( image_separator==0x3b )		// Trailer
			{
				Log("Trailer\n---END---\n\n");
				break;
			}
			else
			{
				Log("Block with unknown separator: $%02hhx\n", image_separator);
				break;
			}
		}

	}
	catch(any_error e)
	{
		Log( "file error: %s\n", e.what() );
	}

	close(fd);
}








