/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#define	SAFE	3
#define	LOG		1

#import "Pixelmap.h"



/*	Copy Creator:
*/	
void Pixelmap::init( cPixelmap& q ) throw(bad_alloc)
{
	data = pixels = new uchar[q.W()*q.H()];
	box  = q.box;
	dy	 = W();		

	for( int y=box.Y1(); y<box.Y2(); y++ ) 
	{
		memcpy( PixelRow(y), q.PixelRow(y), dy/*w*/ );
	}
}



/*	Find highest color index in image:
	returns max_index = 0 for empty Pixmaps
*/
int Pixelmap::MaxColorIndex() const
{
	uchar max = 0;					// maximum collector
	cuptr p = Pixels();				// running pointer

	for( int j=H(); --j>=0; p+=dy ) 
	for( int i=W(); --i>=0; ) 
	{
		if( p[i]>max ) max=p[i]; 
	} 

	return max;
}



/*	Count used colors (color indexes) in image:
	if argument max_color_index != NULL, then also store the max. color index found in the image
	returns 1 used color and sets max_index = 0 for empty Pixmaps
*/
int Pixelmap::CountUsedColors( int* max_color_index ) const
{
	uchar bits[256]; memset(bits,0,256);		// flags for used colors
	cuptr p = Pixels();							// running pointer

	for( int j=H(); --j>=0; p+=dy )
	for( int i=W(); --i>=0; ) 
	{
		bits[p[i]] = 1;							// flag all used colors
	} 
	
	int i = 256;								
	while( i && bits[--i]==0 ) {}
	if( max_color_index ) *max_color_index = i; // store max. color index

	int n = 1;
	while( i ) { n += bits[--i]; }				// count used colors
	return n;
}



/*	Set this Pixmap to the difference Pixmap which converts this Pixmap into the new Pixmap.
	-	the Pixmap size is shrinked to the bounding box of all changed pixels
	-	unchanged pixels are set to the transparent color, except if transp_color<0 ((use Colormap::unset))
	intended for gif animation
*/
void Pixelmap::SetToDiff ( cPixelmap& neu, int transp_color )
{
	ShrinkToRect( neu.box );

	int h  = H();	// height
	int w  = W();	// width
	if( w==0 || h==0 ) return;
	int x1 = X1();
	int y1 = Y1();

	uptr  zp = PixelPtr( x1,y1 );
	cuptr np = neu.PixelPtr( x1,y1 );
	
	while( h>0 )	// skip identical top rows:
	{
		int x; for( x=0; x<w; x++ ) { if( zp[x]!=np[x] ) break; }
		if( x<w ) break; else { h--; y1++; zp+=dy; np+=neu.dy; }
	}
	
	zp += (h-1)*dy;
	np += (h-1)*neu.dy;

	while( h>0 )	// skip identical bottom rows:
	{
		int x; for( x=0; x<w; x++ ) { if( zp[x]!=np[x] ) break; }
		if( x<w ) break; else { h--; zp-=dy; np-=neu.dy; }
	}

	zp -= (h-1)*dy;
	np -= (h-1)*neu.dy;
	
	while( w>0 )	// skip identical left columns:
	{
		int y; for( y=0; y<h; y++ ) { if( zp[y*dy]!=np[y*neu.dy] ) break; }
		if( y<h ) break; else { w--; x1++; np++; zp++; }
	}	

	zp += w-1;
	np += w-1;

	while( w>0 )	// skip identical right columns:
	{
		int y; for( y=0; y<h; y++ ) { if( zp[y*dy]!=np[y*neu.dy] ) break; }
		if( y<h ) break; else { w--; zp--; np--; }
	}

	zp -= w-1;	
	np -= w-1;

	SetP1(x1,y1);
	SetSize(w,h);

// my bounding box x,y,w,h is the bounding box for all changes.
// replace unchanged pixels with transparent color:

	if( transp_color>=0 ) 
	for( int y=0; y<h; y++, zp+=dy, np+=neu.dy )
	for( int x=0; x<w; x++ ) 
	if( zp[x]==np[x] ) zp[x] = transp_color;
}



void Pixelmap::Clear ( int color )
{
	uptr p = Pixels();		// p-> first == top/left pixel
	int  w = W();
	int  h = H();
	uptr e = p + h * dy;

	while( p<e ) { memset(p,color,w); p+=dy; }
}























