/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#ifndef Pixelmap_h
#define	Pixelmap_h

#include "kio/kio.h"
#include "Box/Box (P1,SZ).h"


typedef class Pixelmap;
typedef Pixelmap const cPixelmap;

class Pixelmap
{
protected:
	
	void		kill()				{ delete[] data; }
	void		init( cPixelmap& q ) throw(bad_alloc);

	Box			box;			// frame: x,y,w,h
	int			dy;				// address offset per row
	uptr		data;			// allocated memory for delete[]
	uptr		pixels;			// -> pixel[0,0]


public:	

	Pixelmap	()							: box(),		dy(0),   data(NULL),               pixels(NULL)              {}
	Pixelmap	( cBox& bbox )				: box(bbox),    dy(W()), data(new uchar[W()*H()]), pixels(data-X1()-Y1()*dy) {}
	Pixelmap	( int w,int h )				: box(w,h),     dy(w),   data(new uchar[w*h]),     pixels(data)              {}
	Pixelmap	( int x,int y,int w,int h )	: box(x,y,w,h), dy(w),   data(new uchar[w*h]),     pixels(data-x-y*w)        {}

	Pixelmap	( uptr px, int x,int y, int w,int h, int z=0 )	: box(x,y,w,h), dy(z?z:w), data(px), pixels(px-x-y*dy) {}
	Pixelmap	( uptr px,              int w,int h, int z=0 )	: box(w,h),     dy(z?z:w), data(px), pixels(px) {}

	Pixelmap	( uptr px, cBox& bx,			int z=0 )		: box(bx),   dy(z?z:W()),data(px),pixels(px-X1()-Y1()*dy) {}
	Pixelmap	( uptr px, cLoc& p1, cLoc& p2,  int z=0 )		: box(p1,p2),dy(z?z:W()),data(px),pixels(px-p1.x-p1.y*dy) {}
	Pixelmap	( uptr px, cLoc& p1, cDist& sz, int z=0 )		: box(p1,sz),dy(z?z:W()),data(px),pixels(px-p1.x-p1.y*dy) {}
	Pixelmap	( uptr px,           cDist& sz, int z=0 )		: box(sz),   dy(z?z:W()),data(px),pixels(px) {}
		
				~Pixelmap		()								{ kill(); }
				Pixelmap		( cPixelmap& q )				{ init(q); }
	Pixelmap&	operator=		( cPixelmap& q )				{ if(this!=&q) { kill(); init(q); } return *this; }

	cBox&		Frame			() const						{ return box; }
	int			X1				() const						{ return box.X1(); }
	int			Y1				() const 						{ return box.Y1(); }
	int			X2				() const						{ return box.X2(); }
	int			Y2				() const 						{ return box.Y2(); }
	int			W				() const						{ return box.Width(); }
	int			H				() const						{ return box.Height(); }
	int			DY				() const						{ return dy; }
	cLoc&		P1				() const						{ return box.P1(); }
	Dist		Size			() const						{ return box.Size(); }

	bool		IsEmpty			() const						{ return box.IsEmpty(); }
	bool		IsNotEmpty		() const						{ return box.IsNotEmpty(); }

// Reposition frame:
// This affects which pixels are enclosed. 
// Pixel coordinates are not affected!
	void		SetX			( int x )						{ box.SetX1(x); }
	void		SetY			( int y )						{ box.SetY1(y); }
	void		SetP1			( cLoc& p )						{ box.SetP1(p); }
	void		SetP1			( int x, int y )				{ box.SetP1(x,y); }
	void		SetP2			( cLoc& p )						{ box.SetP2(p); }
	void		SetP2			( int x, int y )				{ box.SetP2(x,y); }
	void		SetW			( int w )						{ box.SetWidth(w); }
	void		SetH			( int h )						{ box.SetHeight(h); }
	void		SetSize			( cDist& d )					{ box.SetSize(d); }	
	void		SetSize			( int w, int h )				{ box.SetSize(w,h); }	
	void		SetFrame		( cBox& b )						{ box = b; }
	void		SetFrame		( cLoc& p1, cLoc& p2 )			{ box.p1=p1; box.sz=p2-p1; }

	void		set_data_ptr	( uptr p )						{ data = p; }		// use with care...
	void		set_pixel_ptr	( uptr p )						{ pixels = p; }		// caveat: check data, p1 and p2!
	void		set_dy			( int z )						{ dy = z; }			// caveat: check pixels, p1 and p2!

// ptr -> pixel[x1,y1]:
	uptr		Pixels			()								{ return PixelPtr(P1()); }
	cuptr		Pixels			() const						{ return PixelPtr(P1()); }

// ptr -> first pixel of row:
	uptr		PixelRow		( int y )						{ return PixelPtr(X1(),y); }
	cuptr		PixelRow		( int y ) const					{ return PixelPtr(X1(),y); }

// ptr -> pixel:
	uptr		PixelPtr		( int x, int y )				{ return pixels +x   +   y*dy;   }
	cuptr		PixelPtr		( int x, int y ) const			{ return pixels +x   +   y*dy;   }

	uptr		PixelPtr		( cLoc& p )						{ return pixels +p.x + p.y*dy;   }
	cuptr		PixelPtr		( cLoc& p ) const				{ return pixels +p.x + p.y*dy;   }
	
	uchar&		Pixel			( int x, int y )				{ return pixels[ x   +   y*dy ]; }
	uchar		Pixel			( int x, int y ) const			{ return pixels[ x   +   y*dy ]; }


// Member Functions:

// Change the pixel coordinate system.
// Thereafter all pixel coordinates, including the frame rect coordinates are shifted!
// This does not reposition the frame rect: it does not affect which pixels are enclosed.
	void		OffsetAddresses	( int DX, int DY )			{ box.Move(DX,DY); pixels -= DX+DY*dy; }
	void		OffsetAddresses	( cDist& d )				{ box.Move(d);     pixels -= d.dx+d.dy*dy; }
	void		SetAddressForP1	( cLoc& p )					{ OffsetAddresses(p-box.p1); }

// Color tools:
	int			MaxColorIndex	() const;
	int			CountUsedColors	( int* max_color_index=NULL ) const;

// Special resizing:
	void		SetToDiff		( cPixelmap& new_pixmap, int transp_index );
	void		ShrinkToRect	( cBox& new_box )			{ box ^= new_box; }
	void		CropBackground	( int bgcolor );

	void		Clear			( int color );
};



#endif











