/*	Copyright  (c)	Günter Woigk   1999-2005
  					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


error number and error text definitions

this file provides:

	raising, reading and clearing of errors.
	uses the global errno.
	errors are stored separately for each thread: kio's dispatcher saves and restores errno!
	note: some OS traps set errno directly, some return an error in a parameter, some require
	you to query the error with a dedicated method. this is not my fault!...


	cstr ErrorStr ( OSErr e )				returns the user-readable text for an OS or
											application defined error

	cstr ErrorStr ( )						returns the user-readable text for the current error

	void SetError ( cstr custommessage );	sets an error with an arbitrary error text

	void SetError ( OSErr e );				sets an OS or application defined error

	OSErr Error ( )							gets the current error

	void ClearError ( )						clears the current error


extending the error code list:

	(1) simply use SetError(cstr custommessage) to set arbitrary errors.
		disadvantage: you can't use errno to detect which error exactly happened.
	(2) #define CUSTOMERRORS "myownerrors.h" in your config.h to include more own errors.
	(3) extend etoe[] in errors.cp to map more OS errors to error numbers in this file.
	(4) add errors to the list in this header file.
		do not add application specific errors to this general header file!


	--------------------------------------------------------------------------------------
*/

#ifdef	ERRORS_PASS			// included from error.cp ?
	#undef	errors_h		// then allow multiple inclusion
#else
	#define	ERRORS_PASS 1
#endif


#ifndef	errors_h
#define	errors_h


// ==============================================================================
#if ERRORS_PASS == 1

	extern	bool	customError;

// methods:
	inline	int&	Error			( )							{ return errno; }								// kio 2004-10-16
	inline	bool	NoError			( )							{ return errno==noerror; }

	inline	void	ClearError		( )							{ errno=noerror; customError=0; }
	extern	void	PushError		( );
	extern	void	PopError		( );
	extern	void	DropError		( );

	inline	void	ForceError		( OSErr e )					{ errno=e; customError=0; }
	inline	void	SetError		( OSErr e )					{ if(NoError()) ForceError(e); }

// cstring methods:
	extern	void	ForceError		( OSErr e, cstr msg );
	inline	void	SetError		( OSErr e, cstr msg )		{ if(NoError()) ForceError(e,msg); }
	inline	void	ForceError		( cstr msg )				{ ForceError(-1,msg); }
	inline	void	SetError		( cstr msg )				{ if(NoError()) ForceError(-1,msg); }

	extern	cstr	ErrorStr		( );
	extern	cstr	ErrorStr		( int/*OSErr*/ );

#ifndef __cplusplus
	#error: C und nicht C++  !!
#endif

// String methods:
// only available in serrors.cp
// declared in sstring.h
/*	extern	void	ForceError		( OSErr e, cString& msg );
	inline	void	SetError		( OSErr e, cString& msg )	{ if(NoError()) ForceError(e,msg); }
	inline	void	ForceError		( cString& msg )			{ ForceError(-1,msg); }
	inline	void	SetError		( cString& msg )			{ if(NoError()) ForceError(-1,msg); }
	extern	void	AppendToError	( cString& msg );
	extern	void	PrependToError	( cString& msg );

	inline	String	ErrorString		( OSErr e )					{ return ErrorStr(e); }
	extern	String	ErrorString		( );
*/

//	#undef errno			TODO: Rework Error() with separate variable 

// base of free range for application error numbers:
	#if defined(_MACINTOSH)
		#define EBAS	128		// os errors all negative
	#elif defined(_MACOSX)
		#define	EBAS	-29700	// guessed free range
	#elif defined(_BSD)
		#define EBAS	-29700	// guessed free range
	#elif defined(_SOLARIS)
		#define EBAS	-29700	// guessed free range
	#elif defined(_LINUX)
		#define EBAS	-29700	// guessed free range
	#else
		#error    platform not yet supported: add base of free range for application error numbers here!
	#endif


// setup macro for all-sources inclusion: error enumerations
	#define 	EMAC(a,b)	a
	enum {



// ==============================================================================
#elif ERRORS_PASS == 2

// setup macro for 2nd inclusion from errors.cp: error texts
	#define		EMAC(a,b)	b
	cstr ETEXT[] = {

#else
	#error there is no pass 3 ...
#endif



// ==============================================================================
  	EMAC( outofmemory=EBAS,	"out of memory"					),
	EMAC( internalerror,	"internal error"				),
	EMAC( notimplemented,	"not implemented"				),
	EMAC( notyetimplemented,"not yet implemented"			),
	EMAC( permissiondenied,	"permission denied"				),
	EMAC( parametererror,	"parameter error to function"	),
	EMAC( indexoorange,		"index out of range"			),

	EMAC( notanumber,		"not a number"					),	// 0.0/0.0 or string is not a number
	EMAC( numbertoobig,		"number too big"				),	// +/-HUGE_VAL
	EMAC( numbertoosmall,	"number too small"				),	// 0.0 for very small non-zero value
	EMAC( notastring,		"not a string"					),	// String::StrVal()
	EMAC( nofinalquotes,	"unterminated text literal"		),	// String::StrVal()

	// device errors
	EMAC( nosuchservice,	"no such service"				),
	EMAC( notadevice,		"not a device"					),
	EMAC( badfilename,		"bad file name"					),
	EMAC( unknownfiletype,	"unknown file type"				),
	EMAC( controlcode,		"control code"					),
	EMAC( inputbufferempty,	"input buffer empty"			),
	EMAC( outputbufferfull,	"output buffer full"			),
	EMAC( inputnotpossible,	"device not open for reading"	),
	EMAC( outputnotpossible,"device not open for writing"	),
	EMAC( notsupportedbydev,"operation not supported by device"	),
	EMAC( buffertoosmall,	"buffer too small"				),

	EMAC( ioerror,			"general i/o error"				),
	EMAC( endoffile,		"end of file"					),
	EMAC( dupfilename,		"duplicate filename"			),
	EMAC( fileisbusy,		"file still in use"				),
	EMAC( filenotopen,		"file/stream not open"			),
	EMAC( filealreadyopen,	"file/stream already open"		),
	EMAC( filewritelocked,	"file already open for writing"	),
	EMAC( diskfull,			"disk full!"					),
	EMAC( volumenotfound,	"volume not found"				),
	EMAC( directorynotfound,"directory not found"			),
	EMAC( filenotfound,		"file not found"				),
	EMAC( filenotexecutable,"file is not executable"		),
	EMAC( emptyfilehandle,	"empty file handle"				),
	EMAC( unsuitedfiletype,	"unsuited file type"			),
	EMAC( wrongfiletype,	"wrong file type"				),

	EMAC( diskwriteprot,	"disk is write protected"		), 	// hardware
	EMAC( volumewriteprot,	"volume is write protected"		),	// software
	EMAC( filewriteprot,	"file is write protected"		),	// software

	EMAC( typemismatch,		"type mismatch in operator"		),	// class Var
	EMAC( variableinuse,	"variable in use"		 		),	// class Var
	EMAC( vtypeislocked,	"variable type is locked"		),	// class Var
	EMAC( vdataislocked,	"variable data is locked"		),	// class Var
	EMAC( listsintersect,	"one variable contains the other" ),// class Var
	EMAC( listrequired,		"list required" 				),	// class Var: Find()
	EMAC( corrupteddata,	"corrupted data"				),	// melt()

	EMAC( unknownconversion,"unknown conversion type"		), 	// utf-8
	EMAC( truncatedchar,	"truncated character"			),	// utf-8
	EMAC( illegaloverlong,	"illegal overlong encoding"		),	// utf-8
	EMAC( illegalchar,		"illegal character"				),	// utf-8
	EMAC( unexpectedfup,	"utf-8 char started with fup"	),	// utf-8
	EMAC( notindestcharset,	"character not in destination charset" ), // utf-8
	EMAC( missingquotes,	"missing quote(s) '\"'"			),	// utf-8
	EMAC( unmaskedamp,		"unmasked ampercent '&'"		),	// utf-8
	EMAC( endofsource,		"end of source buffer"			),	// utf-8

	EMAC( childterminatedbysignal,	"child terminated by signal" ),		// unix/os_utilities.cp
	EMAC( childreturnederror,		"child returned error" ),			// unix/os_utilities.cp


// custom application error messages
// may be empty
#include	"custom_errors.h"
};



#undef EMAC
#endif
#undef ERRORS_PASS













