/*	Copyright  (c)	Günter Woigk   1999-2005
  					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


	error handling

note:
	include errors.cp:  support for cstrings only, does not require class String
	_or_    serrors.cp: support for cstrings and class String strings


	2003-12-06 kio	derived this file from errors.cp
*/

#define	SAFE	2
#define	LOG		1

#include	"config.h"
#include	<string.h>
#include	<unistd.h>
#include	"kio.h"
#include 	"ustring/sstring.h"


#include	"errors.h"		// get enumeration
#define ERRORS_PASS 2
#include	"errors.h"		// get enumeration


bool 	customError = 0;
String 	customErrorMsg;



/* ----	conversion table ----------------------------------------------------
		OS errors -> my error numbers
		see file "MacErrors.h"
*/
#if defined(_MACINTOSH)
	static OSErr etoe[] =
	{
		dupFNErr,		dupfilename,
		fBsyErr,		fileisbusy,
		eofErr,			endoffile,
		fnOpnErr,		filenotopen,
		unitEmptyErr,	ioerror,
		opWrErr,		filewritelocked,		/*file already open with write permission*/
		dskFulErr,		diskfull,
		nsvErr,			volumenotfound,
		fnfErr,			filenotfound,
		wPrErr,			diskwriteprot,			/*disk is write protected.*/
		vLckdErr,		volumewriteprot,		/*volume is locked*/
		wrPermErr,		filewriteprot,			/*write permissions error*/
		dirNFErr,		directorynotfound,
		0,0
	};
#endif


/* ----	convert OS error code to "errors.cp" error code ------------------
*/
#if defined(_MACINTOSH)
	static OSErr ErrorToError ( OSErr err )
	{
		if ( err>=EBAS && err<EBAS+(int)NELEM(ETEXT) )
			return err;						// already my error code

		for ( OSErr* p=etoe; p<etoe+NELEM(etoe); p+=2 )
		{
			if (p[0]==err) return p[1];		// my error code found -> return it
		}
		return err;							// no own found -> return OS error
	}
#endif


/* ----	get default error text for error ------------------------------------------------
*/
cstr ErrorStr ( OSErr err )
{
	if (err==0)  return "no error";

	if (err==-1) return "unknown error (-1)";

	if (err>=EBAS && err<EBAS+(int)NELEM(ETEXT)) return ETEXT[err-EBAS];
#if defined(_MACINTOSH)
	err = ErrorToError(err);
	if (err>=EBAS && err<EBAS+(int)NELEM(ETEXT)) return ETEXT[err-EBAS];
#endif

	return Using("%s (%i)",strerror(err),(int)err);
}


/* ----	get error text for current error ------------------------------------------------
*/
cstr ErrorStr ( )
{
	return customError ? CString(customErrorMsg) : ErrorStr(Error());
}

String ErrorString ( )
{
	return customError ? customErrorMsg : ErrorStr(Error());
}



/* ----	set error code & text ------------------------------------------------
*/
void ForceError ( OSErr err, cString& msg )
{
	Error()        = err;
	customError    = 1;
	customErrorMsg = msg;
}

void ForceError ( OSErr err, cstr msg )
{
	Error()        = err;
	customErrorMsg = msg;
	customError    = 1;
}

void AppendToError ( cString& msg )
{
	customErrorMsg = ErrorString() + msg;
	customError    = 1;
}

void PrependToError ( cString& msg )
{
	customErrorMsg = msg + ErrorString();
	customError    = 1;
}


static OSErr  err[4];
static String msg[4];
static uint   idx;

void PushError ( )
{
	XTRAP(idx==NELEM(err));
	err[idx] = Error();
	msg[idx] = customError ? customErrorMsg : emptyString;
	idx++;
	ClearError();
}


void PopError ( )
{
	XXTRAP(idx==0);
	idx--;
	if (msg[idx].Len()) ForceError(err[idx],msg[idx]);
	else if (err[idx])  ForceError(err[idx]); else ClearError();
}


void DropError ( )
{
	XXTRAP(idx==0);
	idx--;
}















