/*	Copyright  (c)	Günter Woigk 1999 - 2008
  					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


	error handling

this file provides:

	raising, reading and clearing of errors.
	uses the global errno.
	errors are stored separately for each thread: kio's dispatcher saves and restores errno!
	note: some OS traps set errno directly, some return an error in a parameter, some require
	you to query the error with a dedicated method. this is not my fault!...


	cstr ErrorStr ( int errorcode )			returns the user-readable text for an OS or
											application defined error

	cstr ErrorStr ( )						returns the user-readable text for the current error

	void SetError ( cstr custommessage );	sets an error with an arbitrary error text

	void SetError ( int errorcode );		sets an OS or application defined error

	int Error ( )							gets the current error

	void ClearError ( )						clears the current error


extending the error code list:

	1:	simply use SetError("custommessage") to set arbitrary errors.
		disadvantage: you can't use errno to detect which error exactly happened.
	2:	define custom errors in "custom_errors.h"
*/



#ifndef	errors_h
#define	errors_h

#ifndef __cplusplus
	#error	C und nicht C++  !!
#endif

#ifdef __OBJC__
	// there could be some obj-c++ methods here.
#endif


// base of free range for own error numbers:
#if defined(_MACOSX)
	#define	EBAS	-29700		// guessed free range
#elif defined(_BSD)
	#define EBAS	-29700		// guessed free range
#elif defined(_SOLARIS)
	#define EBAS	-29700		// guessed free range
#elif defined(_LINUX)
	#define EBAS	-29700		// guessed free range
#else
	#error	"platform not yet supported: add base of free range for application error numbers here!"
#endif


#define 	EMAC(a,b)	a
enum {
#include	"error_emacs.h"		// include files with EMAC(number,"message") pairs
};



extern	bool	customError;

// methods:
inline	int&	Error			( )									{ return errno; }
inline	bool	NoError			( )									{ return errno==noerror; }

inline	void	ClearError		( )									{ errno=noerror; customError=0; }
inline	void	ForceError		( int/*OSErr*/ e )					{ errno=e; customError=0; }
inline	void	SetError		( int/*OSErr*/ e )					{ if(NoError()) ForceError(e); }

// cstring methods:
// implemented in "errors.cp"
extern	void	ForceError		( int/*OSErr*/ e, cstr msg );
inline	void	SetError		( int/*OSErr*/ e, cstr msg )		{ if(NoError()) ForceError(e,msg); }
inline	void	ForceError		( cstr msg )						{ ForceError(-1,msg); }
inline	void	SetError		( cstr msg )						{ if(NoError()) ForceError(-1,msg); }

extern	cstr	ErrorStr		( );
extern	cstr	ErrorStr		( int/*OSErr*/ );

// String methods:
// implemented in "serrors.cp"
// declared in "sstring.h"
#if 0
extern	void	ForceError		( int/*OSErr*/ e, cString& msg );
inline	void	SetError		( int/*OSErr*/ e, cString& msg )	{ if(NoError()) ForceError(e,msg); }
inline	void	ForceError		( cString& msg )					{ ForceError(-1,msg); }
inline	void	SetError		( cString& msg )					{ if(NoError()) ForceError(-1,msg); }
extern	void	AppendToError	( cString& msg );
extern	void	PrependToError	( cString& msg );

inline	String	ErrorString		( int/*OSErr*/ e )					{ return ErrorStr(e); }
extern	String	ErrorString		( );
#endif




// ------------------------------------------------------------------------
//
//			exception classes
//
// ------------------------------------------------------------------------

#include <new>
typedef class std::bad_alloc bad_alloc;			// #include <new>:	std::bad_alloc
typedef struct std::nothrow_t nothrow_t;		// #include <new>:	std::nothrow_t
extern const nothrow_t nothrow;					// #include <new>:	std::nothrow
extern str Using ( cstr format, ... );			// #include "cstrings/cstrings.h"

/*
namespace std 
{
	class exception								// #include <exception>
	{	public:			exception	()			throw(){}
		virtual			~exception	()			throw();
		virtual cstr	what		() const	throw();
	};

	class exception								// #include <exception>		Basisklasse

	class bad_alloc		: public exception		// #include <new>			out of memory
	class bad_typeid	: public exception		// #include <typeinfo>		Falscher Objekttyp
	class bad_cast		: public exception		// #include <typeinfo>		Falscher Objekttyp bei Typumwandlung
	class bad_exception : public exception		// #include <exception>		unexpected()

	class logic_error   : public exception		// #include <stdexcept>		theoretisch vermeidbare Laufzeitfehler
	class invalid_argument: public logic_error	// #include <stdexcept>		stdc++ lib ((general function argument error))
	class length_error	  : public logic_error	// #include <stdexcept>		stdc++ lib ((data exceeding allowed size))
	class out_of_range    : public logic_error	// #include <stdexcept>		stdc++ lib ((e.g. index))
	class domain_error    : public logic_error	// #include <stdexcept>		stdc++ lib ((domain of math. function))

	class runtime_error : public exception		// #include <stdexcept>		fehlerhafte Daten zur Laufzeit
	class range_error    : public runtime_error	// #include <stdexcept>		arith. Bereichsüberschreitung
	class overflow_error : public runtime_error	// #include <stdexcept>		arith. Überlauf
	class underflow_error: public runtime_error	// #include <stdexcept>		arith. Unterlauf
};
*/

class any_error;			// any_error : exception;	base class for own errors: includes error code
class	internal_error;		// ABORT, TRAP, ASSERT, IERR, TODO	--> "kio/abort.h"
class	index_error;		// index outside array
class	limit_error;		// array too large
class	file_error;			// file error						--> "unix/file_utilities.h"
class omem_error;			// omem_error : bad_alloc: out of memory


class any_error : public std::exception
{
	int /*OSErr*/ err;		// OS error code		err != 0
	cstr	sender;			// e.g. function name	NULL, const or in qstring pool
	cstr	message;		// error message		""

public:
	inline	any_error		( cstr s )							throw()	: err(-1), sender(0), message(s) {}
	inline	any_error		( int /*OSErr*/ e )					throw()	: err(e),  sender(0), message(0) { assert(e); }
	inline	any_error		( int /*OSErr*/ e, cstr s )			throw()	: err(e),  sender(0), message(s) { assert(e); }
	inline	any_error		( int /*OSErr*/ e, cstr a, cstr s )	throw()	: err(e),  sender(a), message(s) { assert(e); }
	virtual	~any_error		()									throw()	{ }

	inline	int /*OSErr*/ Error	() const throw()	{ return err; }
	inline	cstr	Sender	() const throw()	{ return sender; }	// may be NULL
	inline	cstr	Message	() const throw()	{ return message; }	// may be NULL

	virtual	cstr	what	() const throw()	{ return message ? message : ErrorStr(err); }
	inline	cstr	who		() const throw()	{ return sender ? sender : "An error occured"; }

	inline	cstr	Text	() const throw()	{ cstr msg = what(); if(sender) msg = Using("%s: %s", sender,msg); return msg; }
};


class file_error : public any_error
{
public:		file_error	( cstr s )							throw()	: any_error(s)		{ }
			file_error	( int /*OSErr*/ e )					throw()	: any_error(e)		{ } 
			file_error	( int /*OSErr*/ e, cstr s )			throw()	: any_error(e,s)	{ }
			file_error	( int /*OSErr*/ e, cstr s, cstr t )	throw()	: any_error(e,s,t)	{ }
	virtual	~file_error	()									throw()						{ }
};


// internal_error ~ std::logic_error
class internal_error : public any_error
{
protected:
	cstr	file;	// const or in qstring pool
	uint	line;
	cstr	filename() const throw() { cstr p = strrchr(file,'/'); return p?p+1:file; }
	
public:
	inline	internal_error	( cstr file, uint line )			throw();
	inline	internal_error	( cstr file, uint line, cstr text )	throw();	// "kio/abort.h": ABORT()
	inline	internal_error	( cstr file, uint line, int /*OSErr*/ )		throw();
	virtual	~internal_error	()									throw()		{ };
	virtual cstr what		() const							throw();
};


// limit_error ~ std::length_error
class limit_error : public internal_error
{
	size_t	size;
	
public:
	inline	limit_error	( cstr file, uint line, size_t sz )	throw();		// "kio/abort.h": LIMIT()
	virtual	~limit_error()									throw()			{ };
	virtual cstr what	() const							throw();
};


// index_error ~ std::out_of_range
class index_error : public internal_error
{
	long	index;
	
public:
	inline	index_error	( cstr file, uint line, long idx )	throw();		// "kio/abort.h": INDEX()
	virtual	~index_error()									throw()			{ };
	virtual cstr what	() const							throw();
};


// omem_error ~ std::bad_alloc
class omem_error : public bad_alloc
{
protected:
	cstr	file;	// const or in qstring pool
	uint	line;
	cstr	filename() const throw() { cstr p = strrchr(file,'/'); return p?p+1:file; }
	
public:		omem_error	( cstr myfile, long myline )		throw()			: file(myfile), line(myline) { };
	virtual	~omem_error	()									throw()			{ };
	virtual cstr what	() const							throw();
};


// --------------------------------------------------------------------------
// implementations:


// internal_error:

inline internal_error::internal_error ( cstr my_file, uint my_row ) throw()
: any_error(internalerror), file(my_file), line(my_row) { }

inline internal_error::internal_error ( cstr my_file, uint my_row, cstr text ) throw()
: any_error(internalerror,text), file(my_file), line(my_row) { }

inline internal_error::internal_error ( cstr my_file, uint my_row, int /*OSErr*/ err ) throw()
: any_error(err), file(my_file), line(my_row) { }

inline cstr internal_error::what () const throw()
{ return Using( "in file \"%s\" line %u: %s",filename(),line,any_error::what() ); }


// index_error:

inline index_error::index_error ( cstr file, uint line, long idx )	throw()
: internal_error(file,line,indexerror), index(idx) { }

inline cstr index_error::what () const throw() 
{ return Using("in file \"%s\" line %u: bad index: %li",filename(),line,index); }


// limit_error:

inline limit_error::limit_error ( cstr file, uint line, size_t sz )	throw()
: internal_error(file,line,limiterror), size(sz) { }

inline cstr limit_error::what () const throw() 
{ return Using("in file \"%s\" line %u: bad size: %zi",filename(),line,(size_t)size); }


// omem_error:

inline cstr omem_error::what () const throw() 
{ return Using("in file \"%s\" line %u: out of memory",filename(),line); }

#endif











