/*	Copyright  (c)	Günter Woigk 1999 - 2008
					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define SAFE	2
#define LOG 	1

#include	"kio.h"


// prevent recursive calls to log.cp
// prevent recursive calls to errors.cp
// preserve Error()
//
static bool lock = 0;
#define	PushError() if(lock) return; lock = 1; int err = Error();
#define	PopError()  Error() = err;   lock = 0;

// Log indenting:
//
static char lastchar = 0;
static int	indent   = 0;


void VaLog( cstr format, va_list va )
{
	if(!format) return;
	
	PushError();
		if( *format==nl )
		{
			fprintf( stderr, "\n" );
			lastchar = nl;
			format++;
		}
		if( lastchar==nl && indent>0 )
		{
			static char	spaces[21] = "                    ";	// 20 spaces
			int n = 20-indent; if(n<0) n=0; 
			fprintf( stderr, spaces+n );
		}
		vfprintf( stderr, format, va );
		lastchar = *format ? strchr(format,0)[-1] : 0;	// in rare cases (%… at format end) this may be a wrong char
	PopError();								
}


void LogNL ( )
{
	PushError();
		fprintf( stderr, "\n" );
		lastchar = nl;
	PopError();
};


void Log ( cstr format, ... )
{
	va_list va;
	va_start(va,format);
		VaLog( format, va );
	va_end(va);
}


void LogLine ( cstr format, ... )
{
	va_list va;
	va_start(va,format);
		VaLog( format, va );
	va_end(va);
	LogNL();
}


LogIndent::LogIndent ( cstr format, ... )
{
	va_list va;
	va_start(va,format);
		VaLog( format, va );
	va_end(va);
	LogNL();
	indent += 2;
}


LogIndent::~LogIndent ( )
{
	indent -= 2; 
}









