/*	Copyright  (c)	Günter Woigk 2004 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		1

#include <math.h>
#include <sys/time.h>
#include "pthreads.h"
#include "cstrings/cstrings.h"


static pthread_mutexattr_t	attr[3];
static pthread_cond_t		wait_cond;	// eine condition für Wait(), die niemals getriggert wird
static pthread_mutex_t		wait_lock;	// ihr Lock


/* ----	Initialization ---------------------------------
*/
static pthread_once_t once_control = PTHREAD_ONCE_INIT;

static void setup(void)
{
	int e = 0;
	for( int i=0;i<3;i++) e |= pthread_mutexattr_init(attr+i); 

	e |= pthread_mutexattr_settype(attr+PLock::normal,    PTHREAD_MUTEX_NORMAL);
	e |= pthread_mutexattr_settype(attr+PLock::recursive, PTHREAD_MUTEX_RECURSIVE);
	e |= pthread_mutexattr_settype(attr+PLock::errorcheck,PTHREAD_MUTEX_ERRORCHECK);

	e |= pthread_mutex_init( &wait_lock, NULL );	
	e |= pthread_cond_init( &wait_cond, NULL );

	TRAP(e);
}

inline void Setup() 
{
	int e = pthread_once(&once_control,setup); XXXTRAP(e); 
}



/* ----	Utilities -------------------------------------
*/
double CurrentTime ( )		// from unix/os_utilities.cp
{
	struct timeval tv;
	gettimeofday ( &tv, NULL );
	return tv.tv_sec + tv.tv_usec/1000000.0;
}



/* ----	PLock -----------------------------------------
*/
PLock::PLock ( Attr a )
{
	Setup();
	XXINDEX(a,NELEM(attr));
	int e = pthread_mutex_init( this, attr+a );	
	XTRAP(e); 
}



/* ----	PSema ------------------------------------------
*/
PSema::PSema ( ) 
{
	int e;
	sema = 0;		// sema<=0 -> free; sema>0 -> locked
	//who = 0;		// id of blocking thread
	name = "";
	if( (e = pthread_mutex_init( &lock, NULL )) ) goto x;
	if( (e = pthread_cond_init(  &cond, NULL )) ) goto x;
	return;
	
x:	Abort("PSema:PSema::",ErrorStr(e),"\n");
}


PSema::PSema ( cstr s ) 
{
	int e;
	sema = 0;		// sema<=0 -> free; sema>0 -> locked
	//who = 0;		// id of blocking thread
	name = s;
	if( (e = pthread_mutex_init( &lock, NULL )) ) goto x;
	if( (e = pthread_cond_init(  &cond, NULL )) ) goto x;
	return;
	
x:	Abort("PSema:PSema:",CatStr(name,":"),ErrorStr(e));
}


PSema::~PSema ( )
{
	int e;
	if( (e = pthread_mutex_destroy( &lock )) ) goto x;
	if( (e = pthread_cond_destroy ( &cond )) ) goto x;
	return;
	
x:	Abort("PSema:~PSema:",CatStr(name,":"),ErrorStr(e));
}


void PSema::Lock ( )
{
	int e;
	if( (e = pthread_mutex_lock( &lock )) ) goto x;

	// check predicate - go to sleep - recheck predicate on awakening
	//	if( sema>0 /*&& who!=pthread_self()*/ )			// 2007-05-09: recursive multiple locking should be _not_ possible!
			while( sema>0 ) if( (e = pthread_cond_wait( &cond, &lock )) ) goto x;
		sema++; who=pthread_self(); 
	
	if( (e = pthread_mutex_unlock( &lock )) ) goto x;
	return;	
	
x:	Abort("PSema:Lock:",CatStr(name,":"),ErrorStr(e));
}


void PSema::Unlock ( )
{
	int e;
	if( (e = pthread_mutex_lock( &lock )) ) goto x;

		sema -= sema>0;
		if( sema<=0 ) if( (e = pthread_cond_signal( &cond )) ) goto x;

	if( (e = pthread_mutex_unlock( &lock )) ) goto x;
	return;	
	
x:	Abort("PSema:Unlock:",CatStr(name,":"),ErrorStr(e));
}


bool PSema::TryLock ( )
{
	if( sema>0 /*&& who!=pthread_self()*/ ) return no;		// 2007-05-09 recursive multiple locking should be _not_ possible!
	
	int  e;
	bool f;
	if( (e = pthread_mutex_lock( &lock )) ) goto x;

		f = sema<=0 /*|| who==pthread_self()*/;				// 2007-05-09 recursive multiple locking should be _not_ possible!
		if(f) { sema++; who=pthread_self(); }

	if( (e = pthread_mutex_unlock( &lock )) ) goto x;
	return f;	
	
x:	Abort("PSema:TryLock:",CatStr(name,":"),ErrorStr(e));
	return 0;	// dead code
}

 
bool PSema::TryLock ( double timeout )
{
	timeout += CurrentTime();
	double time_fract, time_int; time_fract = modf(timeout,&time_int);		
	timespec time = { time_t(time_int), long(time_fract*1000000000) }; 
	XXXTRAP( time.tv_nsec==1000000000 );

	int  e;
	bool f;
	if( (e = pthread_mutex_lock( &lock )) ) goto x;

	// check predicate - go to sleep - recheck predicate on awakening
		/*if(who!=pthread_self())*/ while( sema>0 )				// 2007-05-09 recursive multiple locking should be _not_ possible!
		{
			e = pthread_cond_timedwait( &cond, &lock, &time );
			if(e) { if(e==ETIMEDOUT) break; else goto x; }
		}

		f = sema<=0 /*|| who==pthread_self()*/;					// 2007-05-09 recursive multiple locking should be _not_ possible!
		if(f) { sema++; who=pthread_self(); }

	if( (e = pthread_mutex_unlock( &lock )) ) goto x;
	return f;	

x:	Abort("PSema:TryLock(timeout):",CatStr(name,":"),ErrorStr(e));
	return 0;	// dead code
}

 
/*	recursive locking _not_ possible => no need for WaitUnlock() => use TryLock()  
bool PSema::WaitUnlock ( double timeout )
{
	timeout += CurrentTime();
	double time_fract, time_int; time_fract = modf(timeout,&time_int);		
	timespec time = { time_t(time_int), long(time_fract*1000000000) }; 
	XXXTRAP( time.tv_nsec==1000000000 );

	int  e;
	bool f;
	if( (e = pthread_mutex_lock( &lock )) ) goto x;

	// check predicate - go to sleep - recheck predicate on awakening
		while( sema>0 )
		{
			e = pthread_cond_timedwait( &cond, &lock, &time );
			if(e) { if(e==ETIMEDOUT) break; else goto x; }
		}

		f = sema<=0; 
		sema+=f;

	if( (e = pthread_mutex_unlock( &lock )) ) goto x;
	return f;	

x:	Abort("PSema:WaitUnlock(timeout):",CatStr(name,":"),ErrorStr(e));
	return 0;	// dead code
}
*/
 

/* ----	PTimer ----------------------------------------
*/
void WaitUntil ( double time )
{
	double time_fract, time_int; time_fract = modf(time,&time_int);		
	timespec wait_time = { time_t(time_int), long(time_fract*1000000000) }; 
	XXXTRAP( wait_time.tv_nsec==1000000000 );

	int e;
	if( (e = pthread_mutex_lock( &wait_lock )) ) goto x;

	e = pthread_cond_timedwait( &wait_cond, &wait_lock, &wait_time );
	if(e!=ETIMEDOUT) goto x; 

	if( (e = pthread_mutex_unlock( &wait_lock )) ) goto x;
	return;	

x:	Abort("WaitUntil():",ErrorStr(e),"");
}


void WaitDelay ( double delay )
{
	WaitUntil( CurrentTime()+delay );
}



struct exec_every_t 
{
	bool(*fu)(void*);
	void*	arg;
	double  delay;
};

struct exec_at_t 
{
	double(*fu)(void*);
	void*	arg;
	double  time;
};


static void* execute_every_proc ( void* arg )
{
	exec_every_t& x = *(exec_every_t*)arg;
	double now = CurrentTime();	
	do { WaitUntil( now += x.delay ); } while( x.fu(x.arg) ); 
	return NULL;
}

static void* execute_with_delay_proc ( void* arg )
{
	exec_every_t& x = *(exec_every_t*)arg;
	do { WaitUntil( CurrentTime() + x.delay ); } while( x.fu(x.arg) ); 
	return NULL;
}

static void* execute_at_proc ( void* arg )
{
	exec_at_t& x = *(exec_at_t*)arg;

	for(;;)
	{
		WaitUntil( x.time ); 
		double dt = x.fu(x.arg);

		if(dt<=0) break;
		if(dt<x.time/2) x.time = CurrentTime() + dt; 
		else		    x.time = dt;
	}
	return NULL;
}


void ExecuteEvery ( double delay, bool(*fu)(void*), void* arg )
{
	pthread_t thread;
	exec_every_t spawn_data = { fu,arg,delay };
	pthread_create( &thread, NULL /*attr*/, execute_every_proc, &spawn_data );
}

void ExecuteWithDelay ( double delay, bool(*fu)(void*), void* arg )
{
	pthread_t thread;
	exec_every_t spawn_data = { fu,arg,delay };
	pthread_create( &thread, NULL /*attr*/, execute_with_delay_proc, &spawn_data );
}

void ExecuteAt ( double time, double(*fu)(void*), void* arg )
{
	pthread_t thread;
	exec_at_t spawn_data = { fu,arg,time };
	pthread_create( &thread, NULL /*attr*/, execute_at_proc, &spawn_data );
}

void ExecuteAfter ( double delay, double(*fu)(void*), void* arg )
{
	ExecuteAt( CurrentTime()+delay, fu, arg );
}















